; ACL2 Version 1.8

; Copyright (C) 1989-95 Computational Logic, Inc. (CLI).  All rights reserved.

; Use of this software constitutes agreement with the terms of ACL2
; license agreement, found in the file LICENSE.

; This file combines events from Bishop Brock, John Cowles, and Matt
; Kaufmann.

(in-package "ACL2")

(include-book "equalities")

; ??? I'm not convinced we should be apply FC to RATIONALP hypotheses,
; but for now I'll go ahead and do so at times.

(local
 (defthm iff-equal
   (implies (and (iff x y)
                 (booleanp x)
                 (booleanp y))
            (equal x y))
   :rule-classes nil))

(defthm /-preserves-positive
  (implies (rationalp x)
	   (iff (< 0 (/ x))
                (< 0 x)))
  :rule-classes ((:rewrite
                  :corollary
                  (implies (fc (rationalp x))
                           (equal (< 0 (/ x))
                                  (< 0 x))))))

(defthm /-preserves-negative
  (implies (rationalp x)
	   (iff (< (/ x) 0)
                (< x 0)))
  :rule-classes ((:rewrite
                  :corollary
                  (implies (fc (rationalp x))
                           (equal (< (/ x) 0)
                                  (< x 0))))))

; The following two lemmas could be extended by adding two more such
; lemmas, i.e. for (< (* x z) (* z y)) and (< (* z x) (* y z)), but
; rather than incur that overhead here and in any other such cases
; (and besides, how about for example (< (* x z a) (* z a y))?), I'll
; wait for metalemmas to handle such things.

(defthm <-*-right-cancel
  (implies (and (fc (rationalp x))
		(fc (rationalp y))
                (fc (rationalp z)))
           (iff (< (* x z) (* y z))
                (cond
                 ((< 0 z)
                  (< x y))
                 ((equal z 0)
                  nil)
                 (t (< y x)))))
  :hints (("Goal" :use
           ((:instance (:theorem
                        (implies (and (rationalp a)
                                      (< 0 a)
                                      (rationalp b)
                                      (< 0 b))
                                 (< 0 (* a b))))
                       (a (abs (- y x)))
                       (b (abs z)))))))

(defthm <-*-left-cancel
  (implies (and (fc (rationalp x))
		(fc (rationalp y))
                (fc (rationalp z)))
           (iff (< (* z x) (* z y))
                (cond
                 ((< 0 z)
                  (< x y))
                 ((equal z 0)
                  nil)
                 (t (< y x))))))

(defthm /-inverts-order-1
  (implies (and (< 0 x)
                (< x y)
		(fc (rationalp x))
		(fc (rationalp y)))
	   (< (/ y) (/ x)))
  :hints (("Goal" :use
           ((:instance <-*-right-cancel
                       (z (/ (* x y))))))))

; ??? Interestingly, the prover fails to prove the following corollary.

#|
(defthm <-0-minus
  (iff (< 0 (- x))
       (< x 0))
  :rule-classes
  ((:rewrite :corollary
             (equal (< 0 (- x))
                    (< x 0)))))
|#

(defthm <-0-minus
  (equal (< 0 (- x))
         (< x 0))
  :hints (("Goal" :use
           ((:instance
             iff-equal
             (x (< 0 (- x)))
             (y (< x 0)))))))

(defthm <-*-0
  (implies (and (rationalp x)
                (rationalp y))
           (iff (< (* x y) 0)
                (and (not (equal x 0))
                     (not (equal y 0))
                     (iff (< x 0)
                          (< 0 y)))))
  :rule-classes
  ((:rewrite :corollary
             (implies (and (fc (rationalp x))
                           (fc (rationalp y)))
                      (equal (< (* x y) 0)
                             (and (not (equal x 0))
                                  (not (equal y 0))
                                  (iff (< x 0)
                                       (< 0 y))))))))

(defthm 0-<-*
  (implies (and (rationalp x)
                (rationalp y))
           (iff (< 0 (* x y))
                (and (not (equal x 0))
                     (not (equal y 0))
                     (iff (< 0 x)
                          (< 0 y)))))
  :rule-classes
  ((:rewrite :corollary
             (implies (and (fc (rationalp x))
                           (fc (rationalp y)))
                      (equal (< 0 (* x y))
                             (and (not (equal x 0))
                                  (not (equal y 0))
                                  (iff (< 0 x)
                                       (< 0 y))))))))

(defthm /-inverts-order-2
  (implies (and (< y 0)
		(< x y)
		(fc (rationalp x))
		(fc (rationalp y)))
	   (< (/ y) (/ x)))
  :hints (("Goal" :use
           ((:instance <-*-right-cancel
                       (z (/ (* x y)))))
           :in-theory (disable <-*-right-cancel))))

(defthm /-inverts-weak-order
  (implies (and (< 0 x)
		(<= x y)
		(fc (rationalp x))
		(fc (rationalp y)))
	   (not (< (/ x) (/ y))))
  :hints (("Goal" :use /-inverts-order-1
           :in-theory (disable /-inverts-order-1))))

(defthm <-unary-/-negative-left
  (implies (and (rationalp x)
		(rationalp y)
		(< x 0))
	   (iff (< (/ x) y)
                (< (* x y) 1)))
  :hints (("Goal" :use
           ((:instance <-*-left-cancel
                       (z x)
                       (x (/ x))
                       (y y)))
           :in-theory (set-difference-theories
                       (enable Uniqueness-of-*-inverses)
                       '(equal-/))))
  :rule-classes
  ((:rewrite :corollary
             (implies (and (< x 0)
                           (fc (rationalp x))
                           (fc (rationalp y)))
                      (equal (< (/ x) y)
                             (< (* x y) 1))))))

(defthm <-unary-/-negative-right
  (implies (and (rationalp x)
		(rationalp y)
		(< x 0))
	   (iff (< y (/ x))
                (< 1 (* x y))))
  :hints (("Goal" :use
           ((:instance <-*-right-cancel
                       (z x)
                       (x (/ x))
                       (y y)))))
  :rule-classes
  ((:rewrite :corollary
             (implies (and (< x 0)
                           (fc (rationalp x))
                           (fc (rationalp y)))
                      (equal (< y (/ x))
                             (< 1 (* x y)))))))

(defthm <-unary-/-positive-left
  (implies (and (rationalp x)
		(rationalp y)
		(< 0 x))
	   (iff (< (/ x) y)
                (< 1 (* x y))))
  :hints (("Goal" :use
           ((:instance <-*-left-cancel
                       (z x)
                       (x (/ x))
                       (y y)))))
  :rule-classes
  ((:rewrite :corollary
             (implies (and (< 0 x)
                           (fc (rationalp x))
                           (fc (rationalp y)))
                      (equal (< (/ x) y)
                             (< 1 (* x y)))))))

(defthm <-unary-/-positive-right
  (implies (and (rationalp x)
		(rationalp y)
		(< 0 x))
	   (iff (< y (/ x))
                (< (* x y) 1)))
  :hints (("Goal" :use
           ((:instance <-*-left-cancel
                       (z x)
                       (x (/ x))
                       (y y)))
           :in-theory (union-theories
                       '(Uniqueness-of-*-inverses)
                       (disable <-unary-/-positive-left
                                equal-/
                                <-*-left-cancel))))
  :rule-classes
  ((:rewrite :corollary
             (implies (and (< 0 x)
                           (fc (rationalp x))
                           (fc (rationalp y)))
                      (equal (< y (/ x))
                             (< (* x y) 1))))))

(defthm <-minus-zero
  (equal (< (- x) 0)
         (< 0 x))
  :hints (("Goal" :use
           ((:instance iff-equal
                       (x (< (- x) 0))
                       (y (< 0 x)))))))

(defthm <-*-/-right
  (implies (and (< 0 y)
                (fc (rationalp a))
                (fc (rationalp x))
                (fc (rationalp y)))
           (equal (< a (* x (/ y)))
                  (< (* a y) x)))
  :hints (("Goal" :use
           ((:instance
             <-*-right-cancel
             (x a)
             (y (* x (/ y)))
             (z y))))))

(defthm <-*-/-right-commuted
  (implies (and (< 0 y)
                (fc (rationalp x))
                (fc (rationalp y))
                (fc (rationalp a)))
           (equal (< x (* (/ y) a))
                  (< (* x y) a))))

(defthm <-*-/-left
  (implies (and (< 0 y)
                (fc (rationalp x))
                (fc (rationalp y))
                (fc (rationalp a)))
           (equal (< (* x (/ y)) a)
                  (< x (* a y))))
  :hints (("Goal" :use
           ((:instance
             <-*-right-cancel
             (y a)
             (x (* x (/ y)))
             (z y))))))

(defthm <-*-/-left-commuted
  (implies (and (< 0 y)
                (fc (rationalp x))
                (fc (rationalp y))
                (fc (rationalp a)))
           (equal (< (* (/ y) x) a)
                  (< x (* y a)))))

(encapsulate
 ()

 (local
  (defthm *-preserves->=-1
    (implies (and (rationalp x1)
                  (rationalp x2)
                  (rationalp y1)
                  (>= x1 x2)
                  (>= x2 0)
                  (>= y1 0))
             (>= (* x1 y1) (* x2 y1)))
    :rule-classes nil))

 (local
  (defthm *-preserves->=-2
    (implies (and (rationalp x2)
                  (rationalp y1)
                  (rationalp y2)
                  (>= y1 y2)
                  (>= x2 0)
                  (>= y2 0))
             (>= (* x2 y1) (* x2 y2)))
    :rule-classes nil))

 (defthm *-preserves->=-for-nonnegatives
   (implies (and (>= x1 x2)
                 (>= y1 y2)
                 (>= x2 0)
                 (>= y2 0)
                 (fc (rationalp x1))
                 (fc (rationalp x2))
                 (fc (rationalp y1))
                 (fc (rationalp y2)))
            (>= (* x1 y1) (* x2 y2)))
   :hints (("Goal" :use
            (*-preserves->=-1 *-preserves->=-2)
            :in-theory (disable <-*-left-cancel <-*-right-cancel))))

 )

#| Already covered by EQUAL-*-/-2 from rationals-equalities.lisp
(defthm equal-binary-quotient
  (implies (and (fc (rationalp x))
                (fc (rationalp y))
                (fc (rationalp z))
                (fc (not (equal z 0))))
           (equal (equal x (/ y z))
                  (equal (* x z) y)))
  :hints (("Goal" :use
           ((:instance
             (:theorem
              (implies (and (rationalp x)
                            (rationalp y)
                            (rationalp z)
                            (not (equal z 0))
                            (equal x y))
                       (equal (* x z) (* y z))))
             (x (* x z))
             (y y)
             (z (/ z)))))))
|#

(encapsulate
 ()

 (local
  (defthm *-preserves->-1
    (implies (and (rationalp x1)
                  (rationalp x2)
                  (rationalp y1)
                  (> x1 x2)
                  (>= x2 0)
                  (> y1 0))
             (>= (* x1 y1) (* x2 y1)))
    :rule-classes nil))

 (local
  (defthm *-preserves->-2
    (implies (and (rationalp x2)
                  (rationalp y1)
                  (rationalp y2)
                  (>= y1 y2)
                  (>= x2 0)
                  (>= y2 0))
             (>= (* x2 y1) (* x2 y2)))
    :rule-classes nil))


 (defthm *-preserves->-for-nonnegatives-1
   (implies (and (> x1 x2)
                 (>= y1 y2)
                 (>= x2 0)
                 (> y2 0)
                 (fc (rationalp x1))
                 (fc (rationalp x2))
                 (fc (rationalp y1))
                 (fc (rationalp y2)))
            (> (* x1 y1) (* x2 y2)))
   :hints (("Goal" :use
            (*-preserves->-1 *-preserves->-2)
            :in-theory (disable <-*-left-cancel <-*-right-cancel
                                *-preserves->=-for-nonnegatives))))

 )

(encapsulate
 ()

 (defthm x*y>1-positive-lemma
   (implies
    (and (> x i)
         (> y j)
         (rationalp i)
         (rationalp j)
         (< 0 i)
         (< 0 j)
         (fc (rationalp x))
         (fc (rationalp y)))
    (> (* x y) (* i j)))
   :hints
   (("Goal"
     :use ((:instance
            0-<-*
            (x (- x i))
            (y (- y j)))))))

 (defthm x*y>1-positive
   (implies
    (and (> x 1)
         (> y 1)
         (fc (rationalp x))
         (fc (rationalp y)))
    (> (* x y) 1))
   :rule-classes (:linear :rewrite)
   :hints
   (("Goal" :use
     ((:instance x*y>1-positive-lemma
                 (i 1) (j 1))))))
 )

(defthm <-*-x-y-y
  (implies (and (fc (rationalp x))
		(fc (rationalp y)))
	   (equal (< (* x y) y)
		  (cond
                   ((equal y 0)
                    nil)
                   ((< 0 y)
                    (< x 1))
                   (t
                    (< 1 x)))))
  :hints (("Goal" :use
           ((:instance <-*-right-cancel
                       (z y)
                       (x x)
                       (y 1))))))

(defthm <-*-y-x-y
  (implies (and (fc (rationalp x))
		(fc (rationalp y)))
	   (equal (< (* y x) y)
		  (cond
                   ((equal y 0)
                    nil)
                   ((< 0 y)
                    (< x 1))
                   (t
                    (< 1 x))))))

(defthm <-y-*-x-y
  (implies (and (fc (rationalp x))
		(fc (rationalp y)))
	   (equal (< y (* x y))
		  (cond
                   ((equal y 0)
                    nil)
                   ((< 0 y)
                    (< 1 x))
                   (t
                    (< x 1)))))
  :hints (("Goal" :use
           ((:instance <-*-right-cancel
                       (z y)
                       (x 1)
                       (y x))))))

(defthm <-y-*-y-x
  (implies (and (fc (rationalp x))
		(fc (rationalp y)))
	   (equal (< y (* y x))
		  (cond
                   ((equal y 0)
                    nil)
                   ((< 0 y)
                    (< 1 x))
                   (t
                    (< x 1))))))

(defthm <-0-+-negative-1
  (equal (< 0 (+ (- y) x))
         (< y x))
  :hints (("Goal" :use
           (:instance iff-equal
                      (x (< 0 (+ (- y) x)))
                      (y (< y x))))))

(defthm <-0-+-negative-2
  (equal (< 0 (+ x (- y)))
         (< y x))
  :hints (("Goal" :use
           (:instance iff-equal
                      (x (< 0 (+ x (- y))))
                      (y (< y x))))))

(defthm <-+-negative-0-1
  (equal (< (+ (- y) x) 0)
         (< x y))
  :hints (("Goal" :use
           (:instance iff-equal
                      (x (< (+ (- y) x) 0))
                      (y (< x y))))))

(defthm <-+-negative-0-2
  (equal (< (+ x (- y)) 0)
         (< x y))
  :hints (("Goal" :use
           (:instance iff-equal
                      (x (< (+ x (- y)) 0))
                      (y (< x y))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Expt
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defthm expt->-1
  (implies
   (and (< 1 r)
        (< 0 i)
	(fc (rationalp r))
	(fc (integerp i)))
   (< 1 (expt r i)))
  :rule-classes :linear)

(encapsulate
 ()

 (local
  (defun
    Math-induction-start-at-k ( k n )
    (declare (xargs :guard (and (integerp k)
                                (integerp n)
                                (not (< n k)))
                    :measure (if (and (integerp k)
                                      (integerp n)
                                      (< k n))
                                 (- n k)
                               0)))
    (if (and (integerp k)
             (integerp n)
             (< k n))
        (Math-induction-start-at-k k (+ -1 n))
      t)))

 (local
  (defthm hack
    (implies (and (< (* x y) z)
                  (rationalp x)
                  (rationalp y)
                  (rationalp z)
                  (< 1 x)
                  (< 0 y))
             (< y z))
    :hints
    (("Goal"
      :use
      ((:instance
        (:theorem
         (implies (and (rationalp x)
                       (rationalp y)
                       (rationalp z)
                       (< 0 y)
                       ;; w = (* x1 y)
                       (rationalp w)
                       (< 0 w)
                       (< (+ y w) z))
                  (< y z)))
        (w (* (- x 1) y))))))))
        
 (defthm expt-is-increasing-for-base>1
   (implies (and (< 1 r)
                 (< i j)
                 (fc (rationalp r))
                 (fc (integerp i))
                 (fc (integerp j)))
            (< (expt r i)
               (expt r j)))
   :rule-classes (:rewrite :linear)
   :hints (("Goal"
            :induct (Math-induction-start-at-k (+ 1 i) j))))
 )

(defthm Expt-is-decreasing-for-pos-base<1
  (implies (and (< 0 r)
                (< r 1)
                (< i j)
                (fc (rationalp r))
                (fc (integerp i))
                (fc (integerp j)))
           (< (expt r j)
              (expt r i)))
  :hints (("Goal" :use
           ((:instance
             expt-is-increasing-for-base>1
             (r (/ r)))))))

(defthm Expt-is-weakly-increasing-for-base>1
  (implies (and (< 1 r)
                (<= i j)
                (fc (rationalp r))
                (fc (integerp i))
                (fc (integerp j)))
           (<= (expt r i)
               (expt r j)))
  :hints (("Goal" :use expt-is-increasing-for-base>1
           :in-theory (disable  expt-is-increasing-for-base>1))))

(defthm Expt-is-weakly-decreasing-for-pos-base<1
  (implies (and (< 0 r)
                (< r 1)
                (<= i j)
                (fc (rationalp r))
                (fc (integerp i))
                (fc (integerp j)))
           (<= (expt r j)
               (expt r i)))
  :hints (("Goal" :use expt-is-decreasing-for-pos-base<1
           :in-theory (disable expt-is-decreasing-for-pos-base<1))))

