
(in-package "ACL2")

;; ------------------------------------------------------------------------------
;; The four functions needed are
;; 
;; (defun parse-infix (file eof) ...)
;; (defun ppr-infix (x width rpc col file eviscp) ...)
;; (defun flat-infix (x file eviscp) ...)
;; (defun flatsize-infix (x j max eviscp) ...)
;; 
;; They need to be in the acl2 package.
;; ------------------------------------------------------------------------------

;; THE TWO VARIABLES THAT I HAVEN'T FIGURE OUT HOW TO AVOID HARDWIRING.

(defparameter *parser* "/usr/cli/projects/cacl2/version20/solaris/parser/parse")
(defparameter *oneparser* "/usr/cli/projects/cacl2/version20/solaris/parser/parse1")

;; Load

(eval-when (load eval compile)

 (if (probe-file "printer/infix.o")
     (load "printer/infix.o")
     (load "printer/infix.lisp"))

 (load "printer/ascii-syntax.lisp")

)

(defvar *terminal-in-file* "Xx123xX.tmp"
  "Cheesy temp file name.  Don't want to probe every time")

(defvar *terminal-in-mirror-file* "Xx123xX.tmp.mirror")

(defparameter *oneparse*
  (format nil "~s ~s < ~s > ~s"
	  *oneparser* (f-get-global 'current-package *the-live-state*) 
	  *terminal-in-file* *terminal-in-mirror-file*))

#-GCL
(error "**********************************************************************

INFIX-INPUT-PARSE depends on the GCL specific function
SI::SYSTEM. Your lisp surely supports a similar feature, allowing
a string to be passed to the shell.  Define an appropriate
version of INFIX-INPUT-PARSE here (in infix-interface.lisp).

**********************************************************************
")

#+GCL
(defun infix-input-parse (eof)
  (multiple-value-bind (unix-status other) (si::system *oneparse*)
    (if (equal unix-status 3)
	(error "Infix parse failed.")
	(with-open-file (str *terminal-in-mirror-file* :direction :input)
	  (read str nil eof)))))

(defvar *brace-count* 0)

(defun collect-terminal-input ()
  (let* ((error '(error))
	 (input (read-line *terminal-io* nil error nil))
	 result
	 (*brace-count* 0))
    (sloop while (and (not (equal input error)) (not-semi-colon-terminated input))
	   do (progn (setq result (concatenate 'string result "
" input))
                     (setq input (read-line *terminal-io* nil error nil))))
    (if (equal input error)
	""
	(concatenate 'string result "
" input))))

(defun not-semi-colon-terminated (str)
  (let ((result t)
	(max (- (length str) 1)))
    (sloop for i from 0 to max
	   while result
	   do (cond ((and (char-equal (aref str i) #\')
			  (> (- max i) 1) 
			  (char-equal (aref str (+ i 2)) #\'))
		     (setq i (+ i 2)))
		    ((char-equal (aref str i) #\;)
		     (if (zerop *brace-count*)
			 (setq result nil)))
		    ((char-equal (aref str i) #\{)
		     (setq *brace-count* (+ *brace-count* 1)))
		    ((char-equal (aref str i) #\})
		     (setq *brace-count* (- *brace-count* 1)))))
    result))

;; (defun parse-infix (a eof) (read a nil eof))

(defun parse-infix-from-terminal (eof)

; Stream is a terminal io stream.  Eof is an arbitrary lisp object.  If file is
; empty, return eof.  Parse one well-formed infix expression from stream and
; return the corresponding s-expr.  If the stream is exhausted before the parse
; finishes or if the parse is unsuccessful, cause a hard lisp error.

; We create a "mirror" stream containing the translation of the infix forms
; in stream.
 
  (let ((input (collect-terminal-input)))
    (with-open-file (out *terminal-in-file* :direction :output)
       (princ input out))
    (infix-input-parse eof)))

;; ------------------------------------------------------------------------------

(defvar *eviscerate* nil)

;; How does this interact with the length of the prompt in acl2?
(defvar *acl2-in-output* nil
  "Controls whether we put space around events or not. If T,
suppress the initial newline in begin-event and set the 
current margin to the current loc.  Then pop in end-event.
The T case is normally used for the iteractive ACL2 loop.")

(defun print-infix (x termp width rpc col file eviscp)

; If termp, x is a term.
; Otherwise, x is either a literal or the cadr of a (quote term).

; X is an s-expression denoting a term.  File is an open output file.
; Prettyprint x in infix notation to file.  End with a newline.  If eviscp is t
; then we are to give special treatment to the :evisceration-mark; otherwise
; not.

; We bind *pacakge* so that lisp and acl2 are in the same package.  Thus
; printing works in a reasonable fashion.

; This hook is modeled after the ACL2 pretty-printer, which has the following
; additional features.  These features need not be implemented in the infix
; prettyprinter.  The printer is assumed to be in column COL, where col=0 means
; it is on the left margin.  We are supposed to print our first character in
; that column.  We are supposed to print in a field of width WIDTH.  That is,
; the largest column into which we might print is col+width-2.  Finally, assume
; that on the last line of the output somebody is going to write RPC additional
; characters and arrange for this not to overflow the col+width-2 limit.  Rpc
; is used when, for example, we plan to print some punctuation, like a comma,
; after a form and want to insure that we can do it without overflowing the
; right margin.  (One might think that the desired effect could be obtained by
; setting width smaller, but that is wrong because it narrows the whole field
; and we only want to guarantee space on the last line.)  Here is an example.
; Use ctrl-x = in emacs to see what columns things are in.  The semi-colons are
; in column 0.  Pretend they are all spaces, as they would be if the printing
; had been done by fmt-ppr.

; (foobar
;   (here is a long arg)
;   a)                  

; Here, col = 2, width = 23, and rpc = 19!

  (declare (ignore rpc))
  (let ((*infix-loc* col)
	(*rightmost-char-number* (- (+ col width) 1))
	(*minimize-event* t)
	(*eviscerate* eviscp)
	(*acl2-in-output* t)
	(*print-case* :downcase)
	(*package* (find-package (acl2-current-package)))
	(*standard-output* file)
	(*infix-loc* col))
    (set-margin)
    (cond ((not termp) (constant-printer x))
	  ((eventp x) (infix-event x))
	  (t (infix-print-term1 x)))
    (pop-margin)))


;; ------------------------------------------------------------------------------

(defvar *infinite-column-width* 1000000)

(defun print-flat-infix (x termp file eviscp)

; If termp, x is a term.
; Otherwise, x is either a literal or the cadr of a (quote term).

; Print x flat (without terpri's) in infix notation to the open output file
; file.  Give special treatment to :evisceration-mark iff eviscp.  We only call
; this function if flatsize-infix assures us that x will fit on the line.

  (let ((*infix-loc* 0)
	(*rightmost-char-number* *infinite-column-width*)
	(*minimize-event* t)
	(*eviscerate* eviscp)
	(*print-case* :downcase)
	(*package* (find-package (acl2-current-package)))
	(*standard-output* file))
     (cond ((not termp) (constant-printer x))
	   ((eventp x) (infix-event x))
	   (t (infix-print-term1 x)))))


;; ------------------------------------------------------------------------------

(defun flatsize-infix (x termp j max eviscp)

; If termp, x is a term.
; Otherwise, x is either a literal or the cadr of a (quote term).

; Suppose that printing x flat in infix notation causes k characters to come
; out.  Then we return j+k.  All answers greater than max are equivalent.

; If you think of j as the column into which you start printing flat,
; then this returns the column you'll print into after printing x.
; If that column exceeds max, which is the right margin, then it doesn't
; matter by how far it exceeds max.

  (let ((*infix-loc* j)
	(*rightmost-char-number* max)
	(*print-case* :downcase)
	(*acl2-in-output* t)
	(*package* (find-package (acl2-current-package)))
	(*eviscerate* eviscp))

    ;; If *testing* is T, then advise-flatsize may return NIL.
    ;; But that should never happen when called in this context.
    ;; I would check *testing*, but I have no idea what a sensible
    ;; return value would be.
    (if (not termp)
	(if (or (numberp x) (stringp x) (characterp x))
	    (advise-flatsize x)
	    (+ (advise-flatsize x) 1))
	(advise-flatsize x))))

