
(in-package "ACL2")

(deflabel infix-prelude-begin)

;;------------------------------------------------------------------
;; (include-book (:relative :back "cowles" "acl2-crg"))
(INCLUDE-BOOK "books/deflist" )
(INCLUDE-BOOK "books/defalist" )
(INCLUDE-BOOK "books/structures" )

;; (INCLUDE-BOOK "/usr/cli/projects/cacl2/version20/solaris/books/sets" )
;;------------------------------------------------------------------

;; Define type functions for Common Lisp types that don't have
;; simple conversions.

(defun bitp (x) (or (equal x 1) (equal x 0)))
(defun ratiop (x) (and (rationalp x) (not (integerp x))))
(defun unsigned-bytep (x) (and (integerp x) (<= 0 x)))

;; All of these foo2 functions are copied form event-printers,
;; where they are defined in raw common lisp and thus interfere
;; with the attempt to define them in acl2.  If the printer
;; was in its own package, this would not be a problem.  BUT
;; then we would hit a bunch of other problems, with getting the
;; acl2 package defined, with the right symbols included, to 
;; :USE in the INFIX package.  We should probably some day clean
;; this up and create the INFIX package, but not now.
;; - MKSd&JW 1/31/97

;; Deleted remove-keys from event-printers.lisp

(defun make-defun2 (name args doc declares body)
  (list
   (append `(defun ,name ,args)
	   (if doc (list doc) nil)
	   (if declares declares nil)
	   (list body))))

;;; (integer p) => (integerp p)
;;; (ratio x) =>  (AND (RATIONALP X) (NOT (INTEGERP X)))
;;; Only use with :lemma. INTEGER, RATIO, and etc. are
;;; not defined outside the declarations, but they
;;; are allowed in the infix form. 

(defun make-type-lemma2 (name args type lemma event-name)
  (let ((new-lemma lemma))
    (if (or type lemma)
	(list (append `(defthm ,(intern-in-package-of-symbol
				 (concatenate 'string 
					      (concatenate 'string 
							   event-name "-")
					      (concatenate 'string 
							   (symbol-name name) 
							   "-TYPE"))
				 'acl2::defthm)
			 ,(if lemma new-lemma 
			    (list type (cons name args))))
		      (if (equal event-name "const")
			  '(:rule-classes nil) nil)))
      nil)))

(defun make-disable2 (disable name)
  (if disable 
      `((in-theory (disable ,name)))
      nil))

(defun aget2 (a l)
  (cond ((not (consp l)) nil)
	((equal (car l) a) (cadr l))
	(t (aget2 a (cdr l)))))

(defun extract-declares2 (l)
  (cond ((not (consp l)) l)
	((and (consp (car l)) (equal (caar l) 'declare))
	 (cons (car l) (extract-declares2 (cdr l))))
	(t nil)))

(defun remove-declares2 (l)
  (cond ((not (consp l)) l)
	((and (consp (car l)) (equal (caar l) 'declare))
	 (remove-declares2 (cdr l)))
	(t l)))

;; All functions get (in-theory (disable function-name))

(defmacro def (name args &rest rest)
  (let* ((doc     (if (stringp (car rest)) (car rest) nil))
	 (rest    (if doc (cdr rest) rest))
	 (declares (extract-declares2 rest)))
    (let* ((rest    (remove-declares2 rest))
	   (body    (car rest)))
      (let* ((rest    (cdr rest))
	     (type    (aget2 :type rest))
	     (lemma   (aget2 :lemma rest))
	     (disable (aget2 :disable rest)))
	(cons 'progn
	      ;; These functions return NIL or a list of their result.
	      (append (make-defun2 name args doc declares body)
		      (make-type-lemma2 name args type lemma "def")
		      (make-disable2 disable name)))))))

;; (def foo (x)
;; 	(declare (type (satisfies integerp) x))
;; 	(+ x 1)
;; 	:type integer 
;; 	:lemma (implies (integerp x) (integerp (foo x)))
;; 	:disable t)
;; 
;; (def fact (n k)
;; 	(declare (type integer n)
;; 		 (type integer k)
;; 		 (xargs :guard (and (integerp n) (>= n 0))))
;; 	(if (zp n) 1 (* n (fact (- n 1))))
;; 	:lemma (implies (and (integerp n) (car k) (integerp n) (>= n 0))
;; 			(integerp (fact n k)))
;; 	:type integerp)

(defun make-label (label)
  (if (null label)
      nil
    `((deflabel ,(intern-in-package-of-symbol
			  (concatenate 'string (symbol-name label) "-START")
			  'acl2::deflabel)))))


;;; USE IN ENCAPSULATION 
(defun make-deftheory (name)
  (if (null name)
      nil
    `((deftheory ,name (SET-DIFFERENCE-THEORIES 
		       (CURRENT-THEORY :HERE )
		       (CURRENT-THEORY '
			,(intern-in-package-of-symbol
			  (concatenate 'string (symbol-name name) "-START")
			  'acl2::deftheory)))))))


(defmacro MODULE (signature &rest term)
  (let* ((name (aget2 :label term))
	 (events (if name (butlast term 2)
		  term)))
    (cons 'progn
	  (append  (make-label name)
		   (list (cons 'encapsulate (cons signature events)))
		   (make-deftheory name)))))

;;;(const *a* 3 :doc xxx :lemma (integerp *a*))

(defmacro const ($name$ value &rest rest)
  (let* ((lemma (aget2 :lemma rest))
	 (name  (intern-in-package-of-symbol (subseq 
					      (symbol-name $name$) 1 
					      (- (length (symbol-name $name$)) 
						 1)) 'acl2::defconst)))
    (cons 'progn
	  (append (list (cons 'DEFCONST  (list $name$ value)))
		  (make-type-lemma2 name nil nil lemma "const")))))



(defun make-defstub (name args body)
  (list
   (append `(defstub ,name ,args)
	   (list body))))

(defun make-defstub-axiom (name  body)
  ;; Should check the types of the mv.
  ;; E.g. given
  ;; signature foo (x:integer) : <integer, character>;
  ;; (implies (integerp x) 
  ;;          (MV-LET (a b) (foo X) (and (integerp a) (characterp b))))
  (if body
      (list `(defaxiom  ,(intern-in-package-of-symbol
			  (concatenate 'string 
				       (concatenate 'string 
						    "signature-"
						   (concatenate 'string 
								(symbol-name name) 
								"-axiom")))
				 'acl2::defaxiom)
	       ,body))
      nil))

(defmacro defstub-signature (name args &rest rest)
  (let* ((doc     (if (stringp (car rest)) (car rest) nil))
	 (rest    (if doc (cdr rest) rest)))
    (let* ((body    (car rest)))
      (let* ((rest    (cdr rest))
	     (axiom    (aget2 :axiom rest)))
	(cons 'progn
	      (append (make-defstub name args body)
		      (make-defstub-axiom name axiom )))))))

(deflabel infix-prelude-end)

