%{

#include "scansubs.h"
#include "semsubs.h"
#include "cacl2-gram.tab.h"
#include <math.h>

%}

%p 4000
%a 2000
%e 2000
%o 3000

%%

 /*  Comments are handled by preprocessor (cpp). */



\'\\n\'   { yylval.charptr = "#\\newline";
	   return TOK_CHARACTER_LITERAL; }

\'\\t\'   { yylval.charptr = "#\\tab";
	   return TOK_CHARACTER_LITERAL; }

\'\\b\'   { yylval.charptr = "#\\rubout";
	   return TOK_CHARACTER_LITERAL; }

\'\\f\'   { yylval.charptr = "#\\page";
	   return TOK_CHARACTER_LITERAL; }

\'\ \'   { yylval.charptr = "#\\space";
	   return TOK_CHARACTER_LITERAL; }

\'\.\'   { yylval.charptr = "#\\.";
	   return TOK_CHARACTER_LITERAL; }

'[!-~]'  { sprintf(yytemp, "#\\");
           yylval.charptr = (char *)strdup(strcat(yytemp,yytext+1));
	   if (yylval.charptr[yyleng] != '\'')   
	        { printf(" unterminated character literal"); }
	   else { yylval.charptr[yyleng] = '\0';
		  return TOK_CHARACTER_LITERAL;} }

\n       { yylineno++; } /* yylineno++;*/

 /* In order to allow mixed case, use flex with the -i option. */

"E0-ORD-<" {return TOK_E0_ORD_LESS;}
"BINARY-*" {return TOK_BINARY_TIME;}
"BINARY-+" {return TOK_BINARY_PLUS;}
"BINARY-@" {return TOK_BINARY_APPEND;}
"SKIP-PROOFS" { return TOK_SKIP_PROOFS;}
"CASE"    {return TOK_CASE;}
"CASE-MATCH"       {return TOK_CASE_MATCH;}
"ENUM"    {return TOK_ENUM;}     
"IF"      {return TOK_IF;}
"IF*"     {return TOK_IF_STAR;}
"THEN"    {return TOK_THEN;}
"ELIF"    {return TOK_ELIF;}
"ELSE"    {return TOK_ELSE;}       
"RECORD"  {return TOK_RECORD;}
"TYPE"    {return TOK_TYPE;}
"TYPEDEF" {return TOK_TYPEDEF;}
"LAMBDA"  {return TOK_LAMBDA;}

"WHEN"    {return TOK_WHEN;}
"UNLESS"  {return TOK_UNLESS;}
"COLLECT" {return TOK_COLLECT;}
"APPEND"  {return TOK_APPEND;}
"RETURN"  {return TOK_RETURN;}

"WORLD"      {return TOK_WORLD;}
"BOOLEAN"    {return TOK_BOOLEAN ; }
"INTEGER"    {return TOK_INTEGER ; }
"ATOM"       {return TOK_ATOM; }
"BIT"        {return TOK_BIT; }
"CHAR"       {return TOK_CHARACTER; }
"COMPLEX"    {return TOK_COMPLEX; }
"CONS"       {return TOK_CONS; }
"ASSOC"      {return TOK_ASSOC; }
"NULL"       {return TOK_NULL; }
"NIL"        {return TOK_NIL; }
"T"          {return TOK_T; }
"RATIO"      {return TOK_RATIO; }
"RATIONAL"   {return TOK_RATIONAL; }
"STRING"     {return TOK_STRING; }
"SYMBOL"     {return TOK_SYMBOL; }
"SIGNED-BYTE"   {return TOK_SIGNED_BYTE; }
"UNSIGNED-BYTE" {return TOK_UNSIGNED_BYTE; }
"STANDARD-CHAR" {return TOK_STANDARD_CHAR; }


 /* Added for ACL2 */

"THEORY-INVARIANT" {return TOK_THEORY_INVARIANT;}  
"AXIOM"   {return TOK_AXIOM;}
"CONGRUENCE" {return TOK_CONG;}
"CONSTANT" {return TOK_CONSTANT;}
"EQUIVALENCE" {return TOK_EQUIV;}
"EVALUATOR" {return TOK_EVALUATOR;}
"THEOREM" {return TOK_THEOREM;}     
"THM"     {return TOK_THM;}   
"MODULE"  {return TOK_MODULE;}     
"THEORY"  {return TOK_THEORY;}     
"MUTUAL-RECURSION"  {return TOK_MUTUAL_RECURSION;}     
"CURRENT" {return TOK_CURRENT;}
"FUNCTION"    {return TOK_FUNCTION;}
"SIGNATURE"   {return TOK_SIGNATURE;}

"LABEL"   {return TOK_LABEL;}
"MACRO"   {return TOK_MACRO;}
"PACKAGE" {return TOK_PACKAGE;}
"IN-PACKAGE" {return TOK_IN_PACKAGE;}
"LOCAL"   {return TOK_LOCAL;}
"LD"      {return TOK_LOAD;}
"INCLUDE" {return TOK_INCLUDE;}
"CERTIFY"   {return TOK_CERTIFY;}
"CERTIFY!" {return TOK_CERTIFY_BAN;}
"BRR@"     {return TOK_BRR;}
"CHAR<"    {return TOK_CHAR_LESS;}
"CHAR<="   {return TOK_CHAR_LESS_EQUAL;}
"CHAR>"     {return TOK_CHAR_GREATER;}
"CHAR>="  {return TOK_CHAR_GREATER_EQUAL;}
"STRING<" {return TOK_STRING_LESS;}
"STRING<="  {return TOK_STRING_LESS_EQUAL;}
"STRING>" {return TOK_STRING_GREATER;}
"STRING>=" {return TOK_STRING_GREATER_EQUAL;}
"SYMBOL-<" {return TOK_SYMBOL_LESS;}
"INT=" {return TOK_INTEGER_EQUAL;}

 /* "FROM"    {return TOK_FROM;} */
"TO"      {return TOK_TO ;} 
 /* "MAPPING" {return TOK_MAPPING;} */

 /* "ALL"     {return TOK_ALL;} */
 /* "SOME"    {return TOK_SOME;} */

"TABLE" {return TOK_TABLE;}
"FORALL"  {return TOK_FORALL;}
"EXISTS"  {return TOK_EXISTS;}
"NEVER"   {return TOK_NEVER;}
"FOR"     {return TOK_FOR;}
"OF"      {return TOK_OF;}
"IN"      {return TOK_IN;}
"~IN"     {return TOK_NOT_IN;}
 /* "MEMBER"  {return TOK_MEMBER;} */
"ON"      {return TOK_ON;}
"IFF"     {return TOK_IFF;}             
"AND"     {return TOK_AND;}
"OR"      {return TOK_OR;}
"IMPLIES" {return TOK_IMPLIES;}
"QUOTE"    {return TOK_QUOTE; }
"NOT "     {return TOK_NOT;} 
"UNION"      {return TOK_UNION;}
"DIFFERENCE" {return TOK_DIFFERENCE;}
"INTERSECT"  {return TOK_INTERSECT;}
"UNIVERSAL-THEORY"  {return TOK_UNIVERSAL_THEORY;}
"CURRENT-THEORY"    {return TOK_CURRENT_THEORY;}
"FUNCTION-THEORY"   {return TOK_FUNCTION_THEORY;}
"INCOMPATIBLE"   {return TOK_INCOMPATIBLE;}

"IGNORE"       {return TOK_IGNORE;}

"VERIFY"      {return TOK_VERIFY ;}
"GUARDS"      {return TOK_GUARDS;} 

"TERMINATION" {return TOK_TERMINATION;}
"DISABLE"     {return TOK_DISABLE;}
"ENABLE"      {return TOK_ENABLE;}

"&OPTIONAL"  {return TOK_MACRO_OPTIONAL;}
"&REST"      {return TOK_MACRO_REST;}
"&KEY"       {return TOK_MACRO_KEY;}
"&WHOLE"      {return TOK_MACRO_WHOLE;}
"&BODY"      {return TOK_MACRO_BODY;}
"&ALLOW-OTHER-KEYS"  {return TOK_MACRO_ALLOW_OTHER_KEYS;}

":DOC"      {return TOK_DOC;}
":HINTS"    {return TOK_HINTS;}
":GUARD"    {return TOK_GUARD;} 
":GUARD-HINTS" {return TOK_GUARD_HINTS;}
":VERIFY-GUARDS" {return TOK_VERIFY_GUARDS;} 
":HERE"     {return TOK_HERE;}
":MEASURE"  {return TOK_MEASURE;}
":MODE"     {return TOK_MODE;}
":WELL-FOUNDED-RELATION"  {return  TOK_WELL_FOUNDED_RELATION;}
":RULE-CLASSES"  {return TOK_RULE_CLASSES;}
":DEFINITION"    {return TOK_DEFINITION;}
":REWRITE"       {return TOK_REWRITE;}
":ELIM"          {return TOK_ELIM;}
":EXECUTABLE-COUNTERPART"  {return TOK_EXECUTABLE;}

":OMIT-DEFUN"  {return TOK_OMIT_DEFUN;}
":PREDICATE"  {return TOK_PREDICATE;}
":CONC-NAME"  {return TOK_CONC_NAME;}
":KEYWORD-CONSTRUCTOR"  {return TOK_KEYWORD_CONSTRUCTOR;}
":KEYWORD-UPDATER"  {return TOK_KEYWORD_UPDATER;}
":CAR-RULE-CLASSES"  {return TOK_CAR_RULE_CLASSES;}
":NTH-RULE-CLASSES"  {return TOK_NTH_RULE_CLASSES;}
":THEORY-NAME"       {return TOK_THEORY_NAME;}

":COROLLARY"        {return TOK_COROLLARY;}
":INSTRUCTIONS"     {return TOK_INSTRUCTIONS;}
":OTF-FLAG"         {return TOK_OTF_FLAG;}
":TRIGGER-FNS"      {return TOK_TRIGGER_FNS;}
":TRIGGER-terms"    {return TOK_TRIGGER_TERMS;}
":TYPE-SET"         {return TOK_TYPE_SET;}
":TYPED-TERM"       {return TOK_TYPED_TERM;}
":CLIQUE"           {return TOK_CLIQUE;}
":CONTROLLER-ALIST" {return TOK_CONTROLLER_ALIST;}
":LOOP-STOPPER"     {return TOK_LOOP_STOPPER;}
":CONDITION"        {return TOK_CONDITION;}
":PATTERN"          {return TOK_PATTERN;}
":SCHEME"           {return TOK_SCHEME;}
":LOGIC"            {return TOK_LOGIC;}
":PROGRAM"          {return TOK_PROGRAM;}

":PCB!"             {return TOK_PCBBANG;}
":PR!"              {return TOK_PRBANG;}
":PE!"              {return TOK_PEBANG;}
":MORE!"            {return TOK_MORE_BANG;}
":PUFF*"            {return TOK_PUFF_STAR;}
":REDEF!"           {return TOK_REDEF_BANG;}
":PCS"               {return TOK_PCS;}

":RELATIVE"         {return TOK_RELATIVE;}
":ABSOLUTE"         {return TOK_ABSOLUTE;}
":BACK"             {return TOK_BACK;}

":THEOREM"  {return TOK_THEOREM_KEY;}     
":INSTANCE" {return TOK_INSTANCE_KEY;}     
":FUNCTIONAL-INSTANCE" {return TOK_FUNCTIONAL_INSTANCE_KEY;}     
":DO-NOT-INDUCT" {return TOK_DO_NOT_INDUCT;}             
":DO-NOT"    {return TOK_DO_NOT;}             
"IN-THEORY"  {return TOK_IN_THEORY;}             
":IN-THEORY" {return TOK_IN_THEORY_KEY;}             
":EXPAND"    {return TOK_EXPAND;}             
":RESTRICT"  {return TOK_RESTRICT;}             
":HANDS-OFF" {return TOK_HANDS_OFF;}             
":INDUCT"    {return TOK_INDUCT;}             
":USE"       {return TOK_USE;}             
":BDD"       {return TOK_BDD;}             
":VARS"       {return TOK_VAR;}
":BDD-CONSTRUCTORS"   {return TOK_BDD_CONSTRUCTORS;}
":PROVE"     {return TOK_PROVE ;}
":LITERAL"   {return TOK_BDD_LITERAL ;}
":ALL"       {return TOK_BDD_ALL ;}
":CONC"      {return TOK_BDD_CONC;} 
":CASES"     {return TOK_CASES;}             
":BY"        {return TOK_BY;}             
":Q"         {return TOK_QUIT;}             
"good-bye"   {return TOK_GOOD_BYE;}

":SEARCH"    {return TOK_SEARCH;}

"1+"  {return TOK_ADD1;}
"1-"  {return TOK_SUB1;}

"::" {return TOK_COLON_COLON;}

[*][0-1][*]  { yylval.symptr = symlook (strdup(yytext), symtab, NSYMS);
                        return TOK_ACL2_CONSTANT;}

[a-zA-Z][-a-zA-Z0-9_:?$*]* { yylval.symptr = symlook (strdup(yytext), symtab, NSYMS);
                        return TOK_IDENTIFIER;}

[*][a-zA-Z][-a-zA-Z0-9_?:$]*[*] { yylval.symptr = symlook (strdup(yytext), symtab, NSYMS);
                        return TOK_IDENTIFIER;}

:[a-zA-Z][-a-zA-Z0-9_:?$]* { yylval.symptr = symlook (strdup(yytext), symtab, NSYMS);
                        return TOK_KEYWORD;}

 /* The following will fail to read "\"" as a string of length one  */
 /* containing a  single dbl-quote.  NO.  The extra test in the new */
 /* rule handles it. */

\"\"       { yylval.charptr = (char *)strdup(yytext);
             return TOK_STRING_LITERAL; }

\"[^"]+\"  { if (yytext[yyleng-2] == '\\')
   
             { yyless(yyleng-1);
                  yymore(); }    
              else { yylval.charptr = (char *)strdup(yytext);
                     return TOK_STRING_LITERAL; }; }


-?[1-9][0-9]*"/"-?[1-9][0-9]* { yylval.charptr = (char *)strdup(yytext);
                              return TOK_RATIONAL_LITERAL; }

0   { yylval.charptr = (char *)strdup(yytext);
            return TOK_DECIMAL_LITERAL; }

-?[1-9][0-9]* { yylval.charptr = (char *)strdup(yytext);
                return TOK_DECIMAL_LITERAL; }

-?0[0-9]+ { yylval.charptr = (char *)strcat(strdup("#o"),yytext);
            return TOK_OCTAL_LITERAL; }



":>"  {return TOK_SYM_CONS;}
"<="  {return TOK_LE;}
">="  {return TOK_GE;}
"~="  {return TOK_SYM_NE;}
 /* The \\ below is a quoted single \. */
"\\"   {return TOK_SYM_ASSOC;}
"/="  {return TOK_SYM_NEQ;}
"="   {return TOK_SYM_EQUAL;}
"=="  {return TOK_SYM_EQUAL_EQUAL;}               /* = */
":="  {return TOK_SYM_ASSIGN;}
"->"  {return TOK_SYM_IMPLIES;}
"<->" {return TOK_SYM_IFF;}
"&"   {return TOK_SYM_AND;}
"|"   {return TOK_SYM_OR;}
"'["  {return TOK_QLIST;}
"`["    {return TOK_BQ_LIST;}

",@"  {return TOK_COMMA_APPEND;}
"`,"  {return TOK_BQ_COMMA;}
"`"       {return TOK_BQUOTE;}


"~"   { return TOK_SYM_NOT;}
"<"   { return TOK_LT;}
">"   { return TOK_GT;}
"'"   { return TOK_SYM_QUOTE;}
"''"   { return TOK_SYM_DOUBLE_QUOTE;} 

"#"   { return yytext[0]; }
"-"  { return TOK_MINUS; }
"+"  { return TOK_PLUS; }
"*"  { return TOK_TIMES; }
"%"  { return TOK_REM; }
"/"  { return TOK_DIV; }
"@"  { return TOK_SYM_APPEND;}

[]!^.,?:;(){}[] {return yytext[0];}

[ \t\r]+ ;


%%

void yyerror (char *s)
{
 fail_flag = 1;
 fprintf (stderr, "\n;; *** %s at %s on line %d",s,yytext,yylineno+1);
};
