

/*

Routines to implement semantic actions.  These do no take into account
data validation which must be in for a general purpose client-useful
tool.

nil
cons
list1
c_list
cons_terminal
ct_list
cons_symbol
append
car
cdr
writetree
topwritetree

*/

#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <stdlib.h> 
#include "scansubs.h"
#include "semsubs.h"

int fail_flag = 0;
int symbol_quote =0; 
char *remove_tag = "--remove--";
PCELL z = NULL;
CTYPE car_err;
int nodecounter = 0; /* force line returns periodically */

PCELL a,b,c,d,e,f; CTYPE g; 

int cdrp (l) PCELL l; /* if cdr is non NULL */
{ if (l->pcdr) return 1; else return 0 ; };

int consp_car (l) PCELL l;
{ return (l->caflg == tlist); };

int stringp_car (l) PCELL l;
{ return ((l->caflg == tstring) || (l->caflg == tsym)); };

int symbolp_car (l) PCELL l;
{ return (l->caflg == tsym); };

/* CTYPE res; can be a string or a node or a string in the symbol table 

CTYPE car (l) PCELL l;
{ if (consp_car(l))   return l->pcar.pcell;    
  if (stringp_car(l)) return strcar(l);
  return car_err;
};

*/

PCELL lcar (l) PCELL l;
{ return l->pcar.pcell; };

PSTRING strcar (l) PCELL l;
{ if (l==NULL) return "";
  if (l->caflg == tsym) return l->pcar.psymbols->id;
  if (l->caflg == tstring) return l->pcar.atom; 
  if (consp_car(l) && stringp_car(lcar(l)) && (string_eq(remove_tag,strcar(lcar(l)))))
    return strcar(cdr(l));
  return ""; };

PSYMBOLS symcar (l) PCELL l;
{ return l->pcar.psymbols; };

PCELL cdr(l) PCELL l;
 { if (l) return l->pcdr; else return NULL; };

PCELL newnil ()
 { PCELL res;
  res = (PCELL) malloc (sizeof (struct cell));
  res->caflg = tlist;
  return res; };

PCELL cons (x,y) PCELL x; PCELL y;
 {
  PCELL res;
  res = (PCELL) malloc (sizeof (struct cell));
  res->caflg = tlist;
  if (x == NULL) res->pcar.pcell = newnil(); 
  else res->pcar.pcell = x;
  res->pcdr = y;

  /*  printf ("\n the tree as we do a cons is:\n");*/
  /*  writetree (res);*/
  /*  printf("\n");*/
  /*  printf ("\n the symbol table as we do a cons is:\n");*/
  /*             write_table(symtab,NSYMS);*/
  return res;
 };

PCELL list1 (x) PCELL x;
 { PCELL res;
   res = cons(x, NULL);
   return res;
 };


PCELL cons_terminal (x,y) PSTRING x; PCELL y;
 { PCELL res;
  
   res = (PCELL) malloc (sizeof (struct cell));
   if (!res) { perror("Cannot allocate memory in cons_terminal"); };
   res->caflg = tstring;
   /* The following malloc is necessary because */
   /* for some strange reasons in the execution */
   /* strcpy would return funny symbols like ^ and */
   /* concatenation of other strings */
   
   res->pcar.atom = (PSTRING) malloc (strlen(x)+1); //Segmentation Fault
  
   strcpy(res->pcar.atom,x) ;
  
   /* not necessary, but it makes sure the new */
   /* string is the same length of copied string */

   *(res->pcar.atom +  strlen(x)) = '\0';

   /*******************************************/
   
   res->pcdr = y;
   //printf("\nexsiting cons_terminal");

   /*  printf ("\n the tree as we do a cons_terminal is:\n");*/
   /*  writetree (res);*/
   /*  printf ("\n the symbol table as we do a cons_terminal is:\n");*/
   /*             write_table(symtab,NSYMS);*/
   return res;
 };


PCELL cons_symbol (x,y) PSYMBOLS x; PCELL y;
 { PCELL res;
   res = (PCELL) malloc (sizeof (struct cell));
   res->caflg= tsym;
   res->pcar.psymbols = x;       /* symlook(x, symtab, NSYMS); */
   res->pcdr = y;
   return res;
 };
 
/* debug print from above - printf ("aux->id = %s = x = %s\n",aux->id,x); */
/*  printf ("\n the tree as we do a cons_symbol is:\n");*/
/*  writetree (res);*/
/*  printf ("\n the symbol table as we do a cons_symbol is:\n"); */
/*             write_table(symtab,NSYMS);*/

/*  
 append(x,y) : if (x.pcdr == NULL) {return y;} else return cons(car(x), append(cdr(x), y)); 
*/

PCELL append (x,y) PCELL x; PCELL y;
 { 
   if (x == NULL)          { return y;}
   if (x->caflg == tstring){ return cons_terminal(x->pcar.atom, append(cdr(x),y));}
   if (x->caflg == tsym)   { return cons_symbol(x->pcar.psymbols, append(cdr(x),y));}
   return cons(lcar(x), append (cdr(x),y));

 };

PCELL nil()
 { return cons_terminal("NIL",NULL);
 };

void topwritetree (p) PCELL  p;
 { printf("\n"); writetree(p); /* freetree(p); */  };

void writetree (p) PCELL  p;
 {
   if (p == z) return;

   /* Jimmy change of string_equal -> string_eq */
   if (p->caflg == tstring && string_eq(p->pcar.atom, remove_tag)) 
     { if (p->pcdr != z) {writenodes(p->pcdr); return; }; }; 

   printf ("(");
   writenodes (p);
   printf (")");
   
 };

int removep (p) PCELL p;
 /* Jimmy change of string_equal -> string_eq */
 { if (p != z && p->caflg == tstring && string_eq(p->pcar.atom, remove_tag)) return 1; else return 0; };

void writenodes (p) PCELL p;
 { PSYMBOLS aux;
   PSTRING s;

   if (p == z) { return; }		/* z is null - did include free(p) */

   /* car is a string. */
   if (p->caflg == tstring) {
     /* Jimmy change of string_equal -> string_eq */
     if (string_eq(p->pcar.atom, remove_tag)) {
       printf(" . "); writenodes(p->pcdr); return; }
     else if (string_eq(p->pcar.atom, "`") ||
	      string_eq(p->pcar.atom, "'") ||
	      string_eq(p->pcar.atom, ",") ||
	      string_eq(p->pcar.atom, ",@") ||
              string_eq(p->pcar.atom, "`,") ||
	      string_eq(p->pcar.atom, "''"))
       printf("%s",p->pcar.atom);
     else printf("%s ",p->pcar.atom);};

   /* car is element of symbol table. */
   if (p->caflg == tsym) printf ("%s ", p->pcar.psymbols->id); 
  
   if (p->caflg == tlist) 
     /* added so that when we print to the terminal we don't exceed the
	maximum line size. */
     { if (nodecounter > 16) { printf("\n "); nodecounter = 0; };
       nodecounter++;
       if ( removep(p->pcar.pcell) )
	 writenodes(p->pcar.pcell->pcdr); 
       else 
	 writetree(p->pcar.pcell);}
 
 
   if (p->pcdr != z) 
     writenodes (p->pcdr);
 };

void topwritenodes (p) PCELL p;
 { PSYMBOLS aux;
   PSTRING s;

   if (p == z) { return; }		/* z is null - did include free(p) */

   printf("\n");

   /* car is a string. */
   if (p->caflg == tstring) 
     /* Jimmy change of string_equal -> string_eq */
     if (string_eq(p->pcar.atom, "`") || string_eq(p->pcar.atom, "'")
	  || string_eq(p->pcar.atom, ","))
       printf("%s",p->pcar.atom);
     else printf("%s ",p->pcar.atom);

   /* car is element of symbol table. */
   if (p->caflg == tsym) printf ("%s ", p->pcar.psymbols->id); 
  
   if (p->caflg == tlist) 
     { if ( removep(p->pcar.pcell) )
	 writenodes(p->pcar.pcell->pcdr); 
       else 
	 writetree(p->pcar.pcell); }
   if (p->pcdr != z) topwritenodes (p->pcdr); 
 };

void freetree (p) PCELL  p;
 { if (p == z) return;
   freenodes (p);
   free (p);
 };

void freenodes (p) PCELL p;
 { PSYMBOLS aux;
   PSTRING s;

   if (p == z) { return; }		/* z is null - did include free(p) */
   if (p->caflg == tstring || p->caflg == tsym) return; /* car is string in symbol table. */
   if (p->pcdr != z) freenodes (p->pcdr); 
   if (p->caflg == tlist) { freetree(p->pcar.pcell); };
   return; };

  
/* Various old code mains & tests.
main()
 { writetree( make_id("A"));
   writetree( cons_terminal("ID",cons_terminal ("A", cons_terminal("NIL",z))));
 }

Consider: yylval.charptr = strdup(yytext + 1);

main()
 { PSTRING atest,btest;
   atest = "abcdefg";
   printf("atest=%s",atest);
   btest = strdup(atest + 1);
   printf("btest=%s",btest);
  }

main()
 { PCELL x = cons_terminal(remove_tag, cons_terminal("A", z));
   if (x != z) printf("X ne z\n");
   if (x->caflg == tstring) printf("X->caflg == tstring\n");
   printf("X->pcar.atom == %s\n", x->pcar.atom);

   if (string_eq(x->pcar.atom,remove_tag)) printf("match\n"); else printf("nomatch\n");
   if (removep (x))
     printf("A");
   else printf("B");
 }

*/

void ptree (p) PCELL  p;
 { if (p == z) return;
   printf ("(");
   pnodes (p);
   printf (")");
 };

void pnodes (p) PCELL p;
 { PSYMBOLS aux;
   PSTRING s;

   if (p == z) { return; }		/* z is null - did include free(p) */

   /* car is a string. */
   if (p->caflg == tstring) printf("%s ",p->pcar.atom); 

   /* car is element of symbol table. */
   if (p->caflg == tsym) printf ("%s ", p->pcar.psymbols->id); 
  
   if (p->caflg == tlist) 
     { ptree(p->pcar.pcell); }

   if (p->pcdr != z) pnodes (p->pcdr); 
 };







