
in-package("ACL2");

/* We need the concrete version for simulation. 
   More accurate at this point, but not simulatable would be:
   type a-caeti-address;   */

type a-caeti-address = integer;

type a-CMA-type = enum [ tell, ask, do, subscribe, sorry ];

type a-CMA-msg =
 record (kind : a-CMA-type,
         sender : a-caeti-address,
         receiver : a-caeti-address,
         content  : string,
         language : string := "text",
         ontology : string := "text",
         reply-with  : string,
         in-reply-to : string);

type a-CMA-history = a-CMA-msg * ;

type an-in-out = record ( inl : a-CMA-history, outl : a-CMA-history );

type a-caeti-system = [ a-caeti-address . an-in-out ] *;

theorem caeti-system-true-list
 { a-caeti-system(x) -> true-listp(x) };

function is-unique-caeti-address (s : a-caeti-address, l : a-caeti-system) : boolean 
 { ~(l \ s) };

function all-caeti-addresss-unique (l)
{ if ~consp(l)
    then t 
  elif a-caeti-system(l) & is-unique-caeti-address(l.car.car, l.cdr)
    then all-caeti-addresss-unique(l.cdr)
  else nil };

verify guards all-caeti-addresss-unique;

function is-caeti-system (l : a-caeti-system) : boolean 
 { all-caeti-addresss-unique(l) };

function update-outgoing-msg (msg : a-CMA-msg ,
			      s : a-caeti-system): a-caeti-system
/* The nil in make-an-in-out in the empty in history */
{ if endp(s) then [ sender(msg) :> make-an-in-out(nil , [msg]) ]
  elif msg.sender = car(s.car)
    then cons( s.car.car :> update-an-in-out(s.car.cdr,
					     :outl := cons(msg, outl(s.car.cdr))),
	       s.cdr)
  else s.car :> update-outgoing-msg(msg, s.cdr) };

function update-incoming-msg (msg : a-CMA-msg ,
			      s : a-caeti-system): a-caeti-system
{ if endp(s) then [ receiver(msg) :> make-an-in-out([msg] , nil) ]
  elif receiver(msg) = s.car.car
    then cons( s.car.car :> update-an-in-out(s.car.cdr, :inl := cons(msg, inl(s.car.cdr))),
	       s.cdr)
  else cons(s.car, update-incoming-msg(msg, s.cdr)) };

function update-caeti-system (msg : a-CMA-msg, 
                              s : a-caeti-system): a-caeti-system
{  update-incoming-msg (msg,
	update-outgoing-msg (msg, s)) };

function caeti-sim ( l : a-CMA-history, s : a-caeti-system): a-caeti-system
{ if endp(l) then s
  else caeti-sim(l.cdr, update-caeti-system(l.car, s)) };

caeti-sim(
 '[[MAKE-A-CMA-MSG tell  1 2 "msg a from 1 to 2" "text" "text" "" ""]
   [MAKE-A-CMA-MSG ask   2 1 "query from 2 to 1" "text" "text" "" ""]
   [MAKE-A-CMA-MSG tell  1 2 "response from 1 to 2" "text" "text" "" ""]
   [MAKE-A-CMA-MSG ask   3 1 "query from 3 to 1" "text" "text" "" ""]
   [MAKE-A-CMA-MSG sorry 3 1 "failure report 3 to 1" "text" "text" "" ""]],
 nil);


