
in-package("ACL2");

FUNCTION bar(pid,thread)
 { < x, y > := f(pid);
   q(x,y,thread) };

FUNCTION step-form(pid,thread,form,time,st,pd,td,cd)
 {if  null(form) then end-fn(pid,thread,st)
      elif integerp(form) then push-v-pop-p(pid,thread,make-integer(form),st)
      elif symbolp(form) then read-var-fn(pid,thread,form,st)
      elif consp(form)
      then CASE-MATCH form {
            [ 'alloc classname] : alloc-fn(pid,thread,classname,st,td);
            [ 'assert expr] : 
             update-st(push-p-pop-p(pid,thread,expr :> [['stack, form]],st),
                       :atomic);
            [ 'assign varnames expr] : 
              varnames : ignore;
              push-p-pop-p(pid,thread,expr :> [['stack, form]],st); 
            [ 'atomic . stmts] : 
             update-st(push-p-pop-p(pid,
                                    thread,
                                    stmts @ ['[stack [end-atomic]]],
                                    st),
                       :atomic);
            [ 'await . awaitarms] : 
             update-st(push-p-pop-p(pid,
                                    thread,
                                    map-car(awaitarms) @ [['stack, form]],
                                    st),
                       :atomic);
            [ 'block . stmts] : block-fn(pid,thread,stmts,st);
            [ 'comment . forms] : forms : ignore; pop-p(pid,thread,st); 
            [ 'dealloc expr] : push-p-pop-p(pid,
                                            thread,
                                            [expr, ['stack, form]],
                                            st);
            [ 'defvar type . vars] : local-fn(pid,thread,type,vars,[],st,td);
            [ 'defconst type . vars] : local-fn(pid,thread,type,vars,t,st,td);
            [ 'dot expr fieldname] : 
              fieldname : ignore;
              push-p-pop-p(pid,thread,expr :> [['stack, form]],st); 
            [ 'eblock stmt expr] : eblock-fn(pid,thread,stmt,expr,st);
            [ 'if expr . forms] : 
              forms : ignore;
              push-p-pop-p(pid,thread,[expr, ['stack, form]],st); 
            [ 'initial type] : 
             if  type-exprp(type,td)
                 then push-v-pop-p(pid,thread,default-tv(type,td),st)
                 else error-fn(pid,['unrecognized, 'type, form],st);
            [[ 'list typex] . exprs] : 
              typex : ignore;
              push-p-pop-p(pid,thread,exprs @ [['stack, form]],st); 
            [[ 'map typename] . pairs] : 
              typename : ignore;
              push-p-pop-p(pid,
                            thread,
                               flatten-alist(pairlis$(domain(pairs),
                                                      map-cadr(pairs)))
                             @ [['stack, form]],
                            st);
            [ 'new classname . exprs] : 
               classname : ignore;
               push-p-pop-p(pid,thread,exprs @ [['stack, form]],st); 
            [[ 'set typex] . exprs] : 
              typex : ignore;
              push-p-pop-p(pid,thread,exprs @ [['stack, form]],st); 

            [ 'objects classname] : objects-fn(pid,thread,classname,st,td);
            [ 'par . stmts] : par-fn(pid,thread,stmts,st);
            [ 'print stringx . exprs] : 
              stringx : ignore;
              update-st(push-p-pop-p(pid,thread,exprs @ [['stack, form]],st),
                         :atomic); 
            [ 'quote x] : 
             if  symbolp(x)
                 then push-v-pop-p(pid,
                                   thread,
                                   make-tv(:val, x, :typex, 'symbol),
                                   st)
                 else error-fn(pid,
                               ['unrecognized, 'quoted, 'form, form],
                               st);
            [[ 'rec typename] . exprs] : 
              typename : ignore;
              push-p-pop-p(pid,thread,exprs @ [['stack, form]],st); 
            [ 'ref expr fieldspec] : 
              fieldspec : ignore;
              push-p-pop-p(pid,thread,expr :> [['stack, form]],st); 
            [[ 'ref ptr methodspec] . exprs] : 
              methodspec : ignore;
              push-p-pop-p(pid,thread,ptr :> (exprs @ [['stack, form]]),st); 
            [ 'return . exprs] : 
             push-p-pop-p(pid,thread,exprs @ [['stack, form]],st);
            [ 'seq . stmts] : push-p-pop-p(pid,thread,stmts,st);
            [ 'setdot expr fieldname value] : 
             { fieldname : ignore;
               push-p-pop-p(pid,thread,expr :> value :> [['stack, form]],st); };
            [ 'setref expr fieldspec value] : 
             { fieldspec : ignore;
               push-p-pop-p(pid,thread,expr :> value :> [['stack, form]],st); };
            [ 'throw tag . exprs] : 
             { tag : ignore;
               push-p-pop-p(pid,thread,exprs @ [['stack, form]],st); };
            [ 'try stmt . handlers] : try-fn(pid,thread,stmt,handlers,st,td);
            [ 'while expr stmt] : 
             { stmt : ignore;
               push-p(pid,thread,[expr, ['stack, form]],st); };
            [qfier [typex var] [ 'in expr] pred] : 
             { typex : ignore;
               if  qfier member '[exists forall choose list-of set-of]
                   then push-p-pop-p(pid,thread,[expr, ['stack, form]],st)
                   else error-fn(pid,['unrecognized, 'form, form],st); };
            [ 'stack form1] : 
             CASE-MATCH form1 {
              [ 'assert expr] : { expr : ignore; assert-fn(pid,thread,st); };
              [ 'assign varnames expr] : 
               { expr : ignore;
                 assign-fn(pid,thread,varnames,st,td); };
              [ 'await . awaitarms] : await-fn(pid,thread,awaitarms,time,st,td);
              [ 'dealloc expr] : { expr : ignore; dealloc-fn(pid,thread,st); };
              [ 'dot expr fieldname] : 
               { expr : ignore;
                 dot-fn(pid,thread,fieldname,st,td); };
              [ 'end-atomic] : 
               update-st(pop-p(pid,thread,st),:atomic, cdr(st-atomic(st)));
              [ 'if expr thenstmt elsestmt] : 
               { expr : ignore;
                 if-fn(pid,thread,thenstmt,elsestmt,st); };
              [[ 'list typex] . exprs] : 
               list-constant-fn(pid,thread,typex,len(exprs),st,td);
              [[ 'map typename] . pairs] : 
               map-constant-fn(pid,thread,typename,2 * len(pairs),st,td);
              [ 'new classname . exprs] : 
               new-fn(pid,thread,classname,len(exprs),st,td,cd);
              [ 'print stringx . exprs] : 
               print-fn(pid,thread,stringx,len(exprs),st);
              [[ 'rec typename] . exprs] : 
               rec-constant-fn(pid,thread,typename,len(exprs),st,td);
              [ 'ref expr fieldspec] : 
               { expr : ignore;
                 ref-fn(pid,thread,fieldspec,st,td); };
              [[ 'ref ptr methodspec] . exprs] : 
               { ptr : ignore;
                 method-fn(pid,thread,methodspec,len(exprs),st,td,cd); };
              [ 'return . exprs] : return-fn(pid,thread,exprs,st);
              [[ 'set typex] . exprs] : 
               set-constant-fn(pid,thread,typex,len(exprs),st,td);
              [ 'setdot expr fieldname value] : 
               { expr : ignore;
                 setdot-fn(pid,thread,fieldname,st,td); };
              [ 'setref expr fieldspec value] : 
               { expr : ignore;
                 setref-fn(pid,thread,fieldspec,st,td); };
              [ 'throw tag . exprs] : 
               { exprs : ignore;
                 throw-fn(pid,thread,tag,st,td); };
              [ 'while expr stmt] : { expr : ignore; while-fn(pid,
                                                              thread,
                                                              stmt,
                                                              st); };
              [qfier [type var] [ 'in expr] pred] : 
               { expr : ignore;
                 quantifier-fn(pid,thread,qfier,type,var,pred,st,td); };
              & : 
               if  bound?(car(form1),*oprs*)
                   then apply-opr-fn(pid,
                                     thread,
                                     car(form1),
                                     len(cdr(form1)),
                                     st,
                                     td)
                   else call-fn(pid,thread,car(form1),st,pd,td);
              };
            & : push-p-pop-p(pid,thread,cdr(form) @ [['stack, form]],st);
            }
      else error-fn(pid,['unrecognized, 'form, form],st)};


