
/* ACL2 Version 1.8 */
/* Copyright (C) 1989-95 Computational Logic, Inc. (CLI).  All rights reserved. */
/* Use of this software constitutes agreement with the terms of ACL2
   license agreement, found in the file LICENSE. */
/* This file combines events from Bishop Brock, John Cowles, and Matt
   Kaufmann, with help from Art Flatau. */

in-package "ACL2";

/* The following book of John Cowles should be included before certifying
   the present book. */

/* (DEFPKG "ACL2-ASG" (UNION-EQ *ACL2-EXPORTS* *COMMON-LISP-SYMBOLS-FROM-MAIN-LISP-PACKAGE*))
   (DEFPKG "ACL2-CRG" (UNION-EQ *ACL2-EXPORTS* *COMMON-LISP-SYMBOLS-FROM-MAIN-LISP-PACKAGE*))
   (DEFPKG "ACL2-AGP" (UNION-EQ *ACL2-EXPORTS* *COMMON-LISP-SYMBOLS-FROM-MAIN-LISP-PACKAGE*))
   Then:(certify-book "rationals-equalities" 3 nil) */

/*** Include (:relative :back "cowles" "acl2-crg"); ***/

/* ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   Force
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; */

/* It's tempting to force all acl2-numberp hypotheses, and the like.
   Since this is a rather new thing to do, let write a macro for force
   that we can easily change if we want to. */

/* (defmacro fc (x)  (list 'force x)) */

/* Or, since in this version we want to try this without forcing: */

Macro fc(x) x;

/* ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Facts about + (and -)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; */

Theorem commutativity-2-of-plus { x + (y + z) = y + (x + z) };

Theorem functional-self-inversion-of-minus { - (- x) = fix(x) };

Theorem distributivity-of-minus-over-plus { - (x + y) = (- x) + (- y) };

Theorem minus-cancellation-on-right { x + y + (- x) = fix(y) };

Theorem minus-cancellation-on-left { x + (- x) + y = fix(y) };

/* Note that the cancellation rules below (and similarly for *) aren't
complete, in the sense that the element to cancel could be on the
left side of one expression and the right side of the other.  But
perhaps those situations rarely arise in practice.  (?) */

Theorem right-cancellation-for-plus
{ (x + z = y + z) = (fix(x) = fix(y)) };

Theorem left-cancellation-for-plus
{ (x + y = x + z) = (fix(y) = fix(z)) };

Theorem equal-minus-0 { (0 = - x) = (0 = fix(x)) };

Theorem inverse-of-plus-as-equal-0 { (a - b = 0) = (fix(a) = fix(b)) };

Theorem equal-minus-minus { (- a = - b) = (fix(a) = fix(b)) };

Theorem fold-consts-in-plus
{ syntaxp(quotep(x)) & syntaxp(quotep(y)) -> x + (y + z) = (x + y) + z };

/* ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Facts about * (and /)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; */

/* The same as Inverse-of-*, from axioms.lisp, but with force.

(defaxiom right-inverse-of-times  (implies (and (acl2-numberp x)                (not (equal x 0)))           (equal (* x (/ x)) 1))) */

/* The following proof of commutativity-2-of-times was originally obtained by using   John Cowles's macro acl2-asg::add-commutativity-2 as follows, and   then editing out package references in the statement of the final   theorem.(acl2-asg::add-commutativity-2 equal			       acl2-numberp 			       * 			       commutativity-of-times			       commutativity-2-of-times) */

Theorem commutativity-2-of-times { x * (y * z) = y * (x * z) }
 :hints [["Goal" :use [:instance [:functional-instance acl2-asg--commutativity-2-of-op
                                                       [[acl2-asg--equiv (equal)]
                                                        [acl2-asg--pred (lambda(x) { t })]
                                                        [acl2-asg--op (binary-times)]]]
                                 [[acl2-asg--x (x)][acl2-asg--y (y)][acl2-asg--z (z)]]]
                 ]];

Theorem functional-self-inversion-of-div { / (/ x) = fix(x) }
 :hints [["Goal" :use [[:instance [:functional-instance acl2-agp--involution-of-inv
                                                        [[acl2-agp--equiv (equal)]
                           [acl2-agp--pred (lambda(x) { acl2-numberp(x) & x ~= 0 })]
                                                         [acl2-agp--op (binary-times)]
                           [acl2-agp--id (lambda() { 1 })]
                                                         [acl2-agp--inv (unary-div)]]]
                                  [[acl2-agp--x (x)]]]
                       ]]];

Theorem distributivity-of-div-over-times { / (x * y) = (/ x) * (/ y) }
 :hints [["Goal" :use [[:instance [:functional-instance acl2-agp--distributivity-of-inv-over-op
                                                        [[acl2-agp--equiv (equal)]
                          [acl2-agp--pred (lambda(x) { acl2-numberp(x) & x ~= 0 })]
                                                         [acl2-agp--op (binary-times)]
                           [acl2-agp--id (lambda() { 1 })]
                                                         [acl2-agp--inv (unary-div)]]]
                                  [[acl2-agp--x (x)][acl2-agp--y (y)]]]
                       ]]];

Theorem div-cancellation-on-right
{ fc(acl2-numberp(x)) & fc(x ~= 0) -> x * y * (/ x) = fix(y) }
 :hints [["Goal" :use [[:instance [:functional-instance acl2-agp--inv-cancellation-on-right
                                                        [[acl2-agp--equiv (equal)]
                           [acl2-agp--pred (lambda(x) { acl2-numberp(x) & x ~= 0 })]
                                                         [acl2-agp--op (binary-times)]
                            [acl2-agp--id (lambda() { 1 })]
                                                         [acl2-agp--inv (unary-div)]]]
                                  [[acl2-agp--x (x)][acl2-agp--y (y)]]]
                       ]]];

Theorem div-cancellation-on-left
{ fc(acl2-numberp(x)) & fc(0 ~= x) -> x * (/ x) * y = fix(y) }
 :hints [["Goal" :use div-cancellation-on-right]];

Module 
{}
{ 
  local Theorem right-cancellation-for-times-lemma
  {      x * z = y * z
       & acl2-numberp(z)
       & 0 ~= z
       & acl2-numberp(x)
       & acl2-numberp(y)
    -> (x = y) = t }
   :hints [["Goal" :use [[:instance [:functional-instance acl2-agp--right-cancellation-for-op
                                                          [[acl2-agp--equiv (equal)]
                             [acl2-agp--pred (lambda(x) { acl2-numberp(x) & x ~= 0 })]
                                                           [acl2-agp--op (binary-times)]
                            [acl2-agp--id (lambda() { 1 })]
                                                           [acl2-agp--inv (unary-div)]]]
                                    [[acl2-agp--x (x)]
                                     [acl2-agp--y (y)]
                                     [acl2-agp--z (z)]]]
                         ]]];
  
  Theorem right-cancellation-for-times
  { (x * z = y * z) = (0 = fix(z) | fix(x) = fix(y)) };
  
   };

Theorem left-cancellation-for-times
{ (z * x = z * y) = (0 = fix(z) | fix(x) = fix(y)) };

Theorem zero-is-only-zero-divisor
{ (x * y = 0) = (fix(x) = 0 | fix(y) = 0) };

Theorem equal-times-x-y-x
{ (x * y = x) = (x = 0 | (y = 1 & acl2-numberp(x))) }
 :hints [["Goal" :use [[:instance right-cancellation-for-times
                                  [[z (x)][x (y)][y (1)]]]
                       ]]];

Theorem equal-times-x-y-y
{ (x * y = y) = (y = 0 | (x = 1 & acl2-numberp(y))) }
 :hints [["Goal" :use [[:instance right-cancellation-for-times
                                  [[z (y)][x (x)][y (1)]]]
                       ]]];

Module 
{}
{ 
  local Theorem equal-div-lemma
  { acl2-numberp(x) & acl2-numberp(y) & x * y = 1 -> y = / x }
   :hints [["Goal" :use [:instance [:functional-instance acl2-agp--uniqueness-of-op-inverses
                                                         [[acl2-agp--equiv (equal)]
 	 [acl2-agp--pred (lambda(x) {   acl2-numberp(x) & x ~= 0 })] 
                                                   [acl2-agp--op (binary-times)]
        
         [acl2-agp--id (lambda() { 1 })] 
                                                          [acl2-agp--inv (unary-div)]]]
                                   [[acl2-agp--x (x)][acl2-agp--y (y)]]]
                   ]];
  
  Theorem equal-div
  { fc(acl2-numberp(x)) & fc(0 ~= x) -> (/ x = y) = (1 = x * y) }
   :hints [["Goal" :use equal-div-lemma]];
  
   };

/* The following hack helps in the application of equal-div when forcing is
turned off. */
Theorem numerator-nonzero-forward
{ numerator(r) ~= 0 -> r ~= 0 & acl2-numberp(r) }
 :rule-classes [[:forward-chaining :trigger-terms [numerator(r)]]] ;

/*  :rule-classes [[:forward-chaining :trigger-terms [numerator(r)]]];
The following loops with the lemma equal-div just proved but is
sometimes useful. */

Module {}
{ 
  local Theorem uniqueness-of-times-inverses-lemma
  { (x * y = 1) = (x ~= 0 & acl2-numberp(x) & y = / x) };
  
  Theorem uniqueness-of-times-inverses
  { (x * y = 1) = (fix(x) ~= 0 & y = / x) }
   :hints [["Goal" :in-theory disable(equal-div)]];
  
   };

in-theory(disable(uniqueness-of-times-inverses));

theory-invariant(~ (  '[:rewrite uniqueness-of-times-inverses] in theoryx
                    & '[:rewrite equal-div] in theoryx));

Module 
{}
{ 
  local Theorem equal-div-div-lemma
  {    fc(acl2-numberp(a)) & fc(acl2-numberp(b)) & fc(a ~= 0) & fc(b ~= 0)
    -> (/ a = / b) = (a = b) }
   :hints [["Goal" :use [[:instance [:theorem      fc(acl2-numberp(a))
                                                 & fc(acl2-numberp(b))
                                                 & fc(a ~= 0)
                                                 & fc(b ~= 0)
                                              -> (a = b -> / a = / b)]
                                    [[a (/ a)][b (/ b)]]]
                         ]]];
  
  Theorem equal-div-div { (/ a = / b) = (fix(a) = fix(b)) }
   :hints [["Goal" :use equal-div-div-lemma]];
  
   };

Theorem equal-times-div-1
{    acl2-numberp(x) & x ~= 0
  -> ((/ x) * y = z) = (acl2-numberp(z) & fix(y) = x * z) };

Theorem equal-times-div-2
{    acl2-numberp(x) & x ~= 0
  -> (y * (/ x) = z) = (acl2-numberp(z) & fix(y) = z * x) };

Theorem fold-consts-in-times
{ syntaxp(quotep(x)) & syntaxp(quotep(y)) -> x * (y * z) = (x * y) * z };

Theorem times-zero { 0 * x = 0 };

/* ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Facts about + (or -) and * (or /) together
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; */

Module 
{}
{ 
  local Theorem functional-commutativity-of-minus-times-right-lemma
  { fc(acl2-numberp(x)) & fc(acl2-numberp(y)) -> x * (- y) = - x * y }
   :hints [["Goal" :use [[:instance [:functional-instance acl2-crg--functional-commutativity-of-minus-times-right
                                                          [[acl2-crg--equiv (equal)]
                                                           [acl2-crg--pred (acl2-numberp)]
                                                           [acl2-crg--plus (binary-plus)]
                                                           [acl2-crg--times (binary-times)]
         [acl2-crg--zero (lambda() { 0 })]
                                                           [acl2-crg--minus (unary--)]]]
                                    [[acl2-crg--x (x)][acl2-crg--y (y)]]]
                         ]]];
  
  Theorem functional-commutativity-of-minus-times-right
  { x * (- y) = - x * y }
   :hints [["Goal" :use functional-commutativity-of-minus-times-right-lemma]];
  
   };

Theorem functional-commutativity-of-minus-times-left
{ (- x) * y = - x * y };

Theorem reciprocal-minus { / (- x) = - / x }
 :hints [["Goal" :cases [fc(acl2-numberp(x)) & fc(x ~= 0)]]];

/* ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Facts about numerator and denominator
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; */

Module 
{}
{ 
  local Theorem numerator-integerp-lemma-1
  {    rationalp(x)
    -> (numerator(x) * (/ denominator(x))) * denominator(x) = numerator(x) }
   :hints [["Goal" :in-theory disable(rational-implies2)]];
  
  local Theorem numerator-integerp-lemma
  {    rationalp(x) & numerator(x) * (/ denominator(x)) = x
    -> numerator(x) = x * denominator(x) }
   :hints [["Goal" :use [numerator-integerp-lemma-1]
                   :in-theory disable(rational-implies2)]];
  
  Theorem numerator-when-integerp { integerp(x) -> numerator(x) = x }
   :hints [["Goal" :in-theory disable(rational-implies2)
                   :use [[:instance lowest-terms
                                    [[r (x)][q (1)][n (denominator(x))]]]
                         rational-implies2
                         numerator-integerp-lemma
                         ]]];
  
   };

Theorem integerp-denominator-1 { integerp(x) -> denominator(x) = 1 }
 :hints [["Goal" :use [rational-implies2 numerator-when-integerp]
                 :in-theory disable(rational-implies2)]];

Theorem equal-denominator-1
{ (denominator(x) = 1) = (integerp(x) | ~ rationalp(x)) }
 :hints [["Goal" :use [rational-implies2 completion-of-denominator]
                 :in-theory disable(rational-implies2)]];

Theorem times-r-denominator-r
{   r * denominator(r)
  = (if rationalp(r) then numerator(r)
        else fix(r)) }
 :hints [["Goal" :use [[:instance rational-implies2 [[x (r)]]]]
                 :in-theory disable(rational-implies2)]];

Theorem div-r-when-abs-numerator-1
{   (numerator(r) = 1 -> / r = denominator(r))
  & (numerator(r) = -1 -> / r = - denominator(r)) }
 :hints [["Goal" :use [[:instance rational-implies2 [[x (r)]]]]
                 :in-theory disable(rational-implies2)]];

/* ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Facts about expt
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; */

/* Much of this is adapted from John Cowles's acl2-exp.lisp book.
There are various modifications, however. */

Theorem expt-type-prescription-rationalp
{ rationalp(r) -> rationalp(expt(r, i)) }
 :rule-classes [:type-prescription :generalize];

Theorem expt-type-prescription-positive
{ 0 < r & rationalp(r) -> 0 < expt(r, i) }
 :rule-classes [:type-prescription :generalize];

Theorem expt-type-prescription-nonzero
{ fc(acl2-numberp(r)) & r ~= 0 -> 0 ~= expt(r, i) }
 :rule-classes [:type-prescription :generalize];

Theorem expt-type-prescription-integerp
{ 0 <= i & integerp(r) -> integerp(expt(r, i)) }
 :rule-classes [:type-prescription :generalize];

Theorem left-nullity-of-1-for-expt { expt(1, i) = 1 };

Theorem right-unicity-of-1-for-expt { expt(r, 1) = fix(r) }
 :hints [["Goal" :expand [expt( r, 1)] ]];

Theorem expt-minus { expt(r, - i) = / expt(r, i) };

/* The following is superseded by exponents-add below, except for the
case that r = 0.  But I'll leave it here; in fact it's quite natural
to have (roughly speaking) two versions of each rule about expt,
based on the disjunction in the guard for expt. */

Theorem exponents-add-for-nonneg-exponents
{    0 <= i & 0 <= j & fc(integerp(i)) & fc(integerp(j))
  -> expt(r, i + j) = expt(r, i) * expt(r, j) };

Module 
{}
{ 
  local Theorem exponents-add-negative-negative
  {    integerp(i) & integerp(j) & i < 0 & j < 0
    -> expt(r, i + j) = expt(r, i) * expt(r, j) };
  
  local Theorem exponents-add-positive-negative
  {    integerp(i) & integerp(j) & acl2-numberp(r) & r ~= 0 & 0 < i & j < 0
    -> expt(r, i + j) = expt(r, i) * expt(r, j) }
   :hints [["Goal" :expand [expt( r, i + j) ]]];
  
  Theorem exponents-add
  {    0 ~= r & fc(acl2-numberp(r)) & fc(integerp(i)) & fc(integerp(j))
    -> expt(r, i + j) = expt(r, i) * expt(r, j) }
   :hints [["Goal" :use [exponents-add-negative-negative
                         exponents-add-positive-negative
                         [:instance exponents-add-positive-negative
                                    [[i (j)][j (i)]]]
                         ]]];
  
   };

Theorem distributivity-of-expt-over-times
{ expt(a * b, i) = expt(a, i) * expt(b, i) };

/* It's not clear to me whether the following rule belongs this way
or the other way around, but I'll leave it this way -- mk. */

Theorem expt-1 { expt(1, x) = 1 };

Theorem exponents-multiply
{ fc(integerp(i)) & fc(integerp(j)) -> expt(expt(r, i), j) = expt(r, i * j) }
 :hints [["Goal" :cases [~ acl2-numberp(r), r = 0]]];

Theorem functional-commutativity-of-expt-div-base
{ expt(/ r, i) = / expt(r, i) }



