
#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <stdlib.h> 
#include "scansubs.h"
#include "semsubs.h"
#include "types.h"

/* guards is list whose elements are
   ((type (satisfies foo) v1 ... vn)
    (type integer v1 ... vn)
    (xargs (:guard pred))) 
*/

PCELL type_axioms = NULL;
PSTRING type_suffix = "-1";
PSTRING package = "ACL2";

PSTRING our_strcat(s1,s2) PSTRING s1; PSTRING s2;
{ PSTRING temp = (PSTRING) malloc (strlen(s1) + strlen(s2) + 1);
  strcpy(temp,s1);
  return strcat(temp, s2); };

PCELL make_verify_guard()
 { return cons_terminal_package("DECLARE",
			list1(cons_terminal_package("XARGS",
				      cons_terminal(":verify-guards",
						    cons_terminal_package("T", NULL)))));  };

PCELL make_guard()
 { PCELL tmp;
   if (guards) tmp = cons_terminal_package("DECLARE", guards); else tmp = NULL;
   guards = NULL;
   return tmp; };

PCELL add_measure(PCELL measure)
 { if (arg_guards) { arg_guards = append(arg_guards,list1(cons_terminal_package("XARGS", measure))); }
   else arg_guards = cons_terminal_package("DECLARE", list1(cons_terminal_package("XARGS", measure)));
   /*  Apr 21 97 MKS  */
   return NULL; };

void save_function_info(name,args,guards)
 PCELL name; PCELL args ; PCELL guards;
 { /* name is symbolp and stringp_car */
   if (!name->caflg == tsym) printf("Saving function name that is not a symbol.\n");
   function_name = name->pcar.psymbols->id;
   function_args = args;
   /*printf("\nRaw guards of %s\n",function_name); */
   /* ptree(guards); */
   function_guards = function_guard(guards);
   /*printf("\nTransformed guards of %s\n",function_name); */
   /*ptree(function_guards);*/
   if (function_guards && cdr(function_guards))
     function_guards = cons_terminal_package("AND", function_guards);
   return; };

void save_function_type(type,annotation)
 PCELL type ; PCELL annotation;
 { function_type = type; 
   function_annotation = annotation;
   return; };

void reset_function_info()
 {
   function_name = ""; 
   function_recursive = 0;
   macro_id_flag = 0;
   set_macro_arg =0;
   function_args = NULL;
   guards = NULL;
   lemma_arg_type =  NULL;
   function_guards = NULL;
   function_measure = NULL;
   function_type = NULL; 
   function_annotation = NULL;
   lambda_value = NULL;
   macro_argument = NULL;
   type_axioms = NULL;
   arg_guards = NULL;}

int recursivep (PCELL p) 
 { if (p != NULL && !strcasecmp(name_of(p),function_name)) return 1;
   return 0; };

/* (defthm ,function-type (implies ,function-guard (,type (,name ,@args)))) */

PCELL make_function_type_lemma(PSTRING kind)
 { return make_function_type_lemma2(kind,
				    function_name,
				    function_args,
				    function_guards,
				    function_annotation,
				    function_type); };

PCELL make_function_type_lemma2(PSTRING kind, PSTRING name, PCELL args, PCELL guards,PCELL annotation, PCELL type)
 { PCELL hypes = NULL;
   PCELL conc = NULL;
   PCELL func = cons_terminal(name,args);
  
   hypes = append(lemma_arg_type,guards);

   /* if (hypes) printf("\nThere are lemma hypes.\n"); */

   if (hypes) {
     if (cdr(hypes)) {
       /* printf("Conjoining hypes\n"); */
       hypes = cons_terminal_package("AND",hypes);}
     else {
       /* printf("\nSingleton hype\n"); */
       hypes = lcar(hypes);}; };

   if (annotation) { 
     if (type) conc = 
		 cons_terminal_package("AND", cons(cons_type_name_x(func,name_of(type)),
						   list1(annotation))); 
     else conc = annotation; }
   else {
     if (type) {
       /* Needs test for MV */
       /* printf("\nSetting type\n"); */
       conc = cons_type_name_x(func, name_of(type)) ; }
     else return NULL; };

   /* printf("\nconc = "); writenodes(conc); */

   if (hypes) {
     conc = cons_terminal_package("IMPLIES", cons(hypes, list1(conc))); };

   name = our_strcat(make_name_suffix(name,"-TYPE"), type_suffix);
   return cons_terminal_package(kind, cons_terminal_package(name, list1(conc))); 
};

PCELL make_function_requirement(PCELL name)
{ /* printf("\nLooking for axiom for %s in\n", name_of(name)); 
     if (type_axioms) ptree(type_axioms); */
  return find_function_requirement(name_of(name),type_axioms); };

PCELL find_function_requirement(PSTRING name,PCELL axioms)
{ if (!axioms) return NULL;
  if (consp_car(axioms))
    { if (stringp_car(lcar(axioms)))
      {if (!strcasecmp(name,strcar(lcar(axioms))))
	{ /* printf("\nFOUND: "); ptree(cdr(lcar(axioms))); */
	  return cdr(lcar(axioms)); }; }
      else { /* printf("\nNOT stringp_car(lcar(axioms))"); ptree(lcar(axioms)); */ }; }
  else { /* printf("\nNOT consp_car(axioms)"); ptree(axioms); */ }; 

  return find_function_requirement(name,cdr(axioms)); };

PCELL make_disable(PSTRING name)
 { return cons_terminal_package("in-theory",
			list1(cons_terminal_package("disable", cons_terminal(name, NULL)))); };

/* (defthm ,id@"-type" (,type ,id)) */

PSTRING make_name_suffix(name,suffix) PSTRING name; PSTRING suffix;
{ return our_strcat(name, suffix); };

PSTRING make_constant_suffix(name,suffix) PSTRING name; PSTRING suffix;
/* What happened to SUFFIX, its not used!!!  Jan  7 97 MKS */
{ return substring(name,1,strlen(name)-2); };

PCELL make_type_lemma(id,type) PCELL id; PCELL type;
 { PSTRING name = "";
   name = make_name_suffix(name_of(id),"-TYPE");
   return cons(make_name_acl2("DEFTHM"), cons_terminal(name, list1(cons(type,cons(id,NULL))))); };

int stringequal (x,y) char *x; char *y; {
  return !strcasecmp(x,y) ; };

PSTRING satisfies_name(l) PCELL l; {
  /* e.g. l = (satisfies foo) */
  PSTRING temp;
/*   printf("\n In satisfies_name: l = "); ptree(l); */
  if (!cdrp(l)) return "";
  if (strcasecmp("satisfies",strcar(l))) return ""; /* car ne satisfies */
  temp = name_of(cdr(l)); 
/*   if (!strcmp(temp,"")) printf("\n  Returning empty string"); */
/*   printf("\n  Returning %s", temp); */
  return temp; };

PSTRING name_of(l) PCELL l; {
  /* e.g. l = (foo) | (--remove-- foo) | ((--remove-- foo)) */
  /*   printf("\n  Entering name_of "); ptree(l); */
  if (l == NULL) { printf("\nName_of(l) == NULL"); return ""; };
  if (stringp_car(l))
    { if (strcasecmp(strcar(l),remove_tag)) return strcar(l); 
      return name_of(cdr(l)); };
  return name_of(lcar(l)); };




PCELL function_name_of(l) PCELL l; {
  /* e.g. l = (defun foo ...) | (defun (--remove-- foo) ...) | 
              (progn (defun foo ...) ...) | (progn (defun (--remove-- foo) ...)  ...) */

  if (l == NULL) { return NULL; };
  if (stringp_car(l)) {
    if (strcasecmp(strcar(l),"DEFUN")) {
      if (strcasecmp(strcar(l),"PROGN")) return l; 
      function_name_of(lcar(cdr(l))); }
    return cdr(l); };
  return lcar(l); };

PCELL type_to_pred(l) PCELL l; {
  /* e.g. (integer x y) | ((satisfies foo) x y) */
  PCELL res;
  PSTRING name; 
  if (l == NULL)         return NULL;
  if (!cdrp(l))          return NULL;
  if (stringp_car(l))    name = strcar(l);
  if (consp_car(l))      name = satisfies_name(lcar(l));
  res = types_to_pred(name,cdr(l)); 
  return res; };

PCELL types_to_pred(id,tail) PSTRING id; PCELL tail;
{ PCELL tmp;
  PCELL res;
  PSTRING name;
  if (!tail) return tail;
  if (consp_car(tail)) tmp = cons_terminal(id, cons(lcar(tail), NULL));
  if (stringp_car(tail)) tmp = cons_terminal(id, cons_terminal(strcar(tail), NULL));
  res = types_to_pred(id,cdr(tail));
  return cons(tmp,res); };

PCELL function_guard (l) PCELL l;
/* L is list of xargs containig guards or type decls */
 { if (l == NULL) return NULL;
   if (consp_car(l)) return append(guard_to_pred(lcar(l)),function_guard(cdr(l)));
   return function_guard(cdr(l)); };

PCELL guard_to_pred (l) PCELL l; {
/* L is an xargs, i.e. l = (xargs :key value ..) */
  char *id = "";
  PCELL res = NULL;
  if (l == NULL) return NULL;
  if (consp_car(l))   return NULL;
  if (stringp_car(l)) id = strcar(l);
  if (!strcasecmp(id, "TYPE")) res = type_to_pred(cdr(l));
  /* (xargs :guard pred) */
  if (!strcasecmp(id, "XARGS")) res = xarg_to_pred(cdr(l)); 
  return res; };

/* (xargs guard form) */
/* (guard form) */

PCELL xarg_to_pred (l) PCELL l; {
  if (!l || !cdr(l)) return l;
  /* cdr past :guard */
  l = cdr(l);
  /* if (consp_car(l)) return lcar(l); */
  return l; };

PSTRING strcat_package(x) PSTRING x;
{ if (!strcasecmp(package , "ACL2"))
    return(x);
  return our_strcat("ACL2::",x); }

PCELL make_name(x) PSTRING x;
{ return cons_terminal(remove_tag, cons_terminal(x, NULL)); };

PCELL make_name_acl2(x) PSTRING x;
{ if (!strcasecmp(package, "ACL2")) 
    return cons_terminal(remove_tag, cons_terminal(x, NULL));
  return cons_terminal(remove_tag, cons_terminal(our_strcat("ACL2::",x), NULL)); };

PCELL cons_terminal_package(x,y) PSTRING x; PCELL y;
{ return cons(make_name_acl2(x), y);};

PCELL make_id_from_key(key) PCELL key;
{ PSTRING name = name_of(key);
  if (name[0] == ':') name[0] = ' '; /* substring(name,2,strlen(name)-1); */
  return make_name_acl2(strdup(name)); };

PCELL make_type_name(x) PSTRING x;
{ if (!strcasecmp(x,"INTEGER")) return make_name_acl2("INTEGERP");
  if (!strcasecmp(x,"ATOM")) return make_name_acl2("ATOM");   
  if (!strcasecmp(x,"BOOLEAN")) return make_name_acl2("BOOLEANP");   
  if (!strcasecmp(x,"BIT")) return make_name_acl2("BITP");   /* Defined in prelude */
  if (!strcasecmp(x,"CHARACTER")) return make_name_acl2("CHARACTERP");   
  if (!strcasecmp(x,"COMPLEX")) return make_name_acl2("COMPLEX-RATIONALP");   
  if (!strcasecmp(x,"CONS")) return make_name_acl2("CONSP");   
  if (!strcasecmp(x,"NIL")) return make_name_acl2("NIL");   
  if (!strcasecmp(x,"RATIO")) return make_name_acl2("RATIOP");       /* In prelude */  
  if (!strcasecmp(x,"RATIONAL")) return make_name_acl2("RATIONALP"); /* Fixed */
  if (!strcasecmp(x,"STRING")) return make_name_acl2(  "STRING");    /* Fixed */
  if (!strcasecmp(x,"SYMBOL")) return make_name_acl2(  "SYMBOLP");   
  if (!strcasecmp(x,"SIGNED-BYTE")) return make_name_acl2( "INTEGERP");   
  if (!strcasecmp(x,"UNSIGNED-BYTE")) return make_name_acl2( "UNSIGNED-BYTEP"); /* In prelude */
  if (!strcasecmp(x,"STANDARD-CHAR")) return make_name_acl2( "STANDARD-CHAR-P"); 
  return make_name(x); };


PCELL cons_type_name_x(x, type)  PCELL x; PSTRING type;
{ return cons(make_type_name(type),list1(x));  };

int ignore_function_types (x) PCELL x; {
  if (x == NULL) return 1;
  if (ignore_function_type(x)) return ignore_function_types (cdr(x));
  return 0; };

int ignore_function_type (x) PCELL x; {
  /* if x is T, STATE, or MV of T and STATE, return true */
  if ((stringequal(strcar(x),"acl2::mv")) || (stringequal(strcar(x),"mv")))
      return ignore_function_types(cdr(x));
  return (   stringequal(name_of(x),"t") 
	  || stringequal(name_of(x),"state")
	  || stringequal(name_of(x),"acl2::t") 
	  || stringequal(name_of(x),"acl2::state"))  ; };

void make_arg_guards()
{ arg_guards = make_guard();
  guards = NULL;};

 /* (defun NAME (x) (cond ((not (consp x)) (null x)) ((TYPE (car x)) (NAME (cdr x))) (t nil))) */

PCELL make_defun_setp(name, id, type, expr) PCELL name; PCELL id; PCELL type; PCELL expr;
 { 
   PCELL tmp = NULL;
   PCELL xlist = cons(id, NULL);
   PCELL carf = cons_terminal_package("CAR", xlist);
   PCELL cdrf = cons_terminal_package("CDR", xlist);
   PCELL endp = cons(cons_terminal_package("NOT",list1(cons_terminal_package("consp", xlist))),
		     list1(cons_terminal_package("NULL", xlist)));
   PCELL tnil = cons_terminal_package("T", cons_terminal_package("NIL",NULL));
   PCELL test = cons_terminal_package("AND",
			       cons(cons(type, list1(carf)),
				    cons(expr, NULL)));
   PCELL typeb = cons(test, list1(cons(name, list1(cdrf))));
   return
     cons_terminal
     ("DEFUN",
      cons(name,
	   cons(xlist,
		cons(make_verify_guard(),
		list1(cons_terminal_package("COND",
				    cons(endp,
					 cons(typeb, cons(tnil, NULL)))))))));};

PCELL make_defun_enum(name, ids) PCELL name; PCELL ids;
 { 
   PCELL tmp = NULL;
   PCELL xlist = cons_terminal_package("X", NULL);
   return
     cons_terminal
     ("DEFUN",
      cons(name,
	   cons(xlist,
		list1(cons_terminal_package("member",
				    cons_terminal_package("X",
						  cons_terminal("'",
								      list1(ids))))))));};

PCELL make_defun_truelistp(name) PCELL name;
 { 
   PCELL tmp = NULL;
   PCELL xlist = cons_terminal_package("X", NULL);
   PCELL test  = cons_terminal_package("TRUE-LISTP", xlist);
   return
     cons_terminal_package
     ("DEFUN",
      cons(name,
	   cons(xlist,
		list1(test)))); };

PCELL make_defun_listp(name, type) PCELL name; PCELL type;
 { PCELL llist = cons_terminal_package("L", NULL);
   return cons_terminal_package("DEFLIST",
			cons(name,
			     cons(llist,
				  cons(make_verify_guard(), cons(type,NULL))))); };

PCELL make_defun_alistp(name, type1, type2) PCELL name; PCELL type1; PCELL type2;
 { PCELL xlist = cons_terminal_package("L", NULL);
   return cons_terminal_package("DEFALIST",
			cons(name,
			     cons(xlist,
				  cons(make_verify_guard(),
				       list1(cons(type1, cons_terminal(".",
								       list1(type2)))))))); };

PCELL make_defun_vector_conjunct(types) PCELL types;
{ PCELL tmp = NULL;
  char n = '0';
  char foo[2] = { n , '\0' };
  tmp = list1(cons_terminal_package("TRUE-LISTP", cons_terminal_package("X", NULL)));
  while ( types != NULL && n <= '9' )
    { tmp = append(tmp,list1(cons_type_name_x(list1(cons_terminal_package("NTH", 
							    cons_terminal(strdup(foo),
						       cons_terminal_package("X", NULL)))),
					      name_of(lcar(types)))));
      n = n+1; foo[0]=n;
      types  = cdr(types); };
  return cons_terminal_package("AND", tmp); };

PCELL make_defun_vectorp(name, types) PCELL name; PCELL types;
 { PCELL xlist = cons_terminal_package("X", NULL);
   PCELL typeb = make_defun_vector_conjunct(types);
   return
     cons_terminal_package
     ("DEFUN",
      cons(name,
	   cons(xlist,
		cons(make_verify_guard(),
		     list1(make_defun_vector_conjunct(types)))))); };

/*
PCELL make_defun_alistp(name, type1, type2) PCELL name; PCELL type1; PCELL type2;
 { 
   PCELL tmp = NULL;
   PCELL xlist = cons_terminal_package("X", NULL);
   PCELL consp = cons_terminal_package("CONSP", list1(cons_terminal_package("CAR", xlist)));
   PCELL carcarf = cons_terminal_package("CAR", list1(cons_terminal_package("CAR", xlist)));
   PCELL cdrcarf = cons_terminal_package("CDR", list1(cons_terminal_package("CAR", xlist)));
   PCELL carf = cons_terminal_package("CAR", xlist);
   PCELL cdrf = cons_terminal_package("CDR", xlist);
   PCELL endp = cons(cons_terminal_package("not", list1(cons_terminal_package("consp", xlist))),
		     list1(cons_terminal_package("NULL", xlist)));
   PCELL tnil = cons_terminal_package("T", cons_terminal_package("NIL",NULL));
   PCELL typeb = cons(cons_terminal_package("AND",
				    cons(consp,
					 cons(cons(type1, list1(carcarf)),
					      list1(cons(type2, list1(cdrcarf)))))),
		      list1(cons(name, list1(cdrf))));
   return
     cons_terminal_package
     ("DEFUN",
      cons(name,
	   cons(xlist,
		cons(make_verify_guard(),
		     list1(cons_terminal_package("COND",
					 cons(endp,
					      cons(typeb, cons(tnil, NULL))))))))); };
					      */

PCELL make_encapsulate(id,dcl,body)PCELL id; PCELL dcl; PCELL body;
{ PCELL stubs;
if (dcl == NULL) { stubs = cons_terminal_package("MODULE",cons_terminal_package("NIL",body)); }
    else stubs = cons_terminal_package("MODULE", cons(dcl, body));
  if (!id) 
    return stubs;
  else
    return append(stubs,cons_terminal(":Label",
				      cons_terminal(name_of(id),NULL)));};


PCELL make_theory_diff(id,from,to) PSTRING id; PSTRING from; PSTRING to;
{ return cons_terminal_package
	  ("DEFTHEORY", 
	   cons_terminal(id,
		list1(cons_terminal_package
		      ("SET-DIFFERENCE-THEORIES",
		       cons(make_current_theory(to),
			    cons(make_current_theory(from),NULL)))))); };

PCELL make_current_theory(th) PSTRING th;
{ if (string_equal (th, ":HERE"))
  return cons_terminal_package("CURRENT-THEORY",cons_terminal(th, NULL));
else
  return cons_terminal_package("CURRENT-THEORY",
		       cons_terminal("'", cons_terminal(th, NULL))); }; 

PCELL replace(old, new, form) PCELL old; PCELL new; PCELL form;
 { PSYMBOLS aux;
   PSTRING s;

   if (old == NULL)  { printf("\nOld === NULL!!!"); }
   if (new == NULL)  { printf("\nNew === NULL!!!"); }

   if (form == NULL)  return form;

   if (forms_equal(old,form)) return new;

   if (symbolp_car(form)) return cons_symbol(symcar(form),replace(old,new,cdr(form)));

   if (stringp_car(form)) {
       if (string_eq(strcar(form) ,"@"))
           return append(cons_terminal(remove_tag,(cons_terminal("@",
		 			    list1((cons_terminal(remove_tag,
			 				   list1(lcar(cdr (form))))))))),
			     replace(old,new,cdr(cdr(form))));
       else
          return cons_terminal(strcar(form),replace(old,new,cdr(form)));
   };
   /* car is list. */
   if ( forms_equal(old,lcar(form)) ) return cons(new,replace(old,new,cdr(form)));
   return cons(replace(old,new,lcar(form)),replace(old,new,cdr(form)));
 };

int forms_equal(a,b) PCELL a; PCELL b;
{
  if (a == NULL) { if (b == NULL) return 1; return 0; };
  if (b == NULL) { return 0; };
  if (cars_equal(a,b)) return forms_equal(cdr(a),cdr(b));
  return 0; };

int cars_equal(a,b) PCELL a; PCELL b;
{ if (symbolp_car(a))
  { if (symbolp_car(b))
    { if (!strcasecmp(a->pcar.psymbols->id,b->pcar.psymbols->id))
      return 1;
    return 0; }
  else return 0;};
  if (stringp_car(a))
    { if (stringp_car(b))
      { if (!strcasecmp(a->pcar.atom,b->pcar.atom))
	return 1;
      return 0; }
    else return 0;};
  if (consp_car(a))
    { if (consp_car(b))
      { if (cars_equal(a,b)) 
	return forms_equal(cdr(a),cdr(b));
      return 0; }; };
  return 0;};

PCELL localize(form) PCELL form;
{ if (form == NULL)         return form;
  if (symbolp_car(form)) return cons_symbol(symcar(form),localize(cdr(form)));
  if (stringp_car(form)) return cons_terminal(strcar(form),localize(cdr(form)));
  /* car is list. */
  return cons(cons_terminal("LOCAL",list1(lcar(form))),localize(cdr(form)));
};


PCELL make_commaized(x) PCELL x;{

if (symbolp_car(x)) {
  return append(cons_terminal(our_strcat(",",x->pcar.psymbols->id),NULL), 
		       cdr(x)); };
if (stringp_car(x)) 
  if (removep(x)) return cons_terminal(strcar(x), make_commaized(cdr(x)));
  else { x->pcar.atom = our_strcat(",",x->pcar.atom); return x; };
return x; };

PCELL replace_args(args, form) PCELL args; PCELL form;
/* replace occurences of elements of args in form by 'arg. */
 { PCELL first;

   if (!args || !consp_car(args)) return form;

   first = lcar(args);

   return replace_args(cdr(args), replace(first,make_commaized(first),form));
   };

PCELL initialize_one(PSTRING lib)
{ return cons_terminal_package("INCLUDE-BOOK", cons_terminal(lib, NULL)); };

/* No longer need initialize_libs */
PCELL initialize_libs(libs)
{ return
    cons(initialize_one("\"/usr/cli/projects/cacl2/version20/solaris/infix-prelude\""),NULL); };



/* main()
 { PCELL x = cons_terminal(remove_tag, cons_symbol("A", NULL));
   if (x != NULL) printf("X ne z\n");
   if (consp_car(x)) printf("consp_car(X)\n");
   if (x->caflg == tstring) printf("X->caflg == tstring\n");
   printf("X->pcar.atom == %s\n", x->pcar.atom);
   if (string_eq(x->pcar.atom,remove_tag)) printf("match\n"); else printf("nomatch\n");
 }								     

main()
 { PCELL x = cons_terminal(remove_tag, cons_symbol("A", NULL));
   ptree(x);
   printf("\n");

   types_to_pred("INTEGER", list1(x));
 }								     

main()
 { PSYMBOLS w;
   PCELL x;
   PCELL y;
   w = symlook( strdup("foobar"), symtab, NSYMS);
   printf(w->id);
   x = cons_symbol(w, NULL);
   printf("\ntok x\n");
   ptree(x);
   x = cons_terminal("B", x);
   y = list1(cons_terminal("ID",cons_terminal ("A", cons_terminal("NIL",NULL))));
   printf("\nx\n");
   ptree(x);
   printf("\ny\n");
   ptree(y);
   printf("\ny @ y\n");
   ptree(append(x,y));
 }

main()
 { PSYMBOLS f;
   PSYMBOLS a;
   PSYMBOLS b;
   PCELL f2;
   PCELL x;
   PCELL y;
   PCELL m;
   f = symlook( strdup("foo"), symtab, NSYMS);
   a = symlook( strdup("a"), symtab, NSYMS);
   b = symlook( strdup("b"), symtab, NSYMS);
   f2 = cons_terminal("--remove--", cons_symbol(f, NULL));
   x = cons_terminal("--remove--", cons_symbol(a, NULL));
   y = cons_terminal("--remove--", cons_symbol(b, NULL));
   m = cons(f2, cons(x, cons(y, NULL)));
   printf("\nm\n");
   ptree(m);
   printf("\nreplace_args(x,y,m)\n");
   ptree(replace_args(cons(x,cons(y,NULL)),m));
   printf("\nAnd now, with strings\n");
   f2 = cons_terminal("--remove--", cons_terminal("f", NULL));
   x = cons_terminal("--remove--", cons_terminal("a", NULL));
   y = cons_terminal("--remove--", cons_terminal("b", NULL));
   m = cons(f2, cons(x, cons(y, NULL)));
   printf("\nm\n");
   ptree(m);
   printf("\nreplace_args(x,y,m)\n");
   ptree(replace_args(cons(x,cons(y,NULL)),m));
 }

main()
 { PSTRING s  = "TESTING";
   PSTRING s1 = "TESTING";
   PSTRING s2 = "TESTINGverylognfasdfasdfasdfasdfasdfsa"; 
   //printf("\n%s\n", our_strcat(s1,s1));
   //fflush(stdout);
   //printf("\n%s\n", our_strcat(s1,s2));
   //fflush(stdout);
   printf("\n0, 0 %s\n", substring(s,0,0));
   fflush(stdout);
   printf("\n0,-1 %s\n", substring(s,0,-1));
   fflush(stdout);
   printf("\n1,-1 %s\n", substring(s,1,-1));
   fflush(stdout);
   printf("\n4, 3 %s\n", substring(s,4,3));
   fflush(stdout);
   printf("\n3, 4 %s\n", substring(s,3,4));
   fflush(stdout);
   printf("\n9, 3 %s\n", substring(s,9,3));
   fflush(stdout);
   printf("\n9, 9 %s\n", substring(s,9,9));
   fflush(stdout);

 };

 */






