; ACL2 Version 1.8

; Copyright (C) 1989-95 Computational Logic, Inc. (CLI).  All rights reserved.

; Use of this software constitutes agreement with the terms of ACL2
; license agreement, found in the file LICENSE.

(in-package "ACL2")

(defparameter *infix-error-flg* nil)

(symbol-name 'acl2-output-channel::standard-error-output-0)
(setf (get 'acl2-output-channel::standard-error-output-0
           *open-output-channel-type-key*) :character)
(setf (get 'acl2-output-channel::standard-error-output-0
           *open-output-channel-key*) *standard-output*)

(defconst *standard-eo* 'acl2-output-channel::standard-error-output-0)

(f-put-global 'standard-eo 'acl2-output-channel::standard-error-output-0 state)

(defun infix-error-fms (hardp ctx str alist state)

; Almost the same as ACL2 error-fms.

  (setq *infix-error-flg* t)

  (let ((channel *standard-eo*))	;(f-get-global 'standard-co state)
    (mv-let (col state)
      (fmt1 (if hardp
                "~%~%HARD ACL2 ERROR"
              "~%~%ACL2 Error")
            nil 0 channel state nil)
      (mv-let (col state)
        (fmt-in-ctx ctx col channel state)
        (mv-let (col state)
          (fmt1 str alist col channel state (default-evisc-tuple state))
          (mv-let (col state)
                  (fmt1 "~%~%" nil col channel state nil)
                  (declare (ignore col))
                  state))))))

(defun string-output-fn (form)
  `(let ((saved-stream (get 'acl2-output-channel::standard-character-output-0
                            *open-output-channel-key*))
         *infix-error-flg*
         (saved-fn (symbol-function 'error-fms)))
     (unwind-protect
         (progn
           (setf (symbol-function 'error-fms)
                 (symbol-function 'infix-error-fms))
           (let ((ans
                  (with-output-to-string
                   (foo)
                   (setf (get 'acl2-output-channel::standard-character-output-0
                              *open-output-channel-key*)
                         foo)
                   ,form)))
             (cons *infix-error-flg* ans)))
       (setf (symbol-function 'error-fms)
             saved-fn)
       (setf (get 'acl2-output-channel::standard-character-output-0
                  *open-output-channel-key*)
             saved-stream))))

(defmacro string-output (form)
  (string-output-fn form))

(defun set-infix-markup-table (tbl state)
  (f-put-global 'infix-markup-table tbl state))

(defun infix-markup-table (state)
  (or (and (f-boundp-global 'infix-markup-table state)
           (f-get-global 'infix-markup-table state))
      (doc-markup-table state)))

(defun set-infix-char-subst-table (tbl state)
  (f-put-global 'infix-char-subst-table tbl state))

(defun infix-char-subst-table (state)
  (or (and (f-boundp-global 'infix-char-subst-table state)
           (f-get-global 'infix-char-subst-table state))
      (doc-char-subst-table state)))

;; ----------------------------------------------------------------------
;; Documentation from history-management.lisp

; What are the symbols in the fmt alist we need?  To find the answer, look
; in the markup table and collect all the fmt vars used.  They are:

; #\p -- the "pointer"     ; only used if flag is t
; #\s -- the print name version of the pointer, e.g., |abc| or ABC
; #\c -- the parent file   ; only used if flag is t
; #\t -- the displayed text
; #\T -- uppercased displayed text
; #\w -- the html anchor for the warning message ; only used on ~WARN[]

; If flag is nil, then we bind just the last three.  In this case, the
; displayed text is all of str.

; If flag is t, then we first ``read'' a symbol from str, effectively
; splitting str into two parts, sym and text.  The split is indicated
; below.  Note that sym is a string, not really a symbol.
; The "pointer" is the symbol-name of sym.  The "print name of the pointer"
; is the symbol-name of sym possibly surrounded by vertical bars.

;    str             #\p    #\s     #\t

; ~?[abc]           "ABC"  "ABC"    "abc"
; ~?[abc ]          "ABC"  "ABC"    ""
; ~?[abc def ghi]   "ABC"  "ABC"    "def ghi"
; ~?[|abc|]         "abc"  "|abc|"  "abc"
; ~?[|abc| ]        "abc"  "|abc|"  ""
; ~?[|abc| def ghi] "abc"  "|abc|"  "def ghi"
; ~?[|abc|def ghi]  "abc"  "|abc|"  "def ghi"

;; ----------------------------------------------------------------------

;; For examples of markup-tables and char-subst-tables see
;; scribe-init.lisp and latex-init.lisp

(defun infix-preprocess-doc (str &key
                              (markup-table (infix-markup-table *the-live-state*))
                              (char-subst-table (infix-char-subst-table *the-live-state*))
                              (prefix "")
                              fmt-alist
                              (name '|<some name>|)
                              par-p
                              &aux (state *the-live-state*))
  (string-output
   (cond
    ((not (fboundp 'acl2::error-fms))
     (print-string str))
    ((doc-stringp str)
     (pprogn
      (print-doc-string-part 0 str prefix markup-table char-subst-table fmt-alist
                             *standard-co* name par-p state)
      (print-doc-string-part 1 str prefix markup-table char-subst-table fmt-alist
                             *standard-co* name par-p state)
      (print-doc-string-part 2 str prefix markup-table char-subst-table fmt-alist
                             *standard-co* name par-p state)))
    ;; Otherwise, print the string, stopping at the first ~/ (if any, else to end of string).
    ;; Note that unlike the other case, no special effort is made to
    ;; strip off leading spaces using get-doc-string-de-indent [see below].
    (t (print-doc-string-part1 str
                               0
                               (length str)
                               0;;(get-doc-string-de-indent str)
                               prefix
                               markup-table
                               char-subst-table
                               fmt-alist
                               *standard-co*
                               name
                               state
                               (if par-p :par 0))))))

;;; Compute args for key.

(defun keyword-command-arg-number (key state)
  (declare (xargs :mode :program))
  (let ((temp (assoc-eq key (f-get-global 'ld-keyword-aliases state))))
    (cond (temp (cadr temp))
	  ((eq key :q) 0)
	  (t
	   (let ((sym (intern (symbol-name key) "ACL2"))
		 (wrld (w state)))
	     (cond
	      ((function-symbolp sym wrld)
	       (length (formals sym wrld)))
	      ((getprop sym 'macro-body nil 'current-acl2-world wrld)
	       (let ((args (getprop sym 'macro-args nil 'current-acl2-world wrld)))
		 (if (no-lambda-keywordsp args)
		     (length args)
		     nil)))
	      (t nil)))))))

(defvar *infix-package-alist* nil)

;; From J Moore. But so grossly modified as to be unrecognizable. 

(defun acl2-defpkg (name contents)
  ;; Name is a string
  (let ((atoms (if (eq contents nil)
		   nil
		   (or (pseudo-eval contents)
		       *common-lisp-symbols-from-main-lisp-package*))))

    (setq *infix-package-alist* (cons (cons name atoms) *infix-package-alist*))
    (defpkg name atoms)
    (import atoms name)
    (find-package name)))

(defun pseudo-eval (l)
  (cond ((symbolp l) (if (boundp l) (eval l) nil))
	((eq (car l) 'quote) (cadr l))
	((eq (car l) 'cons) 
	 (let ((a (pseudo-eval (cadr l)))
	       (b (pseudo-eval (caddr l))))
	   (if (and a b) (cons a b) nil)))
	((eq (car l) 'append) 
	 (let ((a (pseudo-eval (cadr l)))
	       (b (pseudo-eval (caddr l))))
	   (if (and a b) (append a b) nil)))
	((eq (car l) 'union-eq) 
	 (let ((a (pseudo-eval (cadr l)))
	       (b (pseudo-eval (caddr l))))
	   (if (and a b) (union-eq a b) nil)))
	((every #'symbolp l) l)
	(t nil)))


(proclaim '(ftype (function (t t) t)
		 read-keyword-form))

(defvar acl2-markup-table)
(defvar acl2-char-subst-table)

(defun acl2-parse-string (doc)
  ;; If *hands-off*, don't do special char quoting.
  (infix-preprocess-doc doc
			:markup-table acl2-markup-table
			:char-subst-table (if *hands-off* nil acl2-char-subst-table)))

(defun acl2-keywordp (key) (keywordp key))

(defun read-keyword-command (key)
  (read-keyword-form key
			   (keyword-command-arg-number key *the-live-state*)))

(defun acl2-current-package () (current-package *the-live-state*))
