; ACL2 Version 1.9

; Copyright (C) 1989-96 Computational Logic, Inc. (CLI).  All rights reserved.

; Use of this software constitutes agreement with the terms of ACL2
; license agreement, found in the file LICENSE.

;; Init file for infix.lisp in Text mode.
;; Feb 20 1992, by MKSmith
;; Apr 29 96 MKS 

(in-package "ACL2")

(require "infix")

(format *terminal-io* "Loading the infix text-init file.")

(defparameter *begin-comment-env*  "/* ")
(defparameter *end-comment-env*    " */")

;; (format *terminal-io* "~%About to set *comment-format*~%")

;; Mode should actually be set before this file is loaded.

(infix-settings :mode                   "ascii"
		:extension              "x"   
		:op-location            'front  
		:comment-format         'c
		:format-!-in-comments   nil    
		:hints                  t
		:no-keys                nil
		:no-top-locals          nil
		:no-guards              nil
		:no-verify-guards       nil
		:no-index               t
		:no-index-calls         t )


;; Increase this number to more accurately allow for proper output width.
(defparameter *rightmost-char-number* 80)
(defparameter *default-chars-wide* 80)
(defparameter *latex-indent-number-limit* 80)

(defparameter nqread-normal-clause-enders nqread-default-normal-clause-enders)

; THE TEXT PRELUDE.

(defparameter *standard-prelude*
  "
")

(defparameter *standard-postlude*
 "
")

(defparameter *example-prelude*
  (concatenate 'string *standard-prelude*
"

Here is a summary of the conventional syntax (~a) in terms of the
official syntax of the ACL2 logic.

1. Variables are printed in italics, unless specified otherwise in the
table below.

2. Function application.  For any function symbol for which special
syntax is not given below, an application of the symbol is printed
with the usual notation; e.g., the term !v(fn x y z) is printed as
!t(fn x y z).  Note that the function symbol is printed in Roman.  In
the special case that !qc is a function symbol of no arguments, i.e.,
it is a constant, the term !v(c) is printed merely as !t(c), in small
caps, with no trailing parentheses.  Because variables are printed in
italics, there is no confusion between the printing of variables and
constants.

3. Other constants.  Quoted constants are printed in the ordinary
syntax of the ACL2 logic, in a `typewriter font.'  For example, 
'(a b c) is still printed just that way.  \#b001 is printed as
!t#b001, \#o765 is printed as !t#o765, and \#xa9 is
printed as !t#xa9, representing binary, octal and hexadecimal,
respectively."))

(defparameter *begin-example-table* 
"~%--------------------------------------------------------------------~
 ~%ACL2 Syntax                    Conventional Syntax~
 ~%--------------------------------------------------------------------
")

(defparameter *end-example-table* 
"~%-------------------------------------------------------------------- 
")

(defparameter *example-table-size*  32)

(defparameter *example-postlude* "")

;; BASIC BRACKETS AND THEIR QUOTED VERSION.

(defparameter *dots* "...")
(defparameter *equal-sign* " = ")
(defparameter *lb-sign* " # ")

(defparameter *begin* "")
(defparameter *end*   "")

(defparameter *lbrace* "{")
(defparameter *rbrace* "}")

(defparameter *dots* "...")

(defparameter *begin* "{")
(defparameter *end*   "}")

(defparameter *lbracket* "[")
(defparameter *rbracket* "]")

(defparameter *backslash* "\\")

;; NEWLINE PARAMETERS

(defparameter *newline-in-env* "")
(defparameter *newline-in-text* "")

(defparameter *force-newline-in-env* "
")
(defparameter *force-newline-in-text* "
")

;; ENVIRONMENT BEGIN-END PAIRS

(defparameter *begin-index* "")
(defparameter *end-index* "")

(defparameter *begin-text-env* "/* ")
(defparameter *end-text-env*  " */")

(defparameter *begin-verbatim-env* "/* ")
(defparameter *end-verbatim-env*  " */")

(defparameter *begin-verb-env* "")
(defparameter *end-verb-env*  "")

(defparameter *begin-format-env*  "/* ")
(defparameter *end-format-env*    " */")

(defparameter *begin-emphasis-env*  "")
(defparameter *end-emphasis-env*    "")

(defparameter *begin-section-env*  "")
(defparameter *end-section-env*    "")

(defparameter *begin-subsection-env*  "")
(defparameter *end-subsection-env*    "")

(defparameter *begin-tt-env* "")
(defparameter *end-tt-env*   "")

(defparameter *tt-size* 1)

(defparameter *begin-string-env*  "")
(defparameter *end-string-env*    "")

(defparameter *begin-bold-env* "")
(defparameter *end-bold-env*   "")

(defparameter *begin-italic-env* "")
(defparameter *end-italic-env*   "")

(defparameter *begin-sc-env* "")
(defparameter *end-sc-env*   "")

;; These should ONLY be used by print-examples to help us document
;; the syntax in syntax.tex.

(defparameter *column-separator* "	")

(defparameter *begin-enumerate-env* "")
(defparameter *end-enumerate-env* "")

(defparameter *begin-item* "o ")
(defparameter *end-item* "")

(defparameter *mv-bracket-left* "<")
(defparameter *mv-bracket-right* ">")

(defparameter *forall* "all ")
(defparameter *exists* "some ")


;; TABBING ENVIRONMENT AND TAB OPERATIONS

(defparameter *begin-group-tabbing-env* "")

(defparameter *begin-tabbing-env* "")
(defparameter *end-tabbing-env* "")

(defparameter *new-tab-row* "
")

;; Just in case some other mode defined it otherwise.
(defun new-tab-row (&optional followed-by-infix-print-term)
  (declare (ignore followed-by-infix-print-term))
  (pprinc *new-tab-row*))

;; (defun new-tab-row (&optional followed-by-infix-print-term)
;;   (declare (ignore followed-by-infix-print-term))
;;   (pwrite-char #\Newline)
;;   (to-current-margin))

(defparameter *tab* "	")
(defparameter *flush* "        ")

(defparameter *tab-list* nil)

(defparameter *set-margin* "")
(defparameter *set-tab*    "")
(defparameter *pop-margin* "")

(defparameter *op-length* nil)
(defparameter *default-op-tab-space* "      ")
(defparameter *indent-string* "    ")
(defparameter *default-indent* 4)

(defun get-op-width-string (op)
  (length op))

(defparameter *noindent* "")

; (defparameter *testmargin* nil)

(defun begin-tabbing ()
  (setq *tab-list* (cons '(begin-tabs) *tab-list*))
  (princ *begin-tabbing-env*)
  (to-current-margin))

(defun begin-group-tabbing ()
  (setq *tab-list* nil)
  (princ *begin-group-tabbing-env*)
  (to-current-margin))

(defun end-tabbing ()
  (sloop while (and *tab-list* (not (equal (caar *tab-list*) 'begin-tabs)))
	 do (setq *tab-list* (cdr *tab-list*)))
  (if *tab-list* (setq *tab-list* (cdr *tab-list*)))
  (princ *end-tabbing-env*))

(defun begin-flushright ()
  (pprinc *flush*))

(defun end-flushright ()
  (pprinc *tab*))

(defun begin-normal-text ()
  (pprinc *begin-text-env*))

(defun end-normal-text ()
  (pprinc *end-text-env*)
  (to-current-margin))

(defun flushright (form)
  (begin-flushright)
  (pprinc form)
  (end-flushright))

(defun do-tab ()

; The *tab-list* is either NIL, ((LM loc) ...), or ((TAB loc) ...)
; Only tab if there is something to tab to.

  (cond (*do-not-use-tabs* (pprinc " "))
	((and *tab-list* (eq (caar *tab-list*) 'tab))
	 (spacesx (- (cdar *tab-list*) *infix-loc*)) ;(pprinc *tab*)
	 )
	(t (pprinc " "))))

(defun set-tab (&optional op)

; Generate instructions to set a tab at the current location.
; `Punt' if we hit the limit, by throwing all the way out.

  (cond (*do-not-use-tabs* nil)
        (t (cond ((= *tabs-in* *latex-indent-number-limit*) ;Let Latex-Limit hold for Text also.
                  (throw 'taboverflow t)))
           (setq *tabs-in* (1+ *tabs-in*))
	   (cond ((numberp op)
		  (spacesx (+ 1 op)))
		 ((and op (get-op-width-string op))
		  (spacesx (+ 1 (get-op-width-string op))))
		 (t (pprinc *default-op-tab-space*)))
           (push (cons 'tab *infix-loc*) *tab-list*)
	   ;; (if *testmargin* (format t "~%Setting tab : ~a~%" *tab-list*))
           (pprinc *set-tab*))))

(defun pop-tab ()
  ;; We don't really remove tabs from the formatted env.
  ;; Just track them in Lisp.
  ;; Generate command to `tab to one tab less in'.
  ;; Do not pop tabs beyond left margin or past the beginning of a tabbing env.
  (cond (*do-not-use-tabs* nil)
	((and *tab-list* (eq (caar *tab-list*) 'tab))
	 (setq *tabs-in* (1- *tabs-in*))
	 (pop *tab-list*)
	 ;; (if *testmargin* (format t "~%Popped tab : ~a~%" *tab-list*))
	 )
	(t nil)))

;; (defvar *tracing-margins* nil)

(defun set-margin ()
  ;; Generate instructions to set the current indentation.
  ;; (if *tracing-margins* (or *testing* (princ "|[")))
  (cond (*do-not-use-tabs* nil)
        (t (setq *tabs-in* (1+ *tabs-in*))
	   (setq *left-margin* *infix-loc*)
	   ;(format t "~%(set-margin) = ~d." *infix-loc*)
	   (push (cons 'lm *infix-loc*) *tab-list*))))

(defun pop-margin ()
  ;; Generate command to `return to one margin less in'.
  ;; For bookkeepping reasons we pop tabs after the margin.
  ;; (if *tracing-margins* (or *testing* (princ "]|")))
  (cond (*do-not-use-tabs* nil)
	((and *tab-list* (eq (caar *tab-list*) 'tab))
	 (pop-tab)
	 (pop-margin))
	((and *tab-list* (eq (caar *tab-list*) 'lm))
	 (setq *tabs-in* (1- *tabs-in*))
	 (pop *tab-list*)
	 (setq *left-margin* (or (get-margin2 *tab-list*) 0))
	 ;(format t "~%(pop-margin) = ~d." *left-margin*)
	 )
	(t nil)))

(defun clear-margins ()
  (setq *tab-list* nil)
  (setq *tabs-in* 0)
  (setq *infix-loc* 0)
  (set-margin))

;; FONTS

(defparameter *function-font* "")	; Roman.  sc = small caps

(defparameter *begin-event-name* "")
(defparameter *end-event-name* "")

(defun print-function-name (sym) (roman-font sym))

(defun roman-font (sym) (print-atom sym))


;; MATH ENV AND OPERATORS

(defparameter *neg-str*  "~")

;; MUST be a format directive.  Used by make-inifix-op etc.

(defparameter *math-format* "~a")
(defparameter *math-begin* "")
(defparameter *math-end* "")

;; These must be enclosed in a math env.  
(defparameter *math-thick-space* "  ")
(defparameter *math-space* " ")
(defparameter *math-thin-space* "")

(defparameter *subscript* "")

(defparameter *begin-subscript* "")
(defparameter *end-subscript* "")

;; MISC.

(defparameter *newpage* "")

(defparameter *comma-atsign* ",@")
(defparameter *caret* "^")

(defparameter *list-seperator* ", ")
(defparameter *dotted-pair-seperator* " . ")

(defparameter *no-tab-event-trailer* "~%")
(defparameter *print-default-event-header* "")
(defparameter *print-default-lisp-header* "")

(defparameter *print-default-command-header*  "")
(defparameter *no-tab-command-trailer* "~%")

;; OTHER FUNCTIONS

;; Should both be `(#\@ @\}) but we handle them individually in the appropriate 
;; spots in the two following functions.
(defparameter doc-special-chars nil)
(defparameter doc-index-specials nil)

(defparameter doc-other-chars   nil)

;; We didn't compile the following because the compiler declaration
;; in Sinfix, through a bug in AKCL, caused this routine to produce
;; spurious results.

;; The patch to akcl that is loaded in sinfix should fix this problem.
;; Other lisps shouldn't need it.
;; These use to be of the form (eval-when (load) (eval '<defn>))

(defun handle-special-chars (char)
  ;; USED BY PRINT-ATOM.  CHAR is local to print-atom.
  ;; We don't use the global, DOC-SPECIAL-CHARS, since there are 
  ;; only two, @ and }, and we do something different in each case. 
  (pwrite-char (if (eq *print-case* :downcase)
		   (char-downcase char)
		   char)))

(defun handle-special-chars-in-string (char)
  ;; USED BY PRINT-ATOM.  CHAR is local to print-atom.
  ;; (cond ((char= char #\@) (pprinc "@@")	 (incf *infix-loc* 1))
  ;;       ((char= char #\}) (pprinc *rbrace*) (incf *infix-loc* 1))
  ;;       (t 
  (cond ((eql char #\")  
	 (pwrite-char #\\)
	 (pwrite-char char))
	(t (pwrite-char char)))

  ;; ))
  )

;; PRINTING INDEX ENTRIES

(defun index (x &optional subkind) 
 (if *do-not-index*
     nil
     (if subkind
	 (pformat t " ;; INDEX ~s, ~s
" x subkind)
	 (pformat t " ;; INDEX ~s
" x))))

(defun skip-index-entries ()
  ;; We are looking at a backslash. In Tex mode we need to skip to the end 
  ;; of the entry, because we may add !'s.  In Text mode this is just NIL.
  nil)

(defun adjust-tabbing-env ()
  ;; We are looking at a backslash. In Tex mode we may need to check for 
  ;; some special cases.
  ;; In Text mode this is just NIL.
  nil)

(defparameter acl2-char-subst-table nil)

(defparameter acl-markup-table
  '(("-"   NIL .   "-")
    ("B"   NIL .   "~sp")
    ("BF"  NIL .   "~%
") 
    ("BID" NIL .   "")			;begin implementation dependent
    ("BQ"  NIL .   "\"")
    ("BV"  NIL .   "~%
")
    ("C"   NIL .   "~sp")		;originally @code, but we don't want `' in info file
    ("EF"  NIL .   "
")
    ("EID" NIL .   "")			;end implementation dependent
    ("EM"  NIL .   "~st")		;emphasis
    ("EQ"  NIL .   "\"
") 
    ("EV"  NIL .   "
")
    ("GIF" nil . "")			;gif files; e.g., ~gif[\"foo.gif\" align=top]
    ("I"   NIL .   "~st")
    ("ID"  NIL .   "")			;implementation dependent
    ("IL"  NIL .   "~st")
    ("ILC" NIL .  "~st")		;originally @code, but we don't want `' in info file
    ("L"   NIL .  "~st")   
    ("NL"  NIL .  "
")
    ("PAR" NIL .  "
")					;paragraph mark, of no significance for text
    ("PL"  NIL .  "see ~st")		;used for parenthetical crossrefs
    ("SC"  NIL .  "~sT")		;small caps
    ("ST"  NIL .  "~sT")		;strong emphasis
    ("T"   NIL .  "~st")
    ("TERMINAL" NIL . "")		; terminal only, ignore
    ("WARN" nil . "~sw")
    ("CLICK-HERE" t .  "")
    ("PCLICK-HERE" t .  "")
    ("FLY" t .  "")
    ("WALK" t .  "")
    ))

;; Primitive stuff.

(defun get-margin ()
  (or (get-margin2 *tab-list*) *left-margin*))

(defun get-margin2 (tl)
  (let ((setting (car tl)))
    (cond ((null setting) nil)
	  ((eq (car setting) 'lm) (cdr setting))
	  (t (get-margin2 (cdr tl))))))

(defun spacesx (n)
  (dotimes (i n) (pwrite-char #\ ))
  (setq *infix-loc* (+ *infix-loc* n)))

(defun line-return ()
  (setq *prev-line-loc* *infix-loc*)
  (pwrite-char #\Newline)
  (setq *infix-loc* 0)
  (spacesx (get-margin))
  ;(format t "~%After (Line-return) *infix-loc* = ~d and (get-margin) = ~d." *infix-loc* (get-margin))
  )

(defun newline-in-env (&optional force) 
  (if force (line-return)
      (if (not (eql *infix-loc* (get-margin)))
	  (line-return))))

(defun newline-in-text (&optional force) 
  (if force (line-return)
      (if (not (eql *infix-loc* (get-margin)))
	  (line-return))))

(defun newlinex (&optional force)
  (newline-in-env force))

(defun to-current-margin ()
  (let ((margin (get-margin)))
    (cond ((eql *infix-loc* margin) nil)
	  ((< *infix-loc* margin) 
	   (spacesx (- margin *infix-loc*)))
	  (t (line-return)))))
