(in-package "ACL2")

;(require "sloop")
(require "infix")

(defparameter *defining-events* 
  '( defun def defthm defaxiom const defconst defpkg
     defmacro defstub signature deftheory
     defun-sk defloop
     deflist defalist destructure))

; Defined in infix:
; (defvar *begin-event-name* "")
; (defvar *end-event-name* "")
; (defun begin-event-name () (princ *begin-event-name*))
; (defun end-event-name () (princ *end-event-name*))
; (defun event-name-printer (x) ... )

;; EVENT PRINTERS

(defun defcor-printer (event)
  (let* ((name (nth 1 event))
	 (old-event (nth 2 event))
	 (keys (extract-keywords event '(:rule-classes :doc)))
	 (rule-classes (car keys))
	 (doc (cadr keys)))
    (begin-event doc)
    (event-label-printer "Corollary: ")
    (event-name-printer name)
    (pprinci " is a corollary to ")
    (print-atom old-event)
    (pprinci " ")
    (index name)
    (to-current-margin)
    (cond ((and (consp rule-classes)
		(consp (car rule-classes))
		(equal (car (car rule-classes)) :rewrite)
		(consp (cdr (car rule-classes)))
		(equal (car (cdr (car rule-classes))) :corollary))
	   (infix-print-term (cadr (cdr (car rule-classes)))))
	  (t (pprinci "Rule Classes: ")
	     (infix-print-term rule-classes)))
    (end-event)))

(local-print defcor)
(declare-event-printer defcor (function defcor-printer))

(defun defequiv-printer (event)
  (let* ((fn (nth 1 event))
	 (doc (extract-keywords event '(:doc)))
	 (keys (extract-keywords2
		event
		'(:rule-classes :instructions :hints :otf-flg :event-name))))
    (begin-event (car doc))
    (event-label-printer "Equivalence ")
    (event-name-printer fn)
    (index fn "equivalence")
    (key-list-printer keys)
    (end-event)))

(local-print defequiv)
(declare-event-printer defequiv (function defequiv-printer))

(defun defdoc-printer (term)
 (let ((term
         (eviscerate-check term
                           '(defdoc :symbol :string))))
   (if (and (cddr term) (evisceratep (caddr term)))
       (ppformat "~%Document ~s ~s~%" (cadr term) (cdr (caddr term)))
       (progn (ppformat "~%Document ~s ~%" (cadr term))
	      (if (cddr term) (pprinci (caddr term)))))))

(declare-event-printer defdoc (function defdoc-printer))

(defun print-rule-classes (rules)
  (if *print-keys*
      (progn (to-current-margin)
	     (pprinci " ")
	     (set-margin)
	     (pprinci ":rule-classes ")
	     (rule-class-printer rules)
	     (pop-margin))))

(defvar *eviscerate* nil)



(defun evisceratep (term)
  (and *eviscerate* (consp term) (equal (car term) :evisceration-mark)))


(defun eviscerate-check (term pattern)
  ;; If evisceration not applicable or term matches pattern 
  ;; (which indicates the expected number of  elements in term
  ;; and any requirements on them) return nil.
  ;; If not, create something printable at a term.
  ;; Pattern examples
  ;;  : (defun :symbol :list (* :term))
  ;;  : (:symbol :symbol :list (:optional :string) (:optional (* :term)) :term)
  (cond ((or (not *eviscerate*) (type-match term pattern)) term)
	((and (consp term) (equal (car term) :evisceration-mark)) term)
	(t (evisc-transl term pattern))))

(defun true-listp (x) (if (not (consp x)) (null x) (true-listp (cdr x))))

(defun type-match (term pattern)
  ;; Returns two values (t or NIL) and rest.
  (cond ((not (consp pattern))
	 (values (cond ((null pattern) (null term))
		       ((eq pattern :term) t)
		       ((eq pattern :symbol) (symbolp term))
		       ((eq pattern :string) (stringp term))
		       ((eq pattern :number) (numberp term))
		       ((eq pattern :consp) (consp term))
		       ((eq pattern :list)  (true-listp term))
		       (t (equal term pattern))) nil))
	((not (consp term)) nil)
	((not (consp (car pattern)))	
	 (and (type-match (car term) (car pattern))
	      (type-match (cdr term) (cdr pattern))))
	((eq (caar pattern) :*)
	 (cond ((null term) (values t nil))
	       ((not (consp term)) (values nil nil))
	       (t (let ((rest (match-first-n term (cadar pattern))))
		    (type-match rest (cdr pattern))))))
	((eq (caar pattern) :optional)
	 (if (type-match (car term) (cadar pattern))
	     (type-match (cdr term) (cdr pattern))
	     (type-match term (cdr pattern))))
	((type-match (car term) (car pattern))
	 (type-match (cdr term) (cdr pattern)))
	(t nil)))

(defun match-first-n (term pattern)
  (if (not (consp term))
      term
      (sloop for x on term when (or (not (consp x))
				    (not (type-match (car x) pattern)))
		 return x)))

; Mike:  I think the evisc-transl function is wrong.  Suppose the pattern is
; :string but the term is (:evisceration-mark . ???)  Then we ought to return a
; string, as "expected."  I return "...".  This whole function seems random to
; me.

(defun evisc-transl (term pattern)
  ;; Term failed to match pattern.  So we need to add (:evisceration-mark . "") 
  ;; as needed.
  (cond ((not (consp pattern))
	 (if (member pattern '(:consp :list))
	     (if (consp term)
		 (if (eq (car term) :evisceration-mark)
		     (list term)
		     term)
                 '((:evisceration-mark . "...")))
	     term))
	((and (consp term)
	      (eq (car term) :evisceration-mark))
	 (evisc-generate pattern))
	((not (consp (car pattern)))
	 (cond ((consp term)
		(cons (evisc-transl (car term) (car pattern))
		      (evisc-transl (cdr term) (cdr pattern))))
	       ((cdr pattern)
		(cons '(:evisceration-mark . "")
		      (evisc-transl '((:evisceration-mark . "")) (cdr pattern))))
	       ((null term) '((:evisceration-mark . "")))
	       (t (list term))))
	((eq (caar pattern) :*)
	 (cond ((null term) nil)
	       ((not (consp term)) (evisc-generate (cdr pattern)))
	       (t (multiple-value-bind
		      (front rest)
		      (evisc-first-n term (cadar pattern))
		    (append front (evisc-transl rest (cdr pattern)))))))
	((eq (caar pattern) :optional)
	 (cond ((not (consp term)) (evisc-transl term (cdr pattern)))
	       ((type-match (car term) (cadar pattern))
		(cons term (evisc-transl (cdr term) (cdr pattern))))
	       (t (cons '(:evisceration-mark . "")
			(evisc-transl (cdr term) (cdr pattern))))))
	((not (consp term))
	 (evisc-generate pattern))
	(t (cons (evisc-transl (car term) (car pattern))
		 (evisc-transl (cdr term) (cdr pattern))))))



(defun evisc-first-n (term pattern)
  (cond ((null term) (values nil nil))
	((not (consp term)) (values nil nil))
	((not (type-match (car term) pattern))
	 (values nil term))
	(t (multiple-value-bind (front rest)
	       (evisc-first-n (cdr term) pattern)
	     (values (cons (evisc-transl (car term) pattern) front) rest)))))

(defun evisc-generate (pattern)
  (cond ((not (consp pattern))
	 (if (not (keywordp pattern))
	     pattern
	   (if (member pattern '(:consp :list))
	       '((:evisceration-mark . "..."))
	     '(:evisceration-mark . "#"))))
	((not (consp (car pattern)))
	 (cons (evisc-generate (car pattern))
	       (evisc-generate (cdr pattern))))
	((eq (caar pattern) :*)
	 (evisc-generate (cdr pattern)))
	((eq (caar pattern) :optional)
	 (evisc-generate (cdr pattern)))
	(t (cons (evisc-generate (car pattern))
		 (evisc-generate (cdr pattern))))))


#|  HINT SYNTAX

hints  == ( hint* )
hint   == ( Node key-value )
theory == <defined below>

operations      == op  | '(op*)
functions       == fn  | (fn*)
lemma-instances == lmi | (lmi*)

subst == (v term)
fn    == symbol | (lambda (v1...) b)
op    == preprocess | simplify | eliminate-destructors | fertilize |
         generalize | eliminate-irrelevance

expand-term == (let ((v1 t1) ...) b) | (fn t1 ... tn) | (:free (v1 ... vn) pattern)

lmi  == name | rune | (:theorem term) | (:instance lmi . subst*) |
        (:functional-instance lmi . subst*)  { of form (f1 g1) }

bdd-pair  == :vars (v1 ...) | :prove [ T | nil ] |
             :bdd-constructors (symbol*) |
             :literal [ :ALL | :conc | n ]

restriction == ((:rewrite rune) . subst*) | (eventname . subst*)

key-value ==  :do-not-induct [ t | name | nil ] 
            | :do-not operations
            | :expand (expand-term*)
            | :restrict (restriction*)
            | :hands-off functions
            | :in-theory theory
            | :induct [ t | term ]
            | :use lemma-instances
            | :bdd (bdd-pair*)
            | :cases (term+)
            | :by [ lemma-instance | event | nil ]

|
|#

(defvar *print-hints* nil)
(defvar *print-no-guards* nil)

(defun hints-printer (hints &optional(name ':hints))
  (if (and hints *print-hints*)
      (progn (to-current-margin)
	     (if (equal name ':hints)
		 (pprinci " :hints ")
	       (pprinci " :guard-hints "))
	     (pprinci *lbracket*)
	     (set-margin)
	     (if (consp hints)
		 (sloop for x in hints do 
			(progn (to-current-margin)
			       (hint-printer x)))
		 (italic-sym-printer hints))
	     (pprinci *rbracket*)
	     (pop-margin))))

(defun hint-printer (hint)
  (pprinci *lbracket*)
  (print-atom (car hint))
  (pprinci " ")
  (set-margin)
  (sloop for x on (cdr hint)
	 when (keywordp (car x))
	 do (progn (if (not (equal x (car (last hint))))
		       (to-current-margin))
		   (key-printer (car x) (cadr x))))
 
  (pprinci *rbracket*)
  (pop-margin))


(defun infix-print-subst (subst)
  (let ((newline (advise-break (cons 'list subst))))
    (pprinci *lbracket*)
    (set-margin)
    (sloop for pair in subst do (progn (if newline (to-current-margin))
				       (pprinci *lbracket*)
				       (infix-print-term1 (car pair))
				       (pprinci " ")
				       (parens-print-term1 (cadr pair))
				       (pprinci *rbracket*)))
    (pop-margin)
   (pprinci *rbracket*)))

(defun expand-printer (value)
  (if (and (consp value) (eq (car value) :free))
     (progn (pprinci *lbracket*)
	    (pprinci ":free ")
	    (infix-print-subst (cadr value))(pprinci " ")
	    (infix-print-term1 (caddr value))
	    (pprinci " ] "))
     (progn (infix-print-term1 value) (pprinci " "))))

;;textual subst for :Restrict :instance :functoinal-instance
;; jimmy
;; :restrict ((cancel-<-*$free ((:rewrite nil)))
;;               (member-equal ((x 3)))))))
;; ==> :restrict [ cancel-<-*$free [:rewrite := nil; ], member-equal [x := 3; ] ]]]

(defun textual-subst (subst)
  (let ((newline (advise-break (cons 'list subst))))  
    (pprinci *lbracket*)
    (set-margin)
    (sloop for pair in (car subst) do 
	   (progn (if newline (to-current-margin))
		  (pair-printer (car pair) (cadr pair))
		  (if (not (equal pair (car (last (car subst)))))
		      (pprinci ", "))))
    (pop-margin)
    (pprinci " ")
    (pprinci *rbracket*)))

(defun restrict-printer (value)
  (pprinci *lbracket*)
  (pprinci " ")
  (sloop for x in value do
	 (progn (if (and (consp (car x)) (eq (caar x) :rewrite))
		     (progn (pprinci *lbracket*)
			    (pprinci ":rewrite ")
			    (print-rune (cadr (car x))))
		   (progn (print-atom (car x))
			  (pprinci " ")
			  (textual-subst (cdr x))))               ;;;(infix-print-subst (cdr value))
		 (if (equal x (car (last value)))
			     (pprinci " ")
		      (pprinci ", "))))
  
  (pprinci *rbracket*))
  
(defun bdd-printer(term)
  (pprinci *lbracket*)

  (bdd1-printer term)
 
  (pprinci *rbracket*))

(defun bdd1-printer (term)
  (if (consp term)
      (let ((var (car term))
	    (val (nth 1 term)))
	(progn
	  (set-margin)
          (cond ((or (equal var :vars)
		     (equal var :bdd-constructors))
		 (if (consp val)
		     (pair-printer var (cons 'list val))
		   (pair-printer var val)))
		(t (pair-printer var val)))
	  (pop-margin) 
	  (if (cddr term)
	      (progn (pprinci ", ")
		     (to-current-margin)
		     (bdd1-printer (cddr term))))))))
		     

;; lemma-instances == lmi | (lmi*)
;; lmi  == name | rune | (:theorem term) | (:instance lmi . subst*) |
;;         (:functional-instance lmi . subst*)  { of form (f1 g1) }

(defun lmi-print (x &optional sep newline)
  (if newline (to-current-margin))
  (cond ((symbolp x) (print-atom x))
	((eq (car x) ':theorem) 
	 (pprinci *lbracket*)
	 (pprinci ":theorem ") (set-margin)
	 (infix-print-term1 (cadr x))
	 (pop-margin) (pprinci *rbracket*))
	
	((eq (car x) ':instance) 
	 (pprinci *lbracket*)
	 (pprinci ":instance ") (set-margin)
	 (lmi-print (cadr x) sep (advise-break (cadr x)))
	 (if (cddr x) (progn (to-current-margin) (textual-subst 
						  (list (cddr x)))))
	 (pop-margin)	 (pprinci *rbracket*))
	((eq (car x) ':functional-instance) 
	 (pprinci *lbracket*)
	 (pprinci ":functional-instance ") (set-margin)
	 (lmi-print (cadr x) sep (advise-break (cadr x)))
	 (if (cddr x) (progn (to-current-margin) (textual-subst (list (cddr x)))))
	 (pop-margin) (pprinci *rbracket*))

	 ((keywordp (car x)) (set-margin)
	  (pprinci *lbracket*)
	  (print-rune-string x)
	  (pop-margin)	 (pprinci *rbracket*)))

  ;;(if newline (to-current-margin))
  )

(defun lmi-printer (l)
  ;; l is lemma-instance
  (let ((newline (advise-break l)))
    (cond ((or (symbolp l) (keywordp (car l)))
	   (lmi-print l nil newline))
	  (t (pprinci *lbracket*)
	     (set-margin)
	     (print-list l " " nil t (function lmi-print))
	     (pop-margin)
	     (pprinci *rbracket*)))))

(defun key-list-printer (l)
  ;; L is list of dotted pairs
  (cond ((not (consp l)) nil)
	((consp (car l))
	 (to-current-margin)
	 (key-printer (car (car l)) (if (null (cdr (car l))) nil (cdr (car l))))
	 (key-list-printer (cdr l)))
	(t (key-list-printer (cdr l)))))

(defvar *print-keys* t)
(defvar *function-symbol-out*)

(defun key-printer (key value)
  (if *print-keys*
      (let ((*function-symbol-out* t))
	(case key
	  (:rule-classes (print-rule-classes value))
	  (:hints (hints-printer value))
	  (:guard-hints (hints-printer value ':guard-hints))
	  (:guard (or *print-no-guards*
		      (progn (pprinci (format nil "~s " key))
			     (infix-print-term1 value))))
	  (:do-not (pprinci ":do-not ") (quote-printer1 value))
	  (:expand (pprinci ":expand ") (true-list-printer value))
	  (:restrict (pprinci ":restrict ") (restrict-printer value))
	  (:hands-off (pprinci ":hands-off ") (hands-off-printer value))
	  (:use (pprinci ":use ") (lmi-printer value))
	  (:cases (pprinci ":cases ") (true-list-printer value))
	  (:trigger-terms (pprinci ":trigger-terms ") (list-printer (cons 'list value)))
	  (:clique (pprinci ":clique ") (list-printer (cons 'list  value)))
	  (:trigger-fns  (pprinci ":trigger-fns ") (list-printer (cons 'list value)))
	  (:bdd (pprinci ":bdd ") (bdd-printer value))
	  (:by (pprinci ":by ") (lmi-printer value))
	  (otherwise (pprinci (format nil "~s " key)) (infix-print-term1 value))
	  ))
      (let ((*function-symbol-out* t))
	(case key
	  (:hints (hints-printer value))
	  (:guard (or *print-no-guards*
		      (progn (pprinci (format nil "~s " key))
			     (infix-print-term1 value))))
	  (otherwise nil)))))

(defun hands-off-printer (value)
  (if (consp value)
     (list-printer (cons 'list value))
   (infix-print-term1 value)))

(defun defaxiom-printer (term)
  ;; name term :rule-clases clases :doc string
  (let* ((name (nth 1 term))
         (subterm (nth 2 term))
         (keys (extract-keywords term '(:rule-classes :doc)))
	 (rule-classes-p (member :rule-classes term))
         (rule-classes (car keys))
         (doc (cadr keys)))
    (begin-event doc)
    (event-label-printer "Axiom ")
    (event-name-printer name)
    (index name "defined")
    (cond ((let ((*rightmost-char-number* (- *rightmost-char-number* 12)))
            (advise-break subterm))
           ;; No.
           (to-current-margin)		;newline
	   (pprinci " ")  
	   (body-printer subterm))  
	  (t (pprinci " ") (body-printer subterm)))
    (if rule-classes-p (print-rule-classes rule-classes))
    (end-event)))

(local-print defaxiom)
(declare-event-printer defaxiom (function defaxiom-printer))

;;;7/18 jimmy
;;;need pretty print?
;;; doc-section ???

(defun defchoose-printer(term)
  (let* ( (term
          (eviscerate-check term
                            '(defchoose :symbol
                               ((:* :symbol))
			       (:optional :string)
                               (:* :term))))
	  (fn (nth 1 term))
	  (bound-var (nth 2 term))
	  (free-var (nth 3 term))
	  (doc (if (eq (length term) 6) (nth 4 term))) 
	  (body (transform-body (car (last term)))))
    (begin-event doc)
    (event-label-printer "Choose ")
    (event-name-printer fn)
    (index fn)
    
    (print-bound-free-var bound-var free-var)
   
    (cond ((let ((*rightmost-char-number* (- *rightmost-char-number* 12)))
           (advise-break body))
           ;; No.
           (to-current-margin)		;newline
	   (pprinci " ")  
           (infix-print-term1 body))
	   (t (pprinci " ")
	    (infix-print-term1 body)))
    (end-event)))

(local-print defchoose)
(declare-event-printer defchoose (function defchoose-printer))


(defun print-bound-free-var (bound-var free-var)
 (cond ((and (consp bound-var) (consp free-var))
	(pprinci "(")
	(if (atom bound-var) (setf bound-var (list bound-var)))
	(if (atom free-var) (setf free-var (list free-var)))
	(infix-print-args bound-var)
	(pprinci ") ")
	(pprinci "(")
      (infix-print-args free-var)
      (pprinci ") "))))

(defun defcong-printer (term)
  ;; equiv1 k equiv2 fn :rule-classes :instructions :hints :otf-flg :cheat :doc
  (let* ((equiv1 (nth 1 term)) 
         (equiv2 (nth 2 term))
         (Fn (nth 3 term))
         (arity (nth 4 term))
         (doc (extract-keywords term '(:doc)))
         (keys (extract-keywords2
		term
		'(:rule-classes :instructions :hints :otf-flg :event-name :doc))))
    (begin-event (car doc))
    (event-label-printer "Congruence ")
    (event-name-printer equiv1)
    (small-caps-sym-printer " To ")
    (print-atom equiv2)
    (to-current-margin)
    (small-caps-sym-printer " for ")
    (print-atom arity)
    (small-caps-sym-printer " of ")
    (infix-print-term1 fn)
    (index equiv1 "equivalence")
    (index equiv2 "equivalence")
    (if (consp fn) (index (car fn)))
    (key-list-printer keys)
    (end-event)))

(local-print defcong)
(declare-event-printer defcong (function defcong-printer))

(defun const-printer(Event)
    (let* ((name (nth 1 event))
	  (term (nth 2 event))
	  (keys (extract-keywords event '(:type :lemma :doc)))
	  (doc (nth 2 keys))
	  (type (nth 0 keys)))
      (begin-event doc)
      (event-label-printer "Constant ")
      (event-name-printer name)
      (index name "defined")
      (if type (ppformat " : ~a" type))
      (pprinci " = ")
      (cond ((let ((*rightmost-char-number* (- *rightmost-char-number* 12)))
	       (advise-break term))
	     ;; No.
	       (newlinex)
	       (pprinci " ")
	       (set-margin)
	       (to-current-margin)		;newline
	       (infix-print-term1 term)
	       (pop-margin))
	    (t (pprinci " ") (infix-print-term1 term))))
      (end-event))

(local-print const)
(declare-event-printer const (function const-printer))

(defun defconst-printer (event)
  (const-printer event))

;;(let ((name (nth 1 event))
;;        (term (nth 2 event))
;;        (doc  (nth 3 event)))
;;    (begin-event doc)
;;    (event-label-printer "CONSTANT ")
;;    (infix-print-term1 (list 'equal name term))
;;    (end-event)             
;;    ))

(local-print defconst)
(declare-event-printer defconst (function defconst-printer))

(defun defevaluator-printer (term)
  ;; ev ev-list signatures
  (let ((ev (nth 1 term))
	(ev-list (nth 2 term))
	(signatures (cadddr term)))
    (begin-event)
    (event-label-printer  "Evaluator( ")
    (ppformat "~a, ~a )" ev ev-list)
    (to-current-margin)
    (pprinci " ")
    (pprinci *lbracket*)
    (pprinci " ")
    (print-list signatures ", " ". " nil)
    (pprinci " ")
    (pprinci *rbracket*)
    (index ev)
    (index ev-list)
    (end-event)))

(local-print defevaluator)
(declare-event-printer defevaluator (function defevaluator-printer))

(defun deflabel-printer (event)
  (let* ((name (nth 1 event))
         (keys (extract-keywords event '(:doc)))
         (doc (nth 0 keys)))
    (begin-event doc)
    (event-label-printer "Label ")
    (event-name-printer name)
    (index name)
    (end-event)))

(local-print deflabel)
(declare-event-printer deflabel (function deflabel-printer))

(defun defrefinement-printer (term)
  ;; equiv1 equiv2 :rule-classes :instructions :hints :otf-flg :doc
  (let* ((eq1 (nth 1 term))
         (eq2 (nth 2 term))
         (keys (extract-keywords2 term '(:rule-classes :instructions :hints :otf-flg :event-name :doc)))
         (doc (cdr (assoc :doc keys))))
    (begin-event doc)
    (event-label-printer "Refinement ")
    (event-name-printer eq1)
    (small-caps-sym-printer " refines ")
    (print-atom eq2)
    (index eq1)
    (index eq2)
    (key-list-printer keys)
    (end-event)))

(local-print defrefinement)
(declare-event-printer defrefinement (function defrefinement-printer))

(defun defpkg-printer (event)
  (let ((name (nth 1 event))
	(contents (nth 2 event))
        (doc  (if (> (length event) 3) (nth 3 event))))
    ;; Three cases:
    ;; 1. We are in Lisp, without ACL2 loaded at all.
    ;;    In which case we have stubbed out acl2-defpkg to 
    ;;    call make-package.  This case is handled when
    ;;    we load the system (this file doesn't get loaded
    ;;    at all).
    ;; 2. We are in ACL2.  Nothing should be done here, because acl2
    ;;    will eventually evaluate the form.
    ;; 3. We are in ACL2, but out of the LP command loop, running
    ;;    in lisp. This is the problem.  How do we detect it?
    ;;
    ;; We do the following so that we can print and
    ;; read symbols in this package.

    ;; (format t "~%ld-level = ~d" acl2::*ld-level*)
    (if (and (= acl2::*ld-level* 0)
	     (not (find-package name)))

	;; If we are in acl2, the following fakes evaling CONTENTS, 
	;; and does the appropriate acl2 things.
	;; If we are in lisp without acl2, we ignore CONTENTS.  
	;; Will this bite us?  There is nothing else that we can do.
	(if (stringp name)
	    (acl2-defpkg name contents)
	  nil))
    (begin-event doc)
    (event-label-printer "Package ")
    (event-name-printer name)    ;;;(bold-sym-printer name)
    (index name "defined")
    (math-space 1)
    (pprinc "=")
    (math-space 1)
    (infix-print-term1 contents)
    (end-event)))

(local-print defpkg)
(declare-event-printer defpkg (function defpkg-printer))

(defun in-package-printer (event)
  (let ((name (nth 1 event)))
    ;;
    ;; We do the following so that we can print and
    ;; read symbols in this package.
    ;;
    (if (stringp name)
	(if (not (find-package name))

	;; If we are in acl2, the following fakes evaling CONTENTS, 
	;; (thus the NIL is fine below) and does the appropriate acl2 things.
	;; If we are in lisp without acl2, we ignore CONTENTS.  
	;; Will this bite us?  There is nothing else that we can do.
	    (acl2-defpkg name nil)
	  nil))
    
    (begin-event)
    (event-label-printer "In-Package(")
    ;; Was bold-sym-printer, but that doesn't print the string quotes.
    ;;  Sep 29 96 MKS  So I changed bold-sym-printer.  Which broke 
    ;;  numberous other things, so I changed it back to print-atom
    (infix-print-term1 name)
    (pprinci ")")
    (end-event)))

(local-print in-package)
(declare-event-printer in-package (function in-package-printer))

(defun rebuild-printer (event)
  (let ((name (nth 1 event))
	(key  (if (cddr event) (caddr event))))
    (begin-event)
    (event-label-printer "Rebuild ")
    (italic-sym-printer name)
    (cond ((or (null key)
	       (equal key t)	       
	       (equal key :all)
	       (equal key :query))
	   (pprinci "."))
	  ((symbolp key) (ppformat " through  ~s." (symbol-name key)))
	  ((and (consp key)
		(eq (car key) 'quote)
		(symbolp (cadr key)))
	   (ppformat " through  ~s." (symbol-name (cadr key))))
	  (t (ppformat ".")))
    (end-event)))

(local-print rebuild)
(declare-event-printer rebuild (function rebuild-printer))

(defun filename-sym-printer (x)
  (italic-sym-printer (acl2-smash-file-name x)))

;; A file (to acl2) may be a string or
;; (:ABSOLUTE . string* )
;; (:RELATIVE . string* )

(defun acl2-file-name-root (file)
  (cond ((not (consp file)) file)
	(t (car (last file)))))

(defun acl2-smash-file-name (file)
  (cond ((not (consp file))
	 (cond ((stringp file) file)
	       ((symbolp file) (symbol-name file))
	       (t "")))
	((equal (car file) :absolute)
	 (concatenate 'string "/" (acl2-smash-file-name (cdr file))))
	((equal (car file) :relative)
	 (acl2-smash-file-name (cdr file)))
	((equal (car file) :back)
	 (concatenate 'string "../" (acl2-smash-file-name (cdr file))))
	((cdr file) (concatenate 'string (car file) "/" (acl2-smash-file-name (cdr file))))
	(t (car file))))

(defun include-book-printer (event)
  (let* ((file (nth 1 event))
	 (keys (extract-keywords event '(:doc)))
	 (doc (nth 0 keys)))
    (begin-event doc)
    (event-label-printer "Include(")
    (if (evisceratep file)
	(pprinci (cdr file))
      (progn (print-atom file)
	     (index (acl2-file-name-root file) "included")))
    (pprinci ")")
    (end-event)))

;;;(local-print include-book)		;???

(declare-event-printer include-book (function include-book-printer))

;;(defun certify-book-printer (event)
;;  (let* ((file (nth 1 event))
;;	 (k (nth 2 event))
;;	 (flag (nth 3 event)))
;;    (begin-event)
;;    (event-label-printer "Certify-book(")
;;    	(progn (print-atom (cdr file))
;;	       (index (acl2-file-name-root file) "certified")))
;;    (pprinci ")")
;;    (end-event)))

;;;(local-print certify-book)		;???



(defun certify-book-printer (event)
  (let* ((file (nth 1 event))
	 (k (nth 2 event))
	 (flag (nth 3 event)))
    (begin-event)
    (event-label-printer "Certify-book(")
    (progn (print-atom file)
	   (index (acl2-file-name-root file) "certified"))
    (if k (progn (pprinci ", ") 
		 (infix-print-term1 k)))
    (if (> (length event) 3)
	(progn (pprinci ", ")
	       (pprinci flag)))
    
    (pprinci ")")
    (end-event)))

(declare-event-printer certify-book (function certify-book-printer))

(defun infix-print-macro-body (body)
  (cond ((not (consp body)) (infix-print-term1 body))
	((equal (car body) 'quote)
	 (pprinci "'") (infix-print-term1 (cadr body)))
	((equal (car body) '*infix-backquote*)
	 (pprinci "`") (infix-print-term1 (cadr body)))
	(t (infix-print-term1 body))))

(defvar *type-translations*
  '((integerp . integer)
    (stringp . string)
    (characterp . char)
    (booleanp . boolean)
    (bitp . bit)
    (complexp . complex)
    (consp . cons)
    (null . nil)
    (truep . t)
    (rationp . ratio)
    (rationalp . rational)
    (unsigned-bytep . unsigned-byte)
    (standard-char-p . standard-char)))

(defun print-type (symbol)
  (let ((x (assoc symbol *type-translations*)))
    (if x (print-atom (cdr x)) (print-atom symbol))))

(defun macro-member(X l &optional)
  (if (consp l)
      (cond ((equal x (Car l)) t)
	    ((and (consp (car l))
		  (member (caar l) '(ASSIGNMENT)))
	     (or (equal x (nth 1 (car l)))
		 (macro-member (cdr l))))
	    ((member (car l) '(&optional
			       &rest
			      &key
			      &whole
			      &body
			      &allow-other-keys))
	     (if (consp (nth 1 l))
		 (or (member x (nth 1 l)) (macro-member x (cddr l)))
	       (macro-member x (cdr l))))
	    (t (macro-member x (cdr l))))
    nil))


;; Check for substitution macro, which we can parse.
;; Otherwise, macros get printed pretty ugly.

(defun substitution-macro-p (body args &optional flag)
  (cond (flag 
	 (if (not (consp body)) t
	   (and (substitution-macro-p (car body) args nil )
		(substitution-macro-p (cdr body) args t ))))
	((quote-p body) t)
	((null body) t)
	((equal body t) t)
	((keywordp body) t)
	((not (consp body))
	 (if (symbolp body) (macro-member body args) t))
	((or (and (consp (car body)) (equal (caar body) :evisceration-mark))
	     (and (consp (car args)) (equal (caar args) :evisceration-mark)))
	 nil)
	((member (car body) '(CONS LIST))
	 (substitution-macro-p (cdr body) args t))
	(t nil)))

;; (defun substitution-macro-p (body args)
;;   (cond ((and (null args) (quote-p body)) t)
;; 	((null args) nil)
;; 	((not (consp body))
;; 	 (if (symbolp body) (member body args) t))
;; 	((or (and (consp (car body)) (equal (caar body) :evisceration-mark))
;; 	     (and (consp (car args)) (equal (caar args) :evisceration-mark)))
;; 	 nil)
;; 	((member (car body) '(CONS LIST))
;; 	 (substitution2-macro-p (cdr body) args))
;; 	(t nil)))
	
(defun substitution2-macro-p (body args)
  (if (not (consp body))
      t
      (let ((x (car body)))
	(cond ((and (atom x) (member x args)) ;;no quote
	       (substitution2-macro-p (cdr body) args))
	      ((and (atom x) (or (numberp x) (constantp x)))
	       (substitution2-macro-p (cdr body) args))
	      ((and (consp x) (or (equal (car x) 'CONS)
				  (equal (car x) 'LIST)))
	       (substitution-macro-p (car body) args)) ;;(list ...)
	      ((or (single-quote-p x) (double-quote-p x)) ;;single + double quote
	       (and (not (member (cadr x) args))
		    (substitution2-macro-p (cdr body) args)))
	      (t nil)))))
	       
(defun single-quote-p (x)
  (and (consp x) (equal (Car x) 'quote) 
       (if (consp (cadr x)) (not (equal (caadr x) 'quote)) t)))


(defun double-quote-p(x)
  (and (consp x) (equal (car x) 'quote)
       (consp (cdr x))
       (consp (cadr x))
       (equal (caadr x) 'quote)
       (if (consp (cadadr x))
	   ;; But not triple.
	   (not (equal (car (cadadr x)) 'quote)) t)))

(defun quote-p (x)
  (and (consp x) (equal (car x) 'quote)))

(defun reverse-macro (l vars)
  (cond ((not (consp l))
	 (cond ((member l vars) l)
	       ((characterp l) l)
	       ((integerp l) l)
	       ((stringp l) l)
	       ((keywordp l) l)
	       ((null l) l)
	       ((equal l t) l)
	       (t (list 'quote l))))
	((equal (car l) 'quote) (cadr l))

	((equal (car l) *infix-backquote*) (reverse-macro-back (cadr l) vars))
	(t (let* ((result  (mapcar (function (lambda (x) 
					       (reverse-macro x vars)))
				   (Cdr l))))
	     ;; This was commented out in one version and result
	     ;; returned instead.  ??  Apr 18 97 MKS 
	     (apply (car l) result)
	     ))))

(defun reverse-macro-back (l vars)
  (cond ((not (consp l))
	 (cond ((member l vars) l)
	       ((characterp l) l)
	       ((integerp l) l)
	       ((stringp l) l)
	       ((keywordp l) l)
	       ((null l) l)
	       (t (list 'quote l))))
	((equal (car l) *infix-comma*) (reverse-macro (cadr l) vars))
	((equal (car l) *infix-comma-atsign*)
	 (list 'macro-atsign (reverse-macro (cadr l) vars)))
	((equal (car l) *infix-backquote*) (reverse-macro-back (cadr l) vars))
	((equal (car l) 'quote) (cadr l))
	       
	(t (cons (car l)
		 (mapcar (function (lambda (x) (reverse-macro-back x vars))) (cdr l))))))

(defun transform-arg-keyword (arg)
  (if (consp arg)
      (if (Consp (car arg))
	  (if (equal (car arg) '(:evisceration-mark . "..."))
	      arg
	    (cons (cons 'assignment (car arg))
		  (transform-arg-keyword (cdr arg))))
	(cons (car arg) (transform-arg-keyword (cdr arg))))
    arg))

(defvar evisceration-marker '(:evisceration-mark . "..."))

(defvar *equal-sign* " = ")
(defvar *lb-sign* " # ")

(defun defmacro-printer (event)
  (let* ((event
          (eviscerate-check event
			    '(defmacro :symbol :list (:optional :string) (:* :term))))
	 (name (nth 1 event))
	 ;; Note that macro args may include &optional or &rest.
         (old-args (nth 2 event))
         (args (transform-arg-keyword
		old-args));;(a &optional (x 3)) =>  (a &optional (assignment x 3))
	 (doc  (nth 3 event))
         (orig (car (last event)))
	 substitution-macro-p
	 body form dcls)
    
    ;; Fixup for eviserated args.  I think the right thing to
    ;; do is to ensure that when we match :list against (:evisceration-mark . "..")
    ;; we get ((:evisceration-mark . "..")).
    ;; ???? Need better fix.
    
    (if (evisceratep args) (setq args (list args)))
    
    (setq body (cond ((< (length event) 4)
		      '(:evisceration-mark . "..."))
		     ((substitution-macro-p orig args)
		      (setq substitution-macro-p t)
		      (reverse-macro orig args))
		     (t orig)))
    

    (setq form (cons name args))

    (sloop for x in (cdddr event);; extract declare
	   do (if (or (stringp x)
		      (equal x orig))
		  nil
		(setq dcls (append dcls (list x)))))

    (let ((types (extract-declare-types dcls)))
      (begin-event doc)
      (event-label-printer "Macro ")
      (set-margin)
      (signature-header-printer form types)
      (if (not (evisceratep body))
	  (if substitution-macro-p
	      (pprinci *equal-sign*)
	    (pprinci *lb-sign*)))
      nil)
    (index name "defined")
    ;; Can we fit the defun on one line?
    (let ((advice (advise-break (list 'equal form body))))
;;       (cond ((let ((*rightmost-char-number* (- *rightmost-char-number* 12))))
;; 	     (newlinex)
;; 	     (pprinci " ")
;; 	     (set-margin))	; No, won't fit.
;; 	    (t (pprinci " ")))
      (cond (advice (to-current-margin))
	    (t (pprinci " ")))

      ;; If body is an eventp (having been turned into one by REVERSE-MACRO
      ;; then print using event-printer. The T arg to infix-event
      ;; suppresses the event's trailing semi-colon
      (if (eventp body)
	  (infix-event body t)
	(infix-print-term1 body))
      ;; Need to delete type declarations before we perform the following test.
      
    (if dcls (progn (to-current-margin) (pprinci " ")   
		    (set-margin)
		    (print-decls dcls)
		    (pop-margin)))
    (pop-margin)
    (end-event))))

(local-print defmacro)
(declare-event-printer defmacro (function defmacro-printer))

(defun looking(body)
  (nth 2 body))



(defun defstub-signature-printer (event)
   (let ((name (nth 1 event))
	 (args (nth 2 event))
	 (result (nth 3 event))
	 (axiom (aget :axiom event)))
     
     (begin-event)
     (event-label-printer "Signature ")
     (infix-print-term1 (cons name args))
     (pprinci " : ")
     (if axiom 
	 (infix-print-term1 (car axiom))
       (infix-print-term1 result))
     (index name "signature")
     (end-event)))


(local-print defstub-signature)
(declare-event-printer defstub-signature (function defstub-signature-printer))

(defun defstub-printer (event)
  (let ((name (nth 1 event))
        (args (nth 2 event))
        (result (nth 3 event)))
    (begin-event)
    (event-label-printer "Signature ")
    (infix-print-term1 (cons name args))
    (pprinci " : ")
    (infix-print-term1 result)
    (index name "signature")
    (end-event)))

(local-print defstub)
(declare-event-printer defstub (function defstub-printer))

(defun signature-printer (event &optional types)
  (if (eq (car event) 'evisceration-mark)
      (infix-print-term1 event)
      (let ((name (nth 0 event))
	    (args (nth 1 event))
	    (result (nth 2 event)))
	(to-current-margin)
	(cond ((evisceratep args)
	       (infix-print-term1 (cons name '("..."))))
	      (t (infix-print-term1 (cons name (add-types args types)))))
	(pprinci " : ")
	(infix-print-term1 result)
	(pprinci ";"))))


(defun add-types (args types)
  (if (evisceratep args)
      (list args)
    (cond ((not (consp args)) args)
	  ((assoc (car args) types) (cons (cons 'type (assoc (car args) types))
					  (add-types (cdr args) types)))
	  ((consp (car args)) (cons (add-types (car args) types)
				    (add-types (cdr args) types)))
	  (t (cons (car args) (add-types (cdr args) types))))))

;; (foo a b) => foo(a : type, b :type): result

(defun signature-header-printer (event &optional types result)
  (let ((name (nth 0 event))
        (args (cdr event)))
    (begin-event-name)
    (infix-print-term1 (cons name (add-types args types)))
    (end-event-name)
    (if result 
	(progn (pprinci " : ")
	       (infix-print-term1 result)))))

(defun rule-class-printer (classes)
  (let ((advice (advise-break classes)))
  ;; classes is either keyword or a list of keywords or
  ;; lists whose car is a keyword.
    (cond ((consp classes)
	   (if advice
	       (progn (set-margin ) (to-current-margin)))
	   (pprinci *lbracket*)
	   (rule-class-print (car classes))
	   (if advice (to-current-margin))
	   (sloop for x in (Cdr classes)
		  do
		  (progn  (print-string " ") 
			  (rule-class-print x)
			  (if (and advice (not (Equal x (Car (last classes)))))
			      (to-current-margin))))
	   ;;(mapc (function (lambda (x) (print-string " ") 
	   ;;(rule-class-print x))) (cdr classes))

	  
	   (pprinci *rbracket*) (if advice (pop-margin)))
	  (t (rule-class-print classes)))))
  
(defun rule-class-print (class)
  (let(( *function-symbol-out* t))
    (cond ((keywordp class) (italic-sym-printer class))
	  ((not (consp class)) (infix-print-term1 class))
	  ((keywordp (car class))
	   (pprinci *lbracket*)
	   (set-margin)
	   (italic-sym-printer (car class))
	   (pprinci "  ")
	   (let* ((keys (extract-keywords class
					  '(:hints :instructions :otf-flg )))
		  (key-words (extract-keywords2 class '(:COROLLARY :TRIGGER-FNS 
								   :TRIGGER-TERMS
								   :TYPE-SET  :TYPED-TERM 
								   :CLIQUE :CONTROLLER-ALIST  
								   :LOOP-STOPPER :PATTERN 
								   :CONDITION :SCHEME ))))
	     
	     (key-list-printer key-words)   
	     
	     ;; Print :hints :instruction :otf
	     (if (nth 0 keys ) (hints-printer (nth 0 keys )))
	     (if (nth 1 keys ) (simple-key-printer ":instruction"
						   (cons 'list (nth 1 keys ))))
	     (if (nth 2 keys ) (simple-key-printer ":otf-flag"
						   (nth 2 keys ))))
	   (pop-margin) (pprinci *rbracket*) )
	  
	  (t (format t "~%Unhandled rule class, ~a~%" class) nil))))


(defun simple-key-printer (key value)
  (if *print-keys*
      (progn
	(to-current-margin)
	(pprinci " ")
	(set-margin)
	(italic-sym-printer key)
	(pprinci " ")
	(infix-print-term1 value)
	(pop-margin))))

(defun transform-body (term)
  (if (and (consp term) (member (car term) '(MV-LET LET LET*)))
      term
    (list 'body term)))


(defun body-printer (term)
  (let ((body (transform-body term))) 
      (infix-print-term1 body)  ))

(defun defthm-printer (term)
  ;; name term :rule-classes :instructions :hints :otf-flg :doc
  (let* ((name (nth 1 term))
         (subterm (nth 2 term))
         (keys (extract-keywords term
				 '(:rule-classes :instructions :hints :otf-flg :doc)))
         (rule-classes-p (member :rule-classes term))
         (rule-classes (nth 0 keys))
         (instr (nth 1 keys))
         (hints (nth 2 keys))
	 (otf (nth 3 keys))
	 (doc (nth 4 keys)))
    (begin-event doc)
    (event-label-printer "Theorem ")
    (event-name-printer name)
    (index name "theorem")
    ;; Can we fit the defthm on one line?
    (cond ((let ((*rightmost-char-number* (- *rightmost-char-number* 12)))
            (advise-break subterm))
           ;; No.
           (to-current-margin)		;newline
	   (pprinci " ")  
	   (body-printer subterm))
          (t (pprinci " ") (body-printer subterm)))
    (if rule-classes-p (print-rule-classes rule-classes))
    (if instr (simple-key-printer ":instructions" (cons 'list instr)))
    (if hints (hints-printer hints))
    (if otf (simple-key-printer ":otf-flag" otf))
    (end-event)))

(local-print defthm)
(declare-event-printer defthm (function defthm-printer))


(defun thm-printer (term)
  ;; thm term ::hints :otf-flg :doc
  (let* (
         (subterm (nth 1 term))

         (keys (extract-keywords term
				 '(:hints :otf-flg :doc)))
        
         (hints (nth 0 keys))
	 (otf (nth 1 keys))
	 (doc (nth 2 keys)))
    (begin-event doc)
    (event-label-printer "Thm ")
    ;; Can we fit the defthm on one line?
    (cond ((let ((*rightmost-char-number* (- *rightmost-char-number* 12)))
            (advise-break subterm))
           ;; No.
           (to-current-margin)		;newline
	   (pprinci " ")  
           (body-printer subterm))
	  (t (pprinci " ")   (body-printer subterm)))
    (if hints (hints-printer hints))
    (if otf (simple-key-printer ":otf-flag" otf))
    (end-event)))

(declare-event-printer thm (function thm-printer))


(defparameter *noindent* nil)
(defvar *indent-string* " ")
(defvar *default-indent* 2)

(defun encapsulate-signatures-printer (signatures)
  (sloop for signature in signatures do
	 (let* ((name (nth 0 signature))
		(arg (nth 1 signature))
		(form (cons name arg))
		(result-type (nth 2 signature)))
	   (to-current-margin)
	   (signature-header-printer form nil result-type)
	   (pprinci ";")))) 

(defvar *lbrace* "{ ")
(defvar *rbrace* " }")

(defvar *local-handlers* nil
  "List of events that handle printing of 'local' themselves")
(defvar *local-context* nil
  "List of events that handle printing of 'local' themselves")
(defvar *local-in-module* nil)

(defun encapsulate-printer (term &optional name)
  (let ((signatures (nth 1 term))
        (events (cddr term)))
    (begin-event)
    ;; Event-label-printer may print a 'local' prefix.
    (event-label-printer "Module ")
    (let ((*local-context* nil)
	  (*print-top-locals* t)
	  ;; I see a potential recursion problem, but
	  ;; only J will be putting modules in modules. 
	  ;;  Apr 24 97 MKS  ????
	  (*local-in-module* t))
      (if name (progn (event-name-printer name) (index name "module")))
      (if signatures 
	  (cond ((evisceratep signatures)
		 (pprinci *lbrace*)
		 (pprinci "...") 
		 (pprinci *rbrace*))
		(t 
		 (progn (to-current-margin)
			(pprinci *lbrace*)
			(set-margin)
			(encapsulate-signatures-printer signatures)
		      (pop-margin)
		      (pprinci *rbrace*))))
	  (progn (pprinci *lbrace*) (pprinci *rbrace*)))
    
      (if (let ((*rightmost-char-number* (- *rightmost-char-number* 12)))
	    (advise-break events))
	  (to-current-margin))
      (pprinci *lbrace*)
      (set-margin)
      (cond ((evisceratep events)
	     (pprinci "..."))
	    ((and (null (cdr events)) (evisceratep (car events)))
	     (pprinci (cdar events)))
	    (t (mapc #'infix-event (eviscerate-check events '(:* :term)))))
      (pop-margin)
      (pprinci *rbrace*)
      (end-event))))

(local-print encapsulate-printer)
(declare-event-printer encapsulate (function encapsulate-printer))

(defun skip-proofs-printer (term)
  (let ((event (nth 1 term)))
    (ppformat "~%skip-proofs ~%")
    (pprinci *lbrace*)
    (ppformat "~%")
    (newlinex)
    (funcall (get-event-printer (car event)) event)
    (pprinci *rbrace*)
    (ppformat ";~%")))

(defun local-printer (term)
  (if (not *print-top-locals*)
      nil
      (let ((event (cadr term))
	    ;; if *local-context* is set to t,then whenever the
	    ;; event-label-printer is called, "local" is printed
	    ;; in front of the event name
	    (*local-context* t))
	(if (and (consp event) (member (car event) *local-handlers*))
	    (infix-event event)
	    (progn (let ((*local-context* nil))
		     ;; local doesn't have to be used in encapsulate
		     (infix-event event)))))))

(declare-event-printer local (function local-printer))

(defun mutual-recursion-printer (term)
  ;; term was set to
  ;; (eviscerate-check term
  ;;  '(mutual-recursion :symbol :term
  ;;                     ((:* :symbol)) (:optional :string)
  ;;		         (:* :term)))
  ;; But mutural recursion just takes a list of defuns.
  ;; The eviscerated term was not ever used. Sep 29 96 MKS 
  (let* ((events (cdr (eviscerate-check term '(mutual-recursion (:*))))))
    (begin-event)
    (event-label-printer "Mutual-Recursion ")(pprinci *lbrace*)
    ;; if (local (mutual-recrsion (defun f(..)(..)) (defun g(..) (...))))
    ;; *local-context* := nil inside the mutual-recursion
    (let* ((*local-context* nil))
      (newlinex)
      (pprinci " ")
      (set-margin)
      (if (evisceratep (car events))
	  (pprinci "...")
	(progn
	  (mapc (function (lambda (x) (infix-event x))) events)))
      (pop-margin)
      (pprinci *rbrace*)
      (end-event))))

(local-print mutual-recursion)
(declare-event-printer mutual-recursion (function mutual-recursion-printer))

(defun defuns-printer (term)
  (if (or (evisceratep (nth 1 term) ) (evisceratep (nth 2 term)))
      (mutual-recursion-printer (cons 'mutual-recursion (cdr term)))
    (mutual-recursion-printer
     (cons 'mutual-recursion
	   (mapcar (function (lambda (x) (cons 'defun x))) (cdr term))))))

(declare-event-printer defuns (function defuns-printer))

(defun ld-printer (term)
  (let ((keys (cddr term)))
    (begin-event)
    ;; Deleted " the file: "
    (event-label-printer "LD(")
    ;; Replaced filename-sym-printer
    (infix-print-term1 (cadr term))
    (to-current-margin)
    (set-margin)
    (if keys 
	(sloop for x on keys
	       do (progn
		    (if (and (keywordp (car x)) (cdr x) (not (keywordp (cadr x))))
			(progn (key-printer (car x) (cadr x))
			       (setq x (cdr x)))
			(infix-print-term1 (car x)))
		    (if (cdr keys) (to-current-margin)))))
    (pop-margin)
    (pprinci ")")
    (end-event)))

(local-print ld-printer)
(declare-event-printer ld (function ld-printer))

(defun infix-print-setting (a)
  ;; X is pair of form keyword . setting.
  (let ((f (get-keyword-printer (car a))))
    (funcall f a)))

(defun print-settings-list (pairs)
  (sloop for tail on pairs
	 do (cond ((not (consp tail)))
		  ((cddr tail)
		   (infix-print-setting (car tail))
		   (pprinci ", "))
		  ((cdr tail)
		   (infix-print-setting (car tail)))
		  (t
		   (pprinci " and ")
		   (infix-print-setting (car tail)) (pprinci ".")))))

(defun find-result-type-form (x form)
  (cond ((or (not (consp x)) (not (cdr x))) x)
	((and (equal (length x) 2)
	      (symbolp (car x))
	      (equal (cadr x) form)) (car x))
	(t (or (find-result-type-form (car x) form)
	       (find-result-type-form (cdr x) form)))))

(defvar *success*)

(defun remove-result-type (keylist form)
  ;; Replaces first occurence of predicate of the form (foo form) by T.
  (cond ((not (consp keylist)) keylist)
	((equal (car keylist) ':guard)
	 (let (*success* new)
	   (setq new (remove-result-type-form (cadr keylist) form))
	 (append (cons (car keylist) new)
		 (if *success*
		     (cddr keylist)
		     (remove-result-type (cddr keylist) form)))))
	(t (cons (car keylist) (remove-result-type (cddr keylist) form)))))

(defun remove-result-type-form (x form)
  (cond (*success* x)
	((or (not (consp x)) (not (cdr x))) x)
	((and (equal (length x) 2)
	      (symbolp (car x))
	      (equal (cadr x) form)) (setq *success* t) t)
	(t (cons (remove-result-type-form (car x) form)
		 (remove-result-type-form (cdr x) form)))))

(defun extract-declare-xarg (dcls)
  ;; dcls = ((DECLARE arg*)*)
  ;; arg  = xarg || other
  ;; xarg = (XARGS key value ...)
  ;; key  = :guard || :measure || :mode || :verify-guards
  (sloop for declare in dcls
	 appending (if (equal (car declare) 'DECLARE)
		       (extract-xarg (cdr declare))
		       nil)))

(defun extract-xarg (declare)
  (sloop for entry in declare
	 appending (if (equal (car entry) 'XARGS) (list (cdr entry)) nil)))

(defun extract-declare-types (dcls)
  ;; dcls = ((DECLARE arg*))
  ;; arg  = xarg || (type (satisfies int) x) || (type cons x)
  ;; xarg = (XARGS key value ...)
  ;; key  = :guard || :measure || :mode
  (sloop for declare in dcls
	 appending (if (equal (car declare) 'DECLARE)
		       (extract-type (cdr declare))
		       nil)))




(defun extract-type (declare)
  (sloop for entry in declare  
	 appending (if (member (car entry) '(TYPE IGNORE))
		       (let (type)
			 (if (and (consp (cadr entry))
				  (equal (car (cadr entry)) 'SATISFIES))
			     (setq type (cadr (cadr entry)))
			   (if (equal (car entry) 'IGNORE)
			       (setq type (car entry))
			     (setq type (cadr entry))))
			 (if (equal (car entry) 'IGNORE)
			     (sloop for var in (cdr entry)
				  collect (list var type))
			    (sloop for var in (cddr entry)
				  collect (list var type))))
		       nil)))

(defun delete-declare (dcls)
  ;; dcls = (... (DECLARE ...) ...)
  (cond ((null dcls) nil)
	((and (consp (car dcls)) (equal (caar dcls) 'DECLARE))
	 (delete-declare (cdr dcls)))
	(t (cons (car dcls) (delete-declare (cdr dcls))))))

(defun var-type-list (x)
  ;; x = ((satisfies int) x ..) || (cons x ..)
  ;; returns ((var type)*)
  (cond ((not (consp x)) nil)
	((and (consp (car x)) (equal (caar x) 'SATISFIES))
	 (append (pairup (cadr (car x)) (cdr (car x)))
		 (var-type-list (cdr x))))
	((consp (car x))
	 (append (pairup (car (car x)) (cdr (car x)))
		 (var-type-list (cdr x))))
	(t (var-type-list (cdr x)))))

(defun pairup (x l) (if (endp l) nil (cons (list (car l) x) (pairup x (cdr l)))))

;;; ((declare ....)*)
(defun print-decls (dcls)
  (print-decls2 (extract-declare-xarg dcls))) 

(defun print-decls2 (dcls)
  (mapc
   (function
    (lambda (dcl)
      (key-list-printer
	    (extract-keywords2
	     dcl
	     '(:measure :guard :mode :verify-guards
			:well-founded-relation :hints :guard-hints :otf-flag)))))
   dcls))


(defun print-keywords (key)
  (if (null key) nil
      (progn
	(to-current-margin)
	(pprinci " " 1)
	(set-margin)
	(cond ((eq (car key) :hints)
	       (hints-printer (cdr key)))
	      ((eq (Car key) :guard-hints)
	       (hints-printer (cdr key)))
	      (t (simple-key-printer (car key) (cdr key))))
	(pop-margin))))	       

(defun defun-printer1 (term equiv &optional function-type)
  ;; (defun name args doc-string dcl ... dcl body)
  (let* ((name (nth 1 term))
         (args (nth 2 term))
         (doc (if (stringp (nth 3 term)) (nth 3 term)))
         (form (cons name args))
         (body (transform-body (car (last term))))
	 (body1-without-word-body  (cadr body))
	 (eq (list equiv form body))
         (xargs nil)
	 dcls)
    (sloop for x in (cdddr term)        ;; extract declare
	   do (if (or (stringp x)
		      (equal x body1-without-word-body))
		  nil
		(progn (setq dcls (append dcls (list x)))
		       (if (xargs? x)
			   (setq xargs t)))))
    (defun-printer2 name doc form body dcls eq xargs function-type )))

(defun xargs? (decl)
  (if (not (consp decl))
      nil
    (if (not (consp (car decl)))
	(or (Eq (car decl) 'xargs) (xargs? (cdr decl)))
      (or (eq (caar decl) 'xargs) (xargs? (cdr decl))))))
	     
(defun defun-printer2 (name doc form body dcls eq xargs &optional (function-type nil))
  (declare (ignore eq))
  (let ((types (extract-declare-types dcls)))
    (begin-event doc)
    (event-label-printer "Function ") 
    ;; Can we fit the defun on one line?
    (cond ((let ((*rightmost-char-number* (- *rightmost-char-number* 12)))
	     (advise-break body))
	  ;; No.
	   (signature-header-printer form types)
	   (index name "defined")
	   (if function-type (progn (pprinci ": ")
				    (infix-print-term1 function-type)))
	   (to-current-margin)		;newline
	   (pprinci " ")  
	   (infix-print-term1 body))
	  (t (signature-header-printer form types) (pprinci " ")
	     (index name "defined")
	     (if function-type (progn (pprinci " : ")
				      (infix-print-term1 function-type)
				      (pprinci " ")))
	     (infix-print-term1 body)))
    ;; Need to delete type declarations before we perform the following test.
    (if (and xargs dcls)
	(progn (to-current-margin) (pprinci " ")  
	       (set-margin)
	       (print-decls dcls)
	       (pop-margin)))
    ;; End-event is handled here, because defun-printer2 is
    ;; used by other functions (e.g. defloop-printer) too.
    (end-event)
    ))

(defun defun-printer (term &optional function-type)
  ;; Must be at least of the form (defun xxx)
  (let ((new (eviscerate-check term '(defun :symbol :list 
				       (:optional :string) (:* :term))))) 
    (defun-printer1 new 'equal function-type)
    ;;(end-event)
    ))

(local-print defun)
(declare-event-printer defun (function defun-printer))

;(defun remove-keys (l keys)
;  (cond ((not (consp l)) l)
;	((member (car l) keys)
;	 (if (consp (cdr l))
;	     (remove-keys (cddr l) keys)
;	     ;; A singleton key?  Leave.
;	     l))
;	(t (cons (car l) (remove-keys (cdr l) keys)))))

(defun make-defun (name args doc declares body)
  (list
   (append `(defun ,name ,args)
	   (if doc (list doc) nil)
	   (if declares declares nil)
	   (list body))))

(defun make-type-lemma (name args type lemma)
  (if (or type lemma)
      `((defthm ,(intern-in-package-of-symbol
		  (concatenate 'string (symbol-name name) "-TYPE-DEF")
		  'acl2::defthm)
	  ,(if lemma lemma (list type (cons name args)))))
      nil))

(defun make-disable (disable name)
  (if disable 
      `((in-theory (disable ,name)))
      nil))

(defun aget (a l)
  (cond ((not (consp l)) nil)
	((equal (car l) a) (cadr l))
	(t (aget a (cdr l)))))

(defun extract-declares (l)
  (cond ((not (consp l)) l)
	((and (consp (car l)) (equal (caar l) 'declare))
	 (cons (car l) (extract-declares (cdr l))))
	(t nil)))

(defun remove-declares (l)
  (cond ((not (consp l)) l)
	((and (consp (car l)) (equal (caar l) 'declare))
	 (remove-declares (cdr l)))
	(t l)))



(defun def-printer (term)
  (let* ((name (nth 1 term))
	 (args (nth 2 term))
	 (rest (cdddr term)))
   (let* ((doc     (if (stringp (car rest)) (car rest) nil))
	 (rest    (if doc (cdr rest) rest))
	 (declares (extract-declares rest)))
    (let* ((rest    (remove-declares rest))
	   (body    (car rest)))
      (let* ((rest    (cdr rest))
	     (type    (aget :type rest)))
	(defun-printer (car (make-defun name args doc declares body)) type))))))
      
(local-print def)
(declare-event-printer def (function def-printer))


(defun module-printer (term)
  (let* ((signature (nth 1 term))
	 (name (aget :label term))
         (event (if name (butlast (cddr term) 2)
		  (cddr term))))
    (encapsulate-printer (cons 'encapsulate (cons signature event)) name)))

(local-print module)
(declare-event-printer module (function module-printer))

;(local-print define)			
;(declare-event-printer define (function define-printer))


(defun eviscerated-print-dot-dot()
  (pprinci  " ... "))
 
(defun eviscerated-event-printer (name type &optional (mark " ... "))
  (pprinci type) 
  (roman-font name)
  (pprinci mark))

(defun remove-make (f)
  (cond ((stringp f) nil)
	((symbolp f) (setq f (symbol-name f)))
	(t (error "Bad name passed to REMOVE-MAKE")))
  (if (and (> (length f) 5) (string= "MAKE-" (subseq f 0 5)))
      (subseq f 5)
      f))

;;;Skip printing the '{' '}' 

(defun progn-printer (term)
  (sloop for x in (cdr term) do (progn  (infix-event x) 
					(to-current-margin))))

(declare-event-printer progn (function progn-printer))

(defun defstructure-printer (term)
  ;; (defstructure name [doc] fields* options)
  ;; options == (:options :guards (:predicate person )(:conc-name NIL)
  ;;              (:keyword-constructor key-person )
  ;;              (:keyword-updater update-person ))
  (let* ((name (nth 1 term))
         (doc (if (stringp (nth 2 term)) (nth 2 term)))
         (fields (if doc (nthcdr 3 term) (nthcdr 2 term)))
         (options (car (last term))))
    (if (and (consp options) (equal (car options) ':options))
	(setq options (cdr options))
	(setq options nil))
    (if options (setq fields (butlast fields)))

    (begin-event doc)
    (event-label-printer "Type ")
    (print-string (remove-make name))
    (index name "defined")
    (pprinci *equal-sign*)
    (pprinci "record ( ")
    ;; Can we fit the fields on one line?
    (cond ((advise-break (cons 'and fields))
           ;; No.
           (set-margin)		;newline
           (print-record-fields fields t)
	   (if (not options) (pprinci " )"))
	   (pop-margin))
          (t (print-record-fields fields nil)))
    (if options (progn (to-current-margin)
		       (pprinci " ")
		       (set-margin)
		       (print-options options)
		       (pprinci " )")
		       (pop-margin)))
    (end-event)))

(local-print defstructure-printer)			
(declare-event-printer defstructure (function defstructure-printer))

;; options == (:options :guards (:predicate symbol )(:conc-name NIL)
;;              (:keyword-constructor sybmol ) (:keyword-updater symbol ))
;; and from
;; deflist      (:CAR-RULE-CLASSES rule-classes) (:NTH-RULE-CLASSES rule-classes)
;;               :OMIT-DEFUN (:THEORY . list-functions*) (:THEORY-NAME theory-symbol)

(defun print-options (options)
  (declare (ignore options)) nil)

(defun print-record-fields (fields newlines)
  (sloop for f on fields
	 do (print-field (car f) (cdr f) newlines)))

(defun print-field (f nextp newlines)
  ;; name | (name [ (:DEFAULT expr) ] [ (:assert (type f) :rewrite) ])
  (let* ((name (if (symbolp f) f (car f)))
	 (default (if (consp f) (assoc :default (cdr f))))
	 (assert (if (consp f) (assoc :assert (cdr f))))
	 (type (if assert (car (cadr assert)))))
    (event-name-printer name)
    (if type (progn (pprinci " : ") (print-type type)))
    (if (and default (cadr default))
	(progn (pprinci " := ") (infix-print-term1 (cadr default))))
    (if nextp (pprinci ", "))
    (if (and newlines nextp) (to-current-margin))))

;; 
;; DEFLIST name arglist [documentation] {declaration}* predicate [option-list]
;;
;; But in our context, we restrict to
;; DEFLIST name (L) (DECLARE (XARGS :verify-guards T )) predicate 

;; NO NO NO.  We want to be able to print the full form.
;; Mar  3 97 MKS 
;;
;; Name and predicate will be symols.  
;; The declaration does not need to be printed and there will be no options.
;;  predicate -- A symbol designating a predicate to be applied to each element of the list.

(defun deflist-printer (term)
  ;; (DEFLIST name (L) [(DECLARE (XARGS :verify-guards T ))] predicate [options])
  (let* ((name (nth 1 term))
	 ;;(args (nth 2 term))
	 (doc (if (stringp (nth 3 term)) (nth 3 term)))
	 (rev (reverse term))
         (options (if (and (consp (car rev)) (eq (caar rev) :options)) (cdr (car rev)) nil))
         (predicate (if options (cadr rev) (car rev)))
	 dcls xvg)
    (sloop for x in (cdddr term);; extract declare
	   do (cond ((or (stringp x)
		      (equal x predicate)
		      (equal x options))
		     nil)
		    ((equal x '(DECLARE (XARGS :verify-guards T )))
		     (setq xvg t))
		    (t (setq dcls (append dcls (list x))))))
    ;;(if (not xvg)
    ;;(setq options (cons '(:verify-guards nil) options)))
    (begin-event doc)
    (event-label-printer "Type ")
    (event-name-printer name)
    (index name "defined")
    (pprinci *equal-sign*)
    (infix-print-term1 predicate)
    (pprinci " *")
    (if options
	(progn (to-current-margin) (pprinci " ")  
	       (set-margin)
	       (sloop for option on options
		      do (if (cdr option)
			     (progn (option-key-printer (car option)) (to-current-margin))
			     (option-key-printer (car option))))
	       (pop-margin)))
    (end-event)))

(local-print deflist-printer)
(declare-event-printer deflist (function deflist-printer))

(defun option-key-printer (option)
  (let ((key (if (consp option) (car option) option))
	(valuep (if (and (consp option) (cdr option)) t nil))
	(value (if (consp option) (cdr option) nil)))
    (italic-sym-printer key)
    (if valuep
	(progn (pprinci " ")
	       (set-margin)
	       (print-list value)
	       (pop-margin)))))


(defun defalist-printer (term)
  ;; (DEFALIST name (L) [(DECLARE (XARGS :verify-guards T ))] (type . type) [options])
  (let* ((name (nth 1 term))
         ;(args (nth 2 term))
	 (doc (if (stringp (nth 3 term)) (nth 3 term)))
	 (rev (reverse term))
         (options (if (and (consp (car rev)) (eq (caar rev) :options)) (cdr (car rev)) nil))
         (types (if options (cadr rev) (car rev)))
	 dcls xvg)
    (sloop for x in (cdddr term);; extract declare
	   do (cond ((or (stringp x)
			 (equal x types)
			 (equal x options))
		     nil)
		    ((equal x '(DECLARE (XARGS :verify-guards T )))
		     (setq xvg t))
		    (t (setq dcls (append dcls (list x))))))
    (begin-event doc)
    (event-label-printer "Type ")
    (event-name-printer name)
    (index name "defined")
    (pprinci *equal-sign*)
    (pprinci *lbracket*)
    (pprinci " ")
    (print-atom (car types))
    (pprinci " . ")
    (print-atom (cdr types))
    (pprinci " ")
    (pprinci *rbracket*)
    (pprinci " * ")
    (if options
	(progn (to-current-margin) (pprinci " ")  
	       (set-margin)
	       (sloop for option on options
		      do (if (cdr option)
			     (progn (option-key-printer (car option)) (to-current-margin))
			     (option-key-printer (car option))))
	       (pop-margin)))
    (end-event)))

(local-print defalist-printer)			
(declare-event-printer defalist (function defalist-printer))

(defvar *forall* "all")
(defvar *tt-size* 1)

(defun forall-printer (term)
  (let  ((vars (nth 1 term))
	(body (nth 2 term)))
    (pprinci *forall* *tt-size*)
    (cond ((atom vars)
           (funcall (get-atom-printer vars) vars))
          (t (sloop for tail on vars do
                      (funcall (get-atom-printer (car tail)) (car tail))
                      (cond ((cdr tail)
                             (pprinci ", " *tt-size*))))))
    (pprinci ": ")     
    (math-space 1)
    (infix-print-term1 body)))

(defun exists-printer (term)
  (let ((vars (nth 1 term))
	(body (nth 2 term)))
    (pprinci *exists* *tt-size*)
    (cond ((atom vars)
	   (funcall (get-atom-printer vars) vars))
          (t (sloop for tail on vars do
		    (funcall (get-atom-printer (car tail)) (car tail))
		    (cond ((cdr tail)
			   (pprinci ", " *tt-size*))))))
    (pprinci ": ")
    (math-space 1)
    (infix-print-term1 body)))

(declare-fn-printer exists (function exists-printer))
(declare-fn-printer forall (function forall-printer))


;; Rune                           Runic Desig
;;  def = symbolp   :functionp ==  (:DEFINITION symb).
;;  ex  = (symbolp) :functionp ==  (:EXECUTABLE-COUNTERPART symb)
;;  thm = symbolp   :defthm    ==  {...} set introduced
;;  ax  = symbolp   :defaxiom  ==  {...} set introduced
;;  th  = symbolp   :deftheory ==  {...} set introduced

;; runic-desig    = def | ex | thm | ax | theory | ?? string
;; runic-des  ==   sym | (sym)
;; rune       ==   (:DEFINITION sym) | (:EXECUTABLE-COUNTERPART sym)
;;               | (:REWRITE sym)    | (:REWRITE sym)
;;               | (:ELIM sym)       | runic-des

(defun print-rune-string (rune)
  (cond ((eq rune :here)     (pprinci ":here"))
	((not (consp rune))  (print-bare-function-name rune))
	((null (cdr rune))
	 (pprinci ":executable-counterpart ")
	 (print-bare-function-name (car rune)))
	((eq (car rune) :executable-counterpart)
	 (pprinci ":executable-counterpart ")
	 (print-bare-function-name (cadr rune)))
	((eq (car rune) :definition)
	 (pprinci ":definition ")
	 (print-bare-function-name (cadr rune)))
	((eq (car rune) :rewrite)
	 (pprinci ":rewrite ")
	 (print-bare-function-name (cadr rune)))
	((eq (car rune) :elim)
	 (pprinci ":elim ")
	 (print-bare-function-name (cadr rune)))
	(t (print-atom (car rune)) 
	   (pprinci " ")
	   (print-bare-function-name (cadr rune)))))

(defun runes-print (runes)
  (sloop for x on runes
	 do (cond ((cdr x)
		   (print-rune-string (car x))
		   (pprinci ", "))			 
		  (t (print-rune-string (car x))))))

;; name       ==   sym | string | :here   -- event name
;; theory     ==   (rune*)                -- sets of runes to enable/disable in concert
;; theory-des == 
;;      (DISABLE runic-desig*)            => theory        
;;    | (ENABLE  runic-desig*)            => theory        
;; 
;;    | (CURRENT-THEORY   name)           => theory  -- name is atom, string, or :here
;;    | (UNIVERSAL-THEORY name)           => theory  -- as of logical name
;;    | (THEORY           name)           => theory        
;;    | (FUNCTION-THEORY  name)           => theory  -- function symbol rules
;;
;;    | (INTERSECTION-THEORIES th1 th2)   => theory
;;    | (SET-DIFFERENCE-THEORIES th1 th2) => theory
;;    | (UNION-THEORIES th1 th2)          => theory


(defun theory-print (form &optional (tail ""))
  (cond ((not (consp form))
	 (bold-sym-printer form)
	 (pprinci tail))
	;;((eq (car form) 'quote) (runes-print (cadr form)) (pprinci tail))
	(t (infix-print-term1 form)(pprinci tail))))

(defun deftheory-printer (event)
  (let* ((name (nth 1 event))
         (theory (nth 2 event))
         (keys (extract-keywords event '(:doc)))
         (doc (nth 0 keys)))
    (if doc (event-doc-string doc))
    (begin-event)
    (event-label-printer "Theory ")
    (event-name-printer name)
    (index name "theory")
    (pprinci " := ")
    (if (eq (Car theory) 'quote) 
	(progn (pprinci *lbracket*)
	       (print-list (Cadr theory))
	       (pprinci *rbracket*))
      (theory-print theory ""))
    (end-event)))

(local-print deftheory)
(declare-event-printer deftheory (function deftheory-printer))

(defun in-theory-printer (event)
  (let* ((theory (nth 1 event))
         (keys (extract-keywords event '(:doc)))
         (doc (nth 0 keys))
	 *do-not-use-tabs*)
    (event-doc-string doc)
    (begin-event)
    (event-label-printer "In-Theory")
    (if (evisceratep theory)
	(progn
	  (pprinci "(")
	  (pprinci "...")
	  (pprinci ")"))
	(progn
	  (pprinci "(")
	  (theory-print theory "")
	  (pprinci ")")))
    (end-event)))

(local-print in-theory)
(declare-event-printer in-theory (function in-theory-printer))

(defun boot-strap-printer (term)
  (declare (ignore term))
  (begin-event)
  (event-label-printer "boot-strap")
  (end-event))

(local-print boot-strap-printer)
(declare-event-printer boot-strap (function boot-strap-printer))

(defvar *print-verify-guards* t)

(defun verify-guards-printer (event)
  (or (not *print-verify-guards*)
      (let* ((name (nth 1 event))
	     (guard (nth 2 event))
	     (keys (extract-keywords event '(:hints :otf-flg :doc)))
	     (hints (nth 0 keys))     
	     (otf (nth 1 keys))
	     (doc (nth 2 keys)))
	(begin-event doc)
	(cond ((consp name)
	       (mapcar (function
			(lambda (x)
			  (if (consp x)
			      (verify-guards-printer2 (car x) (cadr x))
			      (verify-guards-printer2 x nil)))) name))
	      ((and (symbolp name) (consp guard))
	       (verify-guards-printer2 name guard))
	      ((symbolp name)
	       (verify-guards-printer2 name nil))
	      (t nil))
	(if hints (hints-printer hints))
	(if otf (simple-key-printer ":otf-flg" otf))
	(end-event))))

(defun verify-guards-printer2 (name guard)
  (event-label-printer "Verify Guards" " ")
  (print-bare-function-name name)
  (if guard (progn (pprinci " | ") (infix-print-term1 guard)))
  (index name "guard verification"))

(local-print verify-guards)
(declare-event-printer verify-guards (function verify-guards-printer))

;;    (verify-termination fn dcl ... dcl)
;; or (verify-termination (fn1 dcl ... dcl) (fn2 dcl ... dcl) ...)

(defun verify-termination-printer (event)
  (begin-event)
  (event-label-printer "Verify Termination" " ")
  (set-margin)
  (let ((events (if (symbolp (cadr event))
		    (list (cdr event))
		    (cdr event))))
    (mapl (function (lambda (x)
		      (progn (print-verify-termination (car x))
			     (to-current-margin))))
	  events))
  (pop-margin)
  (end-event))

(local-print verify-termination)
(declare-event-printer verify-termination (function verify-termination-printer))

(defun print-verify-termination (x)
  (cond ((consp x)
	 (pprinci " ")
	 (print-bare-function-name (car x))
	 (print-decls (cdr x)))
	(t (pprinci " ") (print-bare-function-name x))))

(defun current-theory-printer (form)
  (cond ((eq (cadr form) :here)
	 (pprinci "current-theory(:here)"))
	((and (consp (cadr form)) (eq (car (cadr form)) 'quote))
	 (pprinci (format nil "current-theory('~s)" (cadr (cadr form)))))
	(t (pprinci (format nil "current-theory(~s)" (cadr form))))))

(declare-fn-printer current-theory (function current-theory-printer))

(defun universal-theory-printer (form)
  (cond ((eq (cadr form) :here)
	 (pprinci "universal-theory(:here)"))
	((and (consp (cadr form)) (eq (car (cadr form)) 'quote))
	 (pprinci (format nil "universal-theory('~s)" (cadr (cadr form)))))
	(t (pprinci (format nil "universal-theory(~s)" (cadr form))))))

(declare-fn-printer universal-theory (function universal-theory-printer))

(defun theory-printer (form)
  (cond ((and (consp (cadr form)) (eq (car (cadr form)) 'quote))
	 (pprinci (format nil "theory('~s)" (cadr (cadr form)))))
	(t (pprinci (format nil "theory(~s)" (cadr form))))))
 


(declare-fn-printer theory (function theory-printer))

(defun function-theory-printer (form)
  (cond ((eq (cadr form) :here)
	 (pprinci "current-theory(:here) "))
	((and (cadr form) (consp (cadr form)) (eq (car (cadr form)) 'quote))
	 (pprinci "function-theory(")
	 (bold-sym-printer (cadr (cadr form)))
	 (pprinci ")"))
	(t (pprinci (format nil "function-theory(~s)" (cadr form))))))

(declare-fn-printer function-theory (function function-theory-printer))

;; See scribe or latex init to see how the following are handled.
;;  intersection-theories  
;;  set-difference-theories
;;  union-theories         

(defun theory-invariant-printer (form)
  (let ((term (cadr form))
	(key  (if (cddr form) (caddr form) nil)))
    
    (pprinci "Theory-Invariant(")
    (if (evisceratep term) (pprinci "...")
      (progn (infix-print-term1 term)
	     (if key (progn (pprinci ", ") (print-atom key)))))
    (pprinci ")")))
(local-print defcor)
(declare-fn-printer theory-invariant (function theory-invariant-printer))

(defun defun-sk-printer (event)
  ;; (defun-sk name args body &key doc quant-ok skolem-name thm-name)
  (let* ((name (nth 1 event))
         (args (nth 2 event))
         (body (transform-body (nth 3 event))) 
	 (keys (extract-keywords event '(:doc :quant-ok :skolem-name :thm-name)))
	 (doc  (car keys))
         (form (cons name args))
         (eq (list 'equal form body))
	 (dcls (cdr keys)))
    (defun-printer2 name doc form body dcls eq nil)))

(local-print defun-sk)
(declare-event-printer defun-sk (function defun-sk-printer))

#| 
   DEFLOOP name arglist [documentation] {dcls}* loop-form
   See for-printer for more documentation.
|#

(defun defloop-printer (event)
  ;; DEFLOOP name args [doc] {dcls}* loop-form
  (let* ((name (nth 1 event))
         (args (nth 2 event))
         (doc (if (stringp (nth 3 event)) (nth 3 event)))
         (form (cons name args))
         (body (car (last event)))
	 (eq (list 'equal form body))
         dcls)
    (sloop for x in (cdddr event)
	   do (if (or (stringp x)
		      (equal x body))
		  nil
		  (setq dcls (append dcls (list x)))))
    (defun-printer2 name doc form body dcls eq nil)))

(local-print defloop)
(declare-event-printer defloop (function defloop-printer))


;; (  LEFT IN ----------

(defun do-file-printer (term)
  (print-default-event-header)
  (pprinc "Do all forms in the file:
")
  (italic-sym-printer (cadr term))
  (no-tab-event-trailer))

(defun setq-printer (term)
  ;; (print-default-event-header)
  (let (name value)
    (match term (setq name value))
    (tt-sym-printer name)
    (ppformat " := ")
    (tt-sym-printer value)
    (ppformat ";"))
  (no-tab-event-trailer))

(defun comp-printer (term)
  ":comp t          - compile all uncompiled ACL2 functions
:comp foo        - compile the defined function foo
:comp (foo bar)  - compile the defined functions foo and bar"
  (begin-event)
  (cond ((null (cdr term))
	 (pprinc *begin-comment-env*)
         (pprinc " Unknown compile command. ")
	 (pprinc *end-comment-env*))
        ((equal (cadr term) t)
	 (pprinci "Compile t;"))
        ((symbolp (cadr term))
	 (pprinci "Compile ")
           (print-bare-function-name (cadr term))
           (pprinci ";"))
        (t (pprinci "Compile ")
           (print-bare-function-names (cadr term))
           (pprinci ";")))
  (end-event))

(defmacro define-command-printer (name string)
  (let ((fn-name (intern (format nil "~s-COMMAND-PRINTER" name))))
    `(progn
       (clean-up ',name)
       (defun ,fn-name
         (term)
	 (declare (ignore term))
	 (print-default-command-header)
	 (pprinci ,string)
	 (no-tab-command-trailer))
       (declare-command-printer ,name (function ,fn-name)))))

(defun checkpoint-forced-goals-printer (term)
  (print-default-command-header)
  (if (cadr term)
      (pprinci "Checkpoint forced goals.")
      (pprinci "Do not checkpoint forced goals."))
  (no-tab-command-trailer))

(define-command-printer disable-forcing "Disable-forcing;")
(define-command-printer enable-forcing "Enable-forcing;")
(define-command-printer good-bye "good-bye;")
(define-command-printer oops-printer "oops;")

(defun puff-printer (term)
  (begin-event (format nil " Expand ~s" (cadr term)))
  (infix-print-term term)
  (end-event))

(defun puff*-printer (term)
  (begin-event (format nil " Expanding* ~s" (cadr term)))
  (infix-print-term term)
  (end-event))

(defun q-printer (term)
  (begin-event " Quitting ACL2")
  (pprinci ":q\;")
  (end-event))

(define-command-printer redef ":redef ;")
;; !!! ACL2 system hacker's redefinition command. 
(define-command-printer redef! ":redef! ;")

(defun reset-ld-specials-printer (term)
  (begin-event (format nil "Reset LD specials to their initial value~s"
		       (if (cadr term)
			   ", including I/O channels."
			   ".")))
  (infix-print-term term) 
  (end-event))

(defun set-cbd-printer (term)
  (begin-event)
  (pprinci ":set-cdb ")
  ;; Replaced filename-sym-printer
  (print-atom (cadr term))
  (end-event))

(defun set-guard-checking-printer (term)
  (print-default-command-header)
  (if (cadr term)
      (pprinci ":Set-guard-checking t;")
      (pprinci ":Set-guard-checking nil;"))
  (no-tab-command-trailer))

(define-command-printer start-proof-tree ":Start-proof-tree;")
(define-command-printer stop-proof-tree ":Stop-proof-tree;")
(define-command-printer u ":U;")

(defun ubt-printer (term)
  "Note that this only works because print-atom does no tabbing."
  (print-default-command-header)
  (cond ((null (cdr term))
         (pprinci ":U ;"))
        (t (pprinci ":Ubt '")
           (print-atom (cadr term))
           (pprinci ";")))
  (no-tab-command-trailer))

(defun retrieve-printer (term)
  (begin-event)
  (infix-print-term term)
  (end-event))

;; (define-command-printer unsave "unsave")
;; (define-command-printer save "save")

(defun verify-printer (event)
  (begin-event "Verify ")
  (infix-print (cadr event))
  (end-event))

(defun accumulated-persistence-printer (term)
  (begin-event)
  (infix-print-term term)
  (end-event))

(defun show-accumulated-persistence-printer (term)
  (begin-event
   (cond ((equal (cadr term) :frames)
	  "Display statistics ordered by frames built.")
	 ((equal (cadr term) :tries)
	  "Display statistics ordered by times tried. ")
	 ((equal (cadr term) :ratio)
	  "Display statistics ordered by the ratio of frames built to times tried.")
	 (t nil)))
   (if (cadr term)
       (pprinci "Accumulated-persistence(true)")
       (pprinci "Accumulated-persistence(nil)"))
   (end-event))

(defun brr-printer (term)
  (print-default-command-header)
  ;; Just use functional form.
  (if (cadr term)
      (pprinci "Brr(t);")
      (pprinci "Brr(nil);"))
  (no-tab-command-trailer))

(defun monitor-printer (term)
  (begin-event)
  (default-fn-printer term)
  (end-event))

(define-command-printer monitored-runes "Print monitored runes.")

(defun unmonitor-printer (term)
  (let ((form (cadr term)))
    (if (and (consp form) (equal (car form) 'quote))
	(setq form (cadr form)))
    (begin-event)
    (pprinci "Un-monitor(")
    (cond ((equal form :all) (pprinci ":all"))
	  (t (print-rune-string form)))
    (pprinci ")")
    (end-event)))

(defun add-macro-alias-printer (term)
  (begin-event "Add macro alias.")
  (default-fn-printer term)
  (end-event))

(defun remove-macro-alias-printer (term)
  (begin-event "Remove macro alias.")
  (default-fn-printer term)
  (end-event))

(define-command-printer logic "logic;")
(define-command-printer program "program;")

(defun set-compile-fns-printer (term)
  (begin-event)
  (default-fn-printer term)
  (end-event))

(defun set-ignore-ok-printer (term)
  (begin-event
   (cond ((null (cadr term))
	  "Disallow unused formals and locals.")
	 ((eq (cadr term) t)
	  "Allow unused formals and locals.")
	 ((eq (cadr term) :warn)
	  "Allow unused formals and locals, but print warning. ")))
  (default-fn-printer term)
  (end-event))

(define-command-printer set-invisible-fns-alist "Setting the invisible fns alist.")

(defun set-irrelevant-formals-ok-printer (term)
  (begin-event
   (cond ((null (cadr term))
	 "Disallow irrelevant formals in definitions.")
	((eq (cadr term) t)
	 "Allow irrelevant formals in definitions.")
	((eq (cadr term) :warn)
	 "Allow irrelevant formals in definitions, but print warning.")))
  (infix-print-term term)
  (end-event))

(defun set-measure-function-printer (term)
  (begin-event "Set the default measure function.")
  (default-fn-printer term)
  (end-event))

(defun set-verify-guards-eagerness-printer (term)
  (begin-event
   (cond ((equal (cadr term) 0)
	  "No guard verification unless :verify-guards is t.")
	 ((equal (cadr term) 1)
	  "Verify guards, if supplied.")
	 ((equal (cadr term) 2)
	  "Verify guards, unless :verify-guards is NIL.")))
  (default-fn-printer term)
  (end-event))

(defun set-well-founded-ordering-printer (term)
  (begin-event "Set the default well-founded relation.")
  (default-fn-printer term)
  (end-event))

(defun firstn-terms (n l)
  (if (not (consp l))
      nil
    (if (zerop n)
	nil
      (cons (car l) (firstn-terms (1- n) (cdr l))))))

(defun table-printer (Term)
  (let ((op (nth 4 term))
	(term2 (nth 5 term)))
    (begin-event)
    (pprinci "Table(")
    (cond(op
	   (print-list (firstn-terms 3 (cdr term)) ", ")
	   (pprinci ", ")
	   (infix-print-term1 op)
	   (if term2
	       (progn (pprinci " ")
		      (infix-print-term1 term2))))
	 (t (print-list (cdr term) ", ")))
    
    (pprinci ")")
    (end-event)))

(declare-event-printer table (function table-printer))


;; We would like to put this at the top, but the functions need to be defined first.

(setq *event-printer-alist*
      (append
       (list
	(list 'puff (function puff-printer))
	(list 'puff* (function puff*-printer))
	(list 'reset-ld-specials (function reset-ld-specials-printer))
	(list 'set-cbd (function set-cbd-printer))

	(list 'set-guard-checking (function set-guard-checking-printer))
	(list 'ubt (function ubt-printer))
	(list 'ubt! (function ubt-printer))
	(list 'retrieve (function retrieve-printer))
	(list 'skip-proofs (function skip-proofs-printer))
	(list 'accumulated-persistence (function accumulated-persistence-printer))
	(list 'show-accumulated-persistence (function show-accumulated-persistence-printer))
	(list 'brr (function brr-printer))
	(list 'monitor (function monitor-printer))
	(list 'unmonitor (function unmonitor-printer))
	(list 'add-macro-alias (function add-macro-alias-printer))
	(list 'remove-macro-alias (function remove-macro-alias-printer))
	(list 'set-compile-fns (function set-compile-fns-printer))
	(list 'set-ignore-ok (function set-ignore-ok-printer))
	(list 'set-irrelevant-formals-ok (function set-irrelevant-formals-ok-printer))
	(list 'set-measure-function (function set-measure-function-printer))
	(list 'set-verify-guards-eagerness (function set-verify-guards-eagerness-printer))
	(list 'set-well-founded-ordering (function set-well-founded-ordering-printer))
	;; History.  Just print as function

	;; (list 'pbt (function ignore-printer))
	;; (list 'pc  (function pc-printer))
	;; (list 'pcb (function pcb-printer))
	;; (list 'pe  (function pe-printer))

	)
       *event-printer-alist*))



