
;; FONT SYMBOL PRINTERS

(in-package "ACL2")

; (require "sloop")
(require "infix")

(defvar *begin-bold-env*  "")
(defvar *end-bold-env*    "")

(defun bold-sym-printer (x &optional i)         ; Print in bold face.
  (pprinc *begin-bold-env*)
  (cond ((symbolp x) (print-atom x i))
	((characterp x) (print-character x i)) 
	(t (print-string x i)))
  (pprinc *end-bold-env*))

(defun reserved-sym-printer (x &optional i)         ; Print in bold face.
  (pprinc *begin-bold-env*)
  (cond ((symbolp x) (print-symbol-body x i))
	((characterp x) (print-character x i)) 
	(t (print-string x i)))
  (pprinc *end-bold-env*))

(defun nil-printer (x &optional i)         ; Print in bold face.
  (declare (ignore x))
  (pprinc *begin-bold-env*)
  (print-string "[ ]" i)
  (pprinc *end-bold-env*))

(defvar *begin-italic-env*  "")
(defvar *end-italic-env*    "")

(defun italic-sym-printer (x &optional i)               ; Print in italic face.
  (pprinc *begin-italic-env*)
  (cond ((symbolp x)    (print-atom x i))
	((characterp x) (print-character x i)) 
	(t (print-string x i)))
  (pprinc *end-italic-env*))

(defvar *begin-tt-env*  "")
(defvar *end-tt-env*    "")

(defun tt-sym-printer (x &optional i)           ; Print in typewriter font.
  (pprinc *begin-tt-env*)
  (cond ((symbolp x) (print-atom x i))
	((characterp x) (print-character x i)) 
	(t (print-string x i)))
  ;; We charge more for tt characters.
  (incf *infix-loc* (* (- *tt-size* 1) (our-flatc x)))
  (pprinc *end-tt-env*))

(defvar *begin-sc-env*  "")
(defvar *end-sc-env*    "")

(defun small-caps-sym-printer (x &optional i)           ; Print in small caps.
  (pprinc *begin-sc-env*)
  (cond ((symbolp x) (print-atom x i))
	((characterp x) (print-character x i)) 
	(t (print-string x i)))
  (pprinc *end-sc-env*))

(defun font-sym-printer (symbol font)
  (case font
    (bold   (bold-sym-printer symbol))
    (italic (italic-sym-printer symbol))
    (tt     (tt-sym-printer symbol))
    (sc     (small-caps-sym-printer symbol))
    (otherwise (format *terminal-io* "Bad font descriptor (~a) passed to subscript printer.~%" font)
               (tt-sym-printer symbol))))

(defvar *subscript*  nil)
(defvar *begin-subscript* nil)
(defvar *end-subscript* nil)

(defun subscript (x)
  (pprinc *begin-subscript*)
  (print-atom x)
  (pprinc *end-subscript*))

(defparameter *atom-alist*
  (list (list 't   (function reserved-sym-printer))
        (list 'f   (function reserved-sym-printer))
        (list 'nil (function nil-printer))))
