(in-package "ACL2")

(require "infix")

;                                     QUOTE

; The printing of quote terms is intrinsically intertwined with the printing of
; backquoted terms.  The backquoted terms have been read with our
; special-to-infix-file backquote reader.

(defvar *fn-alist* nil)

(defvar *top-parens-eliminable*)
(defvar *infix-comma* (make-symbol "comma"))
(defvar *infix-comma-atsign* (make-symbol "comma-atsign"))
(defvar *infix-backquote* (make-symbol "backquote"))
(defvar *tt-size* 1.3)

(defvar *begin-tt-env*  "")
(defvar *end-tt-env*    "")

(defvar *left-margin* nil)
(defvar *left-margin-tab-context* nil)
(defvar *dotted-pair-separator* " . ")

(defvar *comma-atsign* ",@")
(defvar *caret* "^")

(defvar *math-format* "~a")
(defvar *math-begin* "")
(defvar *math-end* "")

;; These must be enclosed in a math env.  
(defvar *math-thick-space* "  ")
(defvar *math-space* " ")
(defvar *math-thin-space* " ")

(defvar *subscript* "")

(defvar *begin-subscript* "")
(defvar *end-subscript* "")

(defvar *begin* "")
(defvar *end*   "")

(defun quote-printer1 (term)
  (let ((*function-symbol-out* nil))
					; How we output a quoted term.
    
    (cond ((evisceratep term)                   (pprinci (cdr term)))
	  ((stringp term)                       (print-atom term))
	  ((atom term)	                        (tt-sym-printer term))
	  ((eq (car term) 'quote)               (quote-printer term))
	  ((eq (car term) *infix-comma*)        (comma-printer term))
	  ((eq (car term) *infix-comma-atsign*) (comma-atsign-printer term))
	  ((eq (car term) *infix-backquote*)    (backquote-printer term))
	  ((eq (car term) *infix-radix*)        (funcall (function *infix-radix*-printer) term))
	  ((advise-break (list 'quote term))    (quote-printer1-advise-break term))
	  (t (quote-printer1-tt-form term)))))

(defun tt-pprinci (term &optional (len 1))
  (pprinci *begin-tt-env* 0)
  (pprinci term len)
  (pprinci *end-tt-env* 0))   

(defun quote-printer1-tt-form (term)
  (cond ((and (consp term) (or (characterp (nth 0 term))
			       (quote-p (nth 0 term))))
	 (tt-pprinci *lbracket* *tt-size*)
	 (tt-pprinci " " *tt-size*))
	(t  (tt-pprinci *lbracket* *tt-size*)))
  (set-margin)
  (sloop for tail on term do
	 (progn 
	   (quote-printer1 (car tail))
	   (cond ((null (cdr tail)) (tt-pprinci *rbracket* *tt-size*))
		 ((or (atom (cdr tail))
		      (eq (car (cdr tail)) *infix-comma*)
		      (eq (car (cdr tail)) *infix-comma-atsign*)
		      (eq (car (cdr tail)) *infix-backquote*))
		  (tt-pprinci *dotted-pair-separator*  4)
		  (quote-printer1 (cdr tail))
		  (tt-pprinci *rbracket* *tt-size*)
		  ;(if (and (not (equal (car tail) (cdr term)))
		  ;         (not (equal tail (cdr term))))
		  ;    (to-current-margin))
		  )
		 (t (tt-pprinci " " *tt-size*))))
	 until (atom (cdr tail)))
  (pop-margin))
                                

(defun quote-printer1-advise-break (term)
  (cond ((and (consp term) (or (characterp (nth 0 term))
					   (quote-p (nth 0 term))))
	 (tt-pprinci *lbracket* *tt-size*))
	(t  (tt-pprinci *lbracket* *tt-size*)))
  (set-margin)
  
  (sloop for tail on term
	 do (progn (quote-printer1 (car tail))
		   (cond ((null (cdr tail))
				 (tt-pprinci *rbracket* *tt-size*))
			 ((or (atom (cdr tail))
			      (eq (car (cdr tail)) *infix-comma*)
			      (eq (car (cdr tail)) *infix-comma-atsign*)
				     (eq (car (cdr tail)) *infix-backquote*))
			  (to-current-margin)	;newline
			  (tt-pprinci *dotted-pair-separator* 4)
			  (quote-printer1 (cdr tail))
			  (tt-pprinci *rbracket* *tt-size*)
			  (return nil))
			 ((and (or (atom (car tail)) (eq *infix-radix* (car (car tail))))
			       (cdr tail)
			       (or (atom (cadr tail)) (eq *infix-radix* (car (cadr tail))))
                                 (not (advise-break (list 'quote (cadr tail)))))
			  (tt-pprinci " " *tt-size*))
			 (t (if (not (atom tail))
				(to-current-margin))))) ;newline
	 until (atom (cdr tail)))
  (pop-margin))

(defun pending-printer (term)
  (declare (ignore term))
  (italic-sym-printer "pending"))

;; (declare-fn-printer pending (function pending-printer))

(defvar *special-quoted-forms-alist*
  (list (cons 'pending (function pending-printer))))

(defun quote-printer (term)
  (let ((fun (cdr (assoc (cadr term) *special-quoted-forms-alist*)))
	(*function-symbol-out* nil))
    
    (if fun
	(funcall fun (list (cadr term)))
	(progn (tt-pprinci "'" *tt-size*)
	       (quote-printer1 (cadr term))))))

(declare-fn-printer quote (function quote-printer))

;; Special purpose for top level value printing.  
;; Think of term as derived from (quote term), a value to
;; be printed as a result of a top level eval.

(defun constant-printer (term)
  (quote-printer1 term))

(defun backquote-printer (term)
  (tt-pprinci "`" *tt-size*)
  (quote-printer1 (cadr term)))

(setq *fn-alist*
      (cons (list *infix-backquote* (function backquote-printer))
	    *fn-alist*))

(defun comma-printer (term)
  (tt-pprinci "," *tt-size*)
  (quote-printer1 (cadr term)))

(defun comma-atsign-printer (term)
  (tt-pprinci *comma-atsign* 4)
  (quote-printer1 (cadr term)))

(defun macro-atsign-printer (term)
  (pprinci "@")
  (if (= 2 (length term))
      (quote-printer1 (cadr term))
    (infix-print-term1 (cdr term))))

(declare-fn-printer macro-atsign  (function macro-atsign-printer))

(defun *infix-radix*-printer (term)
  ;; (*infix-radix* n c ... c)
  (let ((base (cadr term))
	(sign (if (member (caddr term) '(#\+ #\-)) (caddr term) nil))
	(digits (cdr term)))
    (if sign (setq digits (cdr digits)))
    (pprinc *math-begin*)
    (pprinc *begin-tt-env*)
    (if sign (pwrite-char sign))
    (case base 
      (8 (pprinci "0"))
      (16 (message "Can't handle base 16."))
      (2 (message "Can't handle base 2.")))
    (pprin1i (cadr term))
    (pprinc *end-tt-env*)
    (pprinc *math-end*)))

(setq *fn-alist*
      (cons (list *infix-radix* (function *infix-radix*-printer))
	    *fn-alist*))

(defvar *lbracket* "[")
(defvar *rbracket* "]")

(defun subscript-printer (symbol subscript &optional (font 'tt))
  ;; Font must be one of bold, italic, tt, sc (small caps)
  (pprinc *math-begin*)
  (font-sym-printer symbol font)
  (pprinc *lbracket*)
  (font-sym-printer subscript font)
  (pprinc *rbracket*)
  (pprinc *math-end*))

(defun math-thick-space (&optional (n 1))
  (cond (*do-not-use-tabs* (setq n 1)))
  (pprinc *math-begin*)
  (sloop for i from 1 to n do
           (pprinci *math-thick-space*))
  (pprinc *math-end*))

(defun math-space (&optional (n 1))
  (cond (*do-not-use-tabs* (setq n 1)))
  (pprinc *math-begin*)
  (sloop for i from 1 to n do
           (pprinci *math-space*))
  (pprinc *math-end*))

(defun math-thin-space (&optional (n 1))
  (cond (*do-not-use-tabs* (setq n 1)))
  (pprinc *math-begin*)
  (sloop for i from 1 to n do
           (pprinci *math-thin-space*))
  (pprinc *math-end*))


(defun add-lambda-vars-value (vars value)
  (if (consp vars)
      (cons (list 'assignment (Car vars) (car value))
	    (add-lambda-vars-value (cdr vars) (cdr value)))
    nil))


; A FEW HAND-CODED FORM PRINTERS:  LAMBDA, IF, COND, CASE, LET, FOR, FORALL and EXISTS.
;; ((lambda (vars) body) values)
;; (lambda (Vars) body)

(defun lambda-printer (term)

  (let* ((vars nil) 
	 (body nil) 
	 (value nil)
	 (newline nil)
	 (dcls nil)
	 (types nil))
    (if (and (consp (car term)) (equal (caar term) 'LAMBDA))
	(progn 
	  (setq body (car (last (nth 0 term))))
	  (sloop for x in (cddar term);; extract declare
		 do (if (or (stringp x)
			    (equal x body))
			nil
		      (setq dcls (append dcls (list x)))))
               (setq body (list 'body (car (last(nth 0 term)))))
	       (setq vars (nth 1 (nth 0 term)))
	       (setq value (cdr term))
               ;;(setq vars (add-lambda-vars-value vars value))
	     )
	       
      (progn 
	(setq body (car (last term)))
	(sloop for x in (cddr term);; extract declare
	       do (if (or (stringp x)
			  (equal x body))
		      nil
		    (setq dcls (append dcls (list x)))))
	(setq body (list 'body (car (last term))))
	(setq vars (nth 1 term))))
    
    (setq types (extract-declare-types dcls))
    
    (setq newline (advise-break (list 'equal vars body)))
    ;;alternative-lambda-name = lambda
    ;; it will be converted to "lambda"
    
    (signature-header-printer (cons 'alternative-lambda-name vars) types)
    (if newline
	(progn (to-current-margin) (pprinci " ")
	       (set-margin)
	       (infix-print-term1 body)
	       (if value (default-fn-printer (cons '*do-not-print-fn-name*
						   value)))
	       (pop-margin))
      (progn (pprinci " ") (infix-print-term1 body) 
	     (if value (default-fn-printer (cons '*do-not-print-fn-name* value)) )))))

(declare-fn-printer lambda (function lambda-printer))



;; if term = (if x y z) => (cond (x y) (t z))
;;    term = (if x y (if m n z)) => (cond (x y) (m n) (t z)) 
(defun condify (term)
  (let (x u v)
    (sloop with pairs
	   while (or (match term (if x u v))
		     (match term (if* x u v)))
	     ;; ??? I put the push and setq into a PROGN ???
             do (progn (push (list x u) pairs)         
                       (setq term v))
            finally
             (progn (push (list t term) pairs)
                    (return (cons 'cond (reverse pairs)))))))


;; if term = (if x y z) => (cond (x y) (t z))
;;    term = (if x y (if m n z)) => (cond (x y) (t (cond (m n) (t z)) ))
;;(defun condify (term)
;;   (let (ts tb fb x u v)
;;     (if (or (match term (if x u v))
;;	     (match term (if* x u v)))
;;	 (progn (setq ts (cadr term))
;;		(setq tb (caddr term))
;;		(setq fb (caddr term))
;;		(list 'cond (list ts tb)
;;		(cons 't (list (condify v)))))
;;      term)))


(defun if-printer (term)
  (cond ((null (cddr term))
         (format *terminal-io* "~%Ill formed if-expression - ~a~%" term)
         (cond-printer (condify (append term '(nil nil)) (car term))))
        ((null (cdddr term))
         (format *terminal-io* "~%Ill formed if-expression - ~a~%" term)
         (cond-printer (condify (append term '(nil)) (car term))))
        (t (cond-printer (condify term) (car term)))))

(declare-fn-printer if (function if-printer))
(declare-fn-printer if* (function if-printer))

; We take the attitude that we can print an untranslated term t as we would
; print an untranslated term t' provided that t and t' translate to the same
; term.  This point of view is to be found expressed in our treatment of nested
; if's as though they were conds.

(defun cond-printer (term &optional (name 'if))
  ;; Deleted space in name. Apr 25 97 MKS 
  (cond ((equal name 'if) (setq name "if"))
	((equal name 'if*) (setq name "if*"))
	(t (setq name (format nil "~s " name))))
  (let (some-line-broken)
    (advise-break-if-testing)
    (let ((*top-parens-eliminable* t))
      (cond ((null (cddr term))
	     (infix-print-term1 (cadr (cadr term))))
	    (t (let ((cases (cdr term))
		     (first-case (car (cdr term))))
		 ;;(*left-margin-tab-context* nil)
		 (bold-sym-printer name)
		 ;; Switched order of the following 2 forms.  Jun 22 97 MKS 
		 (set-margin)
		 (pprinci " ")
		 (if (consp (car first-case))
		     (if;;(or (get (caar first-case) 'literalform)
			 (member (caar first-case)
				 '(COND IF IF* CASE))
			 (parens-print-term1 (car first-case))
		       (infix-print-term1 (car first-case)))
		   (infix-print-term1 (car first-case)))
		 ;; then
		 (cond ((let ((*rightmost-char-number* (- *rightmost-char-number* 5)))
			  (advise-break (cadr first-case)))
			(setq some-line-broken t)
			(to-current-margin)))
		 (bold-sym-printer " then " 6)
		 (infix-print-term1 (cadr first-case))
		 ;; else
		 (cond ((let ((*rightmost-char-number* (- *rightmost-char-number* 2)))
			  (advise-break (cons 'cond (cdr cases))))
			(setq some-line-broken t)
			(to-current-margin)))
		 ;;
		 (sloop for tail on (cdr cases) do
                        (cond ((null (cdr tail))
                               (bold-sym-printer " else " 6)
                               (let ((*rightmost-char-number* (- *rightmost-char-number* 6)))
                                 (infix-print-term1 (cadr (car tail))))
			       ;; TESTING.  Inserted (math-space 1) and deleted the following
			       ;; Switched back, using to-current-margin.
			       ;; Tried again to delete the following
			       ;; (if some-line-broken (to-current-margin))
			       )
                              (t (bold-sym-printer " elif " 6)
                                 (infix-print-term1 (caar tail))
                                 (cond ((let ((*rightmost-char-number*
					       (- *rightmost-char-number* 6)))
                                          (advise-break (cadar tail)))
					(setq some-line-broken t)
                                        (to-current-margin)))
                                 (bold-sym-printer " then " 6)
                                 (infix-print-term1 (cadar tail))
                                 (to-current-margin))))
		 (pop-margin)))))))

(declare-fn-printer cond  (function cond-printer))

(defun print-one-case (term)
  (let ((break (advise-break term))
	(case (car term))
	(entries (cdr term)))
    (if (eq case 'otherwise) (setq case 'acl2::else))
    (quote-printer1 case)
    (bold-sym-printer " : " 3)
    ;; If break, always set the margin on the next line.
    (if break
	(progn (to-current-margin)
	       (pprinci " ")
	       (set-margin)))
    (if (consp entries)
	(progn (infix-print-term1 (car entries)) (pprinci ";")))
    ;; There should only be multiple entries if we have
    ;; a bunch of ignore declarations.
    (sloop for entry in (cdr entries)
	   do (progn
		(if break
		    (to-current-margin) ;newline
		    (math-space 1))
		(infix-print-term1 entry)
		(pprinci ";")))
    (if break (pop-margin))
    (to-current-margin)))

(defun declare-printer (term)
  (cond ((and (consp (cdr term)) (consp (cadr term))
	      (eq (car (cadr term)) 'ignore))
	 (infix-print-term1 (cadr (cadr term)))
	 (pprinci " : ignore"))
	(t (default-fn-printer term))))

(declare-fn-printer declare (function declare-printer))

(defun case-printer (term)
  (advise-break-if-testing)
  (let ((*top-parens-eliminable* t))
    (cond ((null (cdddr term))
           (infix-print-term1 (cadr (caddr term))))
          (t (let ((cases (cddr term)))
	       (set-margin)
               (bold-sym-printer "CASE " 5)
	       (infix-print-term1 (cadr term))
	       (pprinci " ")
	       (pprinci *lbrace*)
	       (to-current-margin)
	       (pprinci " " )
	       (set-margin)
               (sloop for case in cases do (print-one-case case))
	       (pop-margin)
	       (pprinci *rbrace*)
	       (pop-margin))))))

(declare-fn-printer case  (function case-printer))

(defun case-match-printer (term)
  (advise-break-if-testing)
  (let ((*top-parens-eliminable* t))
    (cond ((null (cdddr term))
           (infix-print-term1 (cadr (caddr term))))
          (t (let ((cases (cddr term)))
	       (set-margin)
               (bold-sym-printer "CASE-MATCH " 5)
	       ;; NOTE: the guard for case-match requires that
	       ;; cadr term be a symbolp.
	       (infix-print-term1 (cadr term))
	       (pprinci " ")
               (pprinci *lbrace*)
	       (to-current-margin)
	       (pprinci " " )
	       (set-margin)
               (sloop for case in cases do (print-one-case case))
	       (pprinci *rbrace*)
	       (pop-margin)
	       (pop-margin))))))

(declare-fn-printer case-match  (function case-match-printer))

#| (FOR ((var [ IN || ON ] arg [ BY fn ])+) clause)
 
   { for x IN arg BY f , COLLECT x.car }
   { for x ON arg , WHEN foo(x), COLLECT x.car }
   { for x IN arg , ALWAYS p(x.car,x.cdr) }

   arg -- a symbol appearing in the arglist.
   arglist -- an argument list satisfying ACL2::ARGLISTP.
   fn -- a function symbol of a unary function well-defined on true lists.

|#

(defun for-printer (term)
  (let ((*top-parens-eliminable* t)
        (args (cadr term))
	(clause (caddr term))
	(newline (advise-break term)))
    (pprinci *lbrace*)
    (bold-sym-printer "FOR " 5)
    (set-margin)
    (sloop for arg in args
	   do (progn (arg-print arg)
		     (pprinci ", ")
		     (if newline (to-current-margin))))
    (clause-print clause)
    (pprinci *rbrace*)
    (pop-margin)))

(declare-fn-printer for (function for-printer))

(defun arg-print (l)
  (let* ((var (nth 0 l))
	 (in  (nth 1 l))
	 (exp (nth 2 l))
	 (by (if (cdddr l) (nth 3 l)))
	 (fn (if by (nth 4 l))))
    (infix-print-term1 var)
    (pprinci " ")
    (bold-sym-printer in) 
    (pprinci " ")
    (infix-print-term1 exp)
    (if by (progn (pprinci " ")
		  (bold-sym-printer "by " 3) 
		  (print-bare-function-name fn)))))

(defun clause-print (clause)
  ;; clause = ({COLLECT | APPEND} form)
  ;;        | ([ IF | WHEN | UNLESS ] form ({COLLECT | APPEND | RETURN} form)) 
  ;;        | ([ ALWAYS | THEREIS | NEVER ] form)
  (let ((newline (advise-break clause)))
    (cond ((member (car clause)
		   '(COLLECT APPEND ALWAYS THEREIS NEVER RETURN))
	   (bold-sym-printer (car clause))
	   (pprinci " ")
	   (infix-print-term1 (cadr clause)))
	  (t (bold-sym-printer (car clause))
	     (pprinci " ")
	     (infix-print-term1 (cadr clause))
	     (pprinci ", ")
	     (if newline (to-current-margin))
	     (clause-print (caddr clause))))))

;; LET WITH DECLARE
;;
;;(defun let-printer (term)
;;  (let* ((orig (car (last term)))
;;	 (dcls nil))
;;    (sloop for x in (cddr term);; extract declare
;;	   do (if (or (stringp x)
;;		      (equal x orig))
;;		  nil
;;		(setq dcls (append dcls (list x)))))
;;    (let* ((types (extract-declare-types dcls))
;;	  (arg (cadr term)))
;;      (let*-printer `(let* ,(list (car arg))
;;		       
;;		       ,(if (cdr arg)
;;			    `(let ,(cdr arg) ,@(cddr term))
;;			 (cddr term))))  types))))


;; LET W/O DECLARE

(defun let-printer (term)
  (let*-printer `(let* ,(list (car (cadr term)))
		   ,(if (cdr (cadr term))
			`(let ,(cdr (cadr term)) ,(car (last term)))
			(car (last term))))))


(declare-fn-printer let   (function let-printer))

(defun let*-printer (term1 &optional type)
  (let ((lets (cadr term1))
        (term nil ))
    (setq term (if (consp (nth 2 term1))
		   (if (and (consp (car (nth 2 term1)))
			    (= (length (nth 2 term1)) 1))
		       (cons (car term1)
			     (cons (cadr term1)
				   (car (nth 2 term1))))
		     term1)
		 term1))
    (advise-break-if-testing)
    (let ((*top-parens-eliminable* t))
      (if (null lets)
	  (infix-print-term1 (car (last  term)))
	(let ()
	  (pprinci *lbrace*)
	  (set-margin)
	  (if (consp lets)
	      (sloop for let in lets
		     do (progn (assign-printer (car let) (cadr let) type)
			       
			       (to-current-margin))))
	    (let ((*rightmost-char-number* (- *rightmost-char-number* 7)))
	      (infix-print-term1 (car (last term))))
	    (pprinci *rbrace*)
	    (pop-margin))))))

(declare-fn-printer let*  (function let*-printer))

(defun mv-printer (term)
  (mv-printer1 (cdr term)))

(declare-fn-printer mv    (function mv-printer))

(defvar *mv-bracket-left* "<")
(defvar *mv-bracket-right* ">")

(defun mv-printer1 (vars)
  (pprinc *mv-bracket-left*)
  (sloop for tail on vars do
	 (progn
	   (if (symbolp (car tail))
	       (funcall (get-atom-printer (car tail)) (car tail))
	       (infix-print-term1 (car tail)))
	   (if (cdr tail) (pprinci ", " *tt-size*))))
  (pprinc *mv-bracket-right*))

(defun assignx-printer (term)
  (assign-printer (cadr term) (caddr term)))

(defun assign-printer (vars term &optional (types nil))
  (if (and (atom vars)(null types))
      (funcall (get-atom-printer vars) vars)
    (if types
	(mv-printer-with-types (if (atom vars) (list vars) vars) 
			       types)
      (mv-printer1 vars)))
  (math-space 1)
  (pprinc ":=")
  (math-space 1)
  (infix-print-term1 term)
  (pprinci ";"))

;;;(pair-printer a 4) ==> a := 4

(defun pair-printer (vars term)
  (infix-print-term1 vars)
  (math-space 1)
  (pprinci ":=")
  (math-space 1)
  (infix-print-term1 term))

(declare-fn-printer assign (function assignx-printer))

(defun assignment-printer(term)
  (infix-print-term1 (cadr term))
  (math-space 1)
  (pprinci ":=")
  (math-space 1)
  (infix-print-term1 (caddr term)))

(declare-fn-printer assignment (function assignment-printer))

(defun nth-printer (term)
  (let( (n (cadr term))
	(expr (caddr term)))
    (if (consp term)
	(parens-print-term1 expr)
	(infix-print-term1 expr))
    (pprinci *lbracket*)
    (infix-print-term1 n)
    (pprinci *rbracket*)))

(declare-fn-printer nth (function nth-printer))

(defun list-printer (term)
  (pprinci *lbracket*)
  (set-margin)
  (print-list (cdr term))       ;;(list 'a 'b 'c) 
  (pop-margin)
  (pprinci *rbracket*))

(declare-fn-printer list (function list-printer))
 
(defun true-list-printer(terms)
  (cond ((consp (Car terms))
	     	(infix-print-term1 (cons 'list terms)))
	(t (infix-print-term1 terms))))

(defvar *backslash* "\\")

(defun assoc-printer (term)
  (if (and (consp term) (> (length term) 2))
      (default-infix-printer (list (nth 0 term) (nth 2 term) (nth 1 term)) *backslash*)
      ;; The else case should never happen.
    (default-infix-printer term *backslash*)))    

(setf (get 'assoc 'precedence) 10)
(push (list 'assoc (function assoc-printer)) *fn-alist*)
(push 'assoc *infix-ops*)

(defvar *dots* "...")

(defun evisceration-printer (term)
  (pprinci *dots*))

(declare-fn-printer :evisceration-mark (function evisceration-printer))

(defun mv-printer-with-types (vars types)
  (let (type)
    (if (>= (length vars) 2) (pprinc *mv-bracket-left*))
    (sloop for tail on vars do
	   (progn
	     (if (symbolp (car tail))
		 (funcall (get-atom-printer (car tail)) (car tail))
	       (infix-print-term1 (car tail)))
	     (setf type (assoc (car tail) types))
	     (if type 
		 (progn (pprinci " : ")
			(infix-print-term1 (cadr type))))
	     
	     (if (cdr tail) (pprinci ", " *tt-size*))))
	 (if (>= (length vars) 2) (pprinc *mv-bracket-right*))))

(defun mv-let-printer (term)
  ;; (mv-let (arg1 ... argn) value dcl* form)
 
  (advise-break-if-testing)
  (let ((*top-parens-eliminable* t)
	(vars (cadr term))
	(value (caddr term))
	(form (car (last term)))
	(dcls nil)
	(dcls-type nil))
    (sloop for x in (cdddr term)        ;; extract declare
	   do (if (or (stringp x)
		      (equal x form))
		  nil
		(setq dcls (append dcls (list x)))))
    (bold-sym-printer *lbrace* 2)
    (set-margin)
    (if dcls 
	(setf dcls-type (extract-declare-types dcls)))
    (assign-printer vars value dcls-type)
    (to-current-margin)
    (infix-print-term1 form)
    (bold-sym-printer *rbrace* 2)
    (pop-margin)))

(declare-fn-printer mv-let (function mv-let-printer))

(defun make-alist-pairlist (l)
  (if (not (consp l))
      l
      (cons (list (caar l) (cdar l))
            (make-alist-pairlist (cdr l)))))

(defun type-printer (term)
  (infix-print-term1 (cadr term))
  (pprinc " : ")
  (infix-print-term1 (caddr term)))

(declare-fn-printer type   (function type-printer))

(defvar *forall* 'forall)
(defvar *exists* 'exists)

(defun forall-printer (term)
  (let  ((vars (nth 1 term))
	(body (nth 2 term)))
    (pprinci *forall*)
    (cond ((atom vars)
           (funcall (get-atom-printer vars) vars))
          (t (sloop for tail on vars do
                      (funcall (get-atom-printer (car tail)) (car tail))
                      (cond ((cdr tail)
                             (pprinci ", "))))))
    (pprinc ": ")
    (math-space 1)
    (infix-print-term1 body)))

(declare-fn-printer forall (function forall-printer))

(defun exists-printer (term)
  (let ((vars (nth 1 term))
	(body (nth 2 term)))
    (pprinci *exists*)
    (cond ((atom vars)
	   (funcall (get-atom-printer vars) vars))
          (t (sloop for tail on vars do
		    (funcall (get-atom-printer (car tail)) (car tail))
		    (cond ((cdr tail)
			   (pprinci ", " *tt-size*))))))
    (pprinc ": ")
    (math-space 1)
    (infix-print-term1 body)))

(declare-fn-printer exists (function exists-printer))

(defvar *eviscerate* nil)

(defun body-printer (term)
  (pprinci *lbrace*)
  (set-margin)
  (if *eviscerate* (pprinci "...") (infix-print-term1 (cadr term)))
  (pop-margin)
  (pprinci *rbrace*))

(declare-fn-printer body  (function body-printer))

(defun evisc-printer (term)
      (pprinci "..."))

(declare-fn-printer :evisceration-mark (function evisc-printer))

;; Arrays. Jimmy 
;; 

(defun cut-quote (Term)
    term)
;; (cond ((atom term) term)
;;       ((quote-p term) (eval term))
;;       (t term)))    ;;reduntant??

(defun aref1-printer (term)
  (let ((name (nth 1 term)))
    (pprinci "Aref1(")
    (print-list (cons (cut-quote name) (cddr term)) ", ")
    (pprinci ")")))


(defun aref2-printer (term)
  (let ((name (nth 1 term)))
    (pprinci "Aref2(")
    (print-list (cons (cut-quote name) (cddr term)) ", ")
    (pprinci ")")))




(declare-fn-printer aref1 (function aref1-printer))
(declare-fn-printer aref2 (function aref2-printer))

(defun aset1-printer(term)
 (let ((name (nth 1 term)))
    (pprinci "Aset1(")
    (print-list (cons (cut-quote name) (cddr term)) ", ")
    (pprinci ")")))

(defun aset2-printer(term)
 (let ((name (nth 1 term)))
    (pprinci "Aset2(")
    (print-list (cons (cut-quote name) (cddr term)) ", ")
    (pprinci ")")))

(declare-fn-printer aset1 (function aset1-printer))
(declare-fn-printer aset2 (function aset2-printer))
(defun array1p-printer (term)
  ;; (array1p name alist)
  (let ((name (nth 1 term)))
    (pprinci "Array1p(")
    (print-list (cons (cut-quote name) (cddr term)) ", ")
    (pprinci ")")))


(defun array2p-printer (term)
  ;; (array1p name alist)
  (let ((name (nth 1 term)))
    (pprinci "Array2p(")
    (print-list (cons (cut-quote name) (cddr term)) ", ")
    (pprinci ")")))

(declare-fn-printer array1p (function array1p-printer))
(declare-fn-printer array2p (function array2p-printer))

(defun compress-1-printer (term)
  ;;;(compress1 name alist)
  (let ((name (nth 1 term)))
    (pprinci "Compress1(")
    (print-list (cons (cut-quote name) (cddr term)) ", " )
    (pprinc ")")))



(defun compress-2-printer (term)
  ;;;(compress1 name alist)
  (let ((name (nth 1 term)))
    (pprinci "Compress2(")
    (print-list (cons (cut-quote name) (cddr term)) ", " )
    (pprinc ")")))



(declare-fn-printer compress1 (function compress-1-printer))
(declare-fn-printer compress2 (function compress-2-printer))

(defun incompatible-printer(term)
  (pprinci "incompatible( ")
  (print-rune-string (cadr term))
  (pprinci ", ")
  (print-rune-string (caddr term))
  (pprinci ")"))

(declare-fn-printer incompatible (function incompatible-printer))

;; THIS WON'T WORK!  @ = append

(defun @-printer(term)
  (pprinci "@")
  (if (= 2 (length term))
      (infix-print-term1 (cadr term))
    (infix-print-term1 (Cdr term))))

(declare-fn-printer @ (function @-printer))


