; ACL2 Version 1.9

; Copyright (C) 1989-96 Computational Logic, Inc. (CLI).  All rights reserved.

; Use of this software constitutes agreement with the terms of ACL2
; license agreement, found in the file LICENSE.

;; Init file for infix.lisp in HTML mode.
;;  Apr 24 97 MKS 

(in-package "ACL2")

(require "infix")

(format *terminal-io* "Loading the infix html-init file.")

(defparameter *begin-comment-env*  "<!--- ")
(defparameter *end-comment-env*    " --- >")

;; Mode should actually be set before this file is loaded.

(infix-settings :mode                   "html"
		:extension              "html"   
		:op-location            'front  
		:comment-format         'cl
		:format-!-in-comments   nil    
		:no-index               nil
		:no-index-calls         nil )


;; Increase this number to more accurately allow for proper output width.
(defparameter *rightmost-char-number* 90)
(defparameter *default-chars-wide* 90)
(defparameter *latex-indent-number-limit* 80)

(defparameter nqread-normal-clause-enders nqread-default-normal-clause-enders)

; THE HTML PRELUDE.

(defparameter *standard-prelude*
  "<!DOCTYPE HTML PUBLIC>
<!--Converted with IACL2 HTML mode ( Apr 24 97 )
    Michael K. Smith (mksmith@cli.com), Computational Logic, Austin, TX -->
<HTML>
<HEAD>
<TITLE>None</TITLE>
<META NAME=\"description\" CONTENT=\"No Title\">
<META NAME=\"resource-type\" CONTENT=\"document\">
<META NAME=\"distribution\" CONTENT=\"global\">
</HEAD>
<BODY LANG=\"EN\" >
")

(defparameter *standard-postlude*
 "
</BODY>
</HTML>
")

(defparameter *example-prelude*
  (concatenate 'string *standard-prelude*
"
Here is a summary of the conventional syntax (~a) in terms of the official syntax
of the ACL2 logic.<p>

<OL>
<li> Variables are printed in italics, unless specified otherwise in the table below.

<li> Function application.  For any function symbol for which special
syntax is not given below, an application of the symbol is printed with
the usual notation; e.g., the term !v(fn x y z) is
printed as !t(fn x y z).  Note that the function symbol is printed in
Roman.  In the special case that !qc is a function symbol of no
arguments, i.e., it is a constant, the term !v(c) is printed merely as
!t(c), in small caps, with no trailing parentheses.  Because variables are printed in
italics, there is no confusion between the printing of variables and
constants.

<li> Other constants.
Quoted constants are printed in the ordinary syntax of the ACL2 logic,
in a `typewriter font.'  For example,
<tt>'(a b c)</tt> is still printed just that way.  <tt>#b001</tt> is printed
as !t#b001, <tt>#o765</tt> is printed as !t#o765, and <tt>#xa9</tt> is printed as
!t#xa9, representing binary, octal and hexadecimal, respectively."))

(defparameter *begin-example-table* "<table>~%~
<tr><th>ACL2 Syntax & Conventional Syntax</th><th></th></tr>
")

(defparameter *end-example-table* "
</table>")

(defparameter *example-table-size*  32)

(defparameter *example-postlude* "")

;; BASIC BRACKETS AND THEIR QUOTED VERSION.

(defparameter *dots* "...")
(defparameter *equal-sign* " = ")
(defparameter *lb-sign* " # ")

(defparameter *begin* "")
(defparameter *end*   "")

(defparameter *lbrace* "{")
(defparameter *rbrace* "}")

(defparameter *dots* "...")

(defparameter *begin* "{")
(defparameter *end*   "}")

(defparameter *lbracket* "[")
(defparameter *rbracket* "]")

(defparameter *backslash* "\\")

;; NEWLINE PARAMETERS

(defparameter *newline-in-env* "
")
(defparameter *newline-in-text* "
")

(defparameter *force-newline-in-env* "<br>")
(defparameter *force-newline-in-text* "<br>")

;; ENVIRONMENT BEGIN-END PAIRS

(defparameter *begin-index* "")
(defparameter *end-index* "")

(defparameter *begin-label* "<A NAME=\"")
(defparameter *end-label* "\"></A>")

;; Note that ref must insert #
;; Reason: URL#foo vs. #foo
(defparameter *begin-reference* "<A HREF=\"")
(defparameter *end-reference* "\"></A>")

;; Only for print-function-name
(defparameter *mid-reference* "\">")
(defparameter *end-reference2* "</A>")

(defparameter *begin-text-env* "")
(defparameter *end-text-env*  "<p>")

(defparameter *begin-verbatim-env* "<pre>")
(defparameter *end-verbatim-env*  "</pre>")

(defparameter *begin-verb-env* "<code>")
(defparameter *end-verb-env*  "</code>")

(defparameter *begin-format-env*  "<pre>")
(defparameter *end-format-env*    "</pre>")

(defparameter *begin-emphasis-env*  "<em>")
(defparameter *end-emphasis-env*    "</em>")

(defparameter *begin-section-env*  "<h2>")
(defparameter *end-section-env*    "</h2>")

(defparameter *begin-subsection-env*  "<h3>")
(defparameter *end-subsection-env*    "</h3>")

(defparameter *begin-tt-env* "<tt>")
(defparameter *end-tt-env*   "</tt>")

(defparameter *tt-size* 1)

(defparameter *begin-string-env*  "")
(defparameter *end-string-env*    "")

(defparameter *begin-bold-env* "<b>")
(defparameter *end-bold-env*   "</b>")

(defparameter *begin-italic-env* "<i>")
(defparameter *end-italic-env*   "</i>")

(defparameter *begin-sc-env* "")
(defparameter *end-sc-env*   "")

;; These should ONLY be used by print-examples to help us document
;; the syntax in syntax.tex.

(defparameter *column-separator* "</td><td>")

;(defparameter *column-start* "<td>")
;(defparameter *column-end* "</td>")

(defparameter *begin-enumerate-env* "<ol>")
(defparameter *end-enumerate-env* "</ol>")

(defparameter *begin-item* "<li> ")
(defparameter *end-item* "")

(defparameter *mv-bracket-left* "&lt;")
(defparameter *mv-bracket-right* "&gt;")

(defparameter *forall* "all ")
(defparameter *exists* "some ")


;; TABBING ENVIRONMENT AND TAB OPERATIONS

(defparameter *begin-group-tabbing-env* "<pre>")

(defparameter *begin-tabbing-env* "<pre>")
(defparameter *end-tabbing-env* "</pre>")

(defparameter *new-tab-row* "<br>")

;; Just in case some other mode defined it otherwise.
(defun new-tab-row (&optional followed-by-infix-print-term)
  (declare (ignore followed-by-infix-print-term))
  (pprinc *new-tab-row*))

;; (defun new-tab-row (&optional followed-by-infix-print-term)
;;   (declare (ignore followed-by-infix-print-term))
;;   (pwrite-char #\Newline)
;;   (to-current-margin))

(defparameter *tab* "	")
(defparameter *flush* "        ")

(defparameter *tab-list* nil)

(defparameter *set-margin* "")
(defparameter *set-tab*    "")
(defparameter *pop-margin* "")

(defparameter *op-length* nil)
(defparameter *default-op-tab-space* "	")
(defparameter *indent-string* "    ")
(defparameter *default-indent* 4)

(defun get-op-width-string (op)
  (if (numberp op) op (length op)))

(defparameter *noindent* "")

; (defparameter *testmargin* nil)

;; Problem :
;; <pre> <pre> a </pre> b </pre>
;; ->
;; <pre> a b </pre>

(defvar *html-pre-stack* nil)

(defun begin-tabbing ()
  (if (member '(pre) *tab-list*)
      (push nil *html-pre-stack*)
      (progn (princ *begin-tabbing-env*) (push t *html-pre-stack*)))
  (setq *tab-list* (cons '(pre) *tab-list*))
  (to-current-margin))

(defun begin-group-tabbing ()
  (if (member '(pre) *tab-list*)
      (push nil *html-pre-stack*)
      (progn (princ *begin-group-tabbing-env*) (push t *html-pre-stack*)))
  (setq *tab-list* (cons '(pre) *tab-list*))
  (to-current-margin))

(defun end-tabbing ()
  (let ((flag (pop *html-pre-stack*)))
    (setq *tab-list* (cdr (member-equal '(pre) *tab-list*)))
    (if flag (princ *end-tabbing-env*))))

(defun blankline ()
  (cond ((and (eql *infix-loc* (get-margin))
	      (eql *prev-line-loc* *infix-loc*)))
	((eql *infix-loc* (get-margin))
	 (if (and (null *tab-list*)
		  (not (member *current-environment* '(format verbatim))))
	     (princ *end-text-env*))
	 (newlinex t))
	(t (if (and (null *tab-list*)
		  (not (member *current-environment* '(format verbatim))))
	       (princ *end-text-env*))
	   (newlinex t) (newlinex t))))

(defun begin-flushright ()
  (pprinc *flush*))

(defun end-flushright ()
  (pprinc *tab*))

(defun begin-normal-text ()
  (pprinc *begin-text-env*))

(defun end-normal-text ()
  (pprinc *end-text-env*))

(defun flushright (form)
  (begin-flushright)
  (pprinc form)
  (end-flushright))

(defun do-tab ()

; The *tab-list* is either NIL, ((LM loc) ...), or ((TAB loc) ...)
; Only tab if there is something to tab to.

  (cond (*do-not-use-tabs* (pprinc " "))
	((and *tab-list* (eq (caar *tab-list*) 'tab))
	 (spacesx (- (cdar *tab-list*) *infix-loc*)) ;(pprinc *tab*)
	 )
	(t (pprinc " "))))

(defun set-tab (&optional op)

; Generate instructions to set a tab at the current location.
; `Punt' if we hit the limit, by throwing all the way out.

  (cond (*do-not-use-tabs* nil)
        (t (cond ((= *tabs-in* *latex-indent-number-limit*) ;Let Latex-Limit hold for Text also.
                  (throw 'taboverflow t)))
           (setq *tabs-in* (1+ *tabs-in*))
	   (cond ((and op (get-op-width-string op))
		  (spacesx (+ 1 (get-op-width-string op))))
		 (t (pprinc *default-op-tab-space*)))
           (push (cons 'tab *infix-loc*) *tab-list*)
	   ;; (if *testmargin* (format t "~%Setting tab : ~a~%" *tab-list*))
           (pprinc *set-tab*))))

(defun pop-tab ()
  ;; We don't really remove tabs from the formatted env.
  ;; Just track them in Lisp.
  ;; Generate command to `tab to one tab less in'.
  ;; Do not pop tabs beyond left margin or past the beginning of a tabbing env.
  (cond (*do-not-use-tabs* nil)
	((and *tab-list* (eq (caar *tab-list*) 'tab))
	 (setq *tabs-in* (1- *tabs-in*))
	 (pop *tab-list*)
	 ;; (if *testmargin* (format t "~%Popped tab : ~a~%" *tab-list*))
	 )
	(t nil)))

;; (defvar *tracing-margins* nil)

(defun set-margin ()
  ;; Generate instructions to set the current indentation.
  ;; (if *tracing-margins* (or *testing* (princ "|[")))
  (cond (*do-not-use-tabs* nil)
        (t (setq *tabs-in* (1+ *tabs-in*))
	   (setq *left-margin* *infix-loc*)
	   ;(format t "~%(set-margin) = ~d." *infix-loc*)
	   (push (cons 'lm *infix-loc*) *tab-list*))))

(defun pop-margin ()
  ;; Generate command to `return to one margin less in'.
  ;; For bookkeepping reasons we pop tabs after the margin.
  ;; (if *tracing-margins* (or *testing* (princ "]|")))
  (cond (*do-not-use-tabs* nil)
	((and *tab-list* (eq (caar *tab-list*) 'tab))
	 (pop-tab)
	 (pop-margin))
	((and *tab-list* (eq (caar *tab-list*) 'lm))
	 (setq *tabs-in* (1- *tabs-in*))
	 (pop *tab-list*)
	 (setq *left-margin* (or (get-margin2 *tab-list*) 0))
	 ;(format t "~%(pop-margin) = ~d." *left-margin*)
	 )
	(t nil)))

(defun clear-margins ()
  (setq *tab-list* nil)
  (setq *tabs-in* 0)
  (setq *infix-loc* 0)
  (set-margin))

;; FONTS

(defparameter *function-font* "")	; Roman.  sc = small caps
(defvar *do-not-index* nil)

(defun roman-font (term) (print-atom term))

(defparameter *begin-event-name* *begin-bold-env*)
(defparameter *end-event-name* *end-bold-env*)

;; Should be superceded
(defun prefix-event-name-printer (x)
  (princ *begin-bold-env*) (print-atom x) (princ *end-bold-env*))

(defun print-function-name (sym) 
  (cond (*testing*   (roman-font sym)) 
	((or *do-not-index* *do-not-index-calls*) (roman-font sym))
        ((do-not-index-event-of sym) (roman-font sym))
	((symbolp sym)
	 (let ((reference (assoc sym *saved-urls*)))
	   (if (not reference)
	       (roman-font sym)
	       (progn (pprinc *begin-reference*)
		      ;; The xrefs between files are likely 
		      ;; to occur in the same directory.
		      ;; We account for this one special case.
		      ;; More general solution when initial
		      ;; paths partially match is needed.
		      (pprinc
		       (cond ((not (string= (cdr reference) *current-url-file*))
			      (cond ((string=
				      (directory-namestring (cdr reference))
				      (directory-namestring *current-url-file*))
				     (file-namestring (cdr reference)))
				    (t (cdr reference))))
			     (t (cdr reference))))
		      (pprinc "#")
		      (print-atom sym 0) ; note we suppress width here.
		      (pprinc *mid-reference*)
		      (roman-font sym)
		      (pprinc *end-reference2*)))))
	(t (roman-font sym))))


;; MATH ENV AND OPERATORS

(defparameter *neg-str*  "&not;")

;; MUST be a format directive.  Used by make-inifix-op etc.

(defparameter *math-format* "~a")
(defparameter *math-begin* "")
(defparameter *math-end* "")

;; These must be enclosed in a math env.  
(defparameter *math-thick-space* "  ")
(defparameter *math-space* " ")
(defparameter *math-thin-space* "")

(defparameter *subscript* "")

(defparameter *begin-subscript* "<sub>")
(defparameter *end-subscript* "</sub>")

;; MISC.

(defparameter *newpage* "<hr>")

(defparameter *comma-atsign* ",@")
(defparameter *caret* "^")

(defparameter *list-seperator* ", ")
(defparameter *dotted-pair-seperator* " . ")

(defparameter *no-tab-event-trailer* "~%")
(defparameter *print-default-event-header* "")
(defparameter *print-default-lisp-header* "")

(defparameter *print-default-command-header*  "")
(defparameter *no-tab-command-trailer* "~%")

;; OTHER FUNCTIONS

;; Includes <, >, &, " 
(defparameter doc-special-chars  "<>&#")
(defparameter doc-index-specials "<>&#")

(defparameter doc-other-chars   nil)

(defun handle-special-chars (char)
  (cond ((eql char #\") (pprinci "&quot;"))
	((eql char #\<) (pprinci "&lt;"))
	((eql char #\>) (pprinci "&gt;"))
	((eql char #\&) (pprinci "&amp;"))
	(t (pwrite-char (if (eq *print-case* :downcase)
			    (char-downcase char)
			    char)))))

(defun handle-special-chars-in-string (char)
  (cond ((eql char #\") (pprinci "&quot;"))
	((eql char #\<) (pprinci "&lt;"))
	((eql char #\>) (pprinci "&gt;"))
	((eql char #\&) (pprinci "&amp;"))
	(t (pwrite-char char))))

;; Index entries

(defun index (x &optional subkind)
  ;; Subkind must be string or atom.
  (cond (*do-not-index* nil)
	((and subkind
	      (or (string= subkind "defined")
		  (string= subkind "theorem")))
	 (pprinc *begin-label*)
	 (cond ((stringp x) (print-string x 0))
	       (t (print-atom x 0)))
	 (pprinc *end-label*))
	(subkind)
	(t (pprinc *begin-reference*)
	   (cond ((stringp x) (print-string x 0))
		 (t (print-atom x 0)))
	   (pprinc *end-reference*))))

(defun skip-index-entries ()
  ;; We are looking at a backslash. In Tex mode we need to skip to the end 
  ;; of the entry, because we may add !'s.  In Text mode this is just NIL.
  nil)

(defun adjust-tabbing-env ()
  ;; We are looking at a backslash. In Tex mode we may need to check for 
  ;; some special cases.
  ;; In Text mode this is just NIL.
  nil)

(defparameter acl2-char-subst-table 
 '((#\< #\& #\l #\t #\;)
   (#\> #\& #\g #\t #\;)
   (#\" #\&  #\q #\u #\o #\t #\;)
   (#\& #\& #\a #\m #\p #\;)))

(defparameter acl-markup-table
  ;; Note on args: c is the current file name.
  ;; I don't understand w.
  '(("-" nil .    "--")
    ("B" nil .  "<b>~st</B>") ;bold font
    ("BF" nil .  "~%<pre>")   ;begin format --
                              ; -- like verbatim, but if possible don't change font
    ("BID" nil .    "")       ;begin implementation dependent
    ("BQ" nil .  "~%<blockquote>") ;begin quotation
    ("BV" nil .  "~%<pre>")        ;begin verbatim
    ("C" nil .  "<code>~st</code>");code, often preferred even if we have invisible link
    ("EF" nil .  "</pre>~%")       ;end format
    ("EID" nil .    "")            ;end implementation dependent
    ("EM" nil .  "<em>~st</em>")   ;emphasis (normaly italics if available)
    ("EQ" nil .  "</blockquote>~%");end quotation
    ("EV" nil .  "</pre>~%")       ;end verbatim
    ("GIF" nil . "<img src=~st>")  ;gif files; e.g., ~gif[\"foo.gif\" align=top]
    ("I" nil .  "<i>~st</i>")      ;italics font
    ("ID" nil .    "")             ;implementation dependent
    ;;invisible link, for true hypertext environments
    ("IL" t .  "<a href=\"~sc#~sp\">~st</a>") 
    ;;invisible link, but use code if such links don't show
    ("ILC" t .  "<code><a href=\"~sc#~sp\">~st</a></code>")
    ("L" t .  "See <a href=\"~sc#~sp\">~st</a>") ;link at beginning of sentence
    ("NL" nil .  "<br>~%")         ;newline
    ("PAR" nil .  "<p>~%~%")       ;paragraph mark, of no significance at the terminal
    ;;parenthetical link, i.e., link not at beginning of sentence
    ("PL" t .  "see <a href=\"~sc#~sp\">~st</a>")
    ("SC" nil .  "~sT")            ;(small, if possible) caps
    ("ST" nil .  "<strong>~st</strong>") ;strong emphasis
    ("T" nil .  "<tt>~st</tt>")    ;typewriter font
    ("TERMINAL" nil . "")          ;terminal only, otherwise ignore
    ("WARN" nil . "<a href=\"~sw\"><img src=twarning.gif></a>")
    ("CLICK-HERE" t .  "Click <a href=\"~sc#~sp\">here</a>")
    ("PCLICK-HERE" t .  "click <a href=\"~sc#~sp\">here</a>")
    ("FLY" t .  "<a href=\"~sc#~sp\"><img src=flying.gif></a>")
    ("WALK" t .  "<a href=\"~sc#~sp\"><img src=walking.gif></a>")
    ))

;; Primitive stuff.

(defun get-margin ()
  (or (get-margin2 *tab-list*) *left-margin*))

(defun get-margin2 (tl)
  (let ((setting (car tl)))
    (cond ((null setting) nil)
	  ((eq (car setting) 'lm) (cdr setting))
	  (t (get-margin2 (cdr tl))))))

(defun spacesx (n)
  (dotimes (i n) (pwrite-char #\ ))
  (setq *infix-loc* (+ *infix-loc* n)))

(defun line-return ()
  (setq *prev-line-loc* *infix-loc*)
  (pwrite-char #\Newline)
  (setq *infix-loc* 0)
  (spacesx (get-margin)))

(defun newline-in-env (&optional force) 
  (if force (line-return)
      (if (not (eql *infix-loc* (get-margin)))
	  (line-return))))

(defun newline-in-text (&optional force) 
  (if force (line-return)
      (if (not (eql *infix-loc* (get-margin)))
	  (line-return))))

(defun newlinex (&optional force)
  (newline-in-env force))

(defun to-current-margin ()
  (let ((margin (get-margin)))
    (cond ((eql *infix-loc* margin) nil)
	  ((< *infix-loc* margin) 
	   (spacesx (- margin *infix-loc*)))
	  (t (line-return)))))
