; ACL2 Version 1.9

; Copyright (C) 1989-97 Computational Logic, Inc. (CLI).  All rights reserved.

; Use of this software constitutes agreement with the terms of ACL2
; license agreement, found in the file LICENSE.

(in-package "ACL2")

;; The init file should be compiled.
(load-base "latex-init" t)

;                                     INFIX-OPS

; infix-ops (infix operators) should be function symbols of two or more
; arguments for which it is desired that one symbol come out between every
; adjacent pair of arguments.  E.g., invoking (make-infix-op plus "+") causes
; the term (plus a b c d) to be printed as (a $+$ b $+$ c $+$ d).  Invoking
; (make-infix-op equal "=" "\\not=") causes the term (equal x y) to be printed
; as (x $=$ y) and it also causes the term (not (equal x y)) to be printed as
; (x \not= y).

; Thus, for example, if one introduces a new function, say join, and wants to
; print terms of the form (join x y) as (x \bigtriangledown y), cf. p. 44 of
; the Latex manual, then one should invoke:

;   (make-infix-op join "\\bigtriangledown")

; from Lisp.  That is all that need be done to cause infix-file to subsequently
; print `join' terms this way.

; Note that throughout the following examples, we have used two backslashes to
; get one because, in Common Lisp, backslash is a character for quoting other
; characters.

;; PRECEDENCE:
;; These infix operators should already have precedence
;; set from text-syntax.lisp.  Only when you are introducing
;; new operators might you want to provide the optional
;; precedence argument to MAKE-INFIX-OP.  Or do an explicit
;;  (setf (get 'not 'precedence) 16)
;; Even then, its probably a mistake.  Less misleading 
;; to let it default to 0.  

;; Higher numbers represent higher precedence.

;; 20 FUNCTION application (highest precedence)
;; 16 unary prefix, e.g. not, prefix-minus
;; 14 **
;; 12 * /
;; 10 + - union intersect difference
;;  8 >= <= > < e0-ord-<
;;  6 = ~=       
;;  4 & |
;;  2 -> <->               (lowest precedence)

; Equalities:

;;          Text         Latex
;; equal -> x = y        x = y
;; =     -> x == y       x =_{n}
;; eq    -> eq(x,y)      x =_{eq}
;; eql   -> eql(x,y)     eql(x,y)

(make-infix-op equal          "="      "\\not=")
(make-infix-op eq             "=_{eq}" "\\not=_{eq}")
(make-infix-op =              "=_{n}"  "\\not=_{n}")
(make-infix-op /=             "\\not=_{n}" nil)

(make-infix-op <              "<"        "\\not<")
(make-infix-op e0-ord-<       "\\leq_\\epsilon"    "\\not\\leq_\\epsilon")
(make-infix-op <=             "\\leq"    "\\not\\leq")
(make-infix-op >              ">"        "\\not>")
(make-infix-op >=             "\\geq"    "\\not\\geq")

(make-infix-op /              "/" nil)
(make-infix-op *              "*" nil)
(make-infix-op BINARY-*       "*_{2}" nil)
(make-infix-op +              "+" nil)
(make-infix-op BINARY-+       "+_{2}" nil)
(make-infix-op -              "-" nil)

(make-infix-op rem            "\\%" nil)

(make-infix-op cons           ":>" nil)
(make-infix-op append         " @ " nil)
(make-infix-op BINARY-APPEND  "@_{2}" nil)

;; assoc uses *backslash* (assoc a b) => b \ a

(make-infix-op union-theories          "\\cup" nil)
(make-infix-op set-difference-theories "{\\rm\\bf{less}}" nil)
(make-infix-op intersection-theories   "\\cap" nil)
(make-infix-op incompatible "incompatible" nil)

;;(make-infix-op union          "\\cup" nil)

(make-infix-op member-equal   "\\in"     "\\not\\in")
(make-infix-op member         "\\in_{eq}" "\\not\\in_{eq}")

;; Not depends on value of *neg-str*
(make-infix-op and            "\\wedge" nil)
(make-infix-op or             "\\vee" nil)

(make-infix-op implies        "\\rightarrow" nil)
(make-infix-op iff            "\\leftrightarrow" nil)

;; (make-infix-op congruent      "\\cong" nil 2)

(setf (get 'if 'precedence) 2)
(setf (get 'if* 'precedence) 2)
(setf (get 'cond 'precedence) 2)    
(setf (get 'case-match 'precedence) 2)
(setf (get 'case 'precedence) 2)
(setf (get 'elif 'precedence) 2)  

(defun 1+-printer (term)
  (infix-print-term1 (list '+ 1 (cadr term))))

(declare-fn-printer 1+ (function 1+-printer))

(defun 1--printer (term)
  (infix-print-term1 (list '- (cadr term) 1)))

(declare-fn-printer 1- (function 1--printer))
;;(declare-fn-printer sub1 (function 1--printer))


;	UNARY-PREFIX-OPS, UNARY-SUFFIX-OPS, and UNARY-ABS-OPS

; Use make-unary-prefix-op and make-unary-suffix-op only for function symbols
; of one argument.  The string str (or *neg-str*) will be printed before or
; after the argument.

; unary-suffix-ops should be unary function symbols.
; (make-unary-suffix-op foo x str) makes (foo x) print as `x $str$'.

(make-unary-suffix-op numberp "\\in {\\rm\\bf{N}}" "\\not\\in {\\rm\\bf{N}}")
(make-unary-suffix-op zerop   "=_{n} {\\tt{0}}"    "\\not=_{n} {\\tt{0}}")
(make-unary-suffix-op zp   "=_{i} {\\tt{0}}"       "\\not=_{i} {\\tt{0}}")

; Examples of make-unary-suffix-op.
; (make-unary-suffix-op sub1    "-\\;1")

; unary-prefix-ops should be unary function symbols. 

; (make-unary-prefix-op foo str) makes (foo x) print as `$str$ x'.

; Examples of make-unary-prefix-op.

; (make-unary-prefix-op 1+    "1\\;+\\;")
; (make-unary-prefix-op minus "-")

; unary-abs-ops should be unary function symbols.
; (make-unary-abs-op abs "|" "|")

(make-unary-abs-op mod "\\lfloor\\;" "\\;\\rfloor")

; To create syntax like that for absolute value, use 
; (make-unary-absolute-op lhs-str rhs-str)
; where lhs-str and rhs-str are the strings to print on the
; left and right of the argument.  (make-unary-abs-op foo str1 str2) makes (foo
; x) print as `str1 x str2'. 

(defparameter *do-not-index-calls-of*
  (union *do-not-index-calls-of*
	 '(implies and or not if cond
		   implies iff union
		   eq = equal
		   le < > ge leq <= geq >= lessp e0-ord-< 
		   greaterp
		   member append 
		   + - * / rem mod
		   union intersection
		   car cadr cdr cddr caddr cons consp
		   disable force integerp member-equal null
		   stringp symbolp true-listp alistp)))

;; ----------------------------------------------------------------------
;;
;;                           SOME POSSIBLE EXTENSIONS

(defun simple-extension ()

; Here are a few examples of normal mathematical notation for functions not in
; the bootstrap.  Invoke this function to put these into effect.

  (make-unary-abs-op    abs        "\\mid" "\\mid")

  (make-unary-suffix-op fact       "{\\rm{!}}")


  (make-infix-op        subsetp     "\\subset"         "\\not\\subset")
  (make-infix-op        intersect   "\\cap"))
 
(defun dmg-syntax ()

; Here are some examples once tentatively proposed by David Goldschlag for his
; work.  Invoke this function to put these into effect.

; prefix-multiple-op's should be function symbols that take as many arguments as
; make-prefix-multiple-op is given arguments.  (make-prefix-multiple-op foo str1
; str2) makes (foo x y) print as ($str1$ x $str2$ y).  That is, the first string
; comes first.

  (make-prefix-multiple-op invariant         "\\Box"           "{\\rm\\bf{in}}")
  (make-prefix-multiple-op eventually-stable "\\Diamond\\Box"  "{\\rm\\bf{in}}")

; infix-multiple-op's should be function symbols that take one more argument
; than make-infix-multiple-op is given arguments.  (make-infix-multiple-op foo
; str1 str2) makes (foo x y z) print as (x $str1$ y $str2$ z).  That is, the
; strings are placed between adjacent arguments.

  (make-infix-multiple-op leads-to       "\\mapsto"              "{\\rm\\bf{in}}")
  (make-infix-multiple-op unless         "{\\rm\\bf{unless}}"    "{\\rm\\bf{in}}")
  (make-infix-multiple-op ensures        "{\\rm\\bf{ensures}}"   "{\\rm\\bf{in}}")
  (make-infix-multiple-op e-ensures      "\,${\\rm\\bf{e-ensures}}$\," "{\\rm\\bf{for}}"
                          "{\\rm\\bf{in}}")
  (make-infix-multiple-op n              "\\leadsto"             "{\\rm\\bf{by}}")
  (make-infix-multiple-op initial-condition "{\\rm{\\bf{initially\\;in}}}"))

; Undoing.  To cause applications of a function symbol fn to be printed in the
; default way, i.e., fn(x, y), invoke (clean-up 'fn).


