
(in-package "ACL2-AGP")

(include-book "acl2-asg")

(encapsulate
  ((equiv ( x y ) t)
   (pred ( x ) t)
   (op ( x y ) t)
   (id ( ) t)
   (inv ( x ) t))

  (local
    (defun
      equiv ( x y )
      (equal x y)))

  (local
    (defun
      pred ( x )
      (declare (xargs :verify-guards t))
      (or (equal x t)
          (equal x nil))))

  (local
    (defun
      op ( x y )
      (declare (xargs :guard (and (pred x)
                                  (pred y))))
      (and (or x y)
           (not (and x y)))))

  (local
    (defun 
      id ( )
      nil))

  (local
    (defun
      inv ( x )
      (declare (xargs :guard (pred x)))
      x))

  (defthm
    Equiv-is-an-equivalence
    (and (acl2::booleanp (equiv x y))
         (equiv x x)
         (implies (equiv x y) 
                  (equiv y x))
         (implies (and (equiv x y)
                       (equiv y z))
                  (equiv x z)))
    :rule-classes (:EQUIVALENCE
                   (:TYPE-PRESCRIPTION
                    :COROLLARY
                    (or (equal (equiv x y) t)
                        (equal (equiv x y) nil)))))

  (defthm
    Equiv-1-implies-equiv-op
    (implies (equiv x1 x2)
             (equiv (op x1 y)
                    (op x2 y)))
    :rule-classes :CONGRUENCE)

  (defthm
    Equiv-2-implies-equiv-op
    (implies (equiv y1 y2)
             (equiv (op x y1)
                    (op x y2)))
    :rule-classes :CONGRUENCE)

  (defthm 
    Closure-of-op-for-pred
    (implies (and (pred x)
                  (pred y))
             (pred (op x y))))

  (defthm
    Closure-of-id-for-pred
    (pred (id)))

  (defthm
    Closure-of-inv-for-pred
    (implies (pred x)
             (pred (inv x))))
  
  (defthm
    Commutativity-of-op
    (implies (and (pred x)
                  (pred y))
             (equiv (op x y)
                    (op y x))))

  (defthm 
    Associativity-of-op
    (implies (and (pred x)
                  (pred y)
                  (pred z))
             (equiv (op (op x y) z)
                    (op x (op y z)))))

  (defthm
    Left-unicity-of-id-for-op
    (implies (pred x)
             (equiv (op (id) x)
                    x)))

  (defthm
    Right-inverse-of-inv-for-op
    (implies (pred x)
             (equiv (op x (inv x))
                    (id)))))

(acl2-asg::add-commutativity-2 equiv
                               pred
                               op
                               commutativity-of-op
                               commutativity-2-of-op)

(defthm
  Right-unicity-of-id-for-op
  (implies (pred x)
           (equiv (op x (id))
                  x)))

(defthm
  Left-inverse-of-inv-for-op
  (implies (pred x)
           (equiv (op (inv x) x)
                  (id))))

(local
  (defthm
    Right-cancellation-for-op-iff
    (implies (and (pred x)
                  (pred y)
                  (pred z))
             (iff (equiv (op x z)
                         (op y z))
                  (equiv x y)))
    :rule-classes nil
    :hints (("Subgoal 1"
             :in-theory (disable Equiv-1-implies-equiv-op)
             :use (:instance 
                   Equiv-1-implies-equiv-op
                   (x1 (op x z))
                   (x2 (op y z))
                   (y  (inv z)))))))

(defthm
  Right-cancellation-for-op
  (implies (and (pred x)
                (pred y)
                (pred z))
           (equal (equiv (op x z)
                         (op y z))
                  (equiv x y)))
  :rule-classes nil
  :hints (("Goal"
           :use Right-cancellation-for-op-iff)))

(local
  (defthm
    Left-cancellation-for-op-iff
    (implies (and (pred x)
                  (pred y)
                  (pred z))
             (iff (equiv (op x y)
                         (op x z))                       
                  (equiv z y)))
    :rule-classes nil
    :hints (("Goal"
             :use ((:instance 
                    Right-cancellation-for-op
                    (x z)
                    (z x)))))))

(defthm
  Left-cancellation-for-op
  (implies (and (pred x)
                (pred y)
                (pred z))
           (equal (equiv (op x y)
                         (op x z))
                  (equiv y z)))
  ;rule-classes nil
  :hints (("Goal"
           :use Left-cancellation-for-op-iff)))

(defthm
  Uniqueness-of-id-as-op-idempotent
  (implies (and (pred x)
                (equiv (op x x)
                       x))
           (equiv x (id)))
  :rule-classes nil
  :hints (("Goal"
           :use (:instance
                 Right-cancellation-for-op
                 (y (id))
                 (z x)))))

(defthm
  Uniqueness-of-op-inverses
  (implies (and (pred x)
                (pred y)
                (equiv (op x y)
                       (id)))
           (equiv y (inv x)))
  :rule-classes nil 
  :hints (("Goal"
           :use (:instance
                 Right-cancellation-for-op
                 (x y)
                 (y (inv x))
                 (z x)))))

(defthm
  Involution-of-inv
  (implies (pred x)
           (equiv (inv (inv x))
                  x))
  :hints (("Goal"
           :use (:instance
                 Uniqueness-of-op-inverses
                 (x (inv x))
                 (y x)))))

(defthm
  Uniqueness-of-op-inverses-1
  (implies (and (pred x)
                (pred y)
                (equiv (op x (inv y))
                       (id)))
           (equiv x y))
  :rule-classes nil
  :hints (("Goal"
           :use (:instance
                 Uniqueness-of-op-inverses
                 (y x)
                 (x (inv y))))))

(defthm
  Distributivity-of-inv-over-op
  (implies (and (pred x)
                (pred y))
           (equiv (inv (op x y))
                  (op (inv x)
                      (inv y))))
  :hints (("Goal"
           :use (:instance
                 Uniqueness-of-op-inverses
                 (x (op x y))
                 (y (op (inv x)(inv y)))))))

(defthm
  id-is-its-own-invese
  (equiv (inv (id))
         (id))
  :hints (("Goal"
           :use (:instance
                 Uniqueness-of-op-inverses
                 (x (id))
                 (y (id))))))

(local
  (defthm
    obvious-inv-cancellation
    (implies (and (pred x)
                  (pred y))
             (equiv (op (op x (inv x)) y) y))
    :rule-classes nil))

(defthm
  inv-cancellation-on-right
  (implies (and (pred x)
                (pred y))
           (equiv (op x (op y (inv x)))
                  y))
  :hints (("Goal"
           :use obvious-inv-cancellation
           :in-theory (acl2::disable 
                       right-inverse-of-inv-for-op))))

(defthm
  inv-cancellation-on-left
  (implies (and (pred x)
                (pred y))
           (equiv (op x (op (inv x) y))
                  y)))
