

In-Package("ACL2-AGP");

Include("acl2-asg");

Module 
{equiv(x,y) : t;
 pred(x) : t;
 op(x,y) : t;
 id() : t;
 inv(x) : t;}
{
 local FUNCTION equiv(x,y) {x = y};
 
 local FUNCTION pred(x) {x = t | x = nil}
  :verify-guards t;
 
 local FUNCTION op(x,y) {(x | y) & ~(x & y)}
  :guard pred(x) & pred(y);
 
 local FUNCTION id() {nil};
 
 local FUNCTION inv(x) {x}
  :guard pred(x);
 
 THEOREM equiv-is-an-equivalence
  {   acl2::booleanp(equiv(x,y))
    & equiv(x,x)
    & (equiv(x,y) -> equiv(y,x))
    & (equiv(x,y) & equiv(y,z) -> equiv(x,z))}
  :rule-classes [:equivalence
                 [:type-prescription  
                  :corollary equiv(x,y) = t | equiv(x,y) = nil]];
 
 THEOREM equiv-1-implies-equiv-op
  {equiv(x1,x2) -> equiv(op(x1,y),op(x2,y))}
  :rule-classes :congruence;
 
 THEOREM equiv-2-implies-equiv-op
  {equiv(y1,y2) -> equiv(op(x,y1),op(x,y2))}
  :rule-classes :congruence;
 
 THEOREM closure-of-op-for-pred {pred(x) & pred(y) -> pred(op(x,y))};
 
 THEOREM closure-of-id-for-pred {pred(id())};
 
 THEOREM closure-of-inv-for-pred {pred(x) -> pred(inv(x))};
 
 THEOREM commutativity-of-op
  {pred(x) & pred(y) -> equiv(op(x,y),op(y,x))};
 
 THEOREM associativity-of-op
  {pred(x) & pred(y) & pred(z) -> equiv(op(op(x,y),z),op(x,op(y,z)))};
 
 THEOREM left-unicity-of-id-for-op {pred(x) -> equiv(op(id(),x),x)};
 
 THEOREM right-inverse-of-inv-for-op
  {pred(x) -> equiv(op(x,inv(x)),id())};
 
 };

acl2-asg::add-commutativity-2(equiv,
                              pred,
                              op,
                              commutativity-of-op,
                              commutativity-2-of-op);

THEOREM right-unicity-of-id-for-op {pred(x) -> equiv(op(x,id()),x)};

THEOREM left-inverse-of-inv-for-op
 {pred(x) -> equiv(op(inv(x),x),id())};

local THEOREM right-cancellation-for-op-iff
 {pred(x) & pred(y) & pred(z) -> equiv(op(x,z),op(y,z)) <-> equiv(x,y)}
:hints [["Subgoal 1" :in-theory acl2::disable(equiv-1-implies-equiv-op)
                     :use [:instance equiv-1-implies-equiv-op
                                     [x1 := op(x,z), x2 := op(y,z), y := inv(z) ]]]];

THEOREM right-cancellation-for-op
 {pred(x) & pred(y) & pred(z) -> equiv(op(x,z),op(y,z)) = equiv(x,y)}
:hints [["Goal" :use right-cancellation-for-op-iff]];

local THEOREM left-cancellation-for-op-iff
 {pred(x) & pred(y) & pred(z) -> equiv(op(x,y),op(x,z)) <-> equiv(z,y)}
:hints [["Goal" :use [[:instance right-cancellation-for-op
                                 [x := z, z := x ]]]]];

THEOREM left-cancellation-for-op
 {pred(x) & pred(y) & pred(z) -> equiv(op(x,y),op(x,z)) = equiv(y,z)}
:hints [["Goal" :use left-cancellation-for-op-iff]];

THEOREM uniqueness-of-id-as-op-idempotent
 {pred(x) & equiv(op(x,x),x) -> equiv(x,id())}
:hints [["Goal" :use [:instance right-cancellation-for-op
                                [y := id(), z := x ]]]];

THEOREM uniqueness-of-op-inverses
 {pred(x) & pred(y) & equiv(op(x,y),id()) -> equiv(y,inv(x))}
:hints [["Goal" :use [:instance right-cancellation-for-op
                                [x := y, y := inv(x), z := x ]]]];

THEOREM involution-of-inv {pred(x) -> equiv(inv(inv(x)),x)}
:hints [["Goal" :use [:instance uniqueness-of-op-inverses
                                [x := inv(x), y := x ]]]];

THEOREM uniqueness-of-op-inverses-1
 {pred(x) & pred(y) & equiv(op(x,inv(y)),id()) -> equiv(x,y)}
:hints [["Goal" :use [:instance uniqueness-of-op-inverses
                                [y := x, x := inv(y) ]]]];

THEOREM distributivity-of-inv-over-op
 {pred(x) & pred(y) -> equiv(inv(op(x,y)),op(inv(x),inv(y)))}
:hints [["Goal" :use [:instance uniqueness-of-op-inverses
                                [x := op(x,y), y := op(inv(x),inv(y)) ]]]];

THEOREM id-is-its-own-invese {equiv(inv(id()),id())}
:hints [["Goal" :use [:instance uniqueness-of-op-inverses
                                [x := id(), y := id() ]]]];

local THEOREM obvious-inv-cancellation
 {pred(x) & pred(y) -> equiv(op(op(x,inv(x)),y),y)};

THEOREM inv-cancellation-on-right
 {pred(x) & pred(y) -> equiv(op(x,op(y,inv(x))),y)}
:hints [["Goal" :use obvious-inv-cancellation
                :in-theory acl2::disable(right-inverse-of-inv-for-op)]];

THEOREM inv-cancellation-on-left
 {pred(x) & pred(y) -> equiv(op(x,op(inv(x),y)),y)};


