
(in-package "ACL2-ASG")

(encapsulate
  ((equiv ( x y ) t)
   (pred ( x ) t)
   (op ( x y ) t))

  (local
    (defun
      equiv ( x y )
      (equal x y)))

  (local
    (defun
      pred ( x )
      (declare (xargs :verify-guards t))
      (or (equal x t)
          (equal x nil))))

  (local
    (defun
      op ( x y )
      (declare (xargs :guard (and (pred x)
                                  (pred y))))
      (and (or x y)
           (not (and x y)))))

  (defthm
    Equiv-is-an-equivalence
    (and (acl2::booleanp (equiv x y))
         (equiv x x)
         (implies (equiv x y) 
                  (equiv y x))
         (implies (and (equiv x y)
                       (equiv y z))
                  (equiv x z)))
    :rule-classes :EQUIVALENCE)

  (defthm
    Equiv-2-implies-equiv-op
    (implies (equiv y1 y2)
             (equiv (op x y1)
                    (op x y2)))
    :rule-classes :CONGRUENCE)

  (defthm 
    Closure-of-op-for-pred
    (implies (and (pred x)
                  (pred y))
             (pred (op x y))))

  (defthm 
    Associativity-of-op
    (implies (and (pred x)
                  (pred y)
                  (pred z))
             (equiv (op (op x y) z)
                    (op x (op y z)))))

  (defthm
    Commutativity-of-op
    (implies (and (pred x)
                  (pred y))
             (equiv (op x y)
                    (op y x)))))


(local
 (defthm
   commutativity-2-of-op-lemma
   (implies (and (pred x)
                 (pred y)
                 (pred z))
            (equiv (op (op x y) z)
                   (op (op y x) z)))
   :rule-classes nil))

(defthm
  Commutativity-2-of-op
  (implies (and (pred x)
                (pred y)
                (pred z))
           (equiv (op x (op y z))
                  (op y (op x z))))
  :hints (("Goal"
           :in-theory (acl2::disable commutativity-of-op)
           :use commutativity-2-of-op-lemma)))

(defmacro foo(X) `(+ ,x 1));

(defmacro
  add-commutativity-2 ( equiv-name
                        pred-name
                        op-name
                        commutativity-thm-name
                        commutativity-2-thm-name )
  
  `(encapsulate
    nil
    (local
     (defthm
       Associativity-of-op-name
      
       (IMPLIES (AND (,pred-name X)
                     (,pred-name Y)
                     (,pred-name Z))
                (,equiv-name (,op-name (,op-name X Y) Z)
                             (,op-name X (,op-name Y Z))))
       :hints (("Goal"
                :in-theory (disable ,commutativity-thm-name)))))

    (defthm
      ,commutativity-2-thm-name
      (implies (and (,pred-name x)
                    (,pred-name y)
                    (,pred-name z))
               (,equiv-name (,op-name x (,op-name y z))
                            (,op-name y (,op-name x z))))
      :hints (("Goal"
               :use (:functional-instance
                    Commutativity-2-of-op
                     (equiv (lambda (x y)(,equiv-name x y)))
                     (pred (lambda (x)(,pred-name x)))
                     (op (lambda (x y)(,op-name x y)))))))))



