

(in-package "ACL2")

(include-book "alist-defuns")
(include-book "list-defthms")
(include-book "sets")
(local (include-book (:relative :back "arithmetic" "equalities")))

(deflabel alist-defthms-section)



(defthm member-equal-domain
  (implies (alistp a)
	   (iff (member-equal x (domain a))
		(boundp x a)))
  :hints (("Goal" :induct (assoc-equal x a)
	   :in-theory (enable domain boundp))))

(defthm member-equal-binding-range
  (implies (boundp x a)
	   (member-equal (binding x a)
			 (range a)))
  :hints (("Goal" :induct (assoc-equal x a)
	   :in-theory (enable binding range boundp))))


(local (defun double-list-induction (a b)
	 (declare (xargs :guard (and (true-listp a) (true-listp b))))
	 (cond ((endp a) nil)
	       ((endp b) nil)
	       (t (double-list-induction (cdr a) (cdr b))))))

(defthm equal-domain-fwd-to-bound-p
  (implies (and (equal (domain a) (domain b))
		(boundp x a)
		(alistp a)
		(alistp b))
	   (boundp x b))
  :rule-classes :forward-chaining
  :hints (("Goal" :induct (double-list-induction a b)
	   :in-theory (enable boundp domain))))

(defthm bound-p-bind 
  (equal (boundp x (bind y v a))
	 (or (equal x y)
	     (boundp x a)))
  :hints (("Goal" :in-theory (enable boundp))))

(defthm consp-assoc-equal-bind
  (equal (consp (assoc-equal x (bind y v a)))
	 (or (equal x y)
	     (consp (assoc-equal x a))))
  :hints (("Goal" :do-not-induct t
	   :use (:instance boundp-bind)
	   :in-theory (union-theories '(boundp)
				      (set-difference-theories
				       (current-theory :here)
				       '(boundp-bind))))))

(defthm boundp-fbind 
  (equal (boundp x (fbind y v a))
	 (or (equal x y)
	     (boundP x a)))
  :hints (("Goal" :in-theory (enable boundp fbind))))

(defthm boundp?-rembind
  (equal (boundp x (rembind y a))
	 (and (not (equal x y))
	      (boundp? x a)))
  :hints (("Goal" :in-theory (enable boundp))))

(defthm boundp-pairlis
  (iff (boundp x (pairlis a b))
       (member-equal x a))
  :hints (("Goal" :in-theory (enable boundp))))

(defthm boundp-domain-restrict
  (equal (boundp x (domain-restrict l a))
	 (and (member-equal x l)
	      (boundp x a)))
  :hints (("Goal" :induct t :in-theory (enable boundp))))

(defthm bound-domain-anti-restrict
  (equal (boundp x (domain-anti-restrict l a))
	 (and (not (member-equal x l))
	      (boundp x a)))
  :hints (("Goal" :induct t :in-theory (enable boundp))))


(local
 (defthm boundp-alist-compose-wrt-domain-help
   (implies (or (not (consp (assoc-equal (cdr (assoc-equal x a1)) a2)))
		(not (consp (assoc-equal x a1))))
	    (not (consp (assoc-equal x
				     (alist-compose-wrt-domain dom a1 a2)))))
   :hints (("Goal" :in-theory (enable binding boundp)))))

(defthm boundp-alist-compose-wrt-domain
  (iff (boundp x (alist-compose-wrt-domain dom a1 a2))
       (and (member-equal x dom)
	    (boundp x a1)
	    (boundp (binding x a1) a2)))
  :hints (("Goal" :induct t :in-theory (enable boundp binding))))

(defthm boundp-alist-invert
  (iff (boundp x (alist-invert a))
       (member-equal x (range a)))
  :hints (("Goal" :induct t :in-theory (enable boundp range))))


(defthm domain-bind
  (implies (alistp a)
	   (equal (domain (bind x v a))
		  (if (boundp x a)
		      (domain a)
		      (append (domain a) (list x)))))
  :hints (("Goal" :induct t
	   :in-theory (enable boundp domain))))

(defthm domain-fbind 
  (equal (domain (fbind x v a))
	 (cons x (domain a)))
  :hints (("Goal" :do-not-induct t 
	   :in-theory (enable fbind boundp domain))))

(defthm domain-rembind
  (equal (domain (rembind x a))
	 (remove-equal x (domain a)))
  :hints (("Goal" :induct t 
	   :in-theory (enable domain))))

(defthm domain-pairlis
  (implies (true-listp a)
	   (equal (domain (pairlis a b)) a))
  :hints (("Goal" :induct t 
	   :in-theory (enable domain))))

(defthm domain-alist-invert 
  (equal (domain (alist-invert a)) (range a))
  :hints (("Goal" :induct t 
	   :in-theory (enable domain range))))

(defthm domain-domain-restrict
  (equal (domain (domain-restrict l a))
	 (intersection-equal (domain a) l))
  :hints (("Goal" :induct t 
	   :in-theory (enable domain))))

(defthm domain-domain-anti-restrict
  (equal (domain (domain-anti-restrict l a))
	 (set-difference-equal (domain a) l))
  :hints (("Goal" :induct t
	   :in-theory (enable domain))))



(defthm binding-bind
  (equal (binding x (bind y v a))
	 (if (equal x y)
	     v
	     (binding x a)))
  :hints (("Goal" :induct t
	   :in-theory (enable binding))))

(defthm binding-fbind
  (equal (binding x (fbind y v a))
	 (if (equal x y)
	     v
	     (binding x a)))
  :hints (("Goal" :do-not-induct t
	   :in-theory (enable binding fbind))))

(defthm binding-rembind
  (equal (binding x (rembind y a))
	 (if (equal x y)
	     nil
	     (binding x a)))
  :hints (("Goal" :induct t
	   :in-theory (enable binding))))

(defthm binding-pairlis
  (equal (binding x (pairlis a b))
	 (if (member-equal x a)
	     (nth (position x a) b)
	     nil))
  :hints (("Goal" :induct t
	   :in-theory (enable binding))))

(defthm binding-domain-restrict 
  (implies (and (member-equal x l)
		(boundp x a))
	   (equal (binding x (domain-restrict l a))
		  (binding x a)))
  :hints (("Goal" :induct t
	   :in-theory (enable binding boundp))))

(defthm binding-domain-anti-restrict 
  (implies (and (not (member-equal x l))
		(boundp x a))
	   (equal (binding x (domain-anti-restrict l a))
		  (binding x a)))
  :hints (("Goal" :induct t
	   :in-theory (enable binding boundp))))

(defthm binding-alist-compose-wrt-domain
  (implies (and (member-equal x dom)
		(boundp x a1)
		(boundp (binding x a1) a2))
	   (equal (binding x (alist-compose-wrt-domain dom a1 a2))
		  (binding (binding x a1) a2)))
  :hints (("Goal" :induct (alist-compose-wrt-domain dom a1 a2)
	   :in-theory (enable boundp))))

(defthm binding-alist-invert
  (implies (member-equal x (range a))
	   (equal (binding x (alist-invert a))
		  (nth (position x (range a)) (domain a))))
  :hints (("Goal" :induct t
	   :in-theory (enable binding domain range))))

; ---------- RANGE ----------

(DEFTHM RANGE-BIND
  (IMPLIES (and (ALISTP A)
		(not (BOUNDP X A)))
	   (EQUAL (RANGE (BIND X V A))
		  (APPEND (RANGE A) (LIST V))))
  :HINTS
  (("Goal" :INDUCT T :IN-THEORY
    (ENABLE BOUNDP DOMAIN
	    RANGE POSITION BIND ALISTP PUT-NTH))))


; ---------- SUBSETP-EQUAL ----------

(defthm subsetp-equal-all-bindings-range
  (subsetp-equal (all-bindings l a) (range a)))

; ---------- ALL-BOUNDP ----------

(defthm all-boundp-bind
  (implies (all-boundp l a)
	   (all-boundp l (bind x y a))))

(defthm all-boundp-append
  (equal (all-boundp (append x y) a)
	 (and (all-boundp x a)
	      (all-boundp y a))))

(defthm all-boundp-firstn
  (implies (all-boundp l a)
	   (all-boundp (firstn n l) a)))

(defthm all-boundp-nthcdr
  (implies (all-boundp l a)
	   (all-boundp (nthcdr n l) a)))

; ---------- MAKE-ALIST & FLATTEN-ALIST ----------

(defthm make-alist-flatten-alist
  (implies (alistp a)
	   (equal (make-alist (flatten-alist a))
		  a)))



; ========================================

(deftheory alist-defthms
  (set-difference-theories (current-theory :here)
			   (current-theory 'alist-defthms-section)))





