

Include("/usr/cli/projects/cacl2/version13/infix-prelude");

in-package("ACL2");

Include("alist-defuns");

Include("list-defthms");

Include("sets");

local Include((:relative :back "arithmetic" "equalities"));

Label alist-defthms-section;

Theorem member-equal-domain
{ alistp(a) -> x in domain(a) <-> boundp(x,a) }
 :hints [["Goal" :induct assoc-equal(x,a)
                 :in-theory enable(domain,boundp)]];

Theorem member-equal-binding-range
{ boundp(x,a) -> binding(x,a) in range(a) }
 :hints [["Goal" :induct assoc-equal(x,a)
                 :in-theory enable(binding,range,boundp)]];

local Function double-list-induction(a,b)
{ IF endp(a) THEN nil
     ELIF endp(b) THEN nil
     ELSE double-list-induction(cdr(a),cdr(b)) }
 :guard true-listp(a) & true-listp(b);

Theorem equal-domain-fwd-to-bound-p
{ domain(a) = domain(b) & boundp(x,a) & alistp(a) & alistp(b) -> boundp(x,b) }
 :rule-classes :forward-chaining
 :hints [["Goal" :induct double-list-induction(a,b)
                 :in-theory enable(boundp,domain)]];

Theorem bound-p-bind { boundp(x,bind(y,v,a)) = (x = y | boundp(x,a)) }
 :hints [["Goal" :in-theory enable(boundp)]];

Theorem consp-assoc-equal-bind
{ consp(assoc-equal(x,bind(y,v,a))) = (x = y | consp(assoc-equal(x,a))) }
 :hints [["Goal" :do-not-induct t
                 :use [:instance boundp-bind]
                 :in-theory        '[boundp]
                             union (            current-theory(:here)
                                     difference '[boundp-bind])]];

Theorem boundp-fbind { boundp(x,fbind(y,v,a)) = (x = y | boundp(x,a)) }
 :hints [["Goal" :in-theory enable(boundp,fbind)]];

Theorem boundp?-rembind
{ boundp(x,rembind(y,a)) = (x ~= y & boundp?(x,a)) }
 :hints [["Goal" :in-theory enable(boundp)]];

Theorem boundp-pairlis { boundp(x,pairlis(a,b)) <-> x in a }
 :hints [["Goal" :in-theory enable(boundp)]];

Theorem boundp-domain-restrict
{ boundp(x,domain-restrict(l,a)) = (x in l & boundp(x,a)) }
 :hints [["Goal" :induct t
                 :in-theory enable(boundp)]];

Theorem bound-domain-anti-restrict
{ boundp(x,domain-anti-restrict(l,a)) = (x ~in l & boundp(x,a)) }
 :hints [["Goal" :induct t
                 :in-theory enable(boundp)]];

local Theorem boundp-alist-compose-wrt-domain-help
{        ~(consp(assoc-equal(cdr(assoc-equal(x,a1)),a2)))
       | ~(consp(assoc-equal(x,a1)))
   -> ~(consp(assoc-equal(x,alist-compose-wrt-domain(dom,a1,a2)))) }
 :hints [["Goal" :in-theory enable(binding,boundp)]];

Theorem boundp-alist-compose-wrt-domain
{      boundp(x,alist-compose-wrt-domain(dom,a1,a2))
   <-> x in dom & boundp(x,a1) & boundp(binding(x,a1),a2) }
 :hints [["Goal" :induct t
                 :in-theory enable(boundp,binding)]];

Theorem boundp-alist-invert
{ boundp(x,alist-invert(a)) <-> x in range(a) }
 :hints [["Goal" :induct t
                 :in-theory enable(boundp,range)]];

Theorem domain-bind
{     alistp(a)
   ->    domain(bind(x,v,a))
      = (IF boundp(x,a) THEN domain(a)
            ELSE domain(a) @ [x]) }
 :hints [["Goal" :induct t
                 :in-theory enable(boundp,domain)]];

Theorem domain-fbind { domain(fbind(x,v,a)) = x :> domain(a) }
 :hints [["Goal" :do-not-induct t
                 :in-theory enable(fbind,boundp,domain)]];

Theorem domain-rembind
{ domain(rembind(x,a)) = remove-equal(x,domain(a)) }
 :hints [["Goal" :induct t
                 :in-theory enable(domain)]];

Theorem domain-pairlis { true-listp(a) -> domain(pairlis(a,b)) = a }
 :hints [["Goal" :induct t
                 :in-theory enable(domain)]];

Theorem domain-alist-invert { domain(alist-invert(a)) = range(a) }
 :hints [["Goal" :induct t
                 :in-theory enable(domain,range)]];

Theorem domain-domain-restrict
{ domain(domain-restrict(l,a)) = intersection-equal(domain(a),l) }
 :hints [["Goal" :induct t
                 :in-theory enable(domain)]];

Theorem domain-domain-anti-restrict
{ domain(domain-anti-restrict(l,a)) = set-difference-equal(domain(a),l) }
 :hints [["Goal" :induct t
                 :in-theory enable(domain)]];

Theorem binding-bind
{    binding(x,bind(y,v,a))
   = (IF x = y THEN v
         ELSE binding(x,a)) }
 :hints [["Goal" :induct t
                 :in-theory enable(binding)]];

Theorem binding-fbind
{    binding(x,fbind(y,v,a))
   = (IF x = y THEN v
         ELSE binding(x,a)) }
 :hints [["Goal" :do-not-induct t
                 :in-theory enable(binding,fbind)]];

Theorem binding-rembind
{    binding(x,rembind(y,a))
   = (IF x = y THEN nil
         ELSE binding(x,a)) }
 :hints [["Goal" :induct t
                 :in-theory enable(binding)]];

Theorem binding-pairlis
{    binding(x,pairlis(a,b))
   = (IF x in a THEN b[position(x,a)]
         ELSE nil) }
 :hints [["Goal" :induct t
                 :in-theory enable(binding)]];

Theorem binding-domain-restrict
{ x in l & boundp(x,a) -> binding(x,domain-restrict(l,a)) = binding(x,a) }
 :hints [["Goal" :induct t
                 :in-theory enable(binding,boundp)]];

Theorem binding-domain-anti-restrict
{ x ~in l & boundp(x,a) -> binding(x,domain-anti-restrict(l,a)) = binding(x,a) }
 :hints [["Goal" :induct t
                 :in-theory enable(binding,boundp)]];

Theorem binding-alist-compose-wrt-domain
{     x in dom & boundp(x,a1) & boundp(binding(x,a1),a2)
   -> binding(x,alist-compose-wrt-domain(dom,a1,a2)) = binding(binding(x,a1),a2) }
 :hints [["Goal" :induct alist-compose-wrt-domain(dom,a1,a2)
                 :in-theory enable(boundp)]];

Theorem binding-alist-invert
{ x in range(a) -> binding(x,alist-invert(a)) = domain(a)[position(x,range(a))] }
 :hints [["Goal" :induct t
                 :in-theory enable(binding,domain,range)]];

Theorem range-bind
{ alistp(a) & ~(boundp(x,a)) -> range(bind(x,v,a)) = range(a) @ [v] }
 :hints [["Goal" :induct t
                 :in-theory enable(boundp,
                                   domain,
                                   range,
                                   position,
                                   bind,
                                   alistp,
                                   put-nth)]];

Theorem subsetp-equal-all-bindings-range
{ subsetp-equal(all-bindings(l,a),range(a)) };

Theorem all-boundp-bind { all-boundp(l,a) -> all-boundp(l,bind(x,y,a)) };

Theorem all-boundp-append
{ all-boundp(x @ y,a) = (all-boundp(x,a) & all-boundp(y,a)) };

Theorem all-boundp-firstn
{ all-boundp(l,a) -> all-boundp(firstn(n,l),a) };

Theorem all-boundp-nthcdr
{ all-boundp(l,a) -> all-boundp(nthcdr(n,l),a) };

Theorem make-alist-flatten-alist
{ alistp(a) -> make-alist(flatten-alist(a)) = a };

Theory alist-defthms :=             current-theory(:here)
                         difference current-theory('alist-defthms-section);


