

in-package("ACL2");

Label alist-defuns-section;

/* The alist whose domain is A's range, and whose range is A's domain. */

Function alist-invert(a)
{ if endp(a) then nil
      else (cdr(car(a)) :> car(car(a))) :> alist-invert(cdr(a)) }
 :guard alistp(a);

/* The alist derived from A by binding X to Y. */

Function bind(x, y, a)
{ if endp(a) then [x :> y]
      elif x = car(car(a)) then (x :> y) :> cdr(a)
      else car(a) :> bind(x, y, cdr(a)) }
 :guard alistp(a);

/* The value bound to X in alist A. */

Function binding(x, a) { cdr(assoc-equal(x, a)) }
 :guard alistp(a);

/* Is X bound in alist A? */

Function boundp(x, a) { consp(assoc-equal(x, a)) }
 :guard alistp(a);

/* The list of CARs of an alist. */

Function domain(a) { strip-cars(a) }
 :guard alistp(a);

/* An alist containing only those pairs in A whose CAR is not in L. */

Function domain-anti-restrict(l, a)
{ if endp(a) then nil
      elif car(car(a)) in l then domain-anti-restrict(l, cdr(a))
      else car(a) :> domain-anti-restrict(l, cdr(a)) }
 :guard true-listp(l) & alistp(a);

/* An alist containing only those pairs in A whose CAR is in L. */

Function domain-restrict(l, a)
{ if endp(a) then nil
      elif car(car(a)) in l then car(a) :> domain-restrict(l, cdr(a))
      else domain-restrict(l, cdr(a)) }
 :guard true-listp(l) & alistp(a);

/* Bind key X to value V in alist A (fast). */

Function fbind(x, v, a) { (x :> v) :> a }
 :guard alistp(a);

Function flatten-alist(a)
{ if endp(a) then nil
      else caar(a) :> cdar(a) :> flatten-alist(cdr(a)) }
 :guard alistp(a);

Function make-alist(l)
{ if endp(l) then nil
      elif endp(cdr(l)) then nil
      else (car(l) :> cadr(l)) :> make-alist(cddr(l)) }
 :guard true-listp(l);

/* The list of CDRs of an alist. */

Function range(a) { strip-cdrs(a) }
 :guard alistp(a);

/* The alist derived from A by removing any binding of X. */

Function rembind(x, a)
{ if endp(a) then nil
      elif x = car(car(a)) then rembind(x, cdr(a))
      else car(a) :> rembind(x, cdr(a)) }
 :guard alistp(a);

/* The list of bindings of elements of list L in alist A. */

Function all-bindings(l, a)
{ if endp(l) then nil
      elif bound?(car(l), a) then binding(car(l), a) :> all-bindings(cdr(l), a)
      else all-bindings(cdr(l), a) }
 :guard true-listp(l) & alistp(a);

/* Are all elements of list L bound in alist A? */

Function all-boundp(l, a)
{ if endp(l) then t
      else bound?(car(l), a) & all-bound?(cdr(l), a) }
 :guard true-listp(l) & alistp(a);

/* X is bound to Z in (ALIST-COMPOSE DOM A1 A2) if X occurs in
   DOM, X is bound in A1, and there exists a Y such that
   (BINDING X A1) = Y, and (BINDING Y A2) = Z. */

Function alist-compose-wrt-domain(dom, a1, a2)
{ if endp(dom) then nil
      elif bound?(car(dom), a1) & bound?(binding(car(dom), a1), a2)
      then bind(car(dom),
                binding(binding(car(dom), a1), a2),
                alist-compose-wrt-domain(cdr(dom), a1, a2))
      else alist-compose-wrt-domain(cdr(dom), a1, a2) }
 :guard true-listp(dom) & (alistp(a1) & alistp(a2));

Macro alist-compose(a1, a2)
`[alist-compose-wrt-domain [domain ,a1] ,a1 ,a2];

Theorem alistp-bind { alistp(a) -> alistp(bind(x, v, a)) }
 :hints [["Goal" :induct alistp(a)]];

Theorem alistp-fbind { alistp(a) -> alistp(fbind(x, v, a)) }
 :hints [["Goal" :induct alistp(a)]];

Theorem alistp-rembind { alistp(a) -> alistp(rembind(x, a)) }
 :hints [["Goal" :induct rembind(x, a)]];

Theorem alistp-make-alist { alistp(make-alist(l)) }
 :hints [["Goal" :induct make-alist(l)]];

Theorem alistp-pairlis { alistp(pairlis(a, b)) }
 :hints [["Goal" :induct pairli(a, b)]];

Theorem alistp-alist-invert { alistp(alist-invert(a)) };

Theorem alistp-domain-restrict { alistp(a) -> alistp(domain-restrict(l, a)) };

 Theorem alistp-domain-anti-restrict { alistp(a) -> alistp(domain-anti-restrict(l, a)) };

Theorem alistp-alist-compose-wrt-domain { alistp(alist-compose-wrt-domain(dom, a1, a2)) };

Verify guards alist-compose-wrt-domain;

Theory alist-non-recursive-defuns := [binding, bound, domain, range, fbind];

Theorem true-listp-flatten-alist { true-listp(flatten-alist(l)) };

IN-THEORY(disable(alist-non-recursive-defuns));

  Theory alist-defuns := current-theory(:here) difference (current-theory('alist-defuns-section) union theory('alist-non-recursive-defuns));


