

Include("/usr/cli/projects/cacl2/version13/infix-prelude");

in-package("ACL2");

Label alist-defuns-section;

Function alist-invert(a)
{ IF endp(a) THEN nil
     ELSE (cdr(car(a)) :> car(car(a))) :> alist-invert(cdr(a)) }
 :guard alistp(a);

Function bind(x,y,a)
{ IF endp(a) THEN [x :> y]
     ELIF x = car(car(a)) THEN (x :> y) :> cdr(a)
     ELSE car(a) :> bind(x,y,cdr(a)) }
 :guard alistp(a);

Function binding(x,a) { cdr(assoc-equal(x,a)) }
 :guard alistp(a);

Function boundp(x,a) { consp(assoc-equal(x,a)) }
 :guard alistp(a);

Function domain(a) { strip-cars(a) }
 :guard alistp(a);

Function domain-anti-restrict(l,a)
{ IF endp(a) THEN nil
     ELIF car(car(a)) in l THEN domain-anti-restrict(l,cdr(a))
     ELSE car(a) :> domain-anti-restrict(l,cdr(a)) }
 :guard true-listp(l) & alistp(a);

Function domain-restrict(l,a)
{ IF endp(a) THEN nil
     ELIF car(car(a)) in l THEN car(a) :> domain-restrict(l,cdr(a))
     ELSE domain-restrict(l,cdr(a)) }
 :guard true-listp(l) & alistp(a);

Function fbind(x,v,a) { (x :> v) :> a }
 :guard alistp(a);

Function flatten-alist(a)
{ IF endp(a) THEN nil
     ELSE caar(a) :> cdar(a) :> flatten-alist(cdr(a)) }
 :guard alistp(a);

Function make-alist(l)
{ IF endp(l) THEN nil
     ELIF endp(cdr(l)) THEN nil
     ELSE (car(l) :> cadr(l)) :> make-alist(cddr(l)) }
 :guard true-listp(l);

Function range(a) { strip-cdrs(a) }
 :guard alistp(a);

Function rembind(x,a)
{ IF endp(a) THEN nil
     ELIF x = car(car(a)) THEN rembind(x,cdr(a))
     ELSE car(a) :> rembind(x,cdr(a)) }
 :guard alistp(a);

Function all-bindings(l,a)
{ IF endp(l) THEN nil
     ELIF bound?(car(l),a) THEN binding(car(l),a) :> all-bindings(cdr(l),a)
     ELSE all-bindings(cdr(l),a) }
 :guard true-listp(l) & alistp(a);

Function all-boundp(l,a)
{ IF endp(l) THEN t
     ELSE bound?(car(l),a) & all-bound?(cdr(l),a) }
 :guard true-listp(l) & alistp(a);

Function alist-compose-wrt-domain(dom,a1,a2)
{ IF endp(dom) THEN nil
     ELIF bound?(car(dom),a1) & bound?(binding(car(dom),a1),a2)
     THEN bind(car(dom),
               binding(binding(car(dom),a1),a2),
               alist-compose-wrt-domain(cdr(dom),a1,a2))
     ELSE alist-compose-wrt-domain(cdr(dom),a1,a2) }
 :guard true-listp(dom) & alistp(a1) & alistp(a2);

Macro alist-compose(a1,a2) = 
alist-compose-wrt-domain(domain(a1),a1,a2);

Theorem alistp-bind { alistp(a) -> alistp(bind(x,v,a)) }
 :hints [["Goal" :induct alistp(a)]];

Theorem alistp-fbind { alistp(a) -> alistp(fbind(x,v,a)) }
 :hints [["Goal" :induct alistp(a)]];

Theorem alistp-rembind { alistp(a) -> alistp(rembind(x,a)) }
 :hints [["Goal" :induct rembind(x,a)]];

Theorem alistp-make-alist { alistp(make-alist(l)) }
 :hints [["Goal" :induct make-alist(l)]];

Theorem alistp-pairlis { alistp(pairlis(a,b)) }
 :hints [["Goal" :induct pairli(a,b)]];

Theorem alistp-alist-invert { alistp(alist-invert(a)) };

Theorem alistp-domain-restrict
{ alistp(a) -> alistp(domain-restrict(l,a)) };

Theorem alistp-domain-anti-restrict
{ alistp(a) -> alistp(domain-anti-restrict(l,a)) };

Theorem alistp-alist-compose-wrt-domain
{ alistp(alist-compose-wrt-domain(dom,a1,a2)) };

Verify guards alist-compose-wrt-domain;

Theory alist-non-recursive-defuns := [binding, bound, domain, range, fbind];

Theorem true-listp-flatten-alist { true-listp(flatten-alist(l)) };

IN-THEORY(disable(alist-non-recursive-defuns));

Theory alist-defuns :=             current-theory(:here)
                        difference (       current-theory('alist-defuns-section)
                                     union theory('alist-non-recursive-defuns));


