

Include("/usr/cli/projects/cacl2/version13/infix-prelude");

in-package("ACL2");

/* Verify guards, unless :verify-guards is NIL. */
set-verify-guards-eagerness(2);


Module { }
{ 
  Macro implies-x(x,y) = 
  IF y THEN t
     ELSE ~x;
  
  local Theorem impliesx-is-implies { (a -> b) = impliesx(a,b) };
  
   };

Function firstn(n,l)
{ IF endp(l) THEN nil
     ELIF zp(n) THEN nil
     ELSE car(l) :> firstn(n - 1,cdr(l)) }
 :guard true-listp(l) & integerp(n) & 0 <= n;

Theorem firstn-0 { firstn(0,rest) = nil };

Theorem firstn-cons
{    firstn(n,a :> rest)
   = (IF zp(n) THEN nil
         ELSE a :> firstn(n - 1,rest)) };

Function restn(n,l)
{ IF endp(l) THEN l
     ELIF zp(n) THEN l
     ELSE restn(n - 1,cdr(l)) }
 :guard true-listp(l) & integerp(n) & 0 <= n;

Theorem restn-0 { restn(0,rest) = rest };

Theorem restn-cons
{    restn(n,a :> rest)
   = (IF zp(n) THEN a :> rest
         ELSE restn(n - 1,rest)) };

Function tree-size(tree)
{ IF atom(tree) THEN 1
     ELSE tree-size(car(tree)) + tree-size(cdr(tree)) };

Theorem tree-size-step
{ tree-size(a :> b) = tree-size(a) + tree-size(b) };

Theorem tree-size-base { tree-size(0) = 1 };

Function tfirstn(list,tree) { firstn(tree-size(car(tree)),list) }
 :guard consp(tree) & true-listp(list);

Function trestn(list,tree) { restn(tree-size(car(tree)),list) }
 :guard consp(tree) & true-listp(list);

Function t-carry(c,prop,gen) { (c & prop) | gen };

Theorem append-cons { a :> x @ y = a :> (x @ y) };

Theorem append-nil { nil @ y = y };

Macro sub1(x) =  x - 1;

Macro nlistp(x) =  atom(x);

Macro caddddr(x) =  car(cddddr(x));

Macro cdddddr(x) =  cdr(cddddr(x));

Macro cadddddr(x) =  car(cdddddr(x));

Macro cddddddr(x) =  cdr(cdddddr(x));

Macro caddddddr(x) =  car(cddddddr(x));

Macro cdddddddr(x) =  cdr(cddddddr(x));

Macro quotient(x,y) = 
nonnegative-integer-quotient(x,y);

Macro remainder(x,y) =  rem(x,y);

Macro boolp(x) =  booleanp(x);

Macro bvp(x) =  boolean-listp(x);

Function b-buf(x)
{ IF x THEN t
     ELSE nil };

Function b-not(x) { ~x };

Function b-nand(a,b) { ~(a & b) };

Function b-nand3(a,b,c) { ~(a & b & c) };

Function b-nand4(a,b,c,d) { ~(a & b & c & d) };

Function b-xor(a,b)
{ IF a
     THEN IF b THEN nil
             ELSE t
     ELIF b THEN t
     ELSE nil };

Function b-xor3(a,b,c) { b-xor(b-xor(a,b),c) };

Function b-equv(x,y)
{ IF x
     THEN IF y THEN t
             ELSE nil
     ELIF y THEN nil
     ELSE t };

Function b-equv3(a,b,c) { b-equv(a,b-xor(b,c)) };

Function b-and(a,b) { a & b };

Function b-and3(a,b,c) { a & b & c };

Function b-or(a,b) { a | b };

Function b-or3(a,b,c) { a | b | c };

Function b-nor(a,b) { ~(a | b) };

Function b-nor3(a,b,c) { ~(a | b | c) };

Function b-if(c,a,b)
{ IF c
     THEN IF a THEN t
             ELSE nil
     ELIF b THEN t
     ELSE nil };

Theorem b-nand-comm { b-nand(a,b) = b-nand(b,a) };

Theorem b-xor-comm { b-xor(a,b) = b-xor(b,a) };

Theorem b-equv-comm { b-equv(a,b) = b-equv(b,a) };

Theorem b-nor-comm { b-nor(a,b) = b-nor(b,a) };

Function v-if(c,a,b)
{ IF nlistp(a) THEN nil
     ELSE     (IF (IF c THEN car(a)
                       ELSE car(b)) THEN t
                   ELSE nil)
           :> v-if(c,cdr(a),cdr(b)) }
 :guard true-listp(b);

Theorem v-if-base { v-if(c,nil,nil) = nil };

Theorem v-if-step
{ a := a0 :> a1;
  { b := b0 :> b1;
       v-if(c,a,b)
     =     (IF c
               THEN IF a0 THEN t
                       ELSE nil
               ELIF b0 THEN t
               ELSE nil)
       :> v-if(c,a1,b1)}};

Function v-buf(x)
{ IF nlistp(x) THEN nil
     ELSE b-buf(car(x)) :> v-buf(cdr(x)) };

Theorem v-buf-base { v-buf(nil) = nil };

Theorem v-buf-step
{ a := a0 :> a1;
  v-buf(a) = b-buf(a0) :> v-buf(a1)};

Function boolfix(x)
{ IF x THEN t
     ELSE nil };

Function v-nzerop(x)
{ IF nlistp(x) THEN nil
     ELSE car(x) | v-nzerop(cdr(x)) };

Theorem v-nzerop-base { v-nzerop(nil) = nil };

Theorem v-nzerop-step
{ a := a0 :> a1;
  v-nzerop(a) = (a0 | v-nzerop(a1))};

Function v-zerop(x) { ~(v-nzerop(x)) };

Function nat-to-v(x,n)
{ IF zp(n) THEN nil
     ELSE ~(zp(remainder(x,2))) :> nat-to-v(quotient(x,2),sub1(n)) };

Function make-nils(n)
{ IF zp(n) THEN nil
     ELSE nil :> make-nils(n - 1) }
 :guard integerp(n) & ~(n < 0);

Theorem nat-to-v-0 { nat-to-v(0,n) = make-nils(n) };

Theorem fold-plus
{ syntaxp(quotep(i)) & syntaxp(quotep(j)) -> i + j + x = (i + j) + x };

Function v-not(x)
{ IF nlistp(x) THEN nil
     ELSE b-not(car(x)) :> v-not(cdr(x)) };

Theorem v-not-base { v-not(nil) = nil };

Theorem v-not-step
{ a := a0 :> a1;
  v-not(a) = b-not(a0) :> v-not(a1)};

Function v-xor(x,y)
{ IF nlistp(x) THEN nil
     ELSE b-xor(car(x),car(y)) :> v-xor(cdr(x),cdr(y)) }
 :guard true-listp(y);

Theorem v-xor-base { v-xor(nil,x) = nil };

Theorem v-xor-step
{ a := a0 :> a1;
  { b := b0 :> b1;
    v-xor(a,b) = b-xor(a0,b0) :> v-xor(a1,b1)}};

Function v-or(x,y)
{ IF nlistp(x) THEN nil
     ELSE b-or(car(x),car(y)) :> v-or(cdr(x),cdr(y)) }
 :guard true-listp(y);

Theorem v-or-base { v-or(nil,x) = nil };

Theorem v-or-step
{ a := a0 :> a1;
  { b := b0 :> b1;
    v-or(a,b) = b-or(a0,b0) :> v-or(a1,b1)}};

Function v-and(x,y)
{ IF nlistp(x) THEN nil
     ELSE b-and(car(x),car(y)) :> v-and(cdr(x),cdr(y)) }
 :guard true-listp(y);

Theorem v-and-base { v-and(nil,x) = nil };

Theorem v-and-step
{ a := a0 :> a1;
  { b := b0 :> b1;
    v-and(a,b) = b-and(a0,b0) :> v-and(a1,b1)}};

Theorem consp-cons { consp(x :> y) = t };

Theorem nth-cons
{    a :> x[n]
   = IF* zp(n) THEN a
       ELSE x[n - 1] };

Theorem len-cons { len(a :> x) = 1 + len(x) };

Theorem len-nil { len(nil) = 0 };

Theorem ifx-hide
{     x = y
   ->    x
      = IF* test THEN y
          ELSE hide(x) }
 :hints [["Goal" :expand hide(x)]];

Function v-equal(x,y)
{ IF nlistp(x) THEN x = nil & y = nil
     ELSE consp(y) & car(x) = car(y) & v-equal(cdr(x),cdr(y)) };

Theorem v-equal-base { v-equal(nil,x) = (x = nil) };

Theorem v-equal-step
{ a := a0 :> a1;
  { b := b0 :> b1;
    v-equal(a,b) = (a0 = b0 & v-equal(a1,b1))}};

Function vcond-macro(clauses)
{ IF consp(clauses)
     THEN IF eq(car(car(clauses)),t)
             THEN IF cdr(car(clauses)) THEN car(cdr(car(clauses)))
                     ELSE car(car(clauses))
             ELSE ['v-if, 
                   car(car(clauses)), 
                   IF cdr(car(clauses)) THEN car(cdr(car(clauses)))
                      ELSE car(car(clauses)), 
                   vcond-macro(cdr(clauses))]
     ELSE nil }
 :guard cond-clausesp(clauses);

Macro vcond(&rest clauses) # 
vcond-macro(clauses)
 :guard cond-clausesp(clauses);

Function bv2p(a,b) { bvp(a) & bvp(b) & len(a) = len(b) };

Function v-trunc(x,n)
{ IF zp(n) THEN nil
     ELSE boolfix(car(x)) :> v-trunc(cdr(x),n - 1) }
 :guard bvp(x) & integerp(n) & ~(n < 0);

Theorem v-trunc-0 { v-trunc(x,0) = nil };

Theorem v-trunc-cons
{    v-trunc(a :> x,n)
   = (IF zp(n) THEN nil
         ELSE boolfix(a) :> v-trunc(x,n - 1)) };

Theorem len-v-trunc { len(v-trunc(x,y)) = nfix(y) };

Theorem nfix-len { nfix(len(x)) = len(x) };

Theorem v-trunc-v-buf { n = len(a) -> v-trunc(v-buf(a),n) = v-buf(a) };

Theorem len-nat-to-v { len(nat-to-v(i,n)) = nfix(n) };

Theorem len-v-not { len(v-not(v)) = len(v) };

Function v-shift-right(a,si)
{ IF nlistp(a) THEN nil
     ELSE v-buf(cdr(a)) @ boolfix(si) :> nil };

Module { }
{ 
  local Theorem v-trunc-v-shift-right-lemma
  { v-trunc(v-shift-right(a,c),len(a)) = v-shift-right(a,c) };
  
  Theorem v-trunc-v-shift-right
  { n = len(a) -> v-trunc(v-shift-right(a,c),n) = v-shift-right(a,c) }
   :hints [["Goal" :in-theory disable(v-shift-right)]];
  
   };

Theorem v-trunc-v-xor
{ n = len(a) -> v-trunc(v-xor(a,b),n) = v-xor(a,b) };

Module { }
{ 
  local Theorem v-trunc-v-or-lemma
  { bv2p(a,b) -> v-trunc(v-or(a,b),len(a)) = v-or(a,b) };
  
  Theorem v-trunc-v-or
  { bv2p(a,b) & n = len(a) -> v-trunc(v-or(a,b),n) = v-or(a,b) };
  
   };

Theorem v-trunc-v-not { n = len(a) -> v-trunc(v-not(a),n) = v-not(a) };

Theorem len-v-buf { len(v-buf(v)) = len(v) };

Theorem len-v-xor { len(v-xor(a,b)) = len(a) };

Theorem len-v-and { len(v-and(a,b)) = len(a) };

Theorem len-v-or { len(v-or(a,b)) = len(a) };

Function make-tree(n)
{ IF zp(n) THEN 0
     ELIF n == 1 THEN 0
     ELSE make-tree(quotient(n,2)) :> make-tree(n - quotient(n,2)) }
 :guard integerp(n) & ~(n < 0);


