

Include("/usr/cli/projects/cacl2/version13/infix-prelude");

in-package("ACL2");

Include("alist-defuns");

Include("list-defuns");

Include("utilities");

Module 
{domain-elem-type(x) : boolean;
 domain-type(l) : boolean;
 range-elem-type(x) : boolean;
 range-type(l) : boolean;
 alist-type(a) : boolean;} 
{ 
  local IN-THEORY('[ground-zero]);
  
  local Function domain-elem-type(x) { symbolp(x) };
  
  local Function domain-type(l) { symbol-listp(l) };
  
  local Function range-elem-type(x) { integerp(x) };
  
  local Function range-type(l) { integer-listp(l) };
  
  local Function alist-type(a)
  { IF atom(a) THEN eq(a,nil)
       ELSE    consp(car(a))
             & domain-elem-type(caar(a))
             & range-elem-type(cdar(a))
             & alist-type(cdr(a)) };
  
  Theorem domain-type-domain-elem-type
  {    domain-type(nil)
     & domain-type(x :> l) = (domain-elem-type(x) & domain-type(l)) };
  
  Theorem range-type-range-elem-type
  { range-type(nil) & range-type(x :> l) = (range-elem-type(x) & range-type(l)) };
  
  Theorem alist-type-defun
  {      alist-type(l)
     <-> IF atom(l) THEN eq(l,nil)
           ELSE    consp(car(l))
                 & domain-elem-type(caar(l))
                 & range-elem-type(cdar(l))
                 & alist-type(cdr(l)) }
   :rule-classes [[:rewrite  
   :corollary atom(l) -> alist-type(l) = null(l)] [:rewrite  
   :corollary    alist-type(x :> l)
               = (   consp(x)
                   & domain-elem-type(car(x))
                   & range-elem-type(cdr(x))
                   & alist-type(l))]];
  
   };

Function replace-equal(a,b,l)
{ IF endp(l) THEN nil
     ELIF car(l) = a THEN b :> replace-equal(a,b,cdr(l))
     ELSE car(l) :> replace-equal(a,b,cdr(l)) }
 :guard true-listp(l);

Function my-conjoin(termlist1,termlist2)
{ termlist := termlist1 @ termlist2;
  IF len(termlist) == 1 THEN car(termlist)
     ELSE fcons-term('and,termlist)}
 :guard    true-listp(termlist1)
         & true-listp(termlist2)
         & consp(termlist1) | consp(termlist2);

Mutual-Recursion {
 
 Function my-conjuncts(term)
 { IF eq(term,t) THEN nil
      ELIF atom(term) THEN [term]
      ELIF eq(car(term),'and) THEN my-conjuncts-list(cdr(term))
      ELSE [term] };
 
 Function my-conjuncts-list(termlist)
 { IF atom(termlist) THEN nil
      ELSE my-conjuncts(car(termlist)) @ my-conjuncts-list(cdr(termlist)) };
 
 };

Macro alist-type-alistp-lemma(alist-type-fn,
                              dom-type-fn : ignore,
                              ran-type-fn : ignore,
                              dom-elem-type-fn : ignore,
                              ran-elem-type-fn : ignore,
                              formals,
                              &optional 
                              guard := 't) # 
['implies, 
 my-conjoin(my-conjuncts(guard),[list*(alist-type-fn,formals)]), 
 '[alistp l]
 ]
 ;

Theorem alist-type-alistp
{ alist-type-alistp-lemma(alist-type,
                          domain-type,
                          range-type,
                          domain-elem-type,
                          range-elem-type,
                          l()) }
 :hints [["Goal" :induct t]];

IN-THEORY(disable(alist-type-alistp));

Macro alist-type-bind-lemma(alist-type-fn,
                            dom-type-fn : ignore,
                            ran-type-fn : ignore,
                            dom-elem-type-fn,
                            ran-elem-type-fn,
                            formals,
                            &optional 
                            guard := 't) # 
{ vars := u::unique-symbols(2,
                          intern-in-package-of-symbol("X",alist-type-fn),
                          formals);
  var1 := car(vars);
  var2 := cadr(vars);
  ['implies, 
   my-conjoin(my-conjuncts(guard),
              [list*(alist-type-fn,formals), 
               [dom-elem-type-fn, var1], 
               [ran-elem-type-fn, var2]
               ]), 
   list*(alist-type-fn,replace-equal('l,['bind, var1, var2, 'l],formals))
   ]}
 ;

Theorem alist-type-bind
{ alist-type-bind-lemma(alist-type,
                        domain-type,
                        range-type,
                        domain-elem-type,
                        range-elem-type,
                        l()) }
 :hints [["Goal" :induct t]];

Macro alist-type-fbind-lemma(alist-type-fn,
                             dom-type-fn : ignore,
                             ran-type-fn : ignore,
                             dom-elem-type-fn,
                             ran-elem-type-fn,
                             formals,
                             &optional 
                             guard := 't) # 
{ vars := u::unique-symbols(2,
                          intern-in-package-of-symbol("X",alist-type-fn),
                          formals);
  var1 := car(vars);
  var2 := cadr(vars);
  ['implies, 
   my-conjoin(my-conjuncts(guard),
              [list*(alist-type-fn,formals), 
               [dom-elem-type-fn, var1], 
               [ran-elem-type-fn, var2]
               ]), 
   list*(alist-type-fn,replace-equal('l,['fbind, var1, var2, 'l],formals))
   ]}
 ;

Theorem alist-type-fbind
{ alist-type-fbind-lemma(alist-type,
                         domain-type,
                         range-type,
                         domain-elem-type,
                         range-elem-type,
                         l()) }
 :hints [["Goal" :do-not-induct t
                 :in-theory enable(alist-type-alistp,fbind)]];

Macro alist-type-binding-lemma(alist-type-fn,
                               dom-type-fn : ignore,
                               ran-type-fn : ignore,
                               dom-elem-type-fn : ignore,
                               ran-elem-type-fn,
                               formals,
                               &optional 
                               guard := 't) # 
{ vars := u::unique-symbols(1,
                          intern-in-package-of-symbol("X",alist-type-fn),
                          formals);
  var := car(vars);
  ['implies, 
   my-conjoin(my-conjuncts(guard),
              [list*(alist-type-fn,formals), ['bound?, var, 'l]]), 
   [ran-elem-type-fn, ['binding, var, 'l]]
   ]}
 ;

Theorem alist-type-binding
{ alist-type-binding-lemma(alist-type,
                           domain-type,
                           range-type,
                           domain-elem-type,
                           range-elem-type,
                           l()) }
 :hints [["Goal" :in-theory enable(alist-type-alistp,binding,bound?)]];

Macro alist-type-bound?-lemma(alist-type-fn,
                              dom-type-fn : ignore,
                              ran-type-fn : ignore,
                              dom-elem-type-fn,
                              ran-elem-type-fn : ignore,
                              formals,
                              &optional 
                              guard := 't) # 
{ vars := u::unique-symbols(1,
                          intern-in-package-of-symbol("X",alist-type-fn),
                          formals);
  var := car(vars);
  ['implies, 
   my-conjoin(my-conjuncts(guard),
              [list*(alist-type-fn,formals), ['not, [dom-elem-type-fn, var]]]), 
   ['not, ['bound?, var, 'l]]
   ]}
 ;

Theorem alist-type-bound?
{ alist-type-bound?-lemma(alist-type,
                          domain-type,
                          range-type,
                          domain-elem-type,
                          range-elem-type,
                          l()) }
 :hints [["Goal" :in-theory enable(alist-type-alistp,bound?)]];

Macro alist-type-domain-lemma(alist-type-fn,
                              dom-type-fn,
                              ran-type-fn : ignore,
                              dom-elem-type-fn : ignore,
                              ran-elem-type-fn : ignore,
                              formals,
                              &optional 
                              guard := 't) # 
['implies, 
 my-conjoin(my-conjuncts(guard),[list*(alist-type-fn,formals)]), 
 [dom-type-fn, '[domain l]]
 ]
 ;

Theorem alist-type-domain
{ alist-type-domain-lemma(alist-type,
                          domain-type,
                          range-type,
                          domain-elem-type,
                          range-elem-type,
                          l()) }
 :hints [["Goal" :in-theory enable(alist-type-alistp,domain)]];

Macro alist-type-range-lemma(alist-type-fn,
                             dom-type-fn : ignore,
                             ran-type-fn,
                             dom-elem-type-fn : ignore,
                             ran-elem-type-fn : ignore,
                             formals,
                             &optional 
                             guard := 't) # 
['implies, 
 my-conjoin(my-conjuncts(guard),[list*(alist-type-fn,formals)]), 
 [ran-type-fn, '[range l]]
 ]
 ;

Theorem alist-type-range
{ alist-type-range-lemma(alist-type,
                         domain-type,
                         range-type,
                         domain-elem-type,
                         range-elem-type,
                         l()) }
 :hints [["Goal" :in-theory enable(alist-type-alistp,range)]];

Macro alist-type-rembind-lemma(alist-type-fn,
                               dom-type-fn : ignore,
                               ran-type-fn : ignore,
                               dom-elem-type-fn : ignore,
                               ran-elem-type-fn : ignore,
                               formals,
                               &optional 
                               guard := 't) # 
{ vars := u::unique-symbols(1,
                          intern-in-package-of-symbol("X",alist-type-fn),
                          formals);
  var := car(vars);
  ['implies, 
   my-conjoin(my-conjuncts(guard),[list*(alist-type-fn,formals)]), 
   list*(alist-type-fn,replace-equal('l,['rembind, var, 'l],formals))
   ]}
 ;

Theorem alist-type-rembind
{ alist-type-rembind-lemma(alist-type,
                           domain-type,
                           range-type,
                           domain-elem-type,
                           range-elem-type,
                           l()) }
 :hints [["Goal" :induct t]];

Macro alist-type-pairlis$-lemma(alist-type-fn,
                                dom-type-fn,
                                ran-type-fn,
                                dom-elem-type-fn : ignore,
                                ran-elem-type-fn : ignore,
                                formals,
                                &optional 
                                guard := 't) # 
{ vars := u::unique-symbols(2,
                          intern-in-package-of-symbol("X",alist-type-fn),
                          formals);
  var1 := car(vars);
  var2 := cadr(vars);
  ['implies, 
   my-conjoin(my-conjuncts(guard),
              [['true-listp, var1], 
               [dom-type-fn, var1], 
               [ran-type-fn, var2], 
               ['eql, ['len, var1], ['len, var2]]
               ]), 
   list*(alist-type-fn,replace-equal('l,['pairlis$, var1, var2],formals))
   ]}
 ;

Theorem alist-type-pairlis$
{ alist-type-pairlis$-lemma(alist-type,
                            domain-type,
                            range-type,
                            domain-elem-type,
                            range-elem-type,
                            l()) }
 :hints [["Goal" :induct t]];

Macro alist-type-domain-restrict-lemma(alist-type-fn,
                                       dom-type-fn : ignore,
                                       ran-type-fn : ignore,
                                       dom-elem-type-fn : ignore,
                                       ran-elem-type-fn : ignore,
                                       formals,
                                       &optional 
                                       guard := 't) # 
{ vars := u::unique-symbols(1,
                          intern-in-package-of-symbol("L",alist-type-fn),
                          formals);
  var := car(vars);
  ['implies, 
   my-conjoin(my-conjuncts(guard),[list*(alist-type-fn,formals)]), 
   list*(alist-type-fn,replace-equal('l,['domain-restrict, var, 'l],formals))
   ]}
 ;

Theorem alist-type-domain-restrict
{ alist-type-domain-restrict-lemma(alist-type,
                                   domain-type,
                                   range-type,
                                   domain-elem-type,
                                   range-elem-type,
                                   l()) }
 :hints [["Goal" :induct t]];

Macro alist-type-domain-anti-restrict-lemma(alist-type-fn,
                                            dom-type-fn : ignore,
                                            ran-type-fn : ignore,
                                            dom-elem-type-fn : ignore,
                                            ran-elem-type-fn : ignore,
                                            formals,
                                            &optional 
                                            guard := 't) # 
{ vars := u::unique-symbols(1,
                          intern-in-package-of-symbol("L",alist-type-fn),
                          formals);
  var := car(vars);
  ['implies, 
   my-conjoin(my-conjuncts(guard),[list*(alist-type-fn,formals)]), 
   list*(alist-type-fn,
         replace-equal('l,['domain-anti-restrict, var, 'l],formals))
   ]}
 ;

Theorem alist-type-domain-anti-restrict
{ alist-type-domain-anti-restrict-lemma(alist-type,
                                        domain-type,
                                        range-type,
                                        domain-elem-type,
                                        range-elem-type,
                                        l()) }
 :hints [["Goal" :induct t]];


