



(defun car-cdr-symbolp1 (lst)
  (declare (xargs :mode :program))
  (cond ((null lst) nil)
        ((eql (car lst) #\R) (null (cdr lst)))
        (t (and (member-equal (car lst) '(#\A #\D))
                (car-cdr-symbolp1 (cdr lst))))))

(defun car-cdr-symbolp (x)
  (declare (xargs :mode :program))
  (and (symbolp x)
       (equal (symbol-package-name x) "NQTHM")
       (let* ((lst (coerce (symbol-name x) 'list)))
         (and lst
              (eql (car lst) #\C)
              (car-cdr-symbolp1 (cdr lst))))))

(defun make-nqthm-car-cdr-nest1 (lst arg)
  (declare (xargs :mode :program))
  (cond ((null lst) nil)
        ((null (cdr lst)) arg)
        ((eql (car lst) #\A) (list 'nqthm::car (make-nqthm-car-cdr-nest1 (cdr lst) arg)))
        (t  (list 'nqthm::cdr (make-nqthm-car-cdr-nest1 (cdr lst) arg)))))

(defun make-nqthm-car-cdr-nest (sym arg)
  (declare (xargs :mode :program))
  (make-nqthm-car-cdr-nest1 (cdr (coerce (symbol-name sym) 'list)) arg))

(mutual-recursion

(defun nqthm-sublis (alist term)
  (declare (xargs :mode :program))
  (cond ((atom term)
         (cond ((symbolp term)
                (let ((temp (assoc-eq term alist)))
                  (if temp (cdr temp) term)))
               (t term)))
        ((eq (car term) 'quote) term)
        (t (cons (car term) (nqthm-sublis-lst alist (cdr term))))))

(defun nqthm-sublis-lst (alist lst)
  (declare (xargs :mode :program))
  (cond ((null lst) nil)
        (t (cons (nqthm-sublis alist (car lst))
                 (nqthm-sublis-lst alist (cdr lst))))))

)

(deflabel nqthm::nqthm
  :doc
  ":Doc-Section Nqthm

  the Nqthm logic in ACL2~/

  Part of the Nqthm logic has been embedded in ACL2.  By including the
  Nqthm book into ACL2, ACL2 can be made to ``emulate'' Nqthm-1992.

  * This book was primarily intended as a means of testing ACL2 on many
    examples from the Nqthm-1992 benchmarks.  A second objective is as a
    means by which Nqthm users can become familiar with ACL2.  For heavy
    duty system modeling and theorem proving we strongly recommend the
    direct use of ACL2.

  * The ``Nqthm logic'' embedded in ACL2 does not include ~c[V&C$] or any of
    the related functions (e.g., ~c[SUBRP] and ~c[FOR]).

  * Nqthm metafunctions are not supported (though ACL2 metafunctions can
    be defined and used).

  * To convert an Nqthm-1992 script into one that might replay under ACL2,
    certain occurrences of ~c[(*one-*QUOTE btm)] in the Nqthm script must be
    manually replaced by ~c[(*one-*QUOTE (btm))].  The occurrences in question
    are those in which ~c[btm] is the bottom function symbol of a user defined
    shell class and the ~c[(*one*QUOTE btm)] occurs in a quoted Nqthm object.

  * Additional changes to the Nqthm script may be necessary.  There are
    two reasons.  (i) Some aspects of the Nqthm command language either
    could not be reproduced or simply were not reproduced.  (ii) The
    ACL2 theorem prover does not behave exactly the same way the Nqthm
    theorem prover does.

  Relevant ~il[documentation] topics include:~/

  The ACL2 implementation of Nqthm is via a book.  There is nothing
  special about the Nqthm book in the ACL2 system code.  The book
  defines a symbol package, called ~c[\"NQTHM\"], defines the primitive
  Nqthm functions such as ~c[NQTHM::CAR] and ~c[NQTHM::ADD1] within that
  package, provides ACL2 :rewrite and other rules for simplifying
  Nqthm terms, and provides ACL2 macros for implementing certain
  Nqthm-like commands such as ~c[DEFN] and ~c[PROVE-LEMMA].

  Once the book is loaded,
  ~bv[]
  ACL2 !>:start-nqthm
  ~ev[]
  will put ACL2 into its ``Nqthm emulation'' mode.  The ~il[prompt] is
  shown below.  You can type Nqthm-like commands, e.g.,
  ~bv[]
  NQTHM !>(defn length (x) (if (listp x) (add1 (length (cdr x))) 0))

  NQTHM !>(prove-lemma length-append (rewrite)
           (equal (length (append a b)) (plus (length a)(length b))))
  ~ev[]
  You can also evaluate some Nqthm forms, as though you were in
  Nqthm's @c[r-loop]:
  ~bv[]
  NQTHM !>(car 7)
  0
  NQTHM !>(length '(a b c . 7))
  3
  ~ev[]
  However, the evaluation environment is actually ACL2 and some care
  must be taken if the results are to be understood as Nqthm
  computation.  ~l[nqthm-evaluation]

  To exit Nqthm emulation, execute
  ~bv[]
  NQTHM !>:stop-nqthm
  ~ev[]
  To see what actually happens when you start and stop nqthm,
  ~pl[start-nqthm] and ~pl[stop-nqthm].

  However, the main thing ~c[:start-nqthm] does is set ACL2's current
  package to ~c[\"NQTHM\"].  The ``Nqthm emulator'' is just a
  configuration of the top-level ACL2 command loop.  Speaking very
  loosely, we call a command an ``Nqthm command'' if it is typed with
  ~c[\"NQTHM\"] as the current package.

  When you type the Nqthm command
  ~bv[]
  (defn length (x) (if (listp x) (add1 (length (cdr x))) 0))
  ~ev[]
  it is as though you had typed
  ~bv[]
  (NQTHM::DEFN NQTHM::LENGTH (NQTHM::X) ...)
  ~ev[]
  For an extended example of what happens when an Nqthm command form
  is typed ~pl[nqthm-commands].

  The key idea in the embedding of the Nqthm logic in ACL2 is the
  representation within ACL2 of every Nqthm object, including those
  objects created after an ~c[ADD-SHELL] command has axiomatized a new
  data type.  Intuitively, Nqthm numbers are represented as ACL2 integers,
  Nqthm literal atoms as ACL2 symbols, and Nqthm conses as ACL2
  conses.  But things are much more complicated that that.  The
  function ~c[acl2::nqthm-translate] takes an untranslated (but
  translatable) Nqthm term and returns the ACL2 translation of it.
  This function can also be used to see the ACL2 translation of any
  given Nqthm object, of course, by submitting the quoted Nqthm
  object.  The keyword command hack makes it possible to avoid the
  ~c[acl2::] prefix while in the Nqthm package.  Thus,
  ~bv[]
  NQTHM !>:nqthm-translate (PLUS I J K)
  (PLUS I (PLUS J K))
  ~ev[]
  ~l[nqthm-translate] for details of the embedding.")

(deflabel nqthm::nqthm-evaluation
  :doc
  ":Doc-Section Nqthm
  caveats about execution in Nqthm~/
  ~bv[]
  Example Forms:
  NQTHM !>(THM (NEGATIVEP '(*one-*QUOTE MINUS 0)))       ; can be proved

  NQTHM !>(NEGATIVEP '(*one-*QUOTE MINUS 0))             ; evaluates to NIL!
  ~ev[]
  The problem is that ~c[NQTHM::THM] ~c[nqthm-translate]s
  (~pl[nqthm-translate]) its input but no translation is done in
  the second example.~/

  The Nqthm emulator is just the ACL2 read-eval-print loop with the
  current package set to ~c[\"NQTHM\"] and some defaults set
  appropriately.  When a form is typed, the normal ACL2 processing is
  done:  the form is translated via macroexpansion, the result is
  evaluated, and the result of that is printed.  The ``Nqthm command''
  ~pl[nqthm-commands] functions such as ~c[DEFN] and ~c[THM] are defined as
  macros that use ~c[nqthm-translate] to translate the Nqthm terms
  appearing in them.

  Consider what happens when 
  ~bv[]
  NQTHM !>(THM (NEGATIVEP '(*one-*QUOTE MINUS 0)))
  ~ev[]
  is submitted.  Macroexpansion occurs and translates the form, using
  the definition of ~c[NQTHM::THM].  By using the ~c[:]~ilc[trans1] command
  ~bv[]
  NQTHM !>:trans1 (THM (NEGATIVEP '(*one-*QUOTE MINUS 0)))
  ~ev[]
  you can see exactly what ACL2 command is produced by the
  translation, namely,
  ~bv[]
  (acl2::thm (negativep '-1/2)).
  ~ev[]
  ~c[Negativep] is defined to recognize the object ~c[-1/2] as the object that
  Nqthm-1992 calls ~c[(MINUS 0)].  The theorem is proved.

  Now consider what happens when
  ~bv[]
  NQTHM !>(NEGATIVEP '(*one-*QUOTE MINUS 0))
  ~ev[]
  is submitted.  Macroexpansion occurs but no macros are present and
  the result is
  ~bv[]
  (negativep '(*one-*quote minus 0)).
  ~ev[]
  ~c[Negativep] does not recognize the object ~c['(*one-*quote minus 0)] as
  a negative number and so returns ~c[F].

  If one instead did
  ~bv[]
  NQTHM !>:nqthm-translate (NEGATIVEP '(*one-*QUOTE MINUS 0))
  (negativep '-1/2)
  ~ev[]
  and evaluated that,
  ~bv[]
  NQTHM !>(negativep '-1/2)
  ~ev[]
  the result would be ~c[T].

  The user who wishes to do convenient Nqthm evaluation a la Nqthm's
  ~C[R-LOOP] might define a macro such as
  ~bv[]
  NQTHM !>(acl2::defmacro acl2::r (x) (acl2::nqthm-translate (acl2::cadr x)))
  ~ev[]
  Afterwards one could write
  ~bv[]
  NQTHM !>:r (NEGATIVEP '(*one-*QUOTE MINUS 0))
  ~ev[]
  which would expand (by the keyword hack) into ~c[(acl2::r (quote
  (NEGATIVEP '(*one-*QUOTE MINUS 0))))] and thence into ~c[(negativep
  '-1/2)] (by the definition of the ~c[acl2::r] macro).

  One might ask, what exactly is ~c['(*one-*QUOTE MINUS 0)] if it is not a
  ~c[NEGATIVEP].  Tests will show that it is in none of the shell
  classes.  That is, it is not ~c[TRUEP], ~c[FALSEP], a ~c[NUMBERP], a
  ~c[NEGATIVEP], a ~c[LITATOM], or a ~c[LISTP].  Nqthm's logic is
  ``open'' in the sense that there is no axiom that says an object
  must be in one of the known shells.  But as a matter of fact, it is
  impossible to to write an explicit value in Nqthm-1992 that is both
  well-formed and not in one of the known shells.  This was an aspect
  merely of the syntax provided by Nqthm-1992, not of the axioms.
  ~c['(*one-*QUOTE MINUS 0)] is thus a very interesting object to the
  Nqthm user because it is an explicit object that not in any shell
  class.  There are, of course, many others in ACL2.  For example,
  ~c['\"THIS IS A STRING\"] and ~c['#\A] are also explicit objects that
  are not in any Nqthm shell class.")

(deflabel nqthm::nqthm-commands
  :doc
  ":Doc-Section Nqthm
  available Nqthm commands~/

  The following Nqthm-like commands are available from within the
  Nqthm emulator.  Following this enumeration of the commands is a
  discussion of how Nqthm commands are implemented via ACL2 symbol
  packages and macro expansion.~/

  Consider the ``Nqthm command''
  ~bv[]
  NQTHM !>(defn length (x) (if (listp x) (add1 (length (cdr x))) 0))
  ~ev[]
  Because the current package is ~c[\"NQTHM\"], this is read as   
  ~bv[]
  (NQTHM::DEFN NQTHM::LENGTH (NQTHM::X)
    (LISP::IF (NQTHM::LISTP NQTHM::X)
              (NQTHM::ADD1 (NQTHM::LENGTH (NQTHM::CDR NQTHM::X)))
              0))
  ~ev[]
  Indeed, the effect of typing the above Nqthm command can be gotten,
  whether one is in the emulator or not, by typing the latter form,
  which is independent of the current package setting.

  Note that nearly all the symbols typed in the Nqthm emulator are
  interned in the ~c[\"NQTHM\"] package.  But some are not:  ~ilc[if], for
  example, is imported from the ~c[\"ACL2\"] package to the
  ~c[\"NQTHM\"] package, so that ~c[NQTHM::IF] is ~c[acl2::if] (which is
  in fact ~c[lisp::if]).  To see what symbols are imported into the
  ~c[\"NQTHM\"] package, type ~c[:pe \"NQTHM\"].

  ~c[NQTHM::DEFN] is defined in ACL2 as a macro so that the above
  ~c[NQTHM::DEFN] command expands into the ACL2 ~ilc[defun] command:
  ~bv[]
  (defun NQTHM::LENGTH (NQTHM::X)
    (LISP::IF (NQTHM::LISTP NQTHM::X)
              (NQTHM::ADD1 (NQTHM::LENGTH (NQTHM::CDR NQTHM::X)))
              0))
  ~ev[]

  In order for this definition to be admissible under the ACL2
  definitional principle, some measure of the argument must be shown
  to be decreasing.  By arranging the ~ilc[acl2-defaults-table]
  appropriately, ~c[:start-nqthm] makes the default measure and
  relation be ~c[NQTHM::COUNT] and ~c[NQTHM::LESSP].  Rules in the ACL2
  data base loaded by the Nqthm book include
  ~bv[]
  (implies (NQTHM::LISTP x)
           (NQTHM::LESSP (NQTHM::COUNT (NQTHM::CDR x))
                         (NQTHM::COUNT x)))
  ~ev[]
  so that the admission of this definition is straightforward.

  Additional rules include
  ~bv[]
  (equal (NQTHM::CDR (NQTHM::CONS x y)) y)
  ~ev[]
  and
  ~bv[]
  (implies (not (NQTHM::LISTP x))
           (equal (NQTHM::CDR x) 0))
  ~ev[]
  In fact, every ~c[REWRITE] rule available to the Nqthm-1992 theorem
  prover about these primitive Nqthm function symbols is available as
  a rule to the ACL2 theorem prover.  Thus, roughly speaking, the ACL2
  theorem prover will behave the way the Nqthm theorem prover behaves
  on terms composed of Nqthm function symbols.  But this is not
  strictly true and divergences between the ACL2 theorem prover and
  the Nqthm theorem prover may become manisfest if you try to replay
  an Nqthm proof script.  A primary cause of such divergence is the
  fact that the type systems in ACL2 and Nqthm-1992 are different.
  Another difference is that the ~ilc[if]-normalization algorithms are
  different.

  Including the Nqthm book in your session extends the ACL2 logic with
  the primitive functions of the Nqthm logic (defined in the
  ~c[\"NQTHM\"] symbol package).  You remain in that logical extension
  (or further extensions of it) henceforth, whether you are ``in'' the
  Nqthm emulator or not.  The definition of ~ilc[LENGTH] via ~c[DEFN]
  ``in'' the Nqthm emulator actually defines ~c[NQTHM::LENGTH] via
  ~ilc[defun] in the normal ACL2 logic.  The ``normal'' ACL2 functions
  are available to you while you are ``in'' the emulator and the Nqthm
  functions are available to you while you are not ``in'' the
  emulator.  The emulator only makes it convenient to type Nqthm terms
  without having to type the ~c[NQTHM::]  prefix.  But ``mixed-mode''
  terms are well-formed and it is, in principle, possible to prove
  such theorems as
  ~bv[]
  (implies (NQTHM::LISTP x)
           (equal (NQTHM::CAR x)
                  (acl2::car x)))
  ~ev[]
  relating the symbols in the two packages.  However, you will find
  that such proofs are difficult because the rules provided by the
  Nqthm book were not designed for mixed-mode terms.  (For example,
  while ~c[NQTHM::CAR] is defined in terms of ~c[acl2::car,] the
  definition is disabled by the Nqthm book so that ~c[acl2::car] does
  not get introduced into Nqthm proofs.)")

(deflabel nqthm::nqthm-translate
  :doc
  ":Doc-Section Nqthm

  how Nqthm objects are represented in ACL2~/
  ~bv[]
  Example Forms:
  NQTHM !>:nqthm-translate '(abc 23 *one-*true)
  '(ABC 23 LISP::T . NIL)

  NQTHM !>:nqthm-translate (plus i j k)
  (plus i (plus j k))
  ~ev[]
  ~c[Nqthm-translate] takes as input a translatable but untranslated Nqthm-1992
  form and returns as output the translated ACL2 expression representing it.
  In the special case in which the input is a quoted object, the quoted
  object returned illustrates the embedding of the Nqthm universe into the
  ACL2 universe.~/

  We discuss the embedding below, illustrated by nqthm-translating some
  quoted Nqthm objects.

  Nqthm ~c[NUMBERP]s are nonnegative integers.
  ~bv[]
  NQTHM !>:nqthm-translate '37
  '37
  ~ev[]
  Nqthm ~c[NEGATIVEP]s are negative integers, 
  ~bv[]
  NQTHM !>:nqthm-translate '-37
  '-37
  ~ev[]
  but ~c[(MINUS 0)] is the negative rational ~c[-1/2]!

  Conventional Nqthm ~c[LITATOM]s are symbols in the ~c[\"NQTHM\"] package: 
  ~bv[]
  NQTHM !>:nqthm-translate 'ABC
  'ABC                             ; (LISP::QUOTE NQTHM::ABC) = 'NQTHM::ABC

  NQTHM !>:nqthm-translate NIL
  'NIL                             ; 'NQTHM::NIL
  ~ev[]
  But unconventional ones, such as ~c[(PACK 1200)], are represented with the
  ~c[*one-QUOTE] notation ~pl[*one*QUOTE-notation].

  Nqthm truth values are symbols in the ~c[\"LISP\"] package.
  ~bv[]
  NQTHM !>:nqthm-translate T
  'LISP::T

  NQTHM !>:nqthm-translate F
  'LISP::NIL
  ~ev[]

  Nqthm conses are ACL2 conses.  But ~c[nqthm-translate] converts the
  implicit ~c[lisp::nil]s at the ends of quoted list constants to
  ~c[NQTHM::NIL]s.
  ~bv[]
  NQTHM !>:nqthm-translate '(A B C)
  '(A B C . NIL)                   ; '(NQTHM::A NQTHM::B NQTHM::C . NQTHM::NIL)
  ~ev[]

  Finally, user-defined shell objects are represented with a variant
  of Nqthm-1992's ~c[*one*QUOTE] notation.  If ~c[PUSH] is a shell constructor
  of two arguments and ~c[EMPTY] is the bottom function symbol of that
  shell, then the object returned by
  ~bv[]
  (PUSH 2 (PUSH 1 (EMPTY))),
  ~ev[]
  which could be written in Nqthm-1992 as
  ~bv[]
  '(*one*QUOTE PUSH 2 (*one*QUOTE PUSH 1 (*one*QUOTE EMPTY)))
  ~ev[]
  is represented by the ACL2 object
  ~bv[]
  '(*one*QUOTE PUSH 2 (*one*QUOTE PUSH 1 (*one*QUOTE (EMPTY)))).
  ~ev[]
  Note the extra level of parentheses around the bottom function symbol.
  ~l[*one*QUOTE-notation].

  These examples illustrate several deep problems which had to be
  solved in order to embed the Nqthm logic.  One of those problems is
  distinquishing Nqthm's false object, ~c[F], from the symbol used to
  mark the end of lists, ~c[NIL].  These two objects are identified in
  Common Lisp.  It was decided to use ~c[lisp::nil] as the
  representation of Nqthm's ~c[F] so that Nqthm's ~c[IF] could be
  ~c[lisp::if] and Nqthm's ~c[EQUAL] could be ~c[lisp::equal].  Such an
  arrangement was necessary because ACL2 gives special significance to
  its ~c[if] and ~c[equal] (for example, in the notion of ``governance''
  and the parsing of theorems into ~c[:]~ilc[rewrite] rules).  Nqthm's
  ~c[NIL] is represented by ~c[NQTHM::NIL].  This makes nqthm-translated
  list constants look odd, because proper lists end with the unusual
  ``~c[. NIL]'' dotted pair.  The ~c[NIL] shown, of course, is Nqthm's
  ~c[NIL], not LISP's ~c[nil].

  Another deep problem is that Nqthm functions are total ACL2 functions
  and may be applied to explicit objects other than those in the Nqthm
  universe.  For example, Nqthm's ~c[CONS] can be applied to the symbol
  ~c[*one*QUOTE].  Consider for example
  ~bv[]
  NQTHM !>(CONS '*one*QUOTE '(PACK 1200))
  ~ev[]
  If the result were ~c['(*one*QUOTE PACK 1200)] then the embedding would
  not be consistent with Nqthm's axioms because
  ~c['(*one*QUOTE PACK 1200)] is a ~c[LITATOM] and Nqthm's ~c[CONS] can 
  never produce a ~c[LITATOM].  In fact, the ~c[CONS] expression above 
  evaluates to ~c['(*one*QUOTE CONS *one*QUOTE (PACK 1200))].

  To establish that the embedding is consistent with the Nqthm axioms,
  the certification of the Nqthm book includes the verification of all
  the Nqthm axioms (except those dependent upon ~c[V&C$]) as ACL2 theorems.")

(mutual-recursion

(defun nqthm-translate (term)

; We assume term is the result of reading into the NQTHM package an
; s-expression that is translatable by Nqthm-1992.  We call such a term an
; "untranslated translatable Nqthm term."  We return the translated Nqthm term,
; except that evgs are represented differently.

  (declare (xargs :mode :program))
  (cond ((atom term)
         (cond ((eq term 'nqthm::t) *t*)
               ((eq term 'nqthm::f) *nil*)
               ((eq term 'nqthm::nil) '(quote nqthm::nil))
               ((symbolp term) term)
               (t (list 'quote (convert-nqthm-evg term)))))
        ((eq (car term) 'quote)
         (list 'quote (convert-nqthm-evg (cadr term))))
        ((member-equal (car term) '(nqthm::and nqthm::or nqthm::plus nqthm::times))
         (xxxjoin (car term)
                  (nqthm-translate-lst (cdr term))))
        ((car-cdr-symbolp (car term))
         (make-nqthm-car-cdr-nest (car term) (nqthm-translate (cadr term))))
        ((eq (car term) 'nqthm::cond)
         (cond ((eq (car (cadr term)) 'nqthm::t)
                (nqthm-translate (cadr (cadr term))))
               (t (list 'if
                        (nqthm-translate (car (cadr term)))
                        (nqthm-translate (cadr (cadr term)))
                        (nqthm-translate (cons 'nqthm::cond (cddr term)))))))
        ((eq (car term) 'nqthm::case)
         (cond ((eq (car (caddr term)) 'nqthm::otherwise)
                (nqthm-translate (cadr (caddr term))))
               (t (list 'if
                        (nqthm-translate `(equal ,(cadr term)
                                                 (quote ,(car (caddr term)))))
                        (nqthm-translate (cadr (caddr term)))
                        (nqthm-translate `(nqthm::case ,(cadr term)
                                                       ,@(cdddr term)))))))
        ((eq (car term) 'nqthm::let)
         (nqthm-sublis (pairlis$ (strip-cars (cadr term))
                                (nqthm-translate-lst (strip-cadrs (cadr term))))
                       (nqthm-translate (caddr term))))
        ((eq (car term) 'nqthm::list)
         (xxxjoin 'nqthm::cons
                  (append (nqthm-translate-lst (cdr term))
                          (list '(quote nqthm::nil)))))
	;; Was list*
        ((eq (car term) 'nqthm::list)
         (xxxjoin 'nqthm::cons
                  (nqthm-translate-lst (cdr term))))
        (t (cons (car term) (nqthm-translate-lst (cdr term))))))

(defun nqthm-translate-lst (lst)
  (cond ((atom lst) nil)
        (t (cons (nqthm-translate (car lst))
                 (nqthm-translate-lst (cdr lst))))))
)

(deflabel nqthm::one-quote-notation
  :DOC
  ":Doc-Section nqthm-translate

  how odd Nqthm objects are embedded in ACL2~/

  The universe of the Nqthm logic contains some objects that are
  unconventional from the Common Lisp perspective.  These are
  generally represented by ACL2 objects reminiscent of Nqthm's
  ~c[*one*QUOTE] notation.  ~c[Nqthm-translate] (~pl[nqthm-translate])
  is able to handle Nqthm's ~c[*one*QUOTE] notation, but manual
  intervention is required when the bottom objects of user defined
  shells are involved.

  Among the odd Nqthm quoted objects to be translated are
  ~bv[]
  Nqthm-1992     Nqthm-1992              ACL2 Object
    term      *one*QUOTE notation

     T        '*one*TRUE                   lisp::t
     F        '*one*FALSE                  lisp::nil
  (MINUS 0)   '(*one*QUOTE MINUS 0)        -1/2
  (PACK 1200) '(*one*QUOTE PACK 1200)      '(*one*quote pack 1200)
  (EMPTY)     '(*one*QUOTE EMPTY)          '(*one*quote (empty))
  ~ev[]
  In addition, the embedding of Nqthm into ACL2 must assign a value to
  such terms as ~c[(CONS '*one*QUOTE '(PACK 1200))] and that value must
  be consistent with Nqthm's axioms.  Such terms do not arise in
  Nqthm-1992 because it is impossible to write the explicit constant
  ~c['*one*QUOTE] as a symbol.  We discuss some details of the embedding
  here.  We assume the reader is familiar with Nqthm's ~c[*one*QUOTE]
  notation.  We tend to display Nqthm terms and objects in upper
  case and ACL2 terms and objects in lower case, though, technically
  case is insignificant in this dicussion.~/

  ~c[Nqthm-translate] is able to translate quoted Nqthm constants
  involving ~c[*one*QUOTE] notation.  ~l[nqthm-translate].  As
  illustrated there
  ~bv[]
  NQTHM !>:nqthm-translate '(*one*TRUE (*one*QUOTE MINUS 0) . *one*FALSE)
  '(lisp::t -1/2)                  ; '(lisp::t -1/2 . lisp::nil)
  ~ev[]
  In addition,
  ~bv[]
  NQTHM !>:nqthm-translate '(*one*QUOTE PACK 1200)
  '(*one*quote pack 1200)
  ~ev[]
  But Nqthm-1992's ~c[*one*QUOTE] notation is ambiguous (in a certain
  sense) when it comes to representing the bottom objects of user
  defined shells.  One cannot tell, from the structure alone, whether
  ~c['(*one*QUOTE fn)] is a bottom object or an application of a shell
  constructor of no arguments.  To convert an Nqthm-1992 script into
  one that might replay under ACL2 it is necessary to edit the
  Nqthm-1992 manually and distinguish those two cases.  In particular,
  when a bottom object is written in ~c[*one*QUOTE] notation, the bottom
  function symbol must be enclosed in parentheses before the form is
  submitted to ACL2.  Nqthm-translate cannot recognize this situation
  and fix it automatically.  We explain and illustrate below.

  Consider first the classic ~c[PUSH] shell constructor of two
  arguments with ~c[EMPTY] as the bottom function symbol.  Recall that
  the value of the term ~c[(EMPTY)], i.e., the bottom object of the
  shell, may be written as the quoted constant ~c['(*one*QUOTE EMPTY)] in
  Nqthm-1992.  In ACL2 Nqthm scripts, that quoted object must be
  written ~c['(*one*quote (empty))].  This transformation is not
  performed by ~c[nqthm-translate].

  To explain why the Nqthm emulator is so arranged, consider another
  user-defined shell class.  Let the shell constructor be named
  ~c[UNCERTAIN] and suppose it takes no arguments.  In Nqthm-1992's
  ~c[*one*QUOTE] notation, how does one write the quoted form of the term
  ~c[(UNCERTAIN)]?  The answer is ~c['(*one*QUOTE UNCERTAIN)].  Note that
  from syntax alone it is impossible in the Nqthm-1992 notation to
  distinguish a quoted bottom object from a quoted constructed object,
  if the constructor has no arguments.  This causes trouble when one
  considers the fact that Nqthm's ~c[COUNT] is axiomatized so that on a
  bottom object, e.g., ~c['(*one*QUOTE EMPTY)], it returns ~c[0] but on a
  constructed object with no components, e.g.,
  ~c['(*one*QUOTE UNCERTAIN)], it returns ~c[1].  In order to define 
  ~c[NQTHM::COUNT] in ACL2 it was necessary to make the structure of a
  bottom object differ from that of a constructed object.
  Furthermore, ~c[nqthm-translate] could not be defined so as to
  implement this distinction automatically because insufficient
  context is provided ~-[] ~c[nqthm-translate] does not know which
  symbols are bottom objects of user shells and which are
  constructors.  Such contextual information could be recovered from
  the ACL2 ~il[state] at the time of the nqthm-translation, but the
  translation occurs during macroexpansion of ~c[DEFN] and
  ~c[PROVE-LEMMA] forms, and ACL2 macros do not have access to
  ~ilc[state].

  Another problem involving ~c[*one*QUOTE] is that explicit quoted
  objects other than those constructed by Nqthm can be supplied as
  arguments to Nqthm functions.  For example, what ACL2 object is
  constructed by
  ~bv[]
  NQTHM !>(CONS '*one*QUOTE '(PACK 1200))
  ~ev[]
  If it were ~c['(*one*quote pack 1200)], as one might at first expect,
  then the embedded logic would be inconsistent with the Nqthm logic,
  because ~c['(*one*quote pack 1200)] is a ~c[LITATOM] and ~c[CONS] never
  constructs a ~c[LITATOM].  Actually, the ACL2 object constructed is
  ~c['(*one*quote cons *one*quote (pack 1200))].  Thus we see that Nqthm's
  ~c[CONS] is not just ACL2's ~c[cons].")

(defun nqthm::numberp (x)
  (and (integerp x)
       (not (< x 0))))

(defun nqthm::zero () 0)

(defun nqthm::add1 (x)
  (if (and (integerp x)
           (not (< x 0)))
      (+ 1 x)
      1))

(defun nqthm::sub1 (x)
  (if (integerp x)
      (if (= x 0)
          0
          (if (< x 0)
              0
              (- x 1)))
      0))

(defun nqthm::negativep (x)
  (or (and (integerp x)
           (< x 0))
      (equal x *minus-0*)))

(defun nqthm::minus (x)
  (if (and (integerp x)
           (< 0 x))
      (- x)
      *minus-0*))

(defun nqthm::negative-guts (x)
  (if (and (integerp x)
           (< x 0))
      (- x)
      0))

(defun listp-escapep (x)
  (and (consp x)
       (eq (car x) *one-shell-quote-mark*)
       (consp (cdr x))
       (eq (cadr x) 'nqthm::cons)
       (consp (cddr x))
       (eq (caddr x) *one-shell-quote-mark*)
       (consp (cdddr x))
       (null (cddddr x))))

(defun nqthm::listp (x)
  (and (consp x)
       (or (not (eq (car x) *one-shell-quote-mark*))
           (listp-escapep x))))

(defun nqthm::nlistp (x)
  (not (nqthm::listp x)))

(defun nqthm::car (x)
  (if (consp x)
      (if (eq (car x) *one-shell-quote-mark*)
          (if (listp-escapep x)
              *one-shell-quote-mark*
              0)
          (car x))
      0))

(defun nqthm::cdr (x)
  (if (consp x)
      (if (eq (car x) *one-shell-quote-mark*)
          (if (listp-escapep x)
              (cadddr x)
              0)
          (cdr x))
      0))

(defun nqthm::my-cons (x y)
  (if (eq x *one-shell-quote-mark*)
      (list *one-shell-quote-mark* 'nqthm::my-cons x y)
      (cons x y)))

(defun ascii-code (x)
  (declare (xargs :guard (and (characterp x)
                              (standard-char-p x))))


  (the (integer 0 127)
       (char-code (the character x))))

(defun ascii-code-lst (lst)

  (declare (xargs :guard (standard-char-listp lst)
                  :hints (("Goal" :in-theory (disable member)))
                  :guard-hints
                  (("Goal" :in-theory (cons '(:definition STANDARD-CHAR-LISTP)
                                            (disable member))))))
  (if (endp lst)
      0
      (cons (ascii-code (car lst))
            (ascii-code-lst (cdr lst)))))
      
(defun ascii-code-to-char-alist (lst)
  (declare (xargs :guard (standard-char-listp lst)
                  :hints (("Goal" :in-theory (disable member)))
                  :guard-hints
                  (("Goal" :in-theory (cons '(:definition STANDARD-CHAR-LISTP)
                                            (disable member))))))
  (cond ((endp lst) nil)
        (t (cons (cons (ascii-code (car lst)) (car lst))
                 (ascii-code-to-char-alist (cdr lst))))))

(defconst *ascii-code-to-char-table*
  (compress1 'ascii-code-to-char-table
             (cons (list :header
                         :dimensions '(128)
                         :maximum-length 129
                         :default nil
                         :name 'ascii-code-to-char-table)
                   (ascii-code-to-char-alist *standard-chars*))))

(defun ascii-char-ug (n)
  (if (and (integerp n)
           (<= 0 n)
           (<= n 127))
      (aref1 'ascii-code-to-char-table
             *ascii-code-to-char-table*
             n)
      nil))

(defun ascii-char-lst (lst)

  (cond ((atom lst) nil)
        (t (cons (ascii-char-ug (car lst))
                 (ascii-char-lst (cdr lst))))))

(defun nqthm-litatom-char-codep (n)

  (if (member n '(36 38 42 43 45 48 49 50 51 52 53 54 55 56 57 60 61 62 63 65 66
                     67 68 69 70 71 72 73 74 75 76 77 78 79 80 81 82 83 84 85 86
                     87 88 89 90 94 95 123 125 126))
      t
      nil))

(defun all-nqthm-litatom-char-codep (lst)
  (if (atom lst)
      (equal lst 0)
      (and (nqthm-litatom-char-codep (car lst))
           (all-nqthm-litatom-char-codep (cdr lst)))))

(defun unpacked-nqthm-litatomp (lst)
  (and (all-nqthm-litatom-char-codep lst)
       (not (equal lst 0))
       (<= 65 (car lst))
       (<= (car lst) 90)))

(defun litatom-escapep (x)
  (and (consp x)
       (eq (car x) *one-shell-quote-mark*)
       (consp (cdr x))
       (eq (cadr x) 'nqthm::pack)
       (consp (cddr x))
       (not (unpacked-nqthm-litatomp (caddr x)))
       (null (cdddr x))))

(defun nqthm::litatom (x)
  (if (symbolp x)
      (or (eq x 'IF)
          (eq x 'EQUAL)
          (eq x 'NOT)
          (eq x 'IMPLIES)
          (eq x 'IFF)
          (eq x 'QUOTE)
          (and (equal (symbol-package-name x) "NQTHM")
               (standard-char-listp (coerce (symbol-name x) 'list))
               (unpacked-nqthm-litatomp
                (ascii-code-lst (coerce (symbol-name x) 'list)))))
      (litatom-escapep x)))
           
(defun nqthm::unpack (x)
  (declare (xargs :guard-hints (("Goal" :in-theory (disable ascii-code)))))
  (if (nqthm::litatom x)
      (if (symbolp x)
          (ascii-code-lst (coerce (symbol-name x) 'list))
          (caddr x))
      0))

(defthm nqthm-litatom-char-codep-implies-characterp
  (implies (nqthm-litatom-char-codep n)
           (characterp (ascii-char-ug n))))

(defthm nqthm-litatom-char-codep-implies-standard-char-p
  (implies (nqthm-litatom-char-codep n)
           (standard-char-p (ascii-char-ug n))))

(defthm nqthm-litatom-char-codep-implies-ascii-code-inverts-ascii-char-ug
  (implies (nqthm-litatom-char-codep n)
           (equal (ascii-code (ascii-char-ug n)) n)))

(defthm ascii-char-ug-inverts-ascii-code
  (implies (and (characterp ch)
                (standard-char-p ch))
           (equal (ascii-char-ug (ascii-code ch)) ch)))

(in-theory (disable nqthm-litatom-char-codep ascii-char-ug ascii-code))

(defthm all-nqthm-litatom-char-codep-implies-standard-char-listp
  (implies (all-nqthm-litatom-char-codep lst)
           (standard-char-listp (ascii-char-lst lst))))

(defthm unpacked-nqthm-litatomp-implies-standard-char-listp 
  (implies (unpacked-nqthm-litatomp lst)
           (standard-char-listp (ascii-char-lst lst))))

(defun nqthm::pack (lst)
  (declare (xargs :guard-hints
                  (("[1]Goal" :in-theory (enable nqthm-litatom-char-codep)))))
  (if (unpacked-nqthm-litatomp lst)
      (let ((str (coerce (ascii-char-lst lst) 'string)))
        (cond ((equal str "IF") 'IF)
              ((equal str "EQUAL") 'EQUAL)
              ((equal str "NOT") 'NOT)
              ((equal str "IMPLIES") 'IMPLIES)
              ((equal str "IFF") 'IFF)
              ((equal str "QUOTE") 'QUOTE)
              (t (intern-in-package-of-symbol str 'nqthm::pack))))
      (list *one-shell-quote-mark*
            'nqthm::pack
            lst)))


(defthm
  all-nqthm-litatom-char-codep-implies-ascii-code-lst-inverts-ascii-char-lst
  (implies (all-nqthm-litatom-char-codep lst)
           (equal (ascii-code-lst (ascii-char-lst lst)) lst)))

(defthm unpacked-nqthm-litatomp-implies-ascii-code-lst-inverts-ascii-char-lst
  (implies (unpacked-nqthm-litatomp lst)
           (equal (ascii-code-lst (ascii-char-lst lst)) lst)))

(defthm ascii-char-lst-inverts-ascii-code-lst 
  (implies (standard-char-listp lst) 
           (equal (ascii-char-lst (ascii-code-lst lst)) lst)))



(defthm ascii-char-ug-elim
  (implies (ascii-char-ug n)
           (equal (ascii-code (ascii-char-ug n)) n))
  :rule-classes :elim
  :hints (("Goal" :in-theory (enable ascii-char-ug))))

(defthm equal-c-ascii-char-ug-implies-equal-k-n
  (implies (and (characterp c)
                (standard-char-p c))
           (equal (equal c (ascii-char-ug n))
                  (equal (ascii-code c) n))))

(defthm symbol-name-intern-nqthm
  (implies
   (stringp s)
   (and
    (symbolp (intern-in-package-of-symbol s 'NQTHM::PACK))
    (equal (symbol-name (intern-in-package-of-symbol s 'NQTHM::PACK)) s)
    (stringp
     (symbol-package-name (intern-in-package-of-symbol s 'NQTHM::PACK)))))
  :hints (("Goal" :in-theory (disable symbol-name-intern-in-package-of-symbol)
                  :use (:instance symbol-name-intern-in-package-of-symbol
                                  (any-symbol 'NQTHM::PACK)))))


(defun sum-lst (lst)
  (cond ((atom lst) 0)
        (t (+ (if (integerp (car lst))
                  (car lst)
                  0)
              (sum-lst (cdr lst))))))

(defun nqthm-count-symbol (s)
  (declare (xargs :guard (and (nqthm::litatom s) (symbolp s))))
  (+ 1
     (length (symbol-name s))
     (sum-lst (ascii-code-lst (coerce (symbol-name s) 'list)))))

(defthm nqthm-litatom-char-codep-implies-non-negative-integerp
  (implies (nqthm-litatom-char-codep x) (and (integerp x) (< 0 x)))
  :rule-classes :compound-recognizer
  :hints (("Goal" :in-theory (enable nqthm-litatom-char-codep))))


(defun count-fn (flg x)
  (declare (xargs :verify-guards nil))
  (if (eq flg 'sum)
      (if (atom x)
          0
          (+ (count-fn t (car x))
             (count-fn 'sum (cdr x))))
      (cond ((eq x t) 0)
            ((eq x nil) 0)
            ((nqthm::numberp x) x)
            ((nqthm::negativep x) (+ 1 (nqthm::negative-guts x)))
            ((nqthm::listp x)
             (+ 1
                (count-fn t (nqthm::car  x))
                (count-fn t (nqthm::cdr  x))))
            ((nqthm::litatom x)
             (cond ((symbolp x) (nqthm-count-symbol x))
                   (t (+ 1 (count-fn t (nqthm::unpack x))))))
            ((and (consp x)
                  (eq (car x) *one-shell-quote-mark*)
                  (consp (cdr x))
                  (symbolp (cadr x))
                  (true-listp (cddr x)))
             (1+ (count-fn 'sum (cddr x))))
            (t 0))))

(defthm nqthm::numberp-compound-recognizer
  (equal (nqthm::numberp x)
         (and (integerp x) (>= x 0)))
  :rule-classes :compound-recognizer)

(defthm nqthm::numberp-type-set-inverter1
  (equal (and (nqthm::numberp x) (not (equal x 0)))
         (and (integerp x) (< 0 x)))
  :rule-classes ((:type-set-inverter :type-set 2)))

(defthm nqthm::numberp-type-set-inverter2
  (equal (nqthm::numberp x)
         (or (and (nqthm::numberp x) (not (equal x 0))) (equal x 0)))
  :rule-classes ((:type-set-inverter :type-set 3)))

(defthm integerp-count-fn
  (integerp (count-fn flg x))
  :rule-classes :type-prescription
  :hints (("Goal" :in-theory (disable nqthm::listp nqthm::litatom))))

(verify-guards count-fn)


(defthm sum-lst-ascii-code-lst
  (implies (standard-char-listp lst)
           (<= 0 (sum-lst (ascii-code-lst lst))))
  :rule-classes :linear)

(defthm sum-lst-ascii-code-lst-coerce-symbol-name 
  (implies (and (nqthm::litatom x)
                (symbolp x))
           (<= 0 (sum-lst (ascii-code-lst (coerce (symbol-name x) 'list)))))
  :rule-classes :linear)

(defthm non-negative-count-fn
  (and (integerp (count-fn flg x))
       (<= 0 (count-fn flg x)))
  :rule-classes :type-prescription
  :hints (("Goal" :in-theory (disable nqthm::listp nqthm::litatom))))

(defun nqthm::count (x) (count-fn t x))



(defthm len-ascii-code-lst 
  (implies (standard-char-listp x)
           (equal (len (ascii-code-lst x)) (len x))))

(defthm len-plus-sum-lst-is-count-fn
  (implies (all-nqthm-litatom-char-codep lst)
           (equal (+ (len lst) (sum-lst lst))
                  (count-fn t lst)))
  :rule-classes nil
  :hints (("Goal" :in-theory (disable nqthm-litatom-char-codep

                                      ))))

(defthm nqthm-count-symbol-is-count-fn-unpack
  (implies (and (nqthm::litatom x)
                (symbolp x))
           (equal (nqthm-count-symbol x)
                  (+ 1 (count-fn t (nqthm::unpack x)))))
  :hints (("Goal"
           :use (:instance len-plus-sum-lst-is-count-fn
                           (lst (ASCII-CODE-LST (COERCE (SYMBOL-NAME X) 'LIST)))))))



(defmacro nqthm::defn (fn args body &rest rst)

  ":Doc-Section nqthm-commands

  Nqthm ~c[DEFN] command~/
  ~bv[]
  General Forms:
  (DEFN fn args body)
  (DEFN fn args body ((relation measure)))
  (DEFN fn args body :measure measure :hints ...)
  ~ev[]~/
  This ACL2 macro uses ~c[nqthm-translate] (~pl[nqthm-translate])
  to translate body to into the ACL2 fragment of Nqthm and then
  generates a suitable ~c[defun] form.  Note that if one form is
  provided after ~c[body] then it is assumed to be an Nqthm-style hint
  specifying a well-founded relation and measure.  While Nqthm allows
  multiple relations and measures, ACL2 allows only one.  If more than
  one form follows ~c[body] it is assumed to be an ACL2-style keyword
  list and is included within a ~c[(declare (xargs ...))]."

  `(defun ,fn ,(nqthm-translate-lst args)
     ,@(cond ((null rst) nil)
             ((keywordp (car rst))
              `((declare (xargs ,@rst))))
             ((and (null (cdr rst))
                   (null (cdr (car rst))))
              `((declare (xargs :well-founded-relation ,(car (car (car rst)))
                                :measure ,(cadr (car (car rst)))))))
             (t `((declare (xargs :well-founded-relations-and-measures ,rst)))))
     ,(nqthm-translate body)))

; We here develop the functions which guess what the rule class of a term
; should be given that the Nqthm user has said it should be a REWRITE rule.

(mutual-recursion

(defun find-fn-applied-to-distinct-vars (term)
  (declare (xargs :mode :program))
  (cond ((atom term) nil)
        ((eq (car term) 'quote) nil)
        ((and (symbol-listp (cdr term))
              (no-duplicatesp (cdr term)))
         term)
        (t (find-fn-applied-to-distinct-vars-lst (cdr term)))))

(defun find-fn-applied-to-distinct-vars-lst (lst)
  (cond ((endp lst) nil)
        (t (or (find-fn-applied-to-distinct-vars (car lst))
               (find-fn-applied-to-distinct-vars-lst (cdr lst))))))

)

(defun probable-nqthm-type-prescriptionp1 (obj term)
  (declare (xargs :mode :program))
  (cond ((atom term) nil)
        ((eq (car term) 'quote) nil)
        ((member (car term) '(implies nqthm::and nqthm::or))
         (and (probable-nqthm-type-prescriptionp1 obj (cadr term))
              (probable-nqthm-type-prescriptionp1 obj (caddr term))))
        ((eq (car term) 'not)
         (probable-nqthm-type-prescriptionp1 obj (cadr term)))
        ((eq (car term) 'equal)
         (or (and (equal obj (cadr term))
                  (member (caddr term) (cdr obj)))
             (and (equal obj (caddr term))
                  (member (cadr term) (cdr obj)))
             (and (equal obj (cadr term))
                  (consp (caddr term))
                  (null (cdr (caddr term))))
             (and (equal obj (caddr term))
                  (consp (cadr term))
                  (null (cdr (cadr term))))))
        (t (and (null (cddr term))
                (equal (cadr term) obj)))))

(defun probable-nqthm-type-prescriptionp (term)

; We consider a term a probable type-prescription lemma if it contains an
; "object subterm", obj, of the form (fn v1 ... vk) where the vi are distinct
; vars, and except for the propositional connectives IMPLIES, AND, OR, and NOT,
; all other subterms are either of the form (r obj), (equal obj vi), (equal vi
; obj), (equal obj (singleton)), or (equal (singleton) obj).

  (declare (xargs :mode :program))
  (let ((obj (find-fn-applied-to-distinct-vars term)))
    (and obj
         (probable-nqthm-type-prescriptionp1 obj term))))

(defun probable-type-set-term1 (op var term)
  (declare (xargs :mode :program))
  (cond ((atom term) nil)
        ((eq (car term) 'quote) nil)
        ((eq (car term) op)
         (and (probable-type-set-term1 op var (cadr term))
              (probable-type-set-term1 op var (caddr term))))
        ((eq op 'nqthm::or)
         (and (null (cddr term))
              (equal (cadr term) var)))
        (t (and (eq (car term) 'not)
                (null (cddr (cadr term)))
                (equal (cadr (cadr term)) var)))))

(defun probable-type-set-term (var term)
  (declare (xargs :mode :program))
  (or (probable-type-set-term1 'nqthm::or var term)
      (probable-type-set-term1 'nqthm::and var term)))

(defun probable-nqthm-compound-recognizerp (term)

; We consider a term a probable compound-recognizer lemma if it is of one of the
; forms

; (IMPLIES (fn X) type-set-term)
; (IMPLIES (NOT (fn X)) type-set-term)
; (EQUAL (fn X) type-set-term)

; where type-set-term is either
; (OR (r1 X) ... (rk X))
; or
; (AND (NOT (r1 X)) ... (NOT (rk X))).

  (declare (xargs :mode :program))
  (case-match term
              (('implies ('not (& x)) ts)
               (and (symbolp x)
                    (probable-type-set-term x ts)))
              (('implies (& x) ts)
               (and (symbolp x)
                    (probable-type-set-term x ts)))
              (('equal (& x) ts)
               (and (symbolp x)
                    (probable-type-set-term x ts)))
              (& nil)))

(defun probable-nqthm-linearp (term)
  (declare (xargs :mode :program))
  (case-match term
              (('nqthm::lessp & &) t)
              (('not ('nqthm::lessp & &)) t)
              (('implies & ('nqthm::lessp & &)) t)
              (('implies & ('not ('nqthm::lessp & &))) t)
              (& nil)))

(defun nqthm-flatten-ands-in-lit (term)

; We assume that term is an nqthm-translated term.

  (declare (xargs :mode :program))
  (cond ((atom term)
         (if (equal term *t*) nil (list term)))
        ((eq (car term) 'if)
         (let ((c1 (cadr term))
               (c2 (caddr term))
               (c3 (cadddr term)))
           (cond ((equal c2 *nil*)
                  (append (nqthm-flatten-ands-in-lit
                           (if (and (consp c1)
                                    (eq (car c1) 'not))
                               (cadr c1)
                               (list 'not c1)))
                          (nqthm-flatten-ands-in-lit c3)))
                 ((equal c3 *nil*)
                  (append (nqthm-flatten-ands-in-lit c1)
                          (nqthm-flatten-ands-in-lit c2)))
                 (t (list term)))))
        ((eq (car term) 'nqthm::and)
         (append (nqthm-flatten-ands-in-lit (cadr term))
                 (nqthm-flatten-ands-in-lit (caddr term))))
        (t (list term))))

(defun nqthm-unprettyify1 (hyps pairs)
  (declare (xargs :mode :program))
  (cond ((null pairs) nil)
        (t (cons (cons (append hyps (caar pairs)) (cdar pairs))
                 (nqthm-unprettyify1 hyps (cdr pairs))))))

;(defun nqthm-unprettyify (term)
;  (declare (xargs :mode :program))
;  (case-match term
;              (('nqthm::and c1 c2)
;               (append (nqthm-unprettyify c1) (nqthm-unprettyify c2)))
;              (('nqthm::implies hyp concl)
;               (nqthm-unprettyify1 (nqthm-flatten-ands-in-lit hyp)
;                                   (nqthm-unprettyify concl)))
;              (& (list (cons nil term)))))

(defun make-rewrite-corollaries (pairs)

; Pairs is a list of pairs, each of the form (hyps . concl) derived by
; unprettyifying a theorem which is to be stored as a REWRITE rule in Nqthm.
; We have to generate a :corollary for each of these.  We cannot let ACL2
; automatically generate them because the nqthm::AND's in the theorem are not
; recognized by ACL2's unprettyify.  Note also that the hypotheses in the
; :corollary are stitched together by lisp::AND rather than nqthm::AND.  This
; allows ACL2 to recognize them as a conjunction; otherwise, it just stores an
; nqthm::AND and sometimes cannot relieve it.

  (declare (xargs :mode :program))
  (cond ((null pairs) nil)
        (t (cons (list :rewrite
                       :corollary
                       (let ((hyps (caar pairs))
                             (concl (cdar pairs)))
                         (cond ((null hyps) concl)
                               ((null (cdr hyps))
                                (list 'implies (car hyps) concl))
                               (t (list 'implies
                                        (cons 'lisp::and hyps)
                                        concl)))))
                 (make-rewrite-corollaries (cdr pairs))))))

(defun nqthm-to-acl2-rule-classes (term lst)
  (declare (xargs :mode :program))
  (cond ((atom lst) nil)
        ((eq (car lst) 'nqthm::rewrite)
         (append
          (cond
           ((probable-nqthm-type-prescriptionp term) '(:type-prescription))
           ((probable-nqthm-compound-recognizerp term) '(:compound-recognizer))
           ((probable-nqthm-linearp term) '(:linear))
           (t (make-rewrite-corollaries (nqthm-unprettyify term))))
          (nqthm-to-acl2-rule-classes term (cdr lst))))
        ((eq (car lst) 'nqthm::elim)
         (cons :elim (nqthm-to-acl2-rule-classes term (cdr lst))))
        ((eq (car lst) 'nqthm::generalize)
         (cons :generalize (nqthm-to-acl2-rule-classes term (cdr lst))))
        ((and (consp (car lst))
              (eq (caar lst) 'nqthm::meta))
         (cons (cons :meta (cdr (car lst)))  ; meant to cause an error
               (nqthm-to-acl2-rule-classes term (cdr lst))))
        (t (cons (car lst) (nqthm-to-acl2-rule-classes term (cdr lst))))))

(defmacro nqthm::prove-lemma (name rule-classes term &rest rst)

  ":Doc-Section nqthm-commands

  Nqthm ~c[PROVE-LEMMA] command~/
  ~bv[]
  General Forms:
  (PROVE-LEMMA name rule-classes term)
  (PROVE-LEMMA name rule-classes term :hints ... ...)
  ~ev[]~/

  Roughly speaking, this macro translates term with ~c[nqthm-translate]
  (~pl[nqthm-translate]) to the ACL2 fragment of Nqthm and then
  generates a suitable ~c[defthm] form.  ~c[Rule-classes] may be either
  as in Nqthm-1992, e.g., ~c[(REWRITE ELIM)], or as in ACL2, e.g.,
  ~c[((:REWRITE :COROLLARY ...) ...)].  The Nqthm rule-class ~c[REWRITE]
  may generate ACL2 ~c[:]~ilc[rewrite], ~c[:]~ilc[type-prescription],
  ~c[:]~ilc[compound-recognizer], or ~c[:]~ilc[linear] arithmetic rules
  depending on the syntax of term.  The heuristic for generating an
  ACL2 rule class from ~c[REWRITE] is not as sophisticated as in
  Nqthm-1992 and the rule stored by ACL2's ~c[PROVE-LEMMA] may be
  different than Nqthm's.  The problem is that the ACL2's
  ~c[PROVE-LEMMA] macro cannot access ~il[state] and hence cannot
  determine the types of certain subterms in ~c[term].

  Note that the ACL2 ~c[PROVE-LEMMA] command does not support
  Nqthm-1992-style proof ~il[hints].  If you want to give the theorem
  prover ~il[hints], you must do it with ACL2-style ~il[hints].  If
  Nqthm-style hints are provided, the ~c[PROVE-LEMMA] macro will expand
  to a form that will cause an error when evaluated."

  (let ((term (nqthm-translate term)))
    `(defthm ,name ,term :rule-classes
       ,(nqthm-to-acl2-rule-classes term rule-classes) ,@rst)))

(defmacro nqthm::add-axiom (name rule-classes term)

  ":Doc-Section nqthm-commands

  Nqthm ~c[ADD-AXIOM] command~/
  ~bv[]
  General Form:
  (ADD-AXIOM name rule-classes term)
  ~ev[]~/
  This command translates term with ~c[nqthm-translate]
  (~pl[nqthm-translate]) and then generates a suitable ~ilc[defaxiom]
  command.  ~c[Rule-classes] may be either in the Nqthm style or the
  ACL2 style and are translated as in ~c[PROVE-LEMMA]
  (~pl[PROVE-LEMMA])."

  (let ((term (nqthm-translate term)))
    `(defaxiom ,name ,term :rule-classes
       ,(nqthm-to-acl2-rule-classes term rule-classes))))

(defmacro nqthm::disable (&rest rst)

  ":Doc-Section nqthm-commands

  Nqthm ~c[DISABLE] command~/
  ~bv[]
  General Form:
  (DISABLE name1 name2 ...)
  ~ev[]~/
  This macro expands to the event
  ~bv[]
  (in-theory (disable name1 name2 ...)).
  ~ev[]
  The use of this name, i.e., ~c[NQTHM::DISABLE], for this purpose
  makes it possible to execute Nqthm scripts that contain ~c[DISABLE]
  events.  But it makes it more awkward to handle ``local disable''
  hints.  In particular, it is a syntax mistake to write:
  ~bv[]
  (PROVE-LEMMA name rule-classes term
    :hints ((\"Goal\" :in-theory (DISABLE name1 name2 ...))))
  ~ev[]
  because ~c[DISABLE] here means the ``event level disabling.''  To get 
  the desired effect you must write
  ~bv[]
  (PROVE-LEMMA name rule-classes term
    :hints ((\"Goal\" :in-theory (ACL2::DISABLE name1 name2 ...)))).
  ~ev[]"

  `(in-theory (disable ,@rst)))

(defmacro nqthm::enable (&rest rst)

  ":Doc-Section nqthm-commands

  Nqthm ~c[ENABLE] command~/
  ~bv[]
  General Form:
  (ENABLE name1 name2 ...)
  ~ev[]~/
  This macro expands to the event 
  ~bv[]
  (in-theory (enable name1 name2 ...)).
  ~ev[]
  The use of this name, i.e., ~c[NQTHM::ENABLE], for this purpose makes
  it possible to execute Nqthm scripts that contain ~c[ENABLE] events.
  But it makes it more awkward to handle ``local enable'' hints.  In
  particular, it is a syntax mistake to write:
  ~bv[]
  (PROVE-LEMMA name rule-classes term
    :hints ((\"Goal\" :in-theory (ENABLE name1 name2 ...))))
  ~ev[]
  because ~c[ENABLE] here means the ``event level disabling.''  To get 
  the desired effect you must write
  ~bv[]
  (PROVE-LEMMA name rule-classes term
    :hints ((\"Goal\" :in-theory (ACL2::ENABLE name1 name2 ...)))).
  ~ev[]"

  `(in-theory (enable ,@rst)))

; We now reconstruct the shell principle.  We do not check that the invocation
; of add-shell is legal, beyond the necessary shape tests for the arguments.
; If some function symbol is non-new, the attempt to define it will cause an
; error.  If some non-function symbol used as a type-restriction recognizer,
; its use in a theorem will cause an error.  If the default object, (dv), for
; the accessor dest does not satisfy the type-restriction tr then the attempt
; to prove (recog (const ... dest ...))  will fail.  (Without loss of
; generality, we can assume const takes only one argument.  Then consider the
; instance (recog (const (dv))).  It is false, because const puts (dv) into the
; dest slot when it detects that (dv) fails the type restriction, but then
; recog fails.)

; The shell principle implemented here is more generous that Nqthm's vis-a-vis
; the default values allowed.  In particular, any object with count 0 that
; satisfies the type-restriction is allowed.  It must satisfy the
; type-restriction because of the theorem noted above.  It must have count 0
; because of the following argument: Consider the theorem
;   (not (lessp (count x) (count (dest x)))), 
; which is required of the new shell.  Let x be 0.  Because 0 is not of the new
; type, (dest 0) is (dv).  Thus, instantiating the theorem with 0 produces
;   (not (lessp 0 (count (dv))))
; which is equivalent to (count (dv)) = 0.

; Finally, our shell principle takes as its last argument the previously
; executed ADD-SHELL!  Thus, an ADD-SHELL command explicitly contains all
; previous such commands.  This encoding of history in the command has two
; uses.

; (1) The disjointness axioms must enumerate all previous recognizers.
; (2) We must declare all shell accessors (new and old) to be primitive
;     recursive destructors.

; Acl2 forces us to include the previous ADD-SHELL command in the next one
; because of its prohibition against macros being sensitive to state.

; Here are some examples of the shell principle in use.  

#|

 (add-shell push empty-stack stackp
            ((top (one-of numberp litatom) zero)
             (pop (one-of stackp) empty-stack))
; Previous add-shell:
            nil)

 (add-shell cons-if nil if-exprp
            ((test (none-of) zero)
             (tbranch (none-of) zero)
             (fbranch (none-of) zero))

; Previous add-shell:
            (add-shell push empty-stack stackp
                       ((top (one-of numberp litatom) zero)
                        (pop (one-of stackp) empty-stack))
                       nil))

 (add-shell pstate nil pstatep
            ((pc (one-of numberp) zero)
             (regs (none-of) zero)
             (mem (none-of) zero)
             (aflg (none-of) zero)
             (bflg (none-of) zero)
             (cflg (none-of) zero)
             (dflg (none-of) zero)
             (eflg (none-of) zero)
             (halt (none-of) zero)
             (status (one-of truep falsep) true))

; Previous add-shell:
            (add-shell cons-if nil if-exprp
                       ((test (none-of) zero)
                        (tbranch (none-of) zero)
                        (fbranch (none-of) zero))
                       (add-shell push empty-stack stackp
                                  ((top (one-of numberp litatom) zero)
                                   (pop (one-of stackp) empty-stack))
                                  nil)))

|#

(defun type-restriction-term1 (tp x)
  (declare (xargs :mode :program))
  (cond ((null tp) nil)
        (t (cons (list (car tp) x) (type-restriction-term1 (cdr tp) x)))))

(defun type-restriction-term (tp x)
  (declare (xargs :mode :program))
  (cond ((eq (car tp) 'nqthm::one-of)
         (cond ((null (cdr tp)) nil)
               ((null (cddr tp)) (car (type-restriction-term1 (cdr tp) x)))
               (t (cons 'nqthm::or (type-restriction-term1 (cdr tp) x)))))
        ((null (cdr tp)) t)
        ((null (cddr tp))
         (list 'nqthm::not (car (type-restriction-term1 (cdr tp) x))))
        (t (list 'nqthm::not
                 (cons 'nqthm::or (type-restriction-term1 (cdr tp) x))))))

(defun add-shell-const1 (tuples)
  (declare (xargs :mode :program))
  (cond
   ((null tuples) nil)
   (t (let* ((var (car (car tuples)))
             (tr (type-restriction-term (cadr (car tuples)) var))
             (dv (list (caddr (car tuples)))))
        (cons (if (eq tr t)
                  var
                  `(if ,tr ,var ,dv))
              (add-shell-const1 (cdr tuples)))))))
         
(defun add-shell-destructor-defuns (btm recog tuples cdr-term)

; Initially, cdr-term should be '(cddr x).

  (declare (xargs :mode :program))
  (cond
   ((null tuples) nil)
   (t (let ((dest (car (car tuples)))
            (dv (list (caddr (car tuples)))))
        (cons `(defun ,dest (x)
                 (if ,(if btm
                          `(and (,recog x) (not (equal (cadr x) '(,btm))))
                          `(,recog x))
                     (car ,cdr-term)
                     ,dv))
              (add-shell-destructor-defuns btm recog (cdr tuples)
                                           (list 'cdr cdr-term)))))))

(defun type-restriction-terms (tuples x)
  (declare (xargs :mode :program))
  (cond ((null tuples) nil)
        (t (cons (type-restriction-term (cadr (car tuples)) (list 'car x))
                 (type-restriction-terms (cdr tuples) (list 'cdr x))))))

(defun add-shell-routines (const btm recog tuples)
  (declare (xargs :mode :program))
  `(,@(cond
       ((null btm) nil)
       (t `((defun ,btm  nil (list *oneshell-quote-mark* '(,btm))))))
      (defun ,recog (x)
        (declare (xargs :well-founded-relation e0-ord-<
                        :measure (acl2-count x)))
        (and (consp x)
             (eq (car x)  *one-shell-quote-mark*)
             (true-listp (cdr x))
             ,(if btm
                  `(if (equal (cadr x) '(,btm))
                       (null (cddr x))
                       (and (eq (cadr x) ',const)
                            (= (length (cddr x)) ,(length tuples))
                            ,@(type-restriction-terms tuples '(cddr x))))
                  `(and (eq (cadr x) ',const)
                        (= (length (cddr x)) ,(length tuples))
                        ,@(type-restriction-terms tuples '(cddr x))))))

      (defun ,const ,(strip-cars tuples)
        (list *one-shell-quote-mark*
              ',const
              ,@(add-shell-const1 tuples)))
      ,@(add-shell-destructor-defuns btm recog tuples '(cddr x))))

(defun packnqthm (lst)
  (declare (xargs :mode :program))
  (intern-in-package-of-symbol (coerce (packn1 lst) 'string)
                               'NQTHM::PACK))

(defun add-type-set-lemmas-destructors (tuples)

; We do not have an effective way to encode every possible type-restriction
; term.  Therefore, we encode a type-restriction theorem for each destructor
; for which we can do so effectively.

; Here are the type-restrictions:
; (one-of)              - can't happen, so we skip it
; (none-of)             - no type-restriction generated but thats perfect
; (one-of rec)          - (rec (dest x))         :rewrite
; (none-of rec)         - (not (rec (dest x)))   :rewrite
; (one-of truep falsep) - (booleanp (dest x))    :type-prescription
; (one-of falsep truep) - (booleanp (dest x))    :type-prescription
; all others            - skipped

  (declare (xargs :mode :program))
  (cond ((null tuples) nil)
        (t
         (let ((dest (car (car tuples)))
               (tp (cadr (car tuples))))
           (cond
            ((= (length tp) 2)
             (cons `(defthm ,(packnqthm (list dest "-TYPE"))
                      ,(type-restriction-term tp `(,dest nqthm::x)))
                   (add-type-set-lemmas-destructors (cdr tuples))))
            ((or (equal tp '(nqthm::one-of nqthm::truep nqthm::falsep))
                 (equal tp '(nqthm::one-of nqthm::falsep nqthm::truep)))
             (cons `(defthm ,(packnqthm (list dest "-TYPE"))
                      (booleanp (,dest nqthm::x))
                      :rule-classes :type-prescription)
                   (add-type-set-lemmas-destructors (cdr tuples))))
            (t (add-type-set-lemmas-destructors (cdr tuples))))))))

(defun add-type-set-lemmas-for-imperfect-recognizers1
  (past-imperfect-recognizers recog)
  (declare (xargs :mode :program))
  (cond ((null past-imperfect-recognizers) nil)
        (t (cons `(defthm ,(packnqthm (list recog
                                            "-V-"
                                            (car past-imperfect-recognizers)
                                            "!"))
                    (and (implies (,recog nqthm::x)
                                  (not (,(car past-imperfect-recognizers)
                                        nqthm::x)))
                         (implies (,(car past-imperfect-recognizers)
                                   nqthm::x)
                                  (not (,recog nqthm::x))))
                    :hints (("Goal"
                             :in-theory
                             (enable ,(car past-imperfect-recognizers)))))
                 (add-type-set-lemmas-for-imperfect-recognizers1
                  (cdr past-imperfect-recognizers) recog)))))

(defun add-type-set-lemmas-for-imperfect-recognizers
  (past-imperfect-recognizers recog)

; The primitive recognizers truep, falsep, numberp, and negativep are handled
; perfectly by their associated compound recognizers and are thus called
; "perfect recognizers."  Let the "imperfect recognizers" be those recognizers
; whose shell types are represented by list structures.  The imperfect
; recognizers initially include LISTP and LITATOM and will include every user
; defined shell.

; To keep these disjoint from the perfect recognizers we will need
; (implies (recog x) (consp x)) :compound-recognizer

; To keep the imperfect recognizers disjoint from eachother we will need
; (and (implies (recog x) (not (other-imperfect-recog x)))
;      (implies (other-imperfect-recog x) (not (recog x)))),
; for each other imperfect recognizer.

  (declare (xargs :mode :program))
  `((defthm ,(packnqthm (list recog "-COMPOUND-RECOGNIZER!"))
      (implies (,recog nqthm::x) (consp nqthm::x))
      :rule-classes :compound-recognizer)
    ,@(add-type-set-lemmas-for-imperfect-recognizers1
       past-imperfect-recognizers recog)))

(defun add-type-set-lemmas (const recog tuples past-imperfect-recognizers)
  (declare (xargs :mode :program))
  `(,@(add-type-set-lemmas-for-imperfect-recognizers
       past-imperfect-recognizers recog)
      (defthm ,(packnqthm (list recog "-" const "!"))
        (,recog (,const ,@(strip-cars tuples))))
      ,@(add-type-set-lemmas-destructors tuples)))
     
(defun add-destructor-lemmas (const btm recog tuples shell-expr)
  (declare (xargs :mode :program))
  (cond
   ((null tuples) nil)
   (t (let* ((dest (car (car tuples)))
             (tr (type-restriction-term (cadr (car tuples)) dest))
             (dv (list (caddr (car tuples)))))
        (append
         `((defthm ,(packnqthm (list dest "-" const))
             (equal (,dest ,shell-expr)
                    ,(if (eq tr t)
                         dest
                         `(if ,tr ,dest ,dv))))
           (defthm ,(packnqthm (list dest "-N" recog))
             (implies (not (,recog nqthm::x)) (equal (,dest nqthm::x) ,dv)))
           ,@(if (not (eq tr t))
                 `((defthm ,(packnqthm (list dest "-TYPE-RESTRICTION"))
                     (implies (not ,tr)
                              (equal ,shell-expr
                                     ,(subst dv dest shell-expr)))))
                 nil)
           (defthm ,(packnqthm (list dest "-LESSP"))
             (implies ,(if btm
                           `(and (,recog nqthm::x)
                                 (not (equal nqthm::x (,btm))))
                           `(,recog nqthm::x))
                      (nqthm::lessp (nqthm::count (,dest nqthm::x))
                                    (nqthm::count nqthm::x)))
             :rule-classes (:rewrite :built-in-clause :linear)
             :hints (("Goal"
                      :in-theory (enable nqthm::count nqthm::fix lessp-is-<))))
           (defthm ,(packnqthm (list dest "-LESSEQP"))
             (not (nqthm::lessp (nqthm::count nqthm::x)
                                (nqthm::count (,dest nqthm::x))))
             :rule-classes (:rewrite :linear)
             :hints (("Goal"
                      :in-theory (enable nqthm::count nqthm::fix lessp-is-<)))))
         (add-destructor-lemmas const btm recog (cdr tuples) shell-expr))))))

(defun packnqthm-lst (lst str)
  (declare (xargs :mode :program))
  (cond ((null lst) nil)
        (t (cons (packnqthm (list (car lst) str))
                 (packnqthm-lst (cdr lst) str)))))

(defun add-shell-const-equal1 (tuples)
  (declare (xargs :mode :program))
  (cond
   ((null tuples) nil)
   (t (let* ((var1 (car (car tuples)))
             (var2 (packnqthm (list var1 "-")))
             (tr1 (type-restriction-term (cadr (car tuples)) var1))
             (tr2 (type-restriction-term (cadr (car tuples)) var2))
             (dv (list (caddr (car tuples)))))
        (cons
         (if (equal tr1 t)
             `(equal ,var1 ,var2)
             `(if ,tr1
                  (if ,tr2
                      (equal ,var1 ,var2)
                      (equal ,var1 ,dv))
                  (if ,tr2
                      (equal ,dv ,var2)
                      t)))
          (add-shell-const-equal1 (cdr tuples)))))))

(defun interpose (x lst)
  (declare (xargs :mode :program))
  (cond ((null lst) nil)
        (t (cons x (cons (car lst) (interpose x (cdr lst)))))))

(defun plusjoin (lst)
  (declare (xargs :mode :program))
  (cond ((null lst) 0)
        ((null (cdr lst)) (car lst))
        (t (list 'nqthm::plus (car lst) (plusjoin (cdr lst))))))

(defun count-destructor-tuples (tuples)
  (declare (xargs :mode :program))
  (cond ((null tuples) nil)
        (t (let* ((var (car (car tuples)))
                  (tr (type-restriction-term (cadr (car tuples)) var)))
             (cons (if (eq tr t)
                       `(nqthm::count ,var)
                       `(if ,tr (nqthm::count ,var) 0))
                   (count-destructor-tuples (cdr tuples)))))))

(defun past-user-defined-recognizers (prev-event)

; Prev-event is the previously executed ADD-SHELL command.  We recover from it
; all of the recognizers thus far defined by the user.

  (declare (xargs :mode :program))
  (cond ((or (null prev-event)
             (eq prev-event 'nqthm::nil)) nil)
        (t (cons (cadddr prev-event)
                 (past-user-defined-recognizers (cadr (cddddr prev-event)))))))

(defun add-shell0 (const btm recog tuples prev-event)
  (declare (xargs :mode :program))
  `(encapsulate nil
     ,@(add-shell-routines const btm recog tuples)
     ,@(add-type-set-lemmas const recog tuples
                            (append '(nqthm::LISTP nqthm::LITATOM)
                                    (past-user-defined-recognizers prev-event)))
     ,@(add-destructor-lemmas const btm recog tuples
                              `(,const ,@(strip-cars tuples)))
     (defthm ,(packnqthm (list const "-EQUAL"))
       (equal
        (equal (,const ,@(strip-cars tuples))
               (,const ,@(packnqthm-lst (strip-cars tuples) "-")))
        ,(conjoin
          (add-shell-const-equal1 tuples))))
     (local
      (defthm equal-rearraanger
        (implies (and (syntaxp (and (quotep i)
                                    (quotep j)))
                      (rationalp i)
                      (rationalp j)
                      (rationalp x))
                 (equal (equal (+ i x) j)
                        (equal x (- j i))))))
     (local
      (defthm equal-len-0
        (equal (equal (len x) 0) (atom x))))
      (defthm ,(packnqthm (cons const (interpose "-" (strip-cars tuples))))
        (equal (,const ,@(pairlis-x2 (strip-cars tuples) '(nqthm::x)))
               (if ,(if btm
                        `(nqthm::and (,recog nqthm::x)
                                     (not (equal nqthm::x (,btm))))
                        `(,recog nqthm::x))
                   nqthm::x
                   (,const ,@(pairlis-x2 (strip-caddrs tuples) nil)))))
      (defthm ,(packnqthm (cons const
                                (append
                                 (interpose "-"
                                            (strip-cars tuples))
                                 '("-ELIM"))))
        (implies ,(if btm
                      `(nqthm::and (,recog nqthm::x)
                                   (not (equal nqthm::x (,btm))))
                      `(,recog nqthm::x))
                 (equal (,const ,@(pairlis-x2 (strip-cars tuples) '(nqthm::x)))
                        nqthm::x))
        :rule-classes :elim
        :otf-flg t)
     (defthm ,(packnqthm (list "COUNT-" const))
       (equal (nqthm::count (,const ,@(strip-cars tuples)))
              (nqthm::add1 ,(plusjoin (count-destructor-tuples tuples))))
       :hints (("Goal" :in-theory (enable nqthm::count nqthm::add1 plus-is-+))))
     ,@(if (and (null tuples) btm)
           `((defthm ,(packnqthm (list recog "-ELIMINATOR"))
               (equal (,recog nqthm::x)
                      (nqthm::or (equal nqthm::x (,btm))
                                 (equal nqthm::x (,const))))))
           nil)
     (in-theory (disable ,const
                         ,@(if btm `(,btm) nil)
                         ,recog ,@(strip-cars tuples)))))

(defun acceptable-add-shellp1 (tuples)
  (declare (xargs :mode :program))
  (cond ((atom tuples) (eq tuples nil))
        ((and (true-listp (car tuples))
              (equal (length (car tuples)) 3))
         (let ((dest (car (car tuples)))
               (tr (cadr (car tuples)))
               (df (caddr (car tuples))))
           (and (symbolp dest)
                (consp tr)
                (member (car tr) '(nqthm::one-of nqthm::none-of))
                (symbol-listp (cdr tr))
                (symbolp df)
                (acceptable-add-shellp1 (cdr tuples)))))
        (t nil)))

(defun acceptable-add-shellp (const btm recog tuples prev-event)
  (declare (xargs :mode :program))
  (and (symbolp const)
       (symbolp btm)
       (symbolp recog)
       (acceptable-add-shellp1 tuples)
       (or (null prev-event)
           (eq prev-event 'nqthm::nil)
           (and (true-listp prev-event)
                (= (length prev-event) 6)
                (eq (car prev-event) 'nqthm::add-shell)
                (acceptable-add-shellp (cadr prev-event)
                                       (caddr prev-event)
                                       (cadddr prev-event)
                                       (car (cddddr prev-event))
                                       (cadr (cddddr prev-event)))))))

(defmacro nqthm::add-shell (const btm recog tuples prev-event)

  ":Doc-Section nqthm-commands

  Nqthm ~c[ADD-SHELL] command~/
  ~bv[]
  General Form:
  (ADD-SHELL const btm recog 
             ((ac1 type-restriction1 default1)
              ...
              (acn type-restrictionn defaultn))
             prev-add-shell)
  ~ev[]
  Note that the ACL2 version of Nqthm-1992's ~c[ADD-SHELL] command
  takes an additional final argument!  That argument must be the ACL2
  ~c[ADD-SHELL] command most recently executed in the current
  chronology.  We discuss this below.  The ACL2 ~c[ADD-SHELL] command
  is expensive to execute because it defines all of the newly
  introduced functions constructively and proves as theorems the
  axioms that are added by Nqthm-1992.~/

  Here we illustrate the new final argument to ~c[ADD-SHELL].  Suppose the
  first ~c[ADD-SHELL] command in an Nqthm-1992 script is
  ~bv[]
  (ADD-SHELL P-STATE NIL P-STATEP
             ((REGS (NONE-OF) ZERO) (MEM (NONE-OF) ZERO)))
  ~ev[]
  Then the corresponding first ~c[ADD-SHELL] command in the ACL2
  version of the script should be
  ~bv[]
  (ADD-SHELL P-STATE NIL P-STATEP
             ((REGS (NONE-OF) ZERO) (MEM (NONE-OF) ZERO))
             NIL).
  ~ev[]
  Note the final ~c[NIL] indicating that this is the first ~c[ADD-SHELL]
  command.  Suppose the second ~c[ADD-SHELL] command in the Nqthm-1992
  script is the traditional introduction of stacks via the ~c[PUSH]
  shell.  Then the second ACL2 ~c[ADD-SHELL] command should be
  ~bv[]
  (ADD-SHELL PUSH EMPTY STACKP
    ((TOP (ONE-OF NUMBERP) ZERO)
     (POP (ONE-OF STACKP) EMPTY))
    (ADD-SHELL P-STATE NIL P-STATEP
               ((REGS (NONE-OF) ZERO) (MEM (NONE-OF) ZERO))
               NIL)).
  ~ev[]
  Note that the most recent ACL2 ~c[ADD-SHELL] command explicitly
  specifies every ~c[ADD-SHELL] command in the history.  This is
  necessary since the ACL2 macro ~c[ADD-SHELL] cannot access ~il[state]
  but must know all previous shell classes so that it can generate the
  rules for showing disjointness."

  (if (acceptable-add-shellp const btm recog tuples prev-event)
      (add-shell0 const
                  (if (equal btm 'nqthm::nil)
                      nil
                      btm)
                  recog tuples prev-event)
      '(illegal 'add-shell "Bad shell" nil)))

; (add-shell push empty stackp
;   ((top (one-of numberp) zero) (pop (one-of stackp) empty)) nil)

; Now we do the remaining Nqthm commands that we do.

(defmacro nqthm::thm (&rest x)
  
  ":Doc-Section nqthm-commands

  Nqthm ~c[PROVE] command~/
  ~bv[]
  General Form:
  (thm term :hints ...)
  ~ev[]~/

  This command takes the same arguments as ACL2's ~ilc[thm] command but
  it uses ~c[nqthm-translate] (~pl[nqthm-translate]) to translate
  ~c[term] before attempting to prove it."

  `(thm ,(nqthm-translate (car x)) ,@(cdr x)))

(defmacro nqthm::in-package (&rest x)

  ":Doc-Section nqthm-commands

  Nqthm ~c[IN-PACKAGE] command~/
  ~bv[]
  General Form:
  (in-package \"pkg\")
  ~ev[]~/
  This is just ACL2's ~ilc[in-package] command.  However, if you wish to
  shift back to the ACL2 package while in the Nqthm emulator, you should
  probably use ~c[:stop-nqthm] (~pl[stop-nqthm]) instead."

 `(in-package ,@x))

;(defmacro nqthm::dcl (fn args)
    
;  ":Doc-Section nqthm-commands

;  Nqthm ~c[DCL] command~/
;  ~bv[]
;  General Form:
;  (DCL fn args)
;  ~ev[]~/

;  This command expands into ~c[(defstub fn args t)] to introduce an
;  undefined, unconstrained function that returns one result."

;  `(defstub ,fn ,args t))

(defun constrain-alistp (x)
  (declare (xargs :mode :program))
  (cond ((atom x) (equal x nil))
        ((and (consp (car x))
              (symbolp (caar x))
              (consp (cdar x))
              (null (cddar x))
              (consp (cadar x))
              (true-listp (cadar x))
              (eq (car (cadar x)) 'nqthm::lambda)
              (true-listp (cadr (cadar x))))
         (constrain-alistp (cdr x)))
        (t (illegal 'constrain
                    "The alist arg of Nqthm's CONSTRAIN must pair symbols with ~
                     LAMBDA expressions so Acl2 can recover the arity of each ~
                     constrained function."
                    (car x)))))

(defun constrain-sigs (alist)
  (declare (xargs :mode :program))
  (cond ((null alist) nil)
        (t (cons (list (car (car alist))
                       (cadr (cadr (car alist)))
                       t)
                 (constrain-sigs (cdr alist))))))

(defun constrain-witnesses (alist)
  (declare (xargs :mode :program))
  (cond ((null alist) nil)
        (t (cons (let ((formals (nqthm-translate-lst (cadr (cadr (car alist)))))
                       (body (nqthm-translate (caddr (cadr (car alist))))))
                   (list 'local
                         (list 'defun
                               (car (car alist))
                               formals
                               `(declare
                                 (ignore ,@(set-difference-eq formals
                                                              (all-vars body))))
                               body)))
                 (constrain-witnesses (cdr alist))))))

(defmacro nqthm::constrain (name rule-classes term alist &rest rst)
  
  ":Doc-Section nqthm-commands

  Nqthm ~c[CONSTRAIN] command~/
  ~bv[]
  General Form:
  (CONSTRAIN name rule-classes term alist :hints ...)
  ~ev[]~/
  This is like the ~c[CONSTRAIN] event of Nqthm-1992, except that
  ~il[hints] are supplied as for ACL2.  ~c[Term] is translated with
  ~c[nqthm-translate] (~pl[nqthm-translate]).  The ~c[rule-classes]
  are translated to ACL2 rule classes as in ~c[PROVE-LEMMA]
  (~pl[PROVE-LEMMA]) and may thus be either in the Nqthm style or
  the ACL2 style.."

  (declare (xargs :guard (constrain-alistp alist)))
  `(encapsulate ,(constrain-sigs alist)
                ,@(constrain-witnesses alist)
                ,(let ((term (nqthm-translate term)))
                   `(defthm ,name ,term
                      :rule-classes
                      ,(nqthm-to-acl2-rule-classes term rule-classes)
                      ,@rst))))

(defmacro nqthm::note-lib (name &optional compile-flg)

  ":Doc-Section nqthm-commands

  Nqthm ~c[NOTE-LIB] command~/
  ~bv[]
  General Form:
  (NOTE-LIB name compile-flg)
  ~ev[]~/
  This expands into an ~c[include-book] of the named file and loads the
  compiled version if ~c[compile-flg] is provided and is ~c[T].  Note
  that the file must have the name ~c[\"name.lisp\"] and should be an
  ACL2 book."

  `(include-book ,name
                 ,(if (or (null compile-flg)
                          (eq compile-flg 'nqthm::nil))
                      nil
                    t)))

