(in-package "ACL2")




(include-book (:relative :back "cowles" "acl2-crg"))


(defmacro fc (x)
  x)


(defthm commutativity-2-of-plus
  (equal (+ x (+ y z))
         (+ y (+ x z))))

(defthm functional-self-inversion-of-minus
  (equal (- (- x))
         (fix x)))

(defthm distributivity-of-minus-over-plus
  (equal (- (+ x y))
         (+ (- x) (- y))))

(defthm minus-cancellation-on-right
  (equal (+ x y (- x))
         (fix y)))

(defthm minus-cancellation-on-left
  (equal (+ x (- x) y)
         (fix y)))


(defthm right-cancellation-for-plus
  (equal (equal (+ x z)
                (+ y z))
         (equal (fix x) (fix y))))

(defthm left-cancellation-for-plus
  (equal (equal (+ x y)
                (+ x z))
         (equal (fix y) (fix z))))

(defthm equal-minus-zero
  (equal (equal 0 (- x))
         (equal 0 (fix x))))

(defthm inverse-of-plus-as-eqeual-0
  (equal (equal (- a b) 0)
         (equal (fix a) (fix b))))

(defthm equal-minus-minus
  (equal (equal (- a) (- b))
         (equal (fix a) (fix b))))

(defthm fold-consts-in-plus
  (implies (and (syntaxp (quotep x))
		(syntaxp (quotep y)))
	   (equal (+ x (+ y z))
		  (+ (+ x y) z))))




(defthm commutativity-2-of-time
   (equal (* x (* y z))
          (* y (* x z)))
   :hints
   (("Goal"
     :use
     (:instance
      (:functional-instance
       commutativity-2-of-op
       (equiv equal)
       (pred (lambda (x) t))
       (op binary-time))
      (x x)
      (y y)
      (z z)))))

(defthm functional-self-inversion-of-div
  (equal (/ (/ x)) (fix x))
  :hints (("Goal"
	   :use
           ((:instance
             (:functional-instance
              Involution-of-inv
              (equiv equal)
              (pred (lambda 
                                (x) 
                                (and (acl2-numberp x)
                                     (not (equal x 0)))))
              (op binary-time)
              (id (lambda () 1))
              (inv unary-div))
             (x x))))))

(defthm distributivity-of-div-over-time
  (equal (/ (* x y))
	 (* (/ 1 x) (/ 1 y)))
  :hints (("Goal"
	   :use ((:instance
                  (:functional-instance
                   Distributivity-of-inv-over-op
                   (equiv equal)
                   (pred (lambda 
                                     (x)
                                     (and (acl2-numberp x)
                                          (not (equal x 0)))))
                   (op binary-time)
                   (id (lambda () 1))
                   (inv unary-div))
                  (x x)
                  (y y))))))

(defthm div-cancellation-on-right
  (implies (and (fc (acl2-numberp x))
		(fc (not (equal x 0))))
	   (equal (* x y (/ x))
		  (fix y)))
  :hints (("Goal"
	   :use
           ((:instance
             (:functional-instance
              inv-cancellation-on-right
              (equiv equal)
              (pred (lambda 
                                (x) 
                                (and (acl2-numberp x)
                                     (not (equal x 0)))))
              (op binary-time)
              (id (lambda () 1))
              (inv unary-div))
             (x x)
             (y y))))))

(defthm div-cancellation-on-left
  (implies (and (fc (acl2-numberp x))
                (fc (not (equal 0 x))))
           (equal (* x (/ x) y)
                  (fix y)))
  :hints (("Goal" :use div-cancellation-on-right)))

(encapsulate
 ()

 (local
  (defthm right-cancellation-for-time-lemma
    (implies (and (equal (* x z) (* y z))
                  (acl2-numberp z)
                  (not (equal 0 z))
                  (acl2-numberp x)
                  (acl2-numberp y))
             (equal (equal x y) t))
    :hints (("Goal" :use
             ((:instance
               (:functional-instance
                Right-cancellation-for-op
                (equiv equal)
                (pred (lambda 
                                  (x) 
                                  (and (acl2-numberp x)
                                       (not (equal x 0)))))
                (op binary-time)
                (id (lambda () 1))
                (inv unary-div))
               (x x)
               (y y)
               (z z)))))))

 (defthm right-cancellation-for-time
   (equal (equal (* x z) (* y z))
          (or (equal 0 (fix z))
              (equal (fix x) (fix y))))))

(defthm left-cancellation-for-time
  (equal (equal (* z x) (* z y))
         (or (equal 0 (fix z))
             (equal (fix x) (fix y)))))

(defthm
  Zero-is-only-zero-divisor
  (equal (equal (* x y) 0)
         (or (equal (fix x) 0)
             (equal (fix y) 0))))

(defthm equal-time-x-y-x
  (equal (equal (* x y) x)
         (or (equal x 0)
             (and (equal y 1)
                  (acl2-numberp x))))
  :hints (("Goal" :use
           ((:instance right-cancellation-for-time
                       (z x)
                       (x y)
                       (y 1))))))

(defthm equal-time-x-y-y
  (equal (equal (* x y) y)
         (or (equal y 0)
             (and (equal x 1)
                  (acl2-numberp y))))
  :hints (("Goal" :use
           ((:instance right-cancellation-for-time
                       (z y)
                       (x x)
                       (y 1))))))

(encapsulate
 ()

 (local
  (defthm equal-div-lemma
    (implies (and (acl2-numberp x)
                  (acl2-numberp y)
                  (equal (* x y) 1))
             (equal y (/ x)))
    :rule-classes nil 
    :hints (("Goal"
             :use
             (:instance
              (:functional-instance
               Uniqueness-of-op-inverses
               (equiv equal)
               (pred (lambda 
                                 (x) 
                                 (and (acl2-numberp x)
                                      (not (equal x 0)))))
               (nilop binary-time)
               (nilid (lambda () 1))
               (nilinv unary-div))
              (x x)
              (y y))))))

 (defthm equal-div
   (implies (and (fc (acl2-numberp x))
                 (fc (not (equal 0 x))))
            (equal (equal (/ x) y)
                   (equal 1 (* x y))))
   :hints (("Goal" :use equal-div-lemma)))
 )

(defthm numerator-nonzero-forward
  (implies (not (equal (numerator r) 0))
           (and (not (equal r 0))
                (acl2-numberp r)))
  :rule-classes
  ((:forward-chaining :trigger-terms
                      ((numerator r)))))


(encapsulate
 ()

 (local
  (defthm Uniqueness-of-time-inverses-lemma
    (equal (equal (* x y)
                  1)
           (and (not (equal x 0))
                (acl2-numberp x)
                (equal y (/ x))))))

 (defthm Uniqueness-of-time-inverses
   (equal (equal (* x y)
                 1)
          (and (not (equal (fix x) 0))
               (equal y (/ x))))
   :hints (("Goal" :in-theory (disable equal-div)))))

(in-theory (disable Uniqueness-of-time-inverses))

(theory-invariant
 (not (and (member-equal '(:rewrite Uniqueness-of-time-inverses) theory)
           (member-equal '(:rewrite equal-div) theory))))

(encapsulate
 ()

 (local
  (defthm equal-div-div-lemma
    (implies
     (and (fc (acl2-numberp a))
          (fc (acl2-numberp b))
          (fc (not (equal a 0)))
          (fc (not (equal b 0))))
     (equal (equal (/ a) (/ b))
            (equal a b)))
    :hints
    (("Goal"
      :use ((:instance
             (:theorem
              (implies
               (and (fc (acl2-numberp a))
                    (fc (acl2-numberp b))
                    (fc (not (equal a 0)))
                    (fc (not (equal b 0))))
               (implies (equal a b)
                        (equal (/ a) (/ b)))))
             (a (/ a)) (b (/ b))))))
    :rule-classes nil))

 (defthm equal-div-div
   (equal (equal (/ a) (/ b))
          (equal (fix a) (fix b)))
   :hints (("Goal" :use equal-div-div-lemma))))

(defthm equal-time-div-1
   (implies (and (acl2-numberp x)
                 (not (equal x 0)))
            (equal (equal (* (/ x) y) z)
                   (and (acl2-numberp z)
                        (equal (fix y) (* x z))))))

(defthm equal-time-div-2
   (implies (and (acl2-numberp x)
                 (not (equal x 0)))
            (equal (equal (* y (/ x)) z)
                   (and (acl2-numberp z)
                        (equal (fix y) (* z x))))))

(defthm fold-consts-in-time
  (implies (and (syntaxp (quotep x))
		(syntaxp (quotep y)))
	   (equal (* x (* y z))
		  (* (* x y) z))))

(defthm times-zero
   (equal (* 0 x) 0))


(encapsulate
 ()
 
 (local
  (defthm functional-commutativity-of-minus-time-right-lemma
    (implies (and (fc (acl2-numberp x))
                  (fc (acl2-numberp y)))
             (equal (* x (- y))
                    (- (* x y))))
    :hints (("Goal"
             :use
             ((:instance
               (:functional-instance
                functional-commutativity-of-minus-times-right
                (equiv equal)
                (pred acl2-numberp)
                (plus binary-plus)
                (times binary-time)
                (zero (lambda () 0))
                (minus unary-minus))
               (x x)
               (y y)))))
    :rule-classes nil))

 (defthm functional-commutativity-of-minus-time-right
   (equal (* x (- y))
          (- (* x y)))
   :hints (("Goal" :use functional-commutativity-of-minus-time-right-lemma))))

(defthm functional-commutativity-of-minus-time-left
  (equal (* (- x) y)
         (- (* x y))))

(defthm reciprocal-minus
  (equal (/ (- x))
         (- (/ x)))
  :hints (("Goal" :cases
           ((and (fc (acl2-numberp x))
                 (fc (not (equal x 0))))))))


(encapsulate
 ()

 (local
  (defthm numerator-integerp-lemma-1
    (implies (rationalp x)
             (equal (* (* (numerator x) (/ (denominator x))) (denominator x))
                    (numerator x)))
    :rule-classes nil
    :hints (("Goal" :in-theory (disable rational-implies2)))))

 (local
  (defthm numerator-integerp-lemma
    (implies (and (rationalp x)
                  (equal (* (numerator x) (/ (denominator x)))
                         x))
             (equal (numerator x)
                    (* x (denominator x))))
    :rule-classes nil
    :hints (("Goal" :use (numerator-integerp-lemma-1)
             :in-theory (disable rational-implies2)))))

 (defthm numerator-when-integerp
   (implies (integerp x)
            (equal (numerator x)
                   x))
   :hints (("Goal" :in-theory (disable rational-implies2)
            :use ((:instance lowest-terms (r x)
                             (q 1)
                             (n (denominator x)))
                  rational-implies2
                  numerator-integerp-lemma))))
 )

(defthm integerp-implies-denominator-1
  (implies (integerp x)
           (equal (denominator x) 1))
  :hints
  (("Goal"
    :use (rational-implies2 numerator-when-integerp)
    :in-theory (disable rational-implies2))))

(defthm equal-denominator-1
  (equal (equal (denominator x) 1)
         (or (integerp x)
             (not (rationalp x))))
  :hints (("Goal" :use (rational-implies2 completion-of-denominator)
           :in-theory (disable rational-implies2))))

(defthm time-r-denominator-r
  (equal (* r (denominator r))
         (if (rationalp r)
             (numerator r)
           (fix r)))
  :hints (("Goal" :use ((:instance rational-implies2 (x r)))
           :in-theory (disable rational-implies2))))

(defthm r-when-abs-numerator-1
  (and
   (implies
    (equal (numerator r) 1)
    (equal (/ r) (denominator r)))
   (implies
    (equal (numerator r) -1)
    (equal (/ r) (- (denominator r)))))
  :hints (("Goal" :use ((:instance rational-implies2 (x r)))
           :in-theory (disable rational-implies2))))



(defthm expt-type-prescription-rationalp
  (implies (rationalp r)
           (rationalp (expt r i)))
  :rule-classes (:type-prescription :generalize))

(defthm expt-type-prescription-positive
  (implies (and (< 0 r)
                (rationalp r))
           (< 0 (expt r i)))
  :rule-classes (:type-prescription :generalize))

(defthm expt-type-prescription-nonzero
  (implies (and (fc (acl2-numberp r))
                (not (equal r 0)))
           (not (equal 0 (expt r i))))
  :rule-classes (:type-prescription :generalize))

(defthm expt-type-prescription-integerp
  (implies (and (<= 0 i)
                (integerp r))
           (integerp (expt r i)))
  :rule-classes (:type-prescription :generalize))

(defthm Left-nullity-of-1-for-expt
  (equal (expt 1 i) 1))

(defthm Right-unicity-of-1-for-expt
  (equal (expt r 1)
         (fix r))
  :hints (("Goal" :expand (expt r 1))))

(defthm expt-minus
  (equal (expt r (- i))
         (/ (expt r i))))

(defthm Exponents-add-for-nonneg-exponents

  (implies (and (<= 0 i)
                (<= 0 j)
                (fc (integerp i))
                (fc (integerp j)))
           (equal (expt r (+ i j))
                  (* (expt r i)
                     (expt r j)))))

(encapsulate
 ()

 (local
  (defthm Exponents-add-negative-negative
    (implies (and (integerp i)
                  (integerp j)
                  (< i 0)
                  (< j 0))
             (equal (expt r (+ i j))
                    (* (expt r i)
                       (expt r j))))
    :rule-classes nil))

 (local
  (defthm Exponents-add-positive-negative
    (implies (and (integerp i)
                  (integerp j)
                  (acl2-numberp r)
                  (not (equal r 0))
                  (< 0 i)
                  (< j 0))
             (equal (expt r (+ i j))
                    (* (expt r i)
                       (expt r j))))
    :hints (("Goal" :expand (expt r (+ i j))))
    :rule-classes nil))

 (defthm Exponents-add
   (implies (and (not (equal 0 r))
                 (fc (acl2-numberp r))
                 (fc (integerp i))
                 (fc (integerp j)))
            (equal (expt r (+ i j))
                   (* (expt r i)
                      (expt r j))))
   :hints (("Goal" :use
            (Exponents-add-negative-negative
             Exponents-add-positive-negative
             (:instance Exponents-add-positive-negative
                        (i j) (j i)))))))

(defthm Distributivity-of-expt-over-time
  (equal (expt (* a b) i)
         (* (expt a i)
            (expt b i))))



(defthm expt-1
  (equal (expt 1 x) 1))

(defthm Exponents-multiply
  (implies (and (fc (integerp i))
                (fc (integerp j)))
           (equal (expt (expt r i) j)
                  (expt r (* i j))))
  :hints (("Goal" :cases
           ((not (acl2-numberp r))
            (equal r 0)))))

(defthm Functional-commutativity-of-expt-div-base
  (equal (expt (/ r) i)
         (/ (expt r i))))








