

in-package("ACL2");

Include((:relative :back "cowles" "acl2-crg"));

Macro fc(x) =  x;

Theorem commutativity-2-of-plus { x + y + z = y + x + z };

Theorem functional-self-inversion-of-minus { --x = fix(x) };

Theorem distributivity-of-minus-over-plus { -(x + y) = -x + -y };

Theorem minus-cancellation-on-right { x + y + -x = fix(y) };

Theorem minus-cancellation-on-left { x + -x + y = fix(y) };

Theorem right-cancellation-for-plus { x + z = y + z = (fix(x) = fix(y)) };

Theorem left-cancellation-for-plus { x + y = x + z = (fix(y) = fix(z)) };

Theorem equal-minus-zero { 0 = -x = (0 = fix(x)) };

Theorem inverse-of-plus-as-eqeual-0 { a - b = 0 = (fix(a) = fix(b)) };

Theorem equal-minus-minus { -a = -b = (fix(a) = fix(b)) };

Theorem fold-consts-in-plus
{ syntaxp(quotep(x)) & syntaxp(quotep(y)) -> x + y + z = (x + y) + z };

Theorem commutativity-2-of-time { x * y * z = y * x * z }
 :hints [["Goal" :use [:instance [:functional-instance commutativity-2-of-op
                                                       [equiv := =, 
                                                        pred := lambda(x) { t }, 
                                                        op := binary-time ]]
                                 [x := x, y := y, z := z ]]
                 ]];

Theorem functional-self-inversion-of-div { 1/(1/x) = fix(x) }
 :hints [["Goal" :use [[:instance [:functional-instance involution-of-inv
                                                        [equiv := =, 
                                                         pred := lambda(x)
                                                          {    acl2-numberp(x)
                                                             & x ~= 0 }, 
                                                         op := binary-time, 
                                                         id := lambda() { 1 }, 
                                                         inv := unary-div ]]
                                  [x := x ]]
                       ]]];

Theorem distributivity-of-div-over-time { 1/(x * y) = 1/x * 1/y }
 :hints [["Goal" :use [[:instance [:functional-instance distributivity-of-inv-over-op
                                                        [equiv := =, 
                                                         pred := lambda(x)
                                                          {    acl2-numberp(x)
                                                             & x ~= 0 }, 
                                                         op := binary-time, 
                                                         id := lambda() { 1 }, 
                                                         inv := unary-div ]]
                                  [x := x, y := y ]]
                       ]]];

Theorem div-cancellation-on-right
{ fc(acl2-numberp(x)) & fc(x ~= 0) -> x * y * 1/x = fix(y) }
 :hints [["Goal" :use [[:instance [:functional-instance inv-cancellation-on-right
                                                        [equiv := =, 
                                                         pred := lambda(x)
                                                          {    acl2-numberp(x)
                                                             & x ~= 0 }, 
                                                         op := binary-time, 
                                                         id := lambda() { 1 }, 
                                                         inv := unary-div ]]
                                  [x := x, y := y ]]
                       ]]];

Theorem div-cancellation-on-left
{ fc(acl2-numberp(x)) & fc(0 ~= x) -> x * 1/x * y = fix(y) }
 :hints [["Goal" :use div-cancellation-on-right]];

Module { }
{ 
  local Theorem right-cancellation-for-time-lemma
  {        x * z = y * z
          & acl2-numberp(z)
          & 0 ~= z
          & acl2-numberp(x)
          & acl2-numberp(y)
     -> x = y = t }
   :hints [["Goal" :use [[:instance [:functional-instance right-cancellation-for-op
                                                          [equiv := =, 
                                                           pred := lambda(x)
                                                            {    acl2-numberp(x)
                                                               & x ~= 0 }, 
                                                           op := binary-time, 
                                                           id := lambda() { 1 }, 
                                                           inv := unary-div ]]
                                    [x := x, y := y, z := z ]]
                         ]]];
  
  Theorem right-cancellation-for-time
  { x * z = y * z = (0 = fix(z) | fix(x) = fix(y)) };
  
   };

Theorem left-cancellation-for-time
{ z * x = z * y = (0 = fix(z) | fix(x) = fix(y)) };

Theorem zero-is-only-zero-divisor
{ x * y = 0 = (fix(x) = 0 | fix(y) = 0) };

Theorem equal-time-x-y-x
{ x * y = x = (x = 0 | y = 1 & acl2-numberp(x)) }
 :hints [["Goal" :use [[:instance right-cancellation-for-time
                                  [z := x, x := y, y := 1 ]]
                       ]]];

Theorem equal-time-x-y-y
{ x * y = y = (y = 0 | x = 1 & acl2-numberp(y)) }
 :hints [["Goal" :use [[:instance right-cancellation-for-time
                                  [z := y, x := x, y := 1 ]]
                       ]]];

Module { }
{ 
  local Theorem equal-div-lemma
  { acl2-numberp(x) & acl2-numberp(y) & x * y = 1 -> y = 1/x }
   :hints [["Goal" :use [:instance [:functional-instance uniqueness-of-op-inverses
                                                         [equiv := =, 
                                                          pred := lambda(x)
                                                           {    acl2-numberp(x)
                                                              & x ~= 0 }, 
                                                          nilop := binary-time, 
                                                          nilid := lambda() { 1 }, 
                                                          nilinv := unary-div ]]
                                   [x := x, y := y ]]
                   ]];
  
  Theorem equal-div
  { fc(acl2-numberp(x)) & fc(0 ~= x) -> 1/x = y = (1 = x * y) }
   :hints [["Goal" :use equal-div-lemma]];
  
   };

Theorem numerator-nonzero-forward
{ numerator(r) ~= 0 -> r ~= 0 & acl2-numberp(r) }
 :rule-classes [[:forward-chaining  
 :trigger-terms [numerator(r)]]];

Module { }
{ 
  local Theorem uniqueness-of-time-inverses-lemma
  { x * y = 1 = (x ~= 0 & acl2-numberp(x) & y = 1/x) };
  
  Theorem uniqueness-of-time-inverses
  { x * y = 1 = (fix(x) ~= 0 & y = 1/x) }
   :hints [["Goal" :in-theory disable(equal-div)]];
  
   };

IN-THEORY(disable(uniqueness-of-time-inverses));

Theory-Invariant(~(   '[:rewrite uniqueness-of-time-inverses] in theory
                    & '[:rewrite equal-div] in theory));

Module { }
{ 
  local Theorem equal-div-div-lemma
  {     fc(acl2-numberp(a)) & fc(acl2-numberp(b)) & fc(a ~= 0) & fc(b ~= 0)
     -> 1/a = 1/b = (a = b) }
   :hints [["Goal" :use [[:instance [:theorem        fc(acl2-numberp(a))
                                                    & fc(acl2-numberp(b))
                                                    & fc(a ~= 0)
                                                    & fc(b ~= 0)
                                               -> (a = b -> 1/a = 1/b)]
                                    [a := 1/a, b := 1/b ]]
                         ]]];
  
  Theorem equal-div-div { 1/a = 1/b = (fix(a) = fix(b)) }
   :hints [["Goal" :use equal-div-div-lemma]];
  
   };

Theorem equal-time-div-1
{ acl2-numberp(x) & x ~= 0 -> 1/x * y = z = (acl2-numberp(z) & fix(y) = x * z) };

Theorem equal-time-div-2
{ acl2-numberp(x) & x ~= 0 -> y * 1/x = z = (acl2-numberp(z) & fix(y) = z * x) };

Theorem fold-consts-in-time
{ syntaxp(quotep(x)) & syntaxp(quotep(y)) -> x * y * z = (x * y) * z };

Theorem times-zero { 0 * x = 0 };

Module { }
{ 
  local Theorem functional-commutativity-of-minus-time-right-lemma
  { fc(acl2-numberp(x)) & fc(acl2-numberp(y)) -> x * -y = -(x * y) }
   :hints [["Goal" :use [[:instance [:functional-instance functional-commutativity-of-minus-times-right
                                                          [equiv := =, 
                                                           pred := acl2-numberp, 
                                                           plus := binary-plus, 
                                                           times := binary-time, 
                                                           zero := lambda() { 0 }, 
                                                           minus := unary-minus ]]
                                    [x := x, y := y ]]
                         ]]];
  
  Theorem functional-commutativity-of-minus-time-right
  { x * -y = -(x * y) }
   :hints [["Goal" :use functional-commutativity-of-minus-time-right-lemma]];
  
   };

Theorem functional-commutativity-of-minus-time-left { -x * y = -(x * y) };

Theorem reciprocal-minus { 1/-x = -(1/x) }
 :hints [["Goal" :cases [fc(acl2-numberp(x)) & fc(x ~= 0)]]];

Module { }
{ 
  local Theorem numerator-integerp-lemma-1
  {     rationalp(x)
     -> (numerator(x) * 1/denominator(x)) * denominator(x) = numerator(x) }
   :hints [["Goal" :in-theory disable(rational-implies2)]];
  
  local Theorem numerator-integerp-lemma
  {     rationalp(x) & numerator(x) * 1/denominator(x) = x
     -> numerator(x) = x * denominator(x) }
   :hints [["Goal" :use [numerator-integerp-lemma-1
                         ]
                   :in-theory disable(rational-implies2)]];
  
  Theorem numerator-when-integerp { integerp(x) -> numerator(x) = x }
   :hints [["Goal" :in-theory disable(rational-implies2)
                   :use [[:instance lowest-terms
                                    [r := x, q := 1, n := denominator(x) ]]
                         rational-implies2
                         numerator-integerp-lemma
                         ]]];
  
   };

Theorem integerp-implies-denominator-1
{ integerp(x) -> denominator(x) = 1 }
 :hints [["Goal" :use [rational-implies2
                       numerator-when-integerp
                       ]
                 :in-theory disable(rational-implies2)]];

Theorem equal-denominator-1
{ denominator(x) = 1 = (integerp(x) | ~(rationalp(x))) }
 :hints [["Goal" :use [rational-implies2
                       completion-of-denominator
                       ]
                 :in-theory disable(rational-implies2)]];

Theorem time-r-denominator-r
{    r * denominator(r)
   = (IF rationalp(r) THEN numerator(r)
         ELSE fix(r)) }
 :hints [["Goal" :use [[:instance rational-implies2
                                  [x := r ]]
                       ]
                 :in-theory disable(rational-implies2)]];

Theorem r-when-abs-numerator-1
{    (numerator(r) = 1 -> 1/r = denominator(r))
   & (numerator(r) = -1 -> 1/r = -(denominator(r))) }
 :hints [["Goal" :use [[:instance rational-implies2
                                  [x := r ]]
                       ]
                 :in-theory disable(rational-implies2)]];

Theorem expt-type-prescription-rationalp
{ rationalp(r) -> rationalp(expt(r,i)) }
 :rule-classes [:type-prescription :generalize];

Theorem expt-type-prescription-positive
{ 0 < r & rationalp(r) -> 0 < expt(r,i) }
 :rule-classes [:type-prescription :generalize];

Theorem expt-type-prescription-nonzero
{ fc(acl2-numberp(r)) & r ~= 0 -> 0 ~= expt(r,i) }
 :rule-classes [:type-prescription :generalize];

Theorem expt-type-prescription-integerp
{ 0 <= i & integerp(r) -> integerp(expt(r,i)) }
 :rule-classes [:type-prescription :generalize];

Theorem left-nullity-of-1-for-expt { expt(1,i) = 1 };

Theorem right-unicity-of-1-for-expt { expt(r,1) = fix(r) }
 :hints [["Goal" :expand expt(r,1)]];

Theorem expt-minus { expt(r,-i) = 1/expt(r,i) };

Theorem exponents-add-for-nonneg-exponents
{     0 <= i & 0 <= j & fc(integerp(i)) & fc(integerp(j))
   -> expt(r,i + j) = expt(r,i) * expt(r,j) };

Module { }
{ 
  local Theorem exponents-add-negative-negative
  {     integerp(i) & integerp(j) & i < 0 & j < 0
     -> expt(r,i + j) = expt(r,i) * expt(r,j) };
  
  local Theorem exponents-add-positive-negative
  {     integerp(i) & integerp(j) & acl2-numberp(r) & r ~= 0 & 0 < i & j < 0
     -> expt(r,i + j) = expt(r,i) * expt(r,j) }
   :hints [["Goal" :expand expt(r,i + j)]];
  
  Theorem exponents-add
  {     0 ~= r & fc(acl2-numberp(r)) & fc(integerp(i)) & fc(integerp(j))
     -> expt(r,i + j) = expt(r,i) * expt(r,j) }
   :hints [["Goal" :use [exponents-add-negative-negative
                         exponents-add-positive-negative
                         [:instance exponents-add-positive-negative
                                    [i := j, j := i ]]
                         ]]];
  
   };

Theorem distributivity-of-expt-over-time
{ expt(a * b,i) = expt(a,i) * expt(b,i) };

Theorem expt-1 { expt(1,x) = 1 };

Theorem exponents-multiply
{ fc(integerp(i)) & fc(integerp(j)) -> expt(expt(r,i),j) = expt(r,i * j) }
 :hints [["Goal" :cases [~(acl2-numberp(r)), r = 0]]];

Theorem functional-commutativity-of-expt-div-base
{ expt(1/r,i) = 1/expt(r,i) };


