(in-package "NQTHM")

(acl2::set-verify-guards-eagerness 2)

(defn true () t)
(defn false () f)
(defn truep (x) (equal x t))
(defn falsep (x) (equal x f))

(defn and (p q) (if p (if q t f) f))
(defn or (p q) (if p t (if q t f)))

(prove-lemma defn-not ()
  (equal (not p) (if p f t)))

(prove-lemma defn-implies ()
  (equal (implies p q) (if p (if q t f) t)))

(prove-lemma defn-iff ()
  (equal (iff p q) (and (implies p q) (implies q p))))


(prove-lemma ax-12-1 ()
  (OR (EQUAL (NUMBERP X) T)
      (EQUAL (NUMBERP X) F)))

(prove-lemma ax-12-2 ()
  (NUMBERP (ADD1 X1)))

(prove-lemma ax-12-3 ()
  (NUMBERP (ZERO)))

(prove-lemma ax-12-4 ()
  (NOT (EQUAL (ADD1 X1) (ZERO))))

(prove-lemma ax-12-5 ()
  (IMPLIES (AND (NUMBERP X)
                (NOT (EQUAL X (ZERO))))
           (EQUAL (ADD1 (SUB1 X)) X)))

(prove-lemma ax-12-6 ()
  (IMPLIES (NUMBERP X1)
           (EQUAL (SUB1 (ADD1 X1)) X1)))

(prove-lemma ax-12-7 ()
  (IMPLIES (OR (NOT (NUMBERP X))
               (OR (EQUAL X (ZERO))
                   (AND (NOT (NUMBERP X1))
                        (EQUAL X (ADD1 X1)))))
           (EQUAL (SUB1 X) (ZERO))))

(prove-lemma ax-12-8-1 ()
  (IMPLIES (NUMBERP X) (NOT (TRUEP X))))

(prove-lemma ax-12-8-2 ()
  (IMPLIES (NUMBERP X) (NOT (FALSEP X))))

(prove-lemma ax-12-9 ()
  (IMPLIES (NUMBERP X)
           (EQUAL (COUNT X)
                  (IF (EQUAL X (ZERO))
                      (ZERO)
                      (ADD1 (COUNT (SUB1 X)))))))

(defn fix (x) (if (numberp x) x 0))

(defn zerop (x) (or (equal x 0) (not (numberp x))))

(defn plus (x y)
  (if (zerop x)
      (fix y)
      (add1 (plus (sub1 x) y))))

(defn lessp (x y)
  (if (or (equal y 0) (not (numberp y)))
      f
      (if (or (equal x 0) (not (numberp x)))
          t
          (lessp (sub1 x) (sub1 y)))))

(defn difference1 (x y)
  (if (or (equal y 0) (not (numberp y)))
      (fix x)
      (if (or (equal x 0) (not (numberp x)))
          0
          (difference1 (sub1 x) (sub1 y)))))

(defn times (x y)
  (if (zerop x)
      0
      (plus y (times (sub1 x) y))))

(prove-lemma ax-19-1 ()
  (OR (EQUAL (LISTP X) T)
      (EQUAL (LISTP X) F)))

(prove-lemma ax-19-2 ()
  (LISTP (CONS X1 X2)))

(prove-lemma ax-19-5 ()
  (IMPLIES (LISTP X)
           (EQUAL (CONS (CAR X) (CDR X)) X)))

(prove-lemma ax-19-6-1 ()
  (EQUAL (CAR (CONS X1 X2)) X1))

(prove-lemma ax-19-6-2 ()
  (EQUAL (CDR (CONS X1 X2)) X2))

(prove-lemma ax-19-7-1 ()
  (IMPLIES (NOT (LISTP X))
           (EQUAL (CAR X) (ZERO))))  

(prove-lemma ax-19-7-2 ()
  (IMPLIES (NOT (LISTP X))
           (EQUAL (CDR X) (ZERO))))

(prove-lemma ax-19-8-1 ()
  (IMPLIES (LISTP X) (NOT (NUMBERP X))))

(prove-lemma ax-19-8-2 ()
  (IMPLIES (LISTP X) (NOT (TRUEP X))))

(prove-lemma ax-19-8-3 ()
  (IMPLIES (LISTP X) (NOT (FALSEP X))))

; Here is the first really neat fact, an inductive proof that PLUS is +.

(acl2::defthm acl2::plus-is-plus
  (equal (plus x y)
         (acl2::+ (fix x) (fix y))))

(prove-lemma ax-19-9 ()
  (IMPLIES (LISTP X)
           (EQUAL (COUNT X)
                  (ADD1 (PLUS (COUNT (CAR X))
                              (COUNT (CDR X)))))))

(prove-lemma ax-20-1 ()
  (OR (EQUAL (LITATOM X) T)
      (EQUAL (LITATOM X) F)))

(defthm acl2::character-listp-ascii-char-lst
 (implies (acl2::all-nqthm-litatom-char-codep acl2::x)
          (acl2::character-listp (acl2::ascii-char-lst acl2::x))))

(acl2::defthm acl2::character-listp-ascii-char-lst
 (implies (acl2::all-nqthm-litatom-char-codep acl2::x)
          (acl2::character-listp (acl2::ascii-char-lst acl2::x))))

(prove-lemma ax-20-2 ()
  (LITATOM (PACK X1)))
; Keywords not allowed in general.
; :hints (("Goal" :in-theory
;           (acl2::disable acl2::ascii-code 
;                          acl2::unpacked-nqthm-litatomp-implies-standard-char-listp)
;           :use (:instance acl2::unpacked-nqthm-litatomp-implies-standard-char-listp
;                           (acl2::lst x1)))) 

(prove-lemma ax-20-5 ()
  (IMPLIES (LITATOM X)
           (EQUAL (PACK (UNPACK X)) X)))
;;  Keywords not allowed in general.
;;   :hints (("Goal"
;;            :in-theory
;;            (acl2::disable
;;             acl2::ascii-code 
;;             acl2::unpacked-nqthm-litatomp-implies-standard-char-listp)
;;            :use
;;            (:instance acl2::unpacked-nqthm-litatomp-implies-standard-char-listp
;;                       (acl2::lst
;;                        (ACL2::ASCII-CODE-LST (LISP::COERCE (LISP::SYMBOL-NAME X)
;;                                                            'LISP::LIST)))))
          
;;           ("Subgoal 6"
;;            :in-theory
;;            (acl2::disable acl2::intern-in-package-of-symbol-symbol-name)
;;            :use (:instance acl2::intern-in-package-of-symbol-symbol-name
;;                            (acl2::x x)
;;                            (acl2::y 'nqthm::pack)))
;;           ("Subgoal 5"
;;            :in-theory
;;            (acl2::disable acl2::intern-in-package-of-symbol-symbol-name)
;;            :use (:instance acl2::intern-in-package-of-symbol-symbol-name
;;                            (acl2::x x)
;;                            (acl2::y 'nqthm::pack)))
;;           ("Subgoal 4"
;;            :in-theory
;;            (acl2::disable acl2::intern-in-package-of-symbol-symbol-name)
;;            :use (:instance acl2::intern-in-package-of-symbol-symbol-name
;;                            (acl2::x x)
;;                            (acl2::y 'nqthm::pack)))
;;           ("Subgoal 3"
;;            :in-theory
;;            (acl2::disable acl2::intern-in-package-of-symbol-symbol-name)
;;            :use (:instance acl2::intern-in-package-of-symbol-symbol-name
;;                            (acl2::x x)
;;                            (acl2::y 'nqthm::pack)))
;;           ("Subgoal 2"
;;            :in-theory
;;            (acl2::disable acl2::intern-in-package-of-symbol-symbol-name)
;;            :use (:instance acl2::intern-in-package-of-symbol-symbol-name
;;                            (acl2::x x)
;;                            (acl2::y 'nqthm::pack)))
;;           ("Subgoal 1"
;;            :in-theory
;;            (acl2::disable acl2::intern-in-package-of-symbol-symbol-name)
;;            :use (:instance acl2::intern-in-package-of-symbol-symbol-name
;;                            (acl2::x x)
;;                            (acl2::y 'nqthm::pack))))

(prove-lemma ax-20-6 ()
  (EQUAL (UNPACK (PACK X1)) X1))
;;  Keywords not allowed in general.
;;   :hints
;;   (("Goal" :in-theory
;;     (acl2::union-theories
;;      '(acl2::ascii-code-lst)
;;      (acl2::disable acl2::ascii-code 
;;                     acl2::unpacked-nqthm-litatomp-implies-standard-char-listp))
;;     :use (:instance acl2::unpacked-nqthm-litatomp-implies-standard-char-listp
;;                     (acl2::lst x1)))
;;    ("Subgoal 9"
;;     :in-theory (acl2::union-theories '(acl2::unpacked-nqthm-litatomp)
;;                                      (acl2::disable acl2::coerce-inverse-1))
;;     :use
;;     ((:instance acl2::coerce-inverse-1
;;                 (acl2::x (acl2::ascii-char-lst x1)))
;;      (:instance
;;       acl2::all-nqthm-litatom-char-codep-implies-ascii-code-lst-inverts-ascii-char-lst
;;       (acl2::lst x1))))
;;    ("Subgoal 8"
;;     :in-theory (acl2::union-theories '(acl2::unpacked-nqthm-litatomp)
;;                                      (acl2::disable acl2::coerce-inverse-1))
;;     :use
;;     ((:instance acl2::coerce-inverse-1
;;                 (acl2::x (acl2::ascii-char-lst x1)))
;;      (:instance
;;       acl2::all-nqthm-litatom-char-codep-implies-ascii-code-lst-inverts-ascii-char-lst
;;       (acl2::lst x1))))
;;    ("Subgoal 7"
;;     :in-theory (acl2::union-theories '(acl2::unpacked-nqthm-litatomp)
;;                                      (acl2::disable acl2::coerce-inverse-1))
;;     :use
;;     ((:instance acl2::coerce-inverse-1
;;                 (acl2::x (acl2::ascii-char-lst x1)))
;;      (:instance
;;       acl2::all-nqthm-litatom-char-codep-implies-ascii-code-lst-inverts-ascii-char-lst
;;       (acl2::lst x1))))
;;    ("Subgoal 6"
;;     :in-theory (acl2::union-theories '(acl2::unpacked-nqthm-litatomp)
;;                                      (acl2::disable acl2::coerce-inverse-1))
;;     :use
;;     ((:instance acl2::coerce-inverse-1
;;                 (acl2::x (acl2::ascii-char-lst x1)))
;;      (:instance
;;       acl2::all-nqthm-litatom-char-codep-implies-ascii-code-lst-inverts-ascii-char-lst
;;       (acl2::lst x1))))
;;    ("Subgoal 5"
;;     :in-theory (acl2::union-theories '(acl2::unpacked-nqthm-litatomp)
;;                                      (acl2::disable acl2::coerce-inverse-1))
;;     :use
;;     ((:instance acl2::coerce-inverse-1
;;                 (acl2::x (acl2::ascii-char-lst x1)))
;;      (:instance
;;       acl2::all-nqthm-litatom-char-codep-implies-ascii-code-lst-inverts-ascii-char-lst
;;       (acl2::lst x1))))
;;    ("Subgoal 4"
;;     :in-theory (acl2::union-theories '(acl2::unpacked-nqthm-litatomp)
;;                                      (acl2::disable acl2::coerce-inverse-1))
;;     :use
;;     ((:instance acl2::coerce-inverse-1
;;                 (acl2::x (acl2::ascii-char-lst x1)))
;;      (:instance
;;       acl2::all-nqthm-litatom-char-codep-implies-ascii-code-lst-inverts-ascii-char-lst
;;       (acl2::lst x1)))))

(prove-lemma ax-20-7 ()
  (IMPLIES (NOT (LITATOM X))
           (EQUAL (UNPACK X) (ZERO))))

(prove-lemma ax-20-8-1 ()
  (IMPLIES (LITATOM X) (NOT (LISTP X))))

(prove-lemma ax-20-8-2 ()
  (IMPLIES (LITATOM X) (NOT (NUMBERP X))))

(prove-lemma ax-20-8-3 ()
  (IMPLIES (LITATOM X) (NOT (TRUEP X))))

(prove-lemma ax-20-8-4 ()
  (IMPLIES (LITATOM X) (NOT (FALSEP X))))

(prove-lemma ax-20-9 ()
  (IMPLIES (LITATOM X)
           (EQUAL (COUNT X)
                  (ADD1 (COUNT (UNPACK X))))))

(prove-lemma ax-21-1 ()
  (OR (EQUAL (NEGATIVEP X) T)
      (EQUAL (NEGATIVEP X) F)))

(prove-lemma ax-21-2 ()
  (NEGATIVEP (MINUS X1)))

(prove-lemma ax-21-5 ()
  (IMPLIES (NEGATIVEP X)
           (EQUAL (MINUS (NEGATIVE-GUTS X)) X)))

(prove-lemma ax-21-6 ()
  (IMPLIES (NUMBERP X1)
           (EQUAL (NEGATIVE-GUTS (MINUS X1)) X1)))

(prove-lemma ax-21-7 ()
  (IMPLIES (OR (NOT (NEGATIVEP X))
               (AND (NOT (NUMBERP X1))
                    (EQUAL X (MINUS X1))))
           (EQUAL (NEGATIVE-GUTS X) (ZERO))))

(prove-lemma ax-21-8-1 ()
  (IMPLIES (NEGATIVEP X) (NOT (LITATOM X))))

(prove-lemma ax-21-8-2 ()
  (IMPLIES (NEGATIVEP X) (NOT (LISTP X))))

(prove-lemma ax-21-8-3 ()
  (IMPLIES (NEGATIVEP X) (NOT (NUMBERP X))))

(prove-lemma ax-21-8-4 ()
  (IMPLIES (NEGATIVEP X) (NOT (TRUEP X))))

(prove-lemma ax-21-8-5 ()
  (IMPLIES (NEGATIVEP X) (NOT (FALSEP X))))

(prove-lemma ax-21-9 ()
  (IMPLIES (NEGATIVEP X)
           (EQUAL (COUNT X)
                  (ADD1 (COUNT (NEGATIVE-GUTS X))))))

