

Include("/usr/cli/projects/cacl2/version13/infix-prelude");

in-package("ACL2");

Include("equalities");

local Theorem iff-equal
{ (x <-> y) & booleanp(x) & booleanp(y) -> x = y };

Theorem div-preserves-positive { rationalp(x) -> 0 < 1/x <-> 0 < x }
 :rule-classes [[:rewrite  
 :corollary fc(rationalp(x)) -> 0 < 1/x = 0 < x]];

Theorem div-preserves-negative { rationalp(x) -> 1/x < 0 <-> x < 0 }
 :rule-classes [[:rewrite  
 :corollary fc(rationalp(x)) -> 1/x < 0 = x < 0]];

Theorem less-time-right-cancel
{     fc(rationalp(x)) & fc(rationalp(y)) & fc(rationalp(z))
   ->      x * z < y * z
      <-> IF 0 < z THEN x < y
            ELIF z = 0 THEN nil
            ELSE y < x }
 :hints [["Goal" :use [[:instance [:theorem        rationalp(a)
                                                 & 0 < a
                                                 & rationalp(b)
                                                 & 0 < b
                                             -> 0 < a * b]
                                  [a := abs(y - x), b := abs(z) ]]
                       ]]];

Theorem less-time-left-cancel
{     fc(rationalp(x)) & fc(rationalp(y)) & fc(rationalp(z))
   ->      z * x < z * y
      <-> IF 0 < z THEN x < y
            ELIF z = 0 THEN nil
            ELSE y < x };

Theorem div-inverts-order-1
{ 0 < x & x < y & fc(rationalp(x)) & fc(rationalp(y)) -> 1/y < 1/x }
 :hints [["Goal" :use [[:instance less-time-right-cancel
                                  [z := 1/(x * y) ]]
                       ]]];

Theorem less-0-minus { 0 < -x <-> x < 0 }
 :rule-classes [[:rewrite  
 :corollary 0 < -x = x < 0]];

Theorem less-0-minus { 0 < -x = x < 0 }
 :hints [["Goal" :use [[:instance iff-equal
                                  [x := 0 < -x, y := x < 0 ]]
                       ]]];

Theorem less-time-0
{     rationalp(x) & rationalp(y)
   -> x * y < 0 <-> x ~= 0 & y ~= 0 & (x < 0 <-> 0 < y) }
 :rule-classes [[:rewrite  
 :corollary     fc(rationalp(x)) & fc(rationalp(y))
             -> x * y < 0 = (x ~= 0 & y ~= 0 & (x < 0 <-> 0 < y))]];

Theorem zero-less-time
{     rationalp(x) & rationalp(y)
   -> 0 < x * y <-> x ~= 0 & y ~= 0 & (0 < x <-> 0 < y) }
 :rule-classes [[:rewrite  
 :corollary     fc(rationalp(x)) & fc(rationalp(y))
             -> 0 < x * y = (x ~= 0 & y ~= 0 & (0 < x <-> 0 < y))]];

Theorem div-inverts-order-2
{ y < 0 & x < y & fc(rationalp(x)) & fc(rationalp(y)) -> 1/y < 1/x }
 :hints [["Goal" :use [[:instance div-time-right-cancel
                                  [z := 1/(x * y) ]]
                       ]
                 :in-theory disable(div-time-right-cancel)]];

Theorem div-inverts-weak-order
{ 0 < x & x <= y & fc(rationalp(x)) & fc(rationalp(y)) -> ~(1/x < 1/y) }
 :hints [["Goal" :use div-inverts-order-1
                 :in-theory disable(div-inverts-order-1)]];

Theorem less-unary-div-negative-left
{ rationalp(x) & rationalp(y) & x < 0 -> 1/x < y <-> x * y < 1 }
 :rule-classes [[:rewrite  
 :corollary x < 0 & fc(rationalp(x)) & fc(rationalp(y)) -> 1/x < y = x * y < 1]]
 :hints [["Goal" :use [[:instance div-time-left-cancel
                                  [z := x, x := 1/x, y := y ]]
                       ]
                 :in-theory             enable(uniqueness-of-time-inverses)
                             difference '[equal-div]]];

Theorem less-unary-div-negative-right
{ rationalp(x) & rationalp(y) & x < 0 -> y < 1/x <-> 1 < x * y }
 :rule-classes [[:rewrite  
 :corollary x < 0 & fc(rationalp(x)) & fc(rationalp(y)) -> y < 1/x = 1 < x * y]]
 :hints [["Goal" :use [[:instance less-time-right-cancel
                                  [z := x, x := 1/x, y := y ]]
                       ]]];

Theorem less-unary-div-positive-left
{ rationalp(x) & rationalp(y) & 0 < x -> 1/x < y <-> 1 < x * y }
 :rule-classes [[:rewrite  
 :corollary 0 < x & fc(rationalp(x)) & fc(rationalp(y)) -> 1/x < y = 1 < x * y]]
 :hints [["Goal" :use [[:instance less-time-left-cancel
                                  [z := x, x := 1/x, y := y ]]
                       ]]];

Theorem less-unary-time-positive-right
{ rationalp(x) & rationalp(y) & 0 < x -> y < 1/x <-> x * y < 1 }
 :rule-classes [[:rewrite  
 :corollary 0 < x & fc(rationalp(x)) & fc(rationalp(y)) -> y < 1/x = x * y < 1]]
 :hints [["Goal" :use [[:instance less-time-left-cancel
                                  [z := x, x := 1/x, y := y ]]
                       ]
                 :in-theory        '[uniqueness-of-time-inverses]
                             union disable(less-unary-div-positive-left,
                                          equal-div,
                                          less-time-left-cancel)]];

Theorem time-minus-zero { -x < 0 = 0 < x }
 :hints [["Goal" :use [[:instance iff-equal
                                  [x := -x < 0, y := 0 < x ]]
                       ]]];

Theorem less-time-div-right
{     0 < y & fc(rationalp(a)) & fc(rationalp(x)) & fc(rationalp(y))
   -> a < x * 1/y = a * y < x }
 :hints [["Goal" :use [[:instance less-time-right-cancel
                                  [x := a, y := x * 1/y, z := y ]]
                       ]]];

Theorem less-time-div-right-commuted
{     0 < y & fc(rationalp(x)) & fc(rationalp(y)) & fc(rationalp(a))
   -> x < 1/y * a = x * y < a };

Theorem less-time-diov-left
{     0 < y & fc(rationalp(x)) & fc(rationalp(y)) & fc(rationalp(a))
   -> x * 1/y < a = x < a * y }
 :hints [["Goal" :use [[:instance less-time-right-cancel
                                  [y := a, x := x * 1/y, z := y ]]
                       ]]];

Theorem div-time-div-left-commuted
{     0 < y & fc(rationalp(x)) & fc(rationalp(y)) & fc(rationalp(a))
   -> 1/y * x < a = x < y * a };

Module { }
{ 
  local Theorem time-preserves-greater-eq-1
  {        rationalp(x1)
         & rationalp(x2)
         & rationalp(y1)
         & x1 >= x2
         & x2 >= 0
         & y1 >= 0
     -> x1 * y1 >= x2 * y1 };
  
  local Theorem time-preserves-greater-eq-2
  {        rationalp(x2)
         & rationalp(y1)
         & rationalp(y2)
         & y1 >= y2
         & x2 >= 0
         & y2 >= 0
     -> x2 * y1 >= x2 * y2 };
  
  Theorem time-preserves-greater-eq-for-nonnegatives
  {        x1 >= x2
         & y1 >= y2
         & x2 >= 0
         & y2 >= 0
         & fc(rationalp(x1))
         & fc(rationalp(x2))
         & fc(rationalp(y1))
         & fc(rationalp(y2))
     -> x1 * y1 >= x2 * y2 }
   :hints [["Goal" :use [time-preserves-greater-eq-1
                         time-preserves-greater-eq-2
                         ]
                   :in-theory disable(less-time-left-cancel,
                                      less-time-right-cancel)]];
  
   };

Theorem equal-binary-quotient
{     fc(rationalp(x)) & fc(rationalp(y)) & fc(rationalp(z)) & fc(z ~= 0)
   -> x = y/z = (x * z = y) }
 :hints [["Goal" :use [[:instance [:theorem        rationalp(x)
                                                 & rationalp(y)
                                                 & rationalp(z)
                                                 & z ~= 0
                                                 & x = y
                                             -> x * z = y * z]
                                  [x := x * z, y := y, z := 1/z ]]
                       ]]];

Module { }
{ 
  local Theorem time-preserves-greater-1
  {        rationalp(x1)
         & rationalp(x2)
         & rationalp(y1)
         & x1 > x2
         & x2 >= 0
         & y1 > 0
     -> x1 * y1 >= x2 * y1 };
  
  local Theorem time-preserves-greater-2
  {        rationalp(x2)
         & rationalp(y1)
         & rationalp(y2)
         & y1 >= y2
         & x2 >= 0
         & y2 >= 0
     -> x2 * y1 >= x2 * y2 };
  
  Theorem time-preserves-greater-for-nonnegatives-1
  {        x1 > x2
         & y1 >= y2
         & x2 >= 0
         & y2 > 0
         & fc(rationalp(x1))
         & fc(rationalp(x2))
         & fc(rationalp(y1))
         & fc(rationalp(y2))
     -> x1 * y1 > x2 * y2 }
   :hints [["Goal" :use [time-preserves-greater-1
                         time-preserves-greater-2
                         ]
                   :in-theory disable(less-greater-left-cancel,
                                      less-greater-right-cancel,
                                      time-preserves-greater-eq-for-nonnegatives)]];
  
   };

Module { }
{ 
  Theorem x-time-y-greater-1-positive-lemma
  {        x > i
         & y > j
         & rationalp(i)
         & rationalp(j)
         & 0 < i
         & 0 < j
         & fc(rationalp(x))
         & fc(rationalp(y))
     -> x * y > i * j }
   :hints [["Goal" :use [[:instance zero-less-time
                                    [x := x - i, y := y - j ]]
                         ]]];
  
  Theorem x-time-y-greater-1-positive
  { x > 1 & y > 1 & fc(rationalp(x)) & fc(rationalp(y)) -> x * y > 1 }
   :rule-classes [:linear :rewrite]
   :hints [["Goal" :use [[:instance x-time-y-greater-1-positive-lemma
                                    [i := 1, j := 1 ]]
                         ]]];
  
   };

Theorem less-time-x-y-y
{     fc(rationalp(x)) & fc(rationalp(y))
   ->    x * y < y
      = (IF y = 0 THEN nil
            ELIF 0 < y THEN x < 1
            ELSE 1 < x) }
 :hints [["Goal" :use [[:instance time-time-right-cancel
                                  [z := y, x := x, y := 1 ]]
                       ]]];

Theorem less-time-y-x-y
{     fc(rationalp(x)) & fc(rationalp(y))
   ->    y * x < y
      = (IF y = 0 THEN nil
            ELIF 0 < y THEN x < 1
            ELSE 1 < x) };

Theorem less-y-time-x-y
{     fc(rationalp(x)) & fc(rationalp(y))
   ->    y < x * y
      = (IF y = 0 THEN nil
            ELIF 0 < y THEN 1 < x
            ELSE x < 1) }
 :hints [["Goal" :use [[:instance less-time-right-cancel
                                  [z := y, x := 1, y := x ]]
                       ]]];

Theorem less-y-time-y-x
{     fc(rationalp(x)) & fc(rationalp(y))
   ->    y < y * x
      = (IF y = 0 THEN nil
            ELIF 0 < y THEN 1 < x
            ELSE x < 1) };

Theorem less-0-plus-negative-1 { 0 < -y + x = y < x }
 :hints [["Goal" :use [:instance iff-equal
                                 [x := 0 < -y + x, y := y < x ]]]];

Theorem less-0-plus-negative-2 { 0 < x + -y = y < x }
 :hints [["Goal" :use [:instance iff-equal
                                 [x := 0 < x + -y, y := y < x ]]]];

Theorem less-plus-negative-0-1 { -y + x < 0 = x < y }
 :hints [["Goal" :use [:instance iff-equal
                                 [x := -y + x < 0, y := x < y ]]]];

Theorem less-plus-negative-0-2 { x + -y < 0 = x < y }
 :hints [["Goal" :use [:instance iff-equal
                                 [x := x + -y < 0, y := x < y ]]]];

Theorem expt-greater-1
{ 1 < r & 0 < i & fc(rationalp(r)) & fc(integerp(i)) -> 1 < expt(r,i) }
 :rule-classes :linear;

Module { }
{ 
  local Function math-induction-start-at-k(k,n)
  { IF integerp(k) & integerp(n) & k < n THEN math-induction-start-at-k(k,
                                                                        -1 + n)
       ELSE t }
   :measure IF integerp(k) & integerp(n) & k < n THEN n - k
               ELSE 0
   :guard integerp(k) & integerp(n) & ~(n < k);
  
  local Theorem hack
  {     x * y < z & rationalp(x) & rationalp(y) & rationalp(z) & 1 < x & 0 < y
     -> y < z }
   :hints [["Goal" :use [[:instance [:theorem        rationalp(x)
                                                   & rationalp(y)
                                                   & rationalp(z)
                                                   & 0 < y
                                                   & rationalp(w)
                                                   & 0 < w
                                                   & y + w < z
                                               -> y < z]
                                    [w := (x - 1) * y ]]
                         ]]];
  
  Theorem expt-is-increasing-for-base-greater-1
  {     1 < r & i < j & fc(rationalp(r)) & fc(integerp(i)) & fc(integerp(j))
     -> expt(r,i) < expt(r,j) }
   :rule-classes [:rewrite :linear]
   :hints [["Goal" :induct math-induction-start-at-k(1 + i,j)]];
  
   };

Theorem expt-is-decreasing-for-pos-base-less-1
{        0 < r
       & r < 1
       & i < j
       & fc(rationalp(r))
       & fc(integerp(i))
       & fc(integerp(j))
   -> expt(r,j) < expt(r,i) }
 :hints [["Goal" :use [[:instance expt-is-increasing-for-base-greater-1
                                  [r := 1/r ]]
                       ]]];

Theorem expt-is-weakly-increasing-for-base-greater-1
{     1 < r & i <= j & fc(rationalp(r)) & fc(integerp(i)) & fc(integerp(j))
   -> expt(r,i) <= expt(r,j) }
 :hints [["Goal" :use expt-is-increasing-for-base-less-1
                 :in-theory disable(expt-is-increasing-for-base-less-1)]];

Theorem expt-is-weakly-decreasing-for-pos-base-less-1
{        0 < r
       & r < 1
       & i <= j
       & fc(rationalp(r))
       & fc(integerp(i))
       & fc(integerp(j))
   -> expt(r,j) <= expt(r,i) }
 :hints [["Goal" :use expt-is-decreasing-for-pos-base-less-1
                 :in-theory disable(expt-is-decreasing-for-pos-base-less-1)]];


