

(in-package "ACL2")
(include-book "list-defuns")
(local (include-book (:relative :back "arithmetic" "top")))

(deflabel list-defthms-section)


(local (defthm one-plus-opener
  (equal (1+ x)
	 (if (acl2-numberp x)
	     (+ 1 x)
	   (apply 'one-plus (list x))))))

(local (defthm one-minus-opener
  (equal (1- x)
	 (if (acl2-numberp x)
	     (+ -1 x)
	   (apply 'one-minus (list x))))))


(defthm append-nil
  (implies (true-listp a)
	   (equal (append a nil) a)))


(defthm nthcdr-nil
  (implies (and (integerp n)
		(<= 0 n))
	   (not (nthcdr n nil))))

(defthm nth-seg-non-recursive-defun
  (equal (nth-seg i j l)
	 (firstn j (nthcdr i l))))

(defthm put-seg-non-recursive-defun
  (implies (and (true-listp l)
		(integerp i)
		(>= i 0))
	   (equal (put-seg i seg l)
		  (append (firstn i l)
			  (firstn (nfix (- (len l) i)) seg)
			  (nthcdr (+ i (min (nfix (- (len l) i)) (len seg)))
				  l))))
  :hints (("Goal" :induct (put-seg i seg l))))

(local 
 (defthm make-list-ac-rewrite
   (implies (and (integerp i)
		 (< 0 i))
	    (equal (make-list-ac i v ac)
		   (cons v (make-list-ac (1- i) v ac))))
   :hints (("Goal" :expand (make-list-ac 1 v ac)))
   :rule-classes nil))

(defthm put-seg-fill-non-recursive-defun
  (implies (and (integerp i)
		(<= 0 i)
		(true-listp l)
		(true-listp seg))
	   (equal (put-seg-fill i seg l filler)
		  (append (firstn i l)
			  (make-list (max (- i (len l)) 0)
				     :initial-element filler)
			  seg
			  (nthcdr (+ i (len seg)) l))))
  :hints (("Goal" :induct (put-seg-fill i seg l filler))
	  ("Subgoal *1/2.2''" :use (:instance make-list-ac-rewrite
					      (i i) (v filler) (ac nil)))
	  ("Subgoal *1/2.1''" :use (:instance make-list-ac-rewrite
					      (i 1) (v filler) (ac nil)))))

(in-theory (disable nth-seg-non-recursive-defun
		    put-seg-non-recursive-defun
		    put-seg-fill-non-recursive-defun))



(defthm positive-len-fwd-to-consp
  (implies (< 0 (len l))
	   (consp l))
  :rule-classes :forward-chaining)

(defthm len-append 
  (equal (len (append a b))
	 (+ (len a) (len b))))

(defthm len-take
  (equal (len (take n l)) (nfix n)))

(defthm len-butlast
  (implies (and (integerp n)
		(listp l))
	   (equal (len (butlast l n))
		  (if (< (len l) (nfix n))
		      0
		      (- (len l) n))))
  :hints (("Goal" :in-theory (enable butlast))))

(defthm len-firstn
  (equal (len (firstn n l))
	 (if (<= (nfix n) (len l))
	     (nfix n)
	     (len l)))
  :hints (("Goal" :induct (firstn n l))))

(defthm len-last 
  (equal (len (last l))
	 (if (consp l)
	     1
	     0))
  :hints (("Goal" :induct (last l))))

(defthm len-make-list-ac
  (equal (len (make-list-ac n val ac))
	 (+ (nfix n) (len ac)))
  :hints (("Goal" :induct (make-list-ac n val ac))))

(defthm len-nthcdr
  (equal (len (nthcdr n l))
	 (if (<= (nfix n) (len l))
	     (- (len l) (nfix n))
	     0))
  :hints (("Goal" :induct (nthcdr n l))))

(defthm len-nth-seg
  (implies (<= (+ i j) (len l))
	   (equal (len (nth-seg i j l))
		  (min (nfix j) (max (- (len l) (nfix i)) 0))))
  :hints (("Goal" :do-not-induct t
	   :in-theory (union-theories
		       (set-difference-theories
			(current-theory :here)
			'((:definition nth-seg)))
		       '(nth-seg-non-recursive-defun)))))

(defthm len-put-nth
  (equal (len (put-nth n v l))
	 (len l)))

(defthm len-put-nth-fill
  (equal (len (put-nth-fill n v l filler))
	 (max (len l) (1+ (nfix n))))
  :hints (("Goal" :induct (put-nth-fill n v l filler))))

(defthm len-put-seg
  (implies (and (true-listp l)
		(integerp i)
		(not (< i 0)))
	   (equal (len (put-seg i seg l))
		  (len l)))
  :hints (("Goal" :do-not-induct t
	   :in-theory (union-theories
		       (set-difference-theories
			(current-theory :here)
			'((:definition put-seg)))
		       '(put-seg-non-recursive-defun)))))

(defthm len-put-seg-fill
  (implies (and (integerp i)
		(not (< i 0))
		(true-listp l)
		(true-listp seg))
	   (equal (len (put-seg-fill i seg l filler))
		  (max (len l) (+ (nfix i) (len seg)))))
  :hints (("Goal" :do-not-induct t
	   :in-theory (union-theories
		       (set-difference-theories
			(current-theory :here)
			'((:definition put-seg-fill)))
		       '(put-seg-fill-non-recursive-defun)))))

(defthm len-remove-equal 
  (equal (len (remove-equal x l))
	 (- (len l) (count-equal x l))))

(defthm len-revappend 
  (equal (len (revappend a b))
	 (+ (len a) (len b))))

(defthm len-reverse 
  (equal (len (reverse a)) (len a)))

(defthm lessp-count-equal-len 
  (not (< (len l) (count-equal x l)))
  :rule-classes (:rewrite :linear))

; ---------- COUNT-EQUAL ----------

(defthm equal-count-equal-zero
   (implies (not (member-equal x l))
	    (equal (count-equal x l) 0))
   :rule-classes (:rewrite))

(defthm count-equal-append 
  (equal (count-equal x (append a b))
	 (+ (count-equal x a)
	    (count-equal x b)))
  :hints (("Goal" :induct (binary-append a b))))

(defthm count-equal-firstn
  (<= (count-equal x (firstn n l))
      (count-equal x l))
  :rule-classes :linear)

(defthm count-equal-firstn-position
  (equal (count-equal x (firstn (position x l) l)) 0)
  :hints (("Goal" :induct (position x l))))

(defthm count-equal-make-list-ac
  (equal (count-equal x (make-list-ac n val ac))
	 (if (equal x val)
	     (+ (nfix n) (count-equal x ac))
	     (count-equal x ac)))
  :hints (("Goal" :induct (make-list-ac n val ac))))

(defthm count-equal-nthcdr
  (<= (count-equal x (nthcdr n l))
      (count-equal x l))
  :rule-classes :linear)

(defthm count-equal-nthcdr-position
  (equal (count-equal x (nthcdr (position x l) l))
	 (count-equal x l))
  :hints (("Goal" :induct (position x l))))

(defthm count-equal-put-nth
  (equal (count-equal x (put-nth n v l))
	 (if (< (nfix n) (len l))
	     (if (equal x v)
		 (if (equal x (nth n l))
		     (count-equal x l)
		     (1+ (count-equal x l)))
		 (if (equal x (nth n l))
		     (1- (count-equal x l))
		     (count-equal x l)))
	     (count-equal x l)))
  :hints (("Goal" :induct (put-nth n v l)
	   :in-theory (enable nth))))

(defthm count-equal-put-nth-fill
  (equal (count-equal x (put-nth-fill n v l filler))
	 (if (< (nfix n) (len l))
	     (if (equal x v)
		 (if (equal x (nth n l))
		     (count-equal x l)
		     (1+ (count-equal x l)))
		 (if (equal x (nth n l))
		     (1- (count-equal x l))
		     (count-equal x l)))
	     (if (equal x v)
		 (if (equal x filler)
		     (+ (count-equal x l)
			(1+ (-(nfix n) (len l))))
		     (1+ (count-equal x l)))			      
		 (if (equal x filler)
		     (+ (count-equal x l)
			(- (nfix n) (len l)))
		     (count-equal x l)))))
  :hints (("Goal" :induct (put-nth-fill n v l filler)
	   :in-theory (enable nth))))

(defthm count-equal-remove-equal
  (equal (count-equal x (remove-equal y l))
	 (if (equal x y)
	     0
	     (count-equal x l)))
  :hints (("Goal" :induct (remove-equal y l))))
	   
(defthm count-equal-revappend 
  (equal (count-equal x (revappend a b))
	 (+ (count-equal x a)
	    (count-equal x b)))
  :hints (("Goal" :induct (revappend a b))))

(defthm count-equal-reverse
  (equal (count-equal x (reverse l))
	 (count-equal x l)))


(defthm position-iff-member-equal
  (iff (position x l)
       (member-equal x l))
  :rule-classes
  ((:rewrite)
   (:type-prescription
    :corollary
    (implies
     (member-equal x l)
     (and (integerp (position x l))
	  (<= 0 (position x l)))))
   (:type-prescription
    :corollary
    (implies
     (not (member-equal x l))
     (equal (position x l) nil)))))

(defthm equal-position-n-false
  (implies (and (member-equal x l)
		(not (equal (nth n l) x)))
	   (equal (equal (position x l) n) nil))
  :hints (("Goal" :in-theory (enable nth))))

(defthm lessp-position-len 
  (implies (member-equal x l)
	   (< (position x l)
	      (len l)))
  :rule-classes (:rewrite :linear))


(defthm position-append 
  (equal (position x (append a b))
	 (if (member-equal x a)
	     (position x a)
	   (if (member-equal x b)
	       (+ (len a) (position x b))
	     nil))))


(defthm member-equal-append 
  (iff (member-equal x (append a b))
       (or (member-equal x a)
	   (member-equal x b))))

(defthm member-equal-firstn-position
  (not (member-equal x (firstn (position x l) l)))
  :hints (("Goal" :induct (position x l)
	   :do-not '(generalize))))

(defthm member-equal-make-list-ac-instance
  (implies (not (zp n))
	   (member-equal x (make-list-ac n x ac)))
  :hints (("Goal" :induct (make-list-ac n x ac))))

(defthm member-equal-make-list-ac
  (implies (not (member-equal x ac))
	   (iff (member-equal x (make-list-ac n y ac))
		(and (not (zp n))
		     (equal x y)))))

(defthm member-equal-car-last
  (implies (consp l)
	   (member-equal (car (last l)) l)))

(defthm member-equal-nth
  (implies (< (nfix n) (len l))
	   (member-equal (nth n l) l))
  :hints (("Goal" :in-theory (enable nth))))

(defthm member-equal-put-nth
  (implies (< (nfix n) (len l))
	   (member-equal x (put-nth n x l))))

(defthm member-equal-put-nth-fill
  (member-equal x (put-nth-fill n x l filler)))

(defthm member-equal-remove-equal
  (iff (member-equal x (remove-equal y l))
       (if (equal x y)
	   nil
	   (member-equal x l))))

(defthm member-equal-revappend
  (iff (member-equal x (revappend a b))
       (or (member-equal x a)
	   (member-equal x b)))
  :hints
  (("Goal"
    :induct (revappend a b))))

(defthm member-equal-reverse
  (iff (member-equal x (reverse l))
       (member-equal x l)))
	   

(defthm append-right-id
  (implies (force (true-listp a))
	   (equal (append a nil)
		  a)))

(defthm associativity-of-append
  (equal (append (append a b) c)
	 (append a (append b c))))
 
(defthm append-firstn-nthcdr
  (implies (true-listp l)
	   (equal (append (firstn n l) (nthcdr n l)) l)))

(defthm firstn-with-large-index
  (implies (and (<= (len l) (nfix n))
		(true-listp l))
	   (equal (firstn n l) l)))
		
(defthm firstn-append
  (equal (firstn n (append a b))
	 (if (<= (nfix n) (len a))
	     (firstn n a)
	     (append a (firstn (- n (len a)) b)))))

(defthm firstn-firstn 
  (equal (firstn i (firstn j l))
	 (if (< (nfix i) (nfix j))
	     (firstn i l)
	     (firstn j l))))

(defthm firstn-make-list-ac
  (equal (firstn i (make-list-ac j v ac))
	 (if (< (nfix i) (nfix j))
	     (make-list-ac i v nil)
	     (make-list-ac j v (firstn (- i (nfix j)) ac)))))

(local 
 (defthm firstn-cons 
   (equal (firstn n (cons a l))
	  (if (zp n)
	      nil
	      (cons a (firstn (1- n) l))))))

(defthm firstn-put-nth 
  (equal (firstn i (put-nth j v l))
	 (if (<= (nfix i) (nfix j))
	     (firstn i l)
	     (put-nth j v (firstn i l))))
  :rule-classes
  ((:rewrite :corollary
	     (implies (<= (nfix i) (nfix j))
		      (equal (firstn i (put-nth j v l))
			     (firstn i l))))))

(defthm firstn-put-nth-fill 
  (implies (and (<= (nfix i) (nfix j))
		(<= (nfix i) (len l)))
	   (equal (firstn i (put-nth-fill j v l filler))
		  (firstn i l))))

; ---------- NTH ----------

(defthm nth-with-large-index 
  (implies (and (<= (len l) (nfix n))
		(true-listp l))
	   (equal (nth n l) nil)))

;; NTH-BINARY-APPEND and NTH-APPEND had (LENGTH A) instead of (LEN A)
;; in it.  This worked under the hypothesis that A was TRUE-LISTP as
;; was required in ACL2 with guards.  (LEN A) is what we need though.
;; A. Flatau 7-Nov-1994
(defthm nth-binary-append
  (equal (nth n (binary-append a b))
	 (if (< (nfix n) (len a))
	     (nth n a)
	     (nth (- (nfix n) (len a)) b))))

(defthm nth-append
  (equal (nth n (append a b))
	 (if (< (nfix n) (len a))
	     (nth n a)
	     (nth (- (nfix n) (len a)) b))))

(defthm nth-take
  (implies (< (nfix i) (nfix n))
	   (equal (nth i (take n l))
		  (nth i l))))

(defthm nth-butlast
  (implies (and (< (nfix i) (- (len l) n))
		(integerp n)
		(listp n))
	   (equal (nth i (butlast l n))
		  (nth i l)))
  :hints (("Goal" :do-not-induct t
	   :in-theory (enable butlast))))
  
(defthm nth-firstn
  (equal (nth i (firstn j l))
	 (if (< (nfix i) (nfix j))
	     (nth i l)
	     nil))
  :hints (("Goal" :induct (and (nth i (firstn j l))
			       (nth i l)
			       (firstn j l)))))

(defthm nth-make-list-ac
  (equal (nth i (make-list-ac j val ac))
	 (if (< (nfix i) (nfix j))
	     val
	     (nth (- i (nfix j)) ac)))
  :hints (("Goal" :induct (make-list-ac j val ac))))

(defthm nth-nthcdr
  (equal (nth i (nthcdr j l))
	 (nth (+ (nfix j) (nfix i)) l)))

(defthm nth-position
  (equal (nth (position x l) l)
	 (if (member-equal x l)
	     x
	   (car l))))

(defthm nth-put-nth
  (equal (nth i (put-nth j v l))
	 (if (< (nfix i) (len l))
	     (if (equal (nfix i) (nfix j))
		 v
		 (nth i l))
	     (nth i l))))

(local
 (defun double-natural-induction (i j)
   (declare (xargs :guard (and (integerp i)
			       (<= 0 i)
			       (integerp j)
			       (<= 0 j))))
   (if (zp j) t
       (if (zp i) t
	   (double-natural-induction (1- i) (1- j))))))

(local (defthm open-nth-crock
	 (implies (< 0 (nfix n))
		  (equal (nth n (Cons x l))
			 (nth (1- n) l)))))

(local (defthm nth-put-nth-fill-crock-plus-fact
	 (implies (and (integerp n)
		       (integerp m))
		  (equal (equal 0 (+ (- n) m))
			 (equal n m)))))

(defthm nth-put-nth-fill
  (implies (true-listp l)
	   (equal (nth i (put-nth-fill j v l filler))
		  (if (equal (nfix i) (nfix j))
		      v
		      (if (< (nfix i) (len l))
			  (nth i l)
			  (if (< (nfix i) (nfix j))
			      filler
			      nil)))))
  :hints (("Goal" :induct (and (put-nth-fill j v l filler)
			       (double-natural-induction i j)))))

(defthm nth-nth-seg
  (equal (nth n (nth-seg i j l))
	 (if (< (nfix n) (nfix j))
	     (nth (+ (nfix i) (nfix n)) l)
	     nil))
  :hints (("Goal" :do-not-induct t
	   :in-theory (union-theories
		       (set-difference-theories
			(current-theory :here)
			'((:definition nth-seg)))
		       '(nth-seg-non-recursive-defun)))))

(local
 (defthm nth-put-seg-helper-1
   (implies (and (true-listp l)
		 (integerp i)
		 (integerp n)
		 (< (len l) i)
		 (not (< n (len l))))
	    (not (nth (+ i n (- (len l))) l)))
   :hints (("Goal" :use (:instance nth-with-large-index
				   (l l) (n (+ i n (- (len l)))))
	    :in-theory (disable nth-with-large-index)))))

(defthm nth-put-seg
  (implies (and (true-listp l)
		(integerp i)
		(not (< i 0)))
	   (equal (nth n (put-seg i seg l))
		  (if (< (nfix n) i)
		      (nth n l)
		      (if (< (nfix n)
			     (+ i (min (max (- (len l) i) 0) (len seg))))
			  (nth (- (nfix n) i) seg)
			  (nth n l)))))
  :hints (("Goal" :do-not-induct t
	   :in-theory (union-theories
		       (set-difference-theories
			(current-theory :here)
			'((:definition put-seg)))
		       '(put-seg-non-recursive-defun)))))

(defthm nth-put-seg-fill
  (implies (and (integerp i)
		(<= 0 i)
		(true-listp l)
		(true-listp seg))
	   (equal (nth n (put-seg-fill i seg l filler))
		  (if (< (nfix n) i)
		      (if (<= i (len l))
			  (nth n l)
			  (if (< (nfix n) (len l))
			      (nth n l)
			      filler))
		      (if (< (nfix n) (+ i (len seg)))
			  (nth (- (nfix n) i) seg)
			  (nth n l)))))
  :hints (("Goal" :do-not-induct t
	   :in-theory (union-theories
		       (set-difference-theories
			(current-theory :here)
			'((:definition put-seg-fill)))
		       '(put-seg-fill-non-recursive-defun)))))
			  
; ---------- NTHCDR ----------

(defthm nthcdr-with-large-index
  (implies (and (<= (len l) (nfix n))
		(true-listp l))
	   (equal (nthcdr n l) nil)))

(defthm nthcdr-append 
  (equal (nthcdr n (append a b))
	 (if (<= (nfix n) (len a))
	     (append (nthcdr n a) b)
	     (nthcdr (- n (len a)) b))))

(defthm nthcdr-firstn
  (equal (nthcdr i (firstn j l))
	 (if (<= (nfix j) (nfix i))
	     nil
	     (firstn (- (nfix j) (nfix i)) (nthcdr i l))))
  :hints (("Goal" :induct t)))

(defthm nthcdr-make-list-ac
  (equal (nthcdr i (make-list-ac j v ac))
	 (if (<= (nfix j) (nfix i))
	     (nthcdr (- (nfix i) (nfix j)) ac)
	     (make-list-ac (- (nfix j) (nfix i)) v ac)))
  :hints (("Goal" :induct t
	   :expand (make-list-ac (+ (- i) j) v ac))))

(defthm nthcdr-nthcdr
  (equal (nthcdr i (nthcdr j l))
	 (nthcdr (+ (nfix j) (nfix i)) l)))

(defthm nthcdr-nth-seg
  (equal (nthcdr n (nth-seg i j l))
	 (nth-seg (+ (nfix i) (nfix n)) (max (- (nfix j) (nfix n)) 0) l))
  :hints (("Goal" :do-not-induct t
	   :in-theory (union-theories
		       (set-difference-theories
			(current-theory :here)
			'((:definition nth-seg)))
		       '(nth-seg-non-recursive-defun)))))

(defthm nthcdr-put-nth
  (equal (nthcdr i (put-nth j v l))
	 (if (< (nfix j) (nfix i))
	     (nthcdr i l)
	     (put-nth (- (nfix j) (nfix i)) v (nthcdr i l))))
  :hints (("Goal" :induct t))
  :rule-classes ((:rewrite :corollary
			   (implies (< (nfix j) (nfix i))
				    (equal (nthcdr i (put-nth j v l))
					   (nthcdr i l))))))

; ---------- PUT-NTH ----------

(defthm put-nth-with-large-index 
  (implies (and (true-listp l)
		(<= (len l) (nfix n)))
	   (equal (put-nth n v l) l)))

(defthm put-nth-append
  (equal (put-nth n v (append a b))
	 (if (< (nfix n) (len a))
	     (append (put-nth n v a) b)
	     (append a (put-nth (- n (len a)) v b)))))

;; Moved FIRSTN-CONS up further in the file, where it was needed.
;; A. Flatau  7-Nov-1994

(local
 (defun put-nth-firstn-induction (i j l)
   (declare (xargs :guard (and (true-listp l)
			       (integerp i) (<= 0 i)
			       (integerp j) (<= 0 j))))
   (cond ((endp (firstn j l)) 0)
	 ((zp i) 0)
	 (t (put-nth-firstn-induction (1- i) (1- j) (cdr l))))))

(defthm put-nth-firstn
  (equal (put-nth i v (firstn j l))
	 (if (< (nfix i) (nfix j))
	     (firstn j (put-nth i v l))
	     (firstn j l)))
  :hints (("Goal" :induct (put-nth-firstn-induction i j l))))

(defthm put-nth-nthcdr
  (equal (put-nth i v (nthcdr j l))
	 (nthcdr j (put-nth (+ (nfix j) (nfix i)) v l)))
  :hints (("Goal" :induct t)))
 
(defthm put-nth-nth 
  (implies (true-listp l)
	   (equal (put-nth i (nth i l) l) l)))
 
(defthm put-nth-put-nth
  (equal (put-nth j b (put-nth i a l))
	 (if (= (nfix i) (nfix j))
	     (put-nth j b l)
	     (put-nth i a (put-nth j b l))))
  :rule-classes 
  ((:rewrite :corollary
	     (implies (= (nfix i) (nfix j))
		      (equal (put-nth j b (put-nth i a l))
			     (put-nth j b l))))
   (:rewrite :corollary
	     (implies (< (nfix i) (nfix j))
		      (equal (put-nth j b (put-nth i a l))
			     (put-nth i a (put-nth j b l)))))))

; ---------- REMOVE-EQUAL ----------

(defthm remove-equal-non-member-equal 
  (implies (and (true-listp l)
		(not (member-equal x l)))
	   (equal (remove-equal x l) l)))

(defthm remove-equal-append
  (equal (remove-equal x (append a b))
	 (append (remove-equal x a) (remove-equal x b))))

(defthm remove-equal-remove-equal
  (equal (remove-equal y (remove-equal x l))
	 (if (equal x y)
	     (remove-equal x l)
	     (remove-equal x (remove-equal y l))))
  :rule-classes 
  ((:rewrite :corollary
	     (equal (remove-equal x (remove-equal x l))
		    (remove-equal x l)))))

; ---------- REVERSE ----------

(defthm revappend-binary-append
  (equal (revappend (binary-append a b) c)
	 (revappend b (revappend a c))))

(defthm binary-append-revappend
  (equal (binary-append (revappend a b) c)
	 (revappend a (binary-append b c))))

(defthm reverse-append
  (implies (and (listp a)
		(listp b))
	   (equal (reverse (append a b))
		  (append (reverse b) (reverse a))))
  :hints (("Goal" :do-not-induct t)))

(defthm revappend-revappend
  (equal (revappend (revappend a b) c)
	 (revappend b (binary-append a c))))

(defthm reverse-reverse
  (implies (true-listp l)
	   (equal (reverse (reverse l)) l))
  :hints (("Goal" :do-not-induct t)))

(deftheory list-defthms
  (set-difference-theories
   (current-theory :here)
   (union-theories (current-theory 'list-defthms-section)
		   '(nth-seg-non-recursive-defun
		     put-seg-non-recursive-defun
		     put-seg-fill-non-recursive-defun))))

