

Include("/usr/cli/projects/cacl2/version13/infix-prelude");

in-package("ACL2");

Include("list-defuns");

local Include((:relative :back "arithmetic" "top"));

Label list-defthms-section;

local Theorem one-plus-opener
{    1 + x
   = (IF acl2-numberp(x) THEN 1 + x
         ELSE apply('one-plus,[x])) };

local Theorem one-minus-opener
{    x - 1
   = (IF acl2-numberp(x) THEN -1 + x
         ELSE apply('one-minus,[x])) };

Theorem append-nil { true-listp(a) -> a @ nil = a };

Theorem nthcdr-nil { integerp(n) & 0 <= n -> ~(nthcdr(n,nil)) };

Theorem nth-seg-non-recursive-defun
{ nth-seg(i,j,l) = firstn(j,nthcdr(i,l)) };

Theorem put-seg-non-recursive-defun
{     true-listp(l) & integerp(i) & i >= 0
   ->    put-seg(i,seg,l)
      =    firstn(i,l)
        @ firstn(nfix(len(l) - i),seg)
        @ nthcdr(i + min(nfix(len(l) - i),len(seg)),l) }
 :hints [["Goal" :induct put-seg(i,seg,l)]];

local Theorem make-list-ac-rewrite
{ integerp(i) & 0 < i -> make-list-ac(i,v,ac) = v :> make-list-ac(i - 1,v,ac) }
 :hints [["Goal" :expand make-list-ac(1,v,ac)]];

Theorem put-seg-fill-non-recursive-defun
{     integerp(i) & 0 <= i & true-listp(l) & true-listp(seg)
   ->    put-seg-fill(i,seg,l,filler)
      =    firstn(i,l)
        @ make-list(max(i - len(l),0),:initial-element,filler)
        @ seg
        @ nthcdr(i + len(seg),l) }
 :hints [["Goal" :induct put-seg-fill(i,seg,l,filler)]
         ["Subgoal *1/2.2''" :use [:instance make-list-ac-rewrite
                                             [i := i, v := filler, ac := nil ]]]
         ["Subgoal *1/2.1''" :use [:instance make-list-ac-rewrite
                                             [i := 1, v := filler, ac := nil ]]]];

IN-THEORY(disable(nth-seg-non-recursive-defun,
                  put-seg-non-recursive-defun,
                  put-seg-fill-non-recursive-defun));

Theorem positive-len-fwd-to-consp { 0 < len(l) -> consp(l) }
 :rule-classes :forward-chaining;

Theorem len-append { len(a @ b) = len(a) + len(b) };

Theorem len-take { len(take(n,l)) = nfix(n) };

Theorem len-butlast
{     integerp(n) & listp(l)
   ->    len(butlast(l,n))
      = (IF len(l) < nfix(n) THEN 0
            ELSE len(l) - n) }
 :hints [["Goal" :in-theory enable(butlast)]];

Theorem len-firstn
{    len(firstn(n,l))
   = (IF nfix(n) <= len(l) THEN nfix(n)
         ELSE len(l)) }
 :hints [["Goal" :induct firstn(n,l)]];

Theorem len-last
{    len(last(l))
   = (IF consp(l) THEN 1
         ELSE 0) }
 :hints [["Goal" :induct last(l)]];

Theorem len-make-list-ac
{ len(make-list-ac(n,val,ac)) = nfix(n) + len(ac) }
 :hints [["Goal" :induct make-list-ac(n,val,ac)]];

Theorem len-nthcdr
{    len(nthcdr(n,l))
   = (IF nfix(n) <= len(l) THEN len(l) - nfix(n)
         ELSE 0) }
 :hints [["Goal" :induct nthcdr(n,l)]];

Theorem len-nth-seg
{ i + j <= len(l) -> len(nth-seg(i,j,l)) = min(nfix(j),max(len(l) - nfix(i),0)) }
 :hints [["Goal" :do-not-induct t
                 :in-theory                    current-theory(:here)
                                    difference '[[:definition nth-seg]]
                             union '[nth-seg-non-recursive-defun]]];

Theorem len-put-nth { len(put-nth(n,v,l)) = len(l) };

Theorem len-put-nth-fill
{ len(put-nth-fill(n,v,l,filler)) = max(len(l),1 + nfix(n)) }
 :hints [["Goal" :induct put-nth-fill(n,v,l,filler)]];

Theorem len-put-seg
{ true-listp(l) & integerp(i) & ~(i < 0) -> len(put-seg(i,seg,l)) = len(l) }
 :hints [["Goal" :do-not-induct t
                 :in-theory                    current-theory(:here)
                                    difference '[[:definition put-seg]]
                             union '[put-seg-non-recursive-defun]]];

Theorem len-put-seg-fill
{     integerp(i) & ~(i < 0) & true-listp(l) & true-listp(seg)
   -> len(put-seg-fill(i,seg,l,filler)) = max(len(l),nfix(i) + len(seg)) }
 :hints [["Goal" :do-not-induct t
                 :in-theory                    current-theory(:here)
                                    difference '[[:definition put-seg-fill]]
                             union '[put-seg-fill-non-recursive-defun]]];

Theorem len-remove-equal
{ len(remove-equal(x,l)) = len(l) - count-equal(x,l) };

Theorem len-revappend { len(revappend(a,b)) = len(a) + len(b) };

Theorem len-reverse { len(reverse(a)) = len(a) };

Theorem lessp-count-equal-len { ~(len(l) < count-equal(x,l)) }
 :rule-classes [:rewrite :linear];

Theorem equal-count-equal-zero { x ~in l -> count-equal(x,l) = 0 }
 :rule-classes [:rewrite];

Theorem count-equal-append
{ count-equal(x,a @ b) = count-equal(x,a) + count-equal(x,b) }
 :hints [["Goal" :induct a BINARY-@ b]];

Theorem count-equal-firstn
{ count-equal(x,firstn(n,l)) <= count-equal(x,l) }
 :rule-classes :linear;

Theorem count-equal-firstn-position
{ count-equal(x,firstn(position(x,l),l)) = 0 }
 :hints [["Goal" :induct position(x,l)]];

Theorem count-equal-make-list-ac
{    count-equal(x,make-list-ac(n,val,ac))
   = (IF x = val THEN nfix(n) + count-equal(x,ac)
         ELSE count-equal(x,ac)) }
 :hints [["Goal" :induct make-list-ac(n,val,ac)]];

Theorem count-equal-nthcdr
{ count-equal(x,nthcdr(n,l)) <= count-equal(x,l) }
 :rule-classes :linear;

Theorem count-equal-nthcdr-position
{ count-equal(x,nthcdr(position(x,l),l)) = count-equal(x,l) }
 :hints [["Goal" :induct position(x,l)]];

Theorem count-equal-put-nth
{    count-equal(x,put-nth(n,v,l))
   = (IF nfix(n) < len(l)
         THEN IF x = v
                 THEN IF x = l[n] THEN count-equal(x,l)
                         ELSE 1 + count-equal(x,l)
                 ELIF x = l[n] THEN count-equal(x,l) - 1
                 ELSE count-equal(x,l)
         ELSE count-equal(x,l)) }
 :hints [["Goal" :induct put-nth(n,v,l)
                 :in-theory enable(nth)]];

Theorem count-equal-put-nth-fill
{    count-equal(x,put-nth-fill(n,v,l,filler))
   = (IF nfix(n) < len(l)
         THEN IF x = v
                 THEN IF x = l[n] THEN count-equal(x,l)
                         ELSE 1 + count-equal(x,l)
                 ELIF x = l[n] THEN count-equal(x,l) - 1
                 ELSE count-equal(x,l)
         ELIF x = v
         THEN IF x = filler THEN count-equal(x,l) + 1 + (nfix(n) - len(l))
                 ELSE 1 + count-equal(x,l)
         ELIF x = filler THEN count-equal(x,l) + (nfix(n) - len(l))
         ELSE count-equal(x,l)) }
 :hints [["Goal" :induct put-nth-fill(n,v,l,filler)
                 :in-theory enable(nth)]];

Theorem count-equal-remove-equal
{    count-equal(x,remove-equal(y,l))
   = (IF x = y THEN 0
         ELSE count-equal(x,l)) }
 :hints [["Goal" :induct remove-equal(y,l)]];

Theorem count-equal-revappend
{ count-equal(x,revappend(a,b)) = count-equal(x,a) + count-equal(x,b) }
 :hints [["Goal" :induct revappend(a,b)]];

Theorem count-equal-reverse
{ count-equal(x,reverse(l)) = count-equal(x,l) };

Theorem position-iff-member-equal { position(x,l) <-> x in l }
 :rule-classes [[:rewrite  ]
                [:type-prescription  
                 :corollary     x in l
                             -> integerp(position(x,l)) & 0 <= position(x,l)]
                [:type-prescription  
                 :corollary x ~in l -> position(x,l) = nil]];

Theorem equal-position-n-false
{ x in l & l[n] ~= x -> position(x,l) = n = nil }
 :hints [["Goal" :in-theory enable(nth)]];

Theorem lessp-position-len { x in l -> position(x,l) < len(l) }
 :rule-classes [:rewrite :linear];

Theorem position-append
{    position(x,a @ b)
   = (IF x in a THEN position(x,a)
         ELIF x in b THEN len(a) + position(x,b)
         ELSE nil) };

Theorem member-equal-append { x in a @ b <-> x in a | x in b };

Theorem member-equal-firstn-position { x ~in firstn(position(x,l),l) }
 :hints [["Goal" :induct position(x,l)
                 :do-not '[generalize]]];

Theorem member-equal-make-list-ac-instance
{ ~(zp(n)) -> x in make-list-ac(n,x,ac) }
 :hints [["Goal" :induct make-list-ac(n,x,ac)]];

Theorem member-equal-make-list-ac
{ x ~in ac -> x in make-list-ac(n,y,ac) <-> ~(zp(n)) & x = y };

Theorem member-equal-car-last { consp(l) -> car(last(l)) in l };

Theorem member-equal-nth { nfix(n) < len(l) -> l[n] in l }
 :hints [["Goal" :in-theory enable(nth)]];

Theorem member-equal-put-nth { nfix(n) < len(l) -> x in put-nth(n,x,l) };

Theorem member-equal-put-nth-fill { x in put-nth-fill(n,x,l,filler) };

Theorem member-equal-remove-equal
{      x in remove-equal(y,l)
   <-> IF x = y THEN nil
         ELSE x in l };

Theorem member-equal-revappend
{ x in revappend(a,b) <-> x in a | x in b }
 :hints [["Goal" :induct revappend(a,b)]];

Theorem member-equal-reverse { x in reverse(l) <-> x in l };

Theorem append-right-id { force(true-listp(a)) -> a @ nil = a };

Theorem associativity-of-append { (a @ b) @ c = a @ b @ c };

Theorem append-firstn-nthcdr
{ true-listp(l) -> firstn(n,l) @ nthcdr(n,l) = l };

Theorem firstn-with-large-index
{ len(l) <= nfix(n) & true-listp(l) -> firstn(n,l) = l };

Theorem firstn-append
{    firstn(n,a @ b)
   = (IF nfix(n) <= len(a) THEN firstn(n,a)
         ELSE a @ firstn(n - len(a),b)) };

Theorem firstn-firstn
{    firstn(i,firstn(j,l))
   = (IF nfix(i) < nfix(j) THEN firstn(i,l)
         ELSE firstn(j,l)) };

Theorem firstn-make-list-ac
{    firstn(i,make-list-ac(j,v,ac))
   = (IF nfix(i) < nfix(j) THEN make-list-ac(i,v,nil)
         ELSE make-list-ac(j,v,firstn(i - nfix(j),ac))) };

local Theorem firstn-cons
{    firstn(n,a :> l)
   = (IF zp(n) THEN nil
         ELSE a :> firstn(n - 1,l)) };

Theorem firstn-put-nth
{    firstn(i,put-nth(j,v,l))
   = (IF nfix(i) <= nfix(j) THEN firstn(i,l)
         ELSE put-nth(j,v,firstn(i,l))) }
 :rule-classes [[:rewrite  
                 :corollary     nfix(i) <= nfix(j)
                             -> firstn(i,put-nth(j,v,l)) = firstn(i,l)]
               ];

Theorem firstn-put-nth-fill
{     nfix(i) <= nfix(j) & nfix(i) <= len(l)
   -> firstn(i,put-nth-fill(j,v,l,filler)) = firstn(i,l) };

Theorem nth-with-large-index
{ len(l) <= nfix(n) & true-listp(l) -> l[n] = nil };

Theorem nth-binary-append
{    a BINARY-@ b[n]
   = (IF nfix(n) < len(a) THEN a[n]
         ELSE b[nfix(n) - len(a)]) };

Theorem nth-append
{    a @ b[n]
   = (IF nfix(n) < len(a) THEN a[n]
         ELSE b[nfix(n) - len(a)]) };

Theorem nth-take { nfix(i) < nfix(n) -> take(n,l)[i] = l[i] };

Theorem nth-butlast
{ nfix(i) < len(l) - n & integerp(n) & listp(n) -> butlast(l,n)[i] = l[i] }
 :hints [["Goal" :do-not-induct t
                 :in-theory enable(butlast)]];

Theorem nth-firstn
{    firstn(j,l)[i]
   = (IF nfix(i) < nfix(j) THEN l[i]
         ELSE nil) }
 :hints [["Goal" :induct firstn(j,l)[i] & l[i] & firstn(j,l)]];

Theorem nth-make-list-ac
{    make-list-ac(j,val,ac)[i]
   = (IF nfix(i) < nfix(j) THEN val
         ELSE ac[i - nfix(j)]) }
 :hints [["Goal" :induct make-list-ac(j,val,ac)]];

Theorem nth-nthcdr { nthcdr(j,l)[i] = l[nfix(j) + nfix(i)] };

Theorem nth-position
{    l[position(x,l)]
   = (IF x in l THEN x
         ELSE car(l)) };

Theorem nth-put-nth
{    put-nth(j,v,l)[i]
   = (IF nfix(i) < len(l)
         THEN IF nfix(i) = nfix(j) THEN v
                 ELSE l[i]
         ELSE l[i]) };

local Function double-natural-induction(i,j)
{ IF zp(j) THEN t
     ELIF zp(i) THEN t
     ELSE double-natural-induction(i - 1,j - 1) }
 :guard integerp(i) & 0 <= i & integerp(j) & 0 <= j;

local Theorem open-nth-crock { 0 < nfix(n) -> x :> l[n] = l[n - 1] };

local Theorem nth-put-nth-fill-crock-plus-fact
{ integerp(n) & integerp(m) -> 0 = -n + m = (n = m) };

Theorem nth-put-nth-fill
{     true-listp(l)
   ->    put-nth-fill(j,v,l,filler)[i]
      = (IF nfix(i) = nfix(j) THEN v
            ELIF nfix(i) < len(l) THEN l[i]
            ELIF nfix(i) < nfix(j) THEN filler
            ELSE nil) }
 :hints [["Goal" :induct    put-nth-fill(j,v,l,filler)
                          & double-natural-induction(i,j)]];

Theorem nth-nth-seg
{    nth-seg(i,j,l)[n]
   = (IF nfix(n) < nfix(j) THEN l[nfix(i) + nfix(n)]
         ELSE nil) }
 :hints [["Goal" :do-not-induct t
                 :in-theory                    current-theory(:here)
                                    difference '[[:definition nth-seg]]
                             union '[nth-seg-non-recursive-defun]]];

local Theorem nth-put-seg-helper-1
{     true-listp(l) & integerp(i) & integerp(n) & len(l) < i & ~(n < len(l))
   -> ~(l[i + n + -(len(l))]) }
 :hints [["Goal" :use [:instance nth-with-large-index
                                 [l := l, n := i + n + -(len(l)) ]]
                 :in-theory disable(nth-with-large-index)]];

Theorem nth-put-seg
{     true-listp(l) & integerp(i) & ~(i < 0)
   ->    put-seg(i,seg,l)[n]
      = (IF nfix(n) < i THEN l[n]
            ELIF nfix(n) < i + min(max(len(l) - i,0),len(seg))
            THEN seg[nfix(n) - i]
            ELSE l[n]) }
 :hints [["Goal" :do-not-induct t
                 :in-theory                    current-theory(:here)
                                    difference '[[:definition put-seg]]
                             union '[put-seg-non-recursive-defun]]];

Theorem nth-put-seg-fill
{     integerp(i) & 0 <= i & true-listp(l) & true-listp(seg)
   ->    put-seg-fill(i,seg,l,filler)[n]
      = (IF nfix(n) < i
            THEN IF i <= len(l) THEN l[n]
                    ELIF nfix(n) < len(l) THEN l[n]
                    ELSE filler
            ELIF nfix(n) < i + len(seg) THEN seg[nfix(n) - i]
            ELSE l[n]) }
 :hints [["Goal" :do-not-induct t
                 :in-theory                    current-theory(:here)
                                    difference '[[:definition put-seg-fill]]
                             union '[put-seg-fill-non-recursive-defun]]];

Theorem nthcdr-with-large-index
{ len(l) <= nfix(n) & true-listp(l) -> nthcdr(n,l) = nil };

Theorem nthcdr-append
{    nthcdr(n,a @ b)
   = (IF nfix(n) <= len(a) THEN nthcdr(n,a) @ b
         ELSE nthcdr(n - len(a),b)) };

Theorem nthcdr-firstn
{    nthcdr(i,firstn(j,l))
   = (IF nfix(j) <= nfix(i) THEN nil
         ELSE firstn(nfix(j) - nfix(i),nthcdr(i,l))) }
 :hints [["Goal" :induct t]];

Theorem nthcdr-make-list-ac
{    nthcdr(i,make-list-ac(j,v,ac))
   = (IF nfix(j) <= nfix(i) THEN nthcdr(nfix(i) - nfix(j),ac)
         ELSE make-list-ac(nfix(j) - nfix(i),v,ac)) }
 :hints [["Goal" :induct t
                 :expand make-list-ac(-i + j,v,ac)]];

Theorem nthcdr-nthcdr
{ nthcdr(i,nthcdr(j,l)) = nthcdr(nfix(j) + nfix(i),l) };

Theorem nthcdr-nth-seg
{    nthcdr(n,nth-seg(i,j,l))
   = nth-seg(nfix(i) + nfix(n),max(nfix(j) - nfix(n),0),l) }
 :hints [["Goal" :do-not-induct t
                 :in-theory                    current-theory(:here)
                                    difference '[[:definition nth-seg]]
                             union '[nth-seg-non-recursive-defun]]];

Theorem nthcdr-put-nth
{    nthcdr(i,put-nth(j,v,l))
   = (IF nfix(j) < nfix(i) THEN nthcdr(i,l)
         ELSE put-nth(nfix(j) - nfix(i),v,nthcdr(i,l))) }
 :rule-classes [[:rewrite  
                 :corollary     nfix(j) < nfix(i)
                             -> nthcdr(i,put-nth(j,v,l)) = nthcdr(i,l)]
               ]
 :hints [["Goal" :induct t]];

Theorem put-nth-with-large-index
{ true-listp(l) & len(l) <= nfix(n) -> put-nth(n,v,l) = l };

Theorem put-nth-append
{    put-nth(n,v,a @ b)
   = (IF nfix(n) < len(a) THEN put-nth(n,v,a) @ b
         ELSE a @ put-nth(n - len(a),v,b)) };

local Function put-nth-firstn-induction(i,j,l)
{ IF endp(firstn(j,l)) THEN 0
     ELIF zp(i) THEN 0
     ELSE put-nth-firstn-induction(i - 1,j - 1,cdr(l)) }
 :guard true-listp(l) & integerp(i) & 0 <= i & integerp(j) & 0 <= j;

Theorem put-nth-firstn
{    put-nth(i,v,firstn(j,l))
   = (IF nfix(i) < nfix(j) THEN firstn(j,put-nth(i,v,l))
         ELSE firstn(j,l)) }
 :hints [["Goal" :induct put-nth-firstn-induction(i,j,l)]];

Theorem put-nth-nthcdr
{ put-nth(i,v,nthcdr(j,l)) = nthcdr(j,put-nth(nfix(j) + nfix(i),v,l)) }
 :hints [["Goal" :induct t]];

Theorem put-nth-nth { true-listp(l) -> put-nth(i,l[i],l) = l };

Theorem put-nth-put-nth
{    put-nth(j,b,put-nth(i,a,l))
   = (IF nfix(i) == nfix(j) THEN put-nth(j,b,l)
         ELSE put-nth(i,a,put-nth(j,b,l))) }
 :rule-classes [[:rewrite  
                 :corollary     nfix(i) == nfix(j)
                             -> put-nth(j,b,put-nth(i,a,l)) = put-nth(j,b,l)]
                [:rewrite  
                 :corollary     nfix(i) < nfix(j)
                             ->    put-nth(j,b,put-nth(i,a,l))
                                = put-nth(i,a,put-nth(j,b,l))]];

Theorem remove-equal-non-member-equal
{ true-listp(l) & x ~in l -> remove-equal(x,l) = l };

Theorem remove-equal-append
{ remove-equal(x,a @ b) = remove-equal(x,a) @ remove-equal(x,b) };

Theorem remove-equal-remove-equal
{    remove-equal(y,remove-equal(x,l))
   = (IF x = y THEN remove-equal(x,l)
         ELSE remove-equal(x,remove-equal(y,l))) }
 :rule-classes [[:rewrite  
                 :corollary    remove-equal(x,remove-equal(x,l))
                             = remove-equal(x,l)]
               ];

Theorem revappend-binary-append
{ revappend(a BINARY-@ b,c) = revappend(b,revappend(a,c)) };

Theorem binary-append-revappend
{ revappend(a,b) BINARY-@ c = revappend(a,b BINARY-@ c) };

Theorem reverse-append
{ listp(a) & listp(b) -> reverse(a @ b) = reverse(b) @ reverse(a) }
 :hints [["Goal" :do-not-induct t]];

Theorem revappend-revappend
{ revappend(revappend(a,b),c) = revappend(b,a BINARY-@ c) };

Theorem reverse-reverse { true-listp(l) -> reverse(reverse(l)) = l }
 :hints [["Goal" :do-not-induct t]];

Theory list-defthms :=             current-theory(:here)
                        difference (       current-theory('list-defthms-section)
                                     union '[nth-seg-non-recursive-defun
                                            put-seg-non-recursive-defun
                                            put-seg-fill-non-recursive-defun]);


