

in-package("ACL2");

Label list-defuns-section;

/* The number of occurrences of an item in a list, using EQUAL as the test. */

Function count-equal(x, l)
{ if endp(l) then 0
      elif x = car(l) then 1 + count-equal(x, cdr(l))
      else count-equal(x, cdr(l)) }
 :guard true-listp(l);

Function firstn(n, l)
{ if endp(l) then nil
      elif zp(n) then nil
      else car(l) :> firstn(n - 1, cdr(l)) }
 :measure nfix(n)
 :guard integerp(n) & (0 <= n & true-listp(l));

Function nth-seg(i, j, l)
{ if endp(l) then nil
      elif zp(i)
      then if zp(j) then nil
               else car(l) :> nth-seg(i, j - 1, cdr(l))
      else nth-seg(i - 1, j, cdr(l)) }
 :guard integerp(i) & (0 <= i & (integerp(j) & (0 <= j & true-listp(l))));

Function position(x, l)
{ if endp(l) then nil
      elif x = car(l) then 0
      else { n := position(x, cdr(l));
             n & 1 + n} }
 :guard true-listp(l);

Function put-nth(n, v, l)
{ if endp(l) then nil
      elif zp(n) then v :> cdr(l)
      else car(l) :> put-nth(n - 1, v, cdr(l)) }
 :guard integerp(n) & (0 <= n & true-listp(l));

Function put-nth-fill(n, v, l, filler)
{ if endp(l)
      then if zp(n) then [v]
               else filler :> put-nth-fill(n - 1, v, l, filler)
      elif zp(n) then v :> cdr(l)
      else car(l) :> put-nth-fill(n - 1, v, cdr(l), filler) }
 :guard integerp(n) & (0 <= n & true-listp(l));

Function put-seg(i, seg, l)
{ if endp(l) then nil
      elif zp(i)
      then if endp(seg) then l
               else car(seg) :> put-seg(i, cdr(seg), cdr(l))
      else car(l) :> put-seg(i - 1, seg, cdr(l)) }
 :guard integerp(i) & (0 <= i & (true-listp(seg) & true-listp(l)));

Function put-seg-fill(i, seg, l, filler)
{ if endp(l)
      then if zp(i) then seg
               else filler :> put-seg-fill(i - 1, seg, l, filler)
      elif zp(i)
      then if endp(seg) then l
               else car(seg) :> put-seg-fill(i, cdr(seg), cdr(l), filler)
      else car(l) :> put-seg-fill(i - 1, seg, cdr(l), filler) }
 :measure 1 + nfix(i) :> len(l)
 :guard integerp(i) & (0 <= i & (true-listp(seg) & true-listp(l)));

Function remove-equal(x, l)
{ if endp(l) then nil
      elif x = car(l) then remove-equal(x, cdr(l))
      else car(l) :> remove-equal(x, cdr(l)) }
 :guard true-listp(l);

Function take(n, l)
{ if zp(n) then nil
      else car(l) :> take(n - 1, cdr(l)) }
 :guard integerp(n) & (~(n < 0) & true-listp(l));

Function butlast(lst, n)
{ lng := len(lst);
  if lng <= n then nil
      else take(lng - n, lst)}
 :guard true-listp(lst) & (integerp(n) & 0 <= n);

Theorem true-listp-append { true-listp(a @ b) = true-listp(b) }
 :rule-classes [[:rewrite  ] [:type-prescription  
 :corollary true-listp(b) -> true-listp(a @ b)]];

     Theorem true-listp-last { true-listp(last(l)) = true-listp(l) }
 :rule-classes [[:rewrite  ] [:type-prescription  
 :corollary force(true-listp(l)) -> true-listp(last(l))]];

 Theorem true-listp-make-list-ac { true-listp(make-list-ac(n, val, ac)) = true-listp(ac) }
 :rule-classes [[:rewrite  ] [:type-prescription  
 :corollary true-listp(ac) -> true-listp(make-list-ac(n, val, ac))]];

  Theorem true-listp-put-nth { true-listp(l) -> true-listp(put-nth(n, v, l)) }
 :rule-classes [:rewrite :type-prescription];

Theorem true-listp-put-nth-fill
{ true-listp(l) -> true-listp(put-nth-fill(n, v, l, filler)) }
 :rule-classes [:rewrite :type-prescription];

Theorem true-listp-put-seg { true-listp(l) -> true-listp(put-seg(i, seg, l)) }
 :rule-classes [:rewrite :type-prescription];

Theorem true-listp-put-seg-fill
{ true-listp(seg) & true-listp(l) -> true-listp(put-seg-fill(i, seg, l, filler)) }
 :rule-classes [:rewrite :type-prescription];

Theorem true-listp-revappend { true-listp(revappend(a, b)) = true-listp(b) }
 :rule-classes [[:rewrite  ] [:type-prescription  
 :corollary true-listp(b) -> true-listp(revappend(a, b))]];

Theorem true-listp-reverse { listp(a) -> true-listp(reverse(a)) }
 :rule-classes [:type-prescription];

/* ---------- CONSP ---------- */
Theorem consp-append { consp(a @ b) = (consp(a) | consp(b)) }
 :rule-classes [[:rewrite  ] [:type-prescription  
 :corollary consp(a) -> consp(a @ b)] [:type-prescription  
 :corollary consp(b) -> consp(a @ b)]];

Theorem consp-take { consp(take(n, l)) = ~(zp(n)) }
 :rule-classes [[:rewrite  ] [:type-prescription  
 :corollary integerp(n) & n > 0 -> consp(take(n, l))]]
 :hints [["Goal" :induct take(n, l)]];

Theorem consp-butlast { listp(l) & integerp(n) -> consp(butlast(l, n)) = n < len(l) }
 :rule-classes [[:rewrite  ] [:type-prescription  
 :corollary n < len(l) & (listp(l) & integerp(n)) -> consp(butlast(l, n))]];

Theorem consp-firstn { consp(firstn(n, l)) = (~(zp(n)) & consp(l)) }
 :rule-classes [[:rewrite  ] [:type-prescription  
 :corollary ~(zp(n)) & consp(l) -> consp(firstn(n, l))]]
 :hints [["Goal" :expand firstn(n, l)]];

Theorem consp-last { consp(last(l)) = consp(l) }
 :rule-classes [[:rewrite  ] [:type-prescription  
 :corollary consp(l) -> consp(last(l))]]
 :hints [["Goal" :in-theory disable(true-listp-last)]];

Theorem consp-make-list-ac { consp(make-list-ac(n, val, ac)) = (~(zp(n)) | consp(ac)) }
 :rule-classes [[:rewrite  ] [:type-prescription  
 :corollary ~(zp(n)) & consp(ac) -> consp(make-list-ac(n, val, ac))]]
 :hints [["Goal" :expand make-list-ac(n, val, ac)]];

Theorem nthcdr-nil { nthcdr(n, nil) = nil };

Theorem consp-nthcdr { consp(nthcdr(n, l)) <-> nfix(n) < len(l) }
 :rule-classes [[:rewrite  ] [:type-prescription  
 :corollary nfix(n) < len(l) -> consp(nthcdr(n, l))]];

Theorem consp-nth-seg { i + j <= len(l) -> consp(nth-seg(i, j, l)) = (consp(l) & ~(zp(j))) }
 :rule-classes [[:rewrite  ] [:type-prescription  
 :corollary consp(l) & (~(zp(j)) & i + j <= len(l)) -> consp(nth-seg(i, j, l))]];

Theorem consp-put-nth { consp(put-nth(n, v, l)) = consp(l) }
 :rule-classes [[:rewrite  ] [:type-prescription  
 :corollary consp(l) -> consp(put-nth(n, v, l))]];

Theorem consp-put-seg { consp(put-seg(i, seg, l)) = consp(l) }
 :rule-classes [[:rewrite  ] [:type-prescription  
 :corollary consp(l) -> consp(put-seg(i, seg, l))]];

Theorem consp-put-seg-fill
{ consp(put-seg-fill(i, seg, l, filler)) = (consp(l) | (consp(seg) | ~(zp(i)))) }
 :rule-classes [[:rewrite  ] [:type-prescription  
 :corollary consp(l) | (consp(seg) | ~(zp(i))) -> consp(put-seg-fill(i, seg, l, filler))]]
 :hints [["Goal" :induct put-seg-fill(i, seg, l, filler)]];

Theorem consp-revappend { consp(revappend(a, b)) = (consp(a) | consp(b)) }
 :rule-classes [[:rewrite  ] [:type-prescription  
 :corollary consp(a) -> consp(revappend(a, b))] [:type-prescription  
 :corollary consp(b) -> consp(revappend(a, b))]];

Theorem consp-reverse { consp(reverse(a)) = consp(a) }
 :rule-classes [[:rewrite  ] [:type-prescription  
 :corollary consp(a) -> consp(reverse(a))]];

Theory list-defuns := current-theory(:here) difference current-theory('list-defuns-section);

IN-THEORY(disable(butlast));


