

(in-package "ACL2")

(deflabel list-defuns-section)



(defun count-equal (x l)
  "The number of occurrences of an item in a list, using EQUAL as the test."
  (declare (xargs :guard (true-listp l)))
  (cond ((endp l) 0)
	((equal x (car l)) (1+ (count-equal x (cdr l))))
	(t (count-equal x (cdr l)))))

(defun firstn (n l)
    (declare (xargs :measure (nfix n)
		  :guard (and (integerp n) (<= 0 n)
			      (true-listp l))))
  (cond ((endp l) nil)
	((zp n) nil)
	(t (cons (car l) (firstn (1- n) (cdr l))))))

(defun nth-seg (i j l)
 
  (declare (xargs :guard (and (integerp i) (<= 0 i)
			      (integerp j) (<= 0 j)
			      (true-listp l))))
  (cond ((endp l) nil)
	((zp i)
	 (cond ((zp j) nil)
	       (t (cons (car l) (nth-seg i (1- j) (cdr l))))))
	(t (nth-seg (1- i) j (cdr l)))))

(defun position (x l)
 
  (declare (xargs :guard (true-listp l)))
  (cond
   ((endp l) nil)
   ((equal x (car l)) 0)
   (t (let ((n (position x (cdr l))))
        (and n (1+ n))))))

(defun put-nth (n v l)

  (declare (xargs :guard (and (integerp n) (<= 0 n)
			      (true-listp l))))
  (cond ((endp l) nil)
	((zp n) (cons v (cdr l)))
	(t (cons (car l) (put-nth (1- n) v (cdr l))))))

(defun put-nth-fill (n v l filler)

  (declare (xargs :guard (and (integerp n) (<= 0 n)
			      (true-listp l))))
  (cond ((endp l) 
	 (if (zp n)
	     (list v)
	     (cons filler (put-nth-fill (1- n) v l filler))))
	((zp n) (cons v (cdr l)))
	(t (cons (car l) (put-nth-fill (1- n) v (cdr l) filler)))))

(defun put-seg (i seg l)

  (declare (xargs :guard (and (integerp i)
			      (<= 0 i)
			      (true-listp seg)
			      (true-listp l))))
  (cond ((endp l) nil)
	((zp i)
	 (cond ((endp seg) l)
	       (t (cons (car seg) (put-seg i (cdr seg) (cdr l))))))
	(t (cons (car l) (put-seg (1- i) seg (cdr l))))))

(defun put-seg-fill (i seg l filler)

  (declare (xargs :guard (and (integerp i)
			      (<= 0 i)
			      (true-listp seg)
			      (true-listp l))
		  :measure (cons (1+ (nfix i)) (len l))))
  (cond ((endp l)
	 (cond ((zp i) seg)
	       (t (cons filler (put-seg-fill (1- i) seg l filler)))))
	((zp i)
	 (cond ((endp seg) l)
	       (t (cons (car seg) (put-seg-fill i (cdr seg) (cdr l) filler)))))
	(t (cons (car l) (put-seg-fill (1- i) seg (cdr l) filler)))))

(defun remove-equal (x l)

  (declare (xargs :guard (true-listp l)))
  (cond ((endp l) nil)
	((equal x (car l)) (remove-equal x (cdr l)))
	(t (cons (car l) (remove-equal x (cdr l))))))

(defun take (n l)
  (declare (xargs :guard
		   (and (integerp n)
			(not (< n 0))
			(true-listp l))))
  (if (zp n)
      nil
    (cons (car l) (take (1- n) (cdr l)))))

(defun butlast (lst n)
  (declare (xargs :guard (and (true-listp lst)
                              (integerp n)
                              (<= 0 n))))
  (let ((lng (len lst)))
    (if (<= lng n)
        nil
      (take (- lng n) lst))))


(defthm true-listp-append
  (equal (true-listp (append a b))
	 (true-listp b))
  :rule-classes
  ((:rewrite)
   (:type-prescription :corollary
		       (implies (true-listp b)
				(true-listp (append a b))))))
    
 
(defthm true-listp-last
  (equal (true-listp (last l))
	 (true-listp l))
  :rule-classes
  ((:rewrite)
   (:type-prescription :corollary
		       (implies (force (true-listp l))
				(true-listp (last l))))))
 
(defthm true-listp-make-list-ac
  (equal (true-listp (make-list-ac n val ac))
	 (true-listp ac)):rule-classes
  ((:rewrite)
   (:type-prescription :corollary
    (implies (true-listp ac)
	     (true-listp (make-list-ac n val ac))))))
 
 
(defthm true-listp-put-nth
  (implies
   (true-listp l)
   (true-listp (put-nth n v l)))
  :rule-classes (:rewrite :type-prescription))

(defthm true-listp-put-nth-fill
  (implies
   (true-listp l)
   (true-listp (put-nth-fill n v l filler)))
  :rule-classes (:rewrite :type-prescription))

(defthm true-listp-put-seg
  (implies
   (true-listp l)
   (true-listp (put-seg i seg l)))
  :rule-classes (:rewrite :type-prescription))

(defthm true-listp-put-seg-fill
  (implies
   (and (true-listp seg)
	(true-listp l))
   (true-listp (put-seg-fill i seg l filler)))
  :rule-classes (:rewrite :type-prescription))

(defthm true-listp-revappend
  (equal (true-listp (revappend a b))
	 (true-listp b))
  :rule-classes
  ((:rewrite)
   (:type-prescription :corollary
		       (implies (true-listp b)
				(true-listp (revappend a b))))))

(defthm true-listp-reverse
  (implies (listp a)
	   (true-listp (reverse a)))
  :rule-classes (:type-prescription))

; ---------- CONSP ----------

(defthm consp-append
  (equal (consp (append a b))
	 (or (consp a)
	     (consp b)))
  :rule-classes
  ((:rewrite)
   (:type-prescription :corollary
		       (implies (consp a)
				(consp (append a b))))
   (:type-prescription :corollary
		       (implies (consp b)
				(consp (append a b))))))

(defthm consp-take
  (equal (consp (take n l)) (not (zp n)))
  :hints (("Goal" :induct (take n l)))
  :rule-classes
  ((:rewrite)
   (:type-prescription :corollary
		       (implies (and (integerp n)
				     (> n 0))
				(consp (take n l))))))

(defthm consp-butlast
  (implies (and (listp l)
		(integerp n))
	   (equal (consp (butlast l n))
		  (< n (len l))))
  :rule-classes
  ((:rewrite)
   (:type-prescription :corollary
		       (implies (and (< n (len l))
				     (listp l)
				     (integerp n))
				(consp (butlast l n))))))

(defthm consp-firstn
  (equal (consp (firstn n l))
	 (and (not (zp n))
	      (consp l)))
  :hints (("Goal" :expand (firstn n l)))
  :rule-classes
  ((:rewrite)
   (:type-prescription :corollary
		       (implies (and (not (zp n))
				     (consp l))
				(consp (firstn n l))))))

(defthm consp-last
  (equal (consp (last l))
	 (consp l))
  :hints (("Goal" :in-theory (disable true-listp-last)))
  :rule-classes ((:rewrite)
		 (:type-prescription :corollary
				     (implies (consp l)
					      (consp (last l))))))

(defthm consp-make-list-ac
  (equal (consp (make-list-ac n val ac))
	 (or (not (zp n))
	     (consp ac)))
  :hints (("Goal" :expand (make-list-ac n val ac)))
  :rule-classes
  ((:rewrite)
   (:type-prescription :corollary
    (implies (and (not (zp n))
		  (consp ac))
	     (consp (make-list-ac n val ac))))))


(defthm nthcdr-nil
  (equal (nthcdr n nil) nil))

(defthm consp-nthcdr
  (iff (consp (nthcdr n l))
       (< (nfix n) (len l)))
  :rule-classes
  ((:rewrite)
   (:type-prescription :corollary
		       (implies (< (nfix n) (len l))
				(consp (nthcdr n l))))))

(defthm consp-nth-seg
  (implies (<= (+ i j) (len l))
	   (equal (consp (nth-seg i j l))
		  (and (consp l)
		       (not (zp j)))))
  :rule-classes
  ((:rewrite)
   (:type-prescription :corollary
		       (implies (and (consp l)
				     (not (zp j))
				     (<= (+ i j) (len l)))
				(consp (nth-seg i j l))))))

(defthm consp-put-nth
  (equal (consp (put-nth n v l))
	 (consp l))
  :rule-classes
  ((:rewrite)
   (:type-prescription :corollary
		       (implies (consp l)
				(consp (put-nth n v l))))))

(defthm consp-put-seg
  (equal (consp (put-seg i seg l))
	 (consp l))
  :rule-classes ((:rewrite)
   (:type-prescription :corollary
		       (implies (consp l)
				(consp (put-seg i seg l))))))

(defthm consp-put-seg-fill
  (equal (consp (put-seg-fill i seg l filler))
	 (or (consp l)
	     (consp seg)
	     (not (zp i))))
  :hints (("Goal" :induct (put-seg-fill i seg l filler)))
  :rule-classes ((:rewrite)
  (:type-prescription :corollary
		      (implies (or (consp l)
				   (consp seg)
				   (not (zp i)))
			       (consp (put-seg-fill i seg l filler))))))

(defthm consp-revappend 
  (equal (consp (revappend a b))
	 (or (consp a)
	     (consp b)))
  :rule-classes
  ((:rewrite)
   (:type-prescription :corollary
		       (implies (consp a)
				(consp (revappend a b))))
   (:type-prescription :corollary
		       (implies (consp b)
				(consp (revappend a b))))))

(defthm consp-reverse
  (equal (consp (reverse a))
	 (consp a))
  :rule-classes
  ((:rewrite)
   (:type-prescription :corollary
		       (implies (consp a)
				(consp (reverse a))))))

(deftheory list-defuns
  (set-difference-theories (current-theory :here)
			   (current-theory 'list-defuns-section)))

(in-theory (disable butlast))



