#| Here are the meta rules which cancel like terms from two +-terms
   when the two +-terms are the operands to the binary relations EQUAL
   or <.

   Also, here is a meta rule which cancels like terms from two *-terms
   when they are operands to EQUAL.

   Original rule (for +-terms, EQUAL, and RATIONALP hyps) by
   M. Kaufmann & Yuan Yu, April 1994.  Modified June 1994 (for
   +-terms, EQUAL, and Acl2-numberp hyps),(for +-terms, <, and
   RATIONALP hyps), and (for *-terms, EQUAL, and Acl2-numberp hyps) by
   John Cowles, University of Wyoming. Modification finished July
   1994. Modified August 1994 for Version 1.6.  Modified March 1995
   for Version 1.8.

   Modifications include removing duplicate hyps and extra forces
   included in the original rule.

(certify-book "meta" 0)

|#

(in-package "ACL2")

(include-book "term-lemmas" :load-compiled-file nil)

(encapsulate
 ((binop (x y) t)
  (binop-fix (x) t)
  (binop-rel (x y) t)
  (binop-id () t))

 (local (defun binop (x y)
          (+ x y)))

 (local (defun binop-fix (x)
          (fix x)))

 (local (defun binop-rel (x y)
          (equal x y)))

 (local (defun binop-id ()
          0))

 (defthm commutativity-of-binop
   (equal (binop x y)
          (binop y x)))

 (defthm commutativity2-of-binop
   (equal (binop x (binop y z))
          (binop y (binop x z))))

 (defthm associativity-of-binop
   (equal (binop (binop x y) z)
          (binop x (binop y z))))

 (defthm binop-fix-binop
   (equal (binop-fix (binop x y))
          (binop x y)))

 (defthm binop-cancel-binop-rel
   (equal (binop-rel (binop x y) (binop x z))
          (binop-rel (binop-fix y) (binop-fix z))))

 (defthm binop-binop-id
   (equal (binop (binop-id) x)
          (binop-fix x)))

 (defthm binop-fix-binop-id
   (equal (binop-fix (binop-id))
          (binop-id)))

 )

(ENCAPSULATE
 ((ev1 (x a) t)
  (ev1-list (x-lst a) t)
  (binop-name () t)
  (binop-fix-name () t)
  (binop-rel-name () t))
 (SET-INHIBIT-WARNINGS "theory")
 (LOCAL (IN-THEORY *DEFEVALUATOR-FORM-BASE-THEORY*))
 (local (defun binop-name () 'binop))
 (local (defun binop-fix-name () 'binop-fix))
 (local (defun binop-rel-name () 'binop-rel))
 (defthm symbolp-binop-name
   (and (symbolp (binop-name))
        (not (equal (binop-name) t))
        (not (equal (binop-name) nil)))
   :rule-classes :type-prescription)
 (defthm symbolp-binop-fix-name
   (and (symbolp (binop-fix-name))
        (not (equal (binop-fix-name) t))
        (not (equal (binop-fix-name) nil)))
   :rule-classes :type-prescription)
 (defthm symbolp-binop-rel-name
   (and (symbolp (binop-rel-name))
        (not (equal (binop-rel-name) t))
        (not (equal (binop-rel-name) nil)))
   :rule-classes :type-prescription)
 (defthm binop-name-not-quote
   (not (equal (binop-name) 'quote)))
 (defthm binop-rel-name-not-quote
   (not (equal (binop-rel-name) 'quote)))
 (defthm binop-fix-name-not-quote
   (not (equal (binop-fix-name) 'quote)))
 (LOCAL (MUTUAL-RECURSION
         (DEFUN EV1 (X A)
           (DECLARE (XARGS :MEASURE (ACL2-COUNT X)
                           :WELL-FOUNDED-RELATION
                           E0-ORD-< :MODE :LOGIC))
           (COND ((SYMBOLP X) (CDR (ASSOC-EQ X A)))
                 ((EQ (CAR X) 'QUOTE) (CAR (CDR X)))
                 ((CONSP (CAR X))
                  (EV1 (CAR (CDR (CDR (CAR X))))
                       (PAIRLIS$ (CAR (CDR (CAR X)))
                                 (EV1-LIST (CDR X) A))))
                 ((EQUAL (CAR X) (BINOP-NAME))
                  (BINOP (EV1 (CAR (CDR X)) A)
                         (EV1 (CAR (CDR (CDR X))) A)))
                 ((EQUAL (CAR X) (BINOP-FIX-NAME))
                  (BINOP-FIX (EV1 (CAR (CDR X)) A)))
                 ((EQUAL (CAR X) (BINOP-REL-NAME))
                  (BINOP-REL (EV1 (CAR (CDR X)) A)
                             (EV1 (CAR (CDR (CDR X))) A)))
                 ((EQUAL (CAR X) 'IF)
                  (IF (EV1 (CAR (CDR X)) A)
                      (EV1 (CAR (CDR (CDR X))) A)
                      (EV1 (CAR (CDR (CDR (CDR X)))) A)))
                 (T NIL)))
         (DEFUN EV1-LIST (X-LST A)
           (DECLARE (XARGS :MEASURE (ACL2-COUNT X-LST)
                           :WELL-FOUNDED-RELATION E0-ORD-<))
           (COND ((ENDP X-LST) NIL)
                 (T (CONS (EV1 (CAR X-LST) A)
                          (EV1-LIST (CDR X-LST) A)))))))
 (DEFTHM EV1-CONSTRAINT-1
   (IMPLIES (SYMBOLP X)
            (EQUAL (EV1 X A) (CDR (ASSOC-EQ X A)))))
 (DEFTHM EV1-CONSTRAINT-2
   (IMPLIES (AND (CONSP X)
                 (EQUAL (CAR X) 'QUOTE))
            (EQUAL (EV1 X A) (CADR X))))
 (DEFTHM EV1-CONSTRAINT-3
   (IMPLIES (AND (CONSP X)
                 (CONSP (CAR X)))
            (EQUAL (EV1 X A)
                   (EV1 (CADDAR X)
                        (PAIRLIS$ (CADAR X)
                                  (EV1-LIST (CDR X) A))))))
 (DEFTHM EV1-CONSTRAINT-4
   (IMPLIES (NOT (CONSP X-LST))
            (EQUAL (EV1-LIST X-LST A) NIL)))
 (DEFTHM EV1-CONSTRAINT-5
   (IMPLIES (CONSP X-LST)
            (EQUAL (EV1-LIST X-LST A)
                   (CONS (EV1 (CAR X-LST) A)
                         (EV1-LIST (CDR X-LST) A)))))
 (DEFTHM EV1-CONSTRAINT-6
   (IMPLIES (AND (CONSP X)
                 (EQUAL (CAR X) (BINOP-NAME)))
            (EQUAL (EV1 X A)
                   (BINOP (EV1 (CADR X) A)
                          (EV1 (CADDR X) A)))))
 (DEFTHM EV1-CONSTRAINT-7
   (IMPLIES (AND (CONSP X)
                 (EQUAL (CAR X) (BINOP-FIX-NAME)))
            (EQUAL (EV1 X A)
                   (BINOP-FIX (EV1 (CADR X) A)))))
 (DEFTHM EV1-CONSTRAINT-8
   (IMPLIES (AND (CONSP X)
                 (EQUAL (CAR X) (BINOP-REL-NAME)))
            (EQUAL (EV1 X A)
                   (BINOP-REL (EV1 (CADR X) A)
                              (EV1 (CADDR X) A)))))
 (DEFTHM EV1-CONSTRAINT-9
   (IMPLIES (AND (CONSP X)
                 (EQUAL (CAR X) 'IF))
            (EQUAL (EV1 X A)
                   (IF (EV1 (CADR X) A)
                       (EV1 (CADDR X) A)
                       (EV1 (CADDDR X) A))))))

(encapsulate
 ((cancel_binop-rel$1 (x) t))

 (local (defun cancel_binop-rel$1 (x) x))

 (defthm cancel_binop-rel$1-property
   (implies (and (consp x) 
                 (eq (car x) (binop-rel-name)))
            (equal (ev1 (cancel_binop-rel$1 x) a)
                   (ev1 x a)))))

(defun cancel_binop-rel (x)
  (if (and (consp x) 
	   (eq (car x) (binop-rel-name)))
      (cond
       ((and (consp (cadr x))
             (eq (car (cadr x)) (binop-name))
             (consp (caddr x))
             (eq (car (caddr x)) (binop-name)))
        (let* ((lt-side (binary-op_fringe (binop-name) (cadr x)))
               (rt-side (binary-op_fringe (binop-name) (caddr x)))
               (int (bagint lt-side rt-side)))
          (if int
              (list (binop-rel-name)
                    (binary-op_tree (binop-name)
                                    (binop-id)
                                    (binop-fix-name)
                                    (bagdiff lt-side int))
                    (binary-op_tree (binop-name)
                                    (binop-id)
                                    (binop-fix-name)
                                    (bagdiff rt-side int)))
            x)))
       (t (cancel_binop-rel$1 x)))
    x))

;;; Local stuff starts here.
#|

(defun binop-list (lst)
  (cond
   ((atom lst) (binop-id))
   (t (binop (car lst)
             (binop-list (cdr lst))))))

(in-theory (disable (binop-list)))

(encapsulate
 ()

 (local
  (defthm binop-binop-id-reverse
    (equal (binop-fix x)
           (binop (binop-id) x))))

 (local (in-theory (disable binop-binop-id)))

 (local
  (defthm binop-binop-id-binop
    (equal (binop (binop-id) (binop x y))
           (binop x y))
    :hints (("Goal" :use binop-fix-binop))))

 (defthm binop-binop-fix-1
   (equal (binop (binop-fix x) y)
          (binop x y)))

 (defthm binop-binop-fix-2
   (equal (binop x (binop-fix y))
          (binop x y)))
 )

(defthm ev1-binary-op_tree+
  (implies (true-listp fringe)
           (equal (ev1 (binary-op_tree+
                        (binop-name)
                        (binop-id)
                        (binop-fix-name)
                        fringe)
                       a)
                  (binop-list (ev1-list fringe a))))
  :hints (("Goal" :induct (true-listp fringe))))

(defthm binop-fix-binop-id
  (equal (binop-fix (binop-id))
         (binop-id)))

(defthm binop-fix-binop-list
  (equal (binop-fix (binop-list x))
         (binop-list x)))

(defthm binop-list-append
  (equal (binop-list (append x y))
         (binop (binop-list x) (binop-list y))))

(defthm ev1-constraint-5-explicit
  (equal (ev1-list (cons b x-lst) a)
         (cons (ev1 b a)
               (ev1-list x-lst a))))

(defthm ev1-list-append
  (implies (true-listp x)
           (equal (ev1-list (append x y) a)
                  (append (ev1-list x a) (ev1-list y a)))))

(defthm binop-rel-ev1-ev1-list
  (binop-rel (binop-fix (ev1 x a))
             (binop-list (ev1-list (binary-op_fringe (binop-name) x) a))))

|#

(skip-proofs
(defthm cancel_binop-rel-correct-lemma
  (equal
   (binop-rel
    (ev1 (binary-op_tree
          (binop-name)
          (binop-id)
          (binop-fix-name)
          (bagdiff (binary-op_fringe (binop-name) x1)
                   (bagint (binary-op_fringe (binop-name) x1)
                           (binary-op_fringe (binop-name)
                                             x2))))
         a)
    (ev1 (binary-op_tree
          (binop-name)
          (binop-id)
          (binop-fix-name)
          (bagdiff (binary-op_fringe (binop-name)
                                     x2)
                   (bagint (binary-op_fringe (binop-name) x1)
                           (binary-op_fringe (binop-name)
                                             x2))))
         a))
   (binop-rel (binop-fix (ev1 x1 a))
              (binop-fix (ev1 x2 a)))))
)

(defthm cancel_binop-rel-correct
  (equal (ev1 x a)
         (ev1 (cancel_binop-rel x) a))
  :rule-classes nil)








