(in-package "ACL2")

(include-book "meta-generic-plus")


(defevaluator ev2 ev2-list
  ((binary-plus x y)
   (fix x)
   (equal x y)
   (acl2-numberp x)
   (< x y)
   (if x y z)))

(defun cancel_plus1 (x)
  (declare (xargs :guard (and (pseudo-termp x)
                              (consp x))))

  (mv-let
   (elt term)
   (cond ((and (consp (cadr x))
               (eq (car (cadr x)) 'binary-plus))
          (mv (caddr x) (cadr x)))
         ((and (consp (caddr x))
               (eq (car (caddr x)) 'binary-plus))
          (mv (cadr x) (caddr x)))
         (t (mv nil nil)))
   (cond
    ((and elt (fringe-occur 'binary-plus elt term))
     (list 'if
           (list 'acl2-numberp elt)
           (list 'equal
                 *zero*
                 (binary-op_tree 'binary-plus
                                 0
                                 'fix
                                 (del elt (binary-op_fringe 'binary-plus term))))
           *nil*))
    (t x))))

(defun cancel_plus-equal (x)
  (declare (xargs :guard (pseudo-termp x)))
  (if (and (consp x) 
	   (eq (car x) 'equal))
      (cond
       ((and (consp (cadr x))
             (eq (car (cadr x)) 'binary-plus)
             (consp (caddr x))
             (eq (car (caddr x)) 'binary-plus))
        (let* ((lt-side (binary-op_fringe 'binary-plus (cadr x)))
               (rt-side (binary-op_fringe 'binary-plus (caddr x)))
               (int (bagint lt-side rt-side)))
          (if int
              (list 'equal
                    (binary-op_tree 'binary-plus
                                    0
                                    'fix
                                    (bagdiff lt-side int))
                    (binary-op_tree 'binary-plus
                                    0
                                    'fix
                                    (bagdiff rt-side int)))
            x)))
       (t (cancel_plus1 x)))
    x))

(defthm acl2-numberp-ev2-binary-op-tree
  (acl2-numberp (ev2 (binary-op_tree 'binary-plus
                                     0 'fix
                                     fringe)
                     a)))

(in-theory (disable binary-op_tree))

(defthm ev2-binary-op_tree-append
  (equal (ev2 (binary-op_tree 'binary-plus
                              0 'fix
                              (append fringe1 fringe2))
              a)
         (+ (ev2 (binary-op_tree 'binary-plus
                                 0 'fix
                                 fringe1)
                 a)
            (ev2 (binary-op_tree 'binary-plus
                                 0 'fix
                                 fringe2)
                 a)))
  :hints (("Goal" :induct (append fringe1 fringe2))))

(defthm ev2-binary-op_tree-fringe
  (equal (ev2 (binary-op_tree 'binary-plus
                              0 'fix
                              (binary-op_fringe 'binary-plus x))
              a)
         (fix (ev2 x a))))

(defthm plus-cancel-left
  (equal (equal (+ x y) (+ x z))
         (equal (fix y) (fix z))))

(defthm binary-op_tree-plus-fringe-del-lemma
  (implies (memb summand fringe)
           (equal (+ (ev2 summand a)
                     (ev2 (binary-op_tree 'binary-plus
                                          0 'fix
                                          (del summand fringe))
                          a))
                  (ev2 (binary-op_tree 'binary-plus
                                       0 'fix
                                       fringe)
                       a)))
  :rule-classes nil
  :hints (("Goal" :expand ((binary-op_tree 'binary-plus
                                           0 'fix
                                           fringe2)))))

(defthm binary-op_tree-plus-fringe-del
  (implies (memb summand fringe)
           (equal (equal 0
                         (ev2 (binary-op_tree 'binary-plus
                                              0 'fix
                                              (del summand fringe))
                              a))
                  (equal (fix (ev2 summand a))
                         (ev2 (binary-op_tree 'binary-plus
                                              0 'fix
                                              fringe)
                              a))))
  :hints (("Goal" :use binary-op_tree-plus-fringe-del-lemma)))

(defthm ev2-acl2-numberp-hack-1
  (implies (equal (car y) 'binary-plus)
           (acl2-numberp (ev2 y a))))

(defthm ev2-acl2-numberp-hack-2
  (implies (and (equal (ev2 x a)
                       (ev2 y a))
                (equal (car y) 'binary-plus))
           (acl2-numberp (ev2 x a))))

(defthm cancel_plus1-property
  (implies (and (consp x)
                (equal (car x) 'equal))
           (equal (ev2 (cancel_plus1 x) a)
                  (ev2 x a)))
  :hints (("Goal" :in-theory (disable binary-op_tree-opener
                                      binary-op_fringe
                                      ev2-constraint-6))))

(in-theory (disable cancel_plus1))

(defthm cancel_plus-equal-correct
  (equal (ev2 x a)
         (ev2 (cancel_plus-equal x) a))
  :rule-classes nil
  :hints (("Goal" :by
           (:functional-instance
            cancel_binop-rel-correct
            (ev1 ev2)
            (ev1-list ev2-list)
            (binop binary-plus)
            (binop-fix fix)
            (binop-rel equal)
            (binop-id (lambda () 0))
            (binop-name (lambda () 'binary-plus))
            (binop-fix-name (lambda () 'fix))
            (binop-rel-name (lambda () 'equal))
            (cancel_binop-rel$1 cancel_plus$1)
            (cancel_binop-rel cancel_plus-equal)))))





