

Include("/usr/cli/projects/cacl2/version13/infix-prelude");

in-package("ACL2");

Include("bdd-primitives");

/* Verify guards, unless :verify-guards is NIL. */
set-verify-guards-eagerness(2);


Function v-propagate(a,b)
{ IF endp(a) THEN t
     ELSE (car(a) | car(b)) & v-propagate(cdr(a),cdr(b)) }
 :guard true-listp(a) & true-listp(b) & length(a) = length(b);

Theorem v-propagate-step
{ v-propagate(a0 :> a1,b0 :> b1) = ((a0 | b0) & v-propagate(a1,b1)) };

Theorem v-propagate-base { v-propagate(nil,nil) = t };

Function v-generate(a,b)
{ IF endp(a) THEN nil
     ELSE    v-generate(cdr(a),cdr(b))
           | car(a) & car(b) & v-propagate(cdr(a),cdr(b)) }
 :guard true-listp(a) & true-listp(b) & length(a) = length(b);

Theorem v-generate-step
{    v-generate(a0 :> a1,b0 :> b1)
   = (v-generate(a1,b1) | a0 & b0 & v-propagate(a1,b1)) };

Theorem v-generate-base { v-generate(nil,nil) = nil };

Function tv-adder(c,a,b,tree)
{ IF atom(tree)
     THEN [car(a) | car(b), car(a) & car(b), b-xor(car(a),b-xor(car(b),c))]
     ELSE { lhs := tv-adder(c,tfirstn(a,tree),tfirstn(b,tree),car(tree));
            { lhs-p := car(lhs);
              { lhs-g := cadr(lhs);
                { lhs-sum := cddr(lhs);
                  { rhs-c := t-carry(c,
                                   lhs-p,
                                   lhs-g);
                    { rhs := tv-adder(rhs-c,
                                    trestn(a,
                                           tree),
                                    trestn(b,
                                           tree),
                                    cdr(tree));
                      { rhs-p := car(rhs);
                        { rhs-g := cadr(rhs);
                          { rhs-sum := cddr(rhs);
                            { p :=    lhs-p
                                  & rhs-p;
                              { g := t-carry(lhs-g,
                                           rhs-p,
                                           rhs-g);
                                    p
                                 :>     g
                                    :> (   lhs-sum
                                         @ rhs-sum)}}}}}}}}}}} };

Theorem tv-adder-step
{ tree := left :> right;
     tv-adder(c,a,b,tree)
   = { lhs := tv-adder(c,tfirstn(a,tree),tfirstn(b,tree),car(tree));
       { lhs-p := car(lhs);
         { lhs-g := cadr(lhs);
           { lhs-sum := cddr(lhs);
             { rhs-c := t-carry(c,lhs-p,lhs-g);
               { rhs := tv-adder(rhs-c,
                               trestn(a,
                                      tree),
                               trestn(b,
                                      tree),
                               cdr(tree));
                 { rhs-p := car(rhs);
                   { rhs-g := cadr(rhs);
                     { rhs-sum := cddr(rhs);
                       { p :=    lhs-p
                             & rhs-p;
                         { g := t-carry(lhs-g,
                                      rhs-p,
                                      rhs-g);
                               p
                            :>     g
                               :> (   lhs-sum
                                    @ rhs-sum)}}}}}}}}}}}};

Theorem tv-adder-base
{    tv-adder(c,a,b,0)
   = [car(a) | car(b), car(a) & car(b), b-xor(car(a),b-xor(car(b),c))] };

Function v-sum(c,a,b)
{ IF endp(a) THEN nil
     ELSE     b-xor(c,b-xor(car(a),car(b)))
           :> v-sum((car(a) & car(b)) | (car(a) & c) | car(b) & c,
                   cdr(a),
                   cdr(b)) }
 :guard true-listp(a) & true-listp(b) & length(a) = length(b);

Theorem v-sum-step
{    v-sum(c,a0 :> a1,b0 :> b1)
   = b-xor(c,b-xor(a0,b0)) :> v-sum((a0 & b0) | (a0 & c) | b0 & c,a1,b1) };

Theorem v-sum-base { v-sum(c,nil,nil) = nil };

Module { }
{ 
  Theorem tv-adder-correct-8
  { a := [a0, a1, a2, a3, a4, a5, a6, a7];
    b := [b0, b1, b2, b3, b4, b5, b6, b7];
    tree4 := '[[0 . 0] 0 . 0];
    tree8 := tree4 :> tree4;
        boolean-listp(a) & boolean-listp(b) & booleanp(c)
     ->    tv-adder(c,a,b,tree8)
        = v-propagate(a,b) :> v-generate(a,b) :> v-sum(c,a,b)}
   :hints [["Goal" :bdd [:vars := [c, 
                                   a0, 
                                   b0, 
                                   a1, 
                                   b1, 
                                   a2, 
                                   b2, 
                                   a3, 
                                   b3, 
                                   a4, 
                                   b4, 
                                   a5, 
                                   b5, 
                                   a6, 
                                   b6, 
                                   a7, 
                                   b7]
                   ]]];
  
  Theorem tv-adder-correct-16
  { a := [a0, a1, a2, a3, a4, a5, a6, a7, a10, a11, a12, a13, a14, a15, a16, a17];
    b := [b0, b1, b2, b3, b4, b5, b6, b7, b10, b11, b12, b13, b14, b15, b16, b17];
    tree4 := '[[0 . 0] 0 . 0];
    tree8 := tree4 :> tree4;
    tree16 := tree8 :> tree8;
        boolean-listp(a) & boolean-listp(b) & booleanp(c)
     ->    tv-adder(c,a,b,tree16)
        = v-propagate(a,b) :> v-generate(a,b) :> v-sum(c,a,b)}
   :hints [["Goal" :bdd [:vars := [c, 
                                   a0, 
                                   b0, 
                                   a1, 
                                   b1, 
                                   a2, 
                                   b2, 
                                   a3, 
                                   b3, 
                                   a4, 
                                   b4, 
                                   a5, 
                                   b5, 
                                   a6, 
                                   b6, 
                                   a7, 
                                   b7, 
                                   a10, 
                                   b10, 
                                   a11, 
                                   b11, 
                                   a12, 
                                   b12, 
                                   a13, 
                                   b13, 
                                   a14, 
                                   b14, 
                                   a15, 
                                   b15, 
                                   a16, 
                                   b16, 
                                   a17, 
                                   b17]
                   ]]];
  
  Theorem tv-adder-correct-32
  { a := [a0, 
        a1, 
        a2, 
        a3, 
        a4, 
        a5, 
        a6, 
        a7, 
        a10, 
        a11, 
        a12, 
        a13, 
        a14, 
        a15, 
        a16, 
        a17, 
        a20, 
        a21, 
        a22, 
        a23, 
        a24, 
        a25, 
        a26, 
        a27, 
        a30, 
        a31, 
        a32, 
        a33, 
        a34, 
        a35, 
        a36, 
        a37];
    b := [b0, 
        b1, 
        b2, 
        b3, 
        b4, 
        b5, 
        b6, 
        b7, 
        b10, 
        b11, 
        b12, 
        b13, 
        b14, 
        b15, 
        b16, 
        b17, 
        b20, 
        b21, 
        b22, 
        b23, 
        b24, 
        b25, 
        b26, 
        b27, 
        b30, 
        b31, 
        b32, 
        b33, 
        b34, 
        b35, 
        b36, 
        b37];
    tree4 := '[[0 . 0] 0 . 0];
    tree8 := tree4 :> tree4;
    tree16 := tree8 :> tree8;
    tree32 := tree16 :> tree16;
        boolean-listp(a) & boolean-listp(b) & booleanp(c)
     ->    tv-adder(c,a,b,tree32)
        = v-propagate(a,b) :> v-generate(a,b) :> v-sum(c,a,b)}
   :hints [["Goal" :bdd [:vars := [c, 
                                   a0, 
                                   b0, 
                                   a1, 
                                   b1, 
                                   a2, 
                                   b2, 
                                   a3, 
                                   b3, 
                                   a4, 
                                   b4, 
                                   a5, 
                                   b5, 
                                   a6, 
                                   b6, 
                                   a7, 
                                   b7, 
                                   a10, 
                                   b10, 
                                   a11, 
                                   b11, 
                                   a12, 
                                   b12, 
                                   a13, 
                                   b13, 
                                   a14, 
                                   b14, 
                                   a15, 
                                   b15, 
                                   a16, 
                                   b16, 
                                   a17, 
                                   b17, 
                                   a20, 
                                   b20, 
                                   a21, 
                                   b21, 
                                   a22, 
                                   b22, 
                                   a23, 
                                   b23, 
                                   a24, 
                                   b24, 
                                   a25, 
                                   b25, 
                                   a26, 
                                   b26, 
                                   a27, 
                                   b27, 
                                   a30, 
                                   b30, 
                                   a31, 
                                   b31, 
                                   a32, 
                                   b32, 
                                   a33, 
                                   b33, 
                                   a34, 
                                   b34, 
                                   a35, 
                                   b35, 
                                   a36, 
                                   b36, 
                                   a37, 
                                   b37]
                   ]]];
  
  Theorem tv-adder-correct-64
  { a := [a0, 
        a1, 
        a2, 
        a3, 
        a4, 
        a5, 
        a6, 
        a7, 
        a10, 
        a11, 
        a12, 
        a13, 
        a14, 
        a15, 
        a16, 
        a17, 
        a20, 
        a21, 
        a22, 
        a23, 
        a24, 
        a25, 
        a26, 
        a27, 
        a30, 
        a31, 
        a32, 
        a33, 
        a34, 
        a35, 
        a36, 
        a37, 
        a40, 
        a41, 
        a42, 
        a43, 
        a44, 
        a45, 
        a46, 
        a47, 
        a50, 
        a51, 
        a52, 
        a53, 
        a54, 
        a55, 
        a56, 
        a57, 
        a60, 
        a61, 
        a62, 
        a63, 
        a64, 
        a65, 
        a66, 
        a67, 
        a70, 
        a71, 
        a72, 
        a73, 
        a74, 
        a75, 
        a76, 
        a77];
    b := [b0, 
        b1, 
        b2, 
        b3, 
        b4, 
        b5, 
        b6, 
        b7, 
        b10, 
        b11, 
        b12, 
        b13, 
        b14, 
        b15, 
        b16, 
        b17, 
        b20, 
        b21, 
        b22, 
        b23, 
        b24, 
        b25, 
        b26, 
        b27, 
        b30, 
        b31, 
        b32, 
        b33, 
        b34, 
        b35, 
        b36, 
        b37, 
        b40, 
        b41, 
        b42, 
        b43, 
        b44, 
        b45, 
        b46, 
        b47, 
        b50, 
        b51, 
        b52, 
        b53, 
        b54, 
        b55, 
        b56, 
        b57, 
        b60, 
        b61, 
        b62, 
        b63, 
        b64, 
        b65, 
        b66, 
        b67, 
        b70, 
        b71, 
        b72, 
        b73, 
        b74, 
        b75, 
        b76, 
        b77];
    tree4 := '[[0 . 0] 0 . 0];
    tree8 := tree4 :> tree4;
    tree16 := tree8 :> tree8;
    tree32 := tree16 :> tree16;
    tree64 := tree32 :> tree32;
    implie(boolean-listp(a) & boolean-listp(b) & booleanp(c),
              tv-adder(c,a,b,tree64)
            = v-propagate(a,b) :> v-generate(a,b) :> v-sum(c,a,b))}
   :hints [["Goal" :bdd [:vars := [c, 
                                   a0, 
                                   b0, 
                                   a1, 
                                   b1, 
                                   a2, 
                                   b2, 
                                   a3, 
                                   b3, 
                                   a4, 
                                   b4, 
                                   a5, 
                                   b5, 
                                   a6, 
                                   b6, 
                                   a7, 
                                   b7, 
                                   a10, 
                                   b10, 
                                   a11, 
                                   b11, 
                                   a12, 
                                   b12, 
                                   a13, 
                                   b13, 
                                   a14, 
                                   b14, 
                                   a15, 
                                   b15, 
                                   a16, 
                                   b16, 
                                   a17, 
                                   b17, 
                                   a20, 
                                   b20, 
                                   a21, 
                                   b21, 
                                   a22, 
                                   b22, 
                                   a23, 
                                   b23, 
                                   a24, 
                                   b24, 
                                   a25, 
                                   b25, 
                                   a26, 
                                   b26, 
                                   a27, 
                                   b27, 
                                   a30, 
                                   b30, 
                                   a31, 
                                   b31, 
                                   a32, 
                                   b32, 
                                   a33, 
                                   b33, 
                                   a34, 
                                   b34, 
                                   a35, 
                                   b35, 
                                   a36, 
                                   b36, 
                                   a37, 
                                   b37, 
                                   a40, 
                                   b40, 
                                   a41, 
                                   b41, 
                                   a42, 
                                   b42, 
                                   a43, 
                                   b43, 
                                   a44, 
                                   b44, 
                                   a45, 
                                   b45, 
                                   a46, 
                                   b46, 
                                   a47, 
                                   b47, 
                                   a50, 
                                   b50, 
                                   a51, 
                                   b51, 
                                   a52, 
                                   b52, 
                                   a53, 
                                   b53, 
                                   a54, 
                                   b54, 
                                   a55, 
                                   b55, 
                                   a56, 
                                   b56, 
                                   a57, 
                                   b57, 
                                   a60, 
                                   b60, 
                                   a61, 
                                   b61, 
                                   a62, 
                                   b62, 
                                   a63, 
                                   b63, 
                                   a64, 
                                   b64, 
                                   a65, 
                                   b65, 
                                   a66, 
                                   b66, 
                                   a67, 
                                   b67, 
                                   a70, 
                                   b70, 
                                   a71, 
                                   b71, 
                                   a72, 
                                   b72, 
                                   a73, 
                                   b73, 
                                   a74, 
                                   b74, 
                                   a75, 
                                   b75, 
                                   a76, 
                                   b76, 
                                   a77, 
                                   b77]
                   ]]];
  
   };


