
/* ACL2 Version 1.8 */
/* Copyright (C) 1989-95 Computational Logic, Inc. (CLI).  All rights reserved. */
/* Use of this software constitutes agreement with the terms of ACL2
license agreement, found in the file LICENSE. */

in-package("ACL2");

Include("inequalities");

/* The following three axioms seemed crucial for some of Matt
Kaufmann's work, both in non-standard analysis and in a termination
and guard verification exercise for axioms.lisp.  After these three
axioms is an informal argument that justifies them.  Perhaps some
day, someone will translate this informal proof to an Acl2 proof. */
Axiom denominator-unary-minus { rationalp(x) -> denominator(-x) = denominator(x) };

Axiom numerator-goes-down-by-integer-division
{ integerp(x) & (0 < x & (rationalp(r) & 0 <= r)) -> numerator(1/x * r) <= numerator(r) }
 :rule-classes :linear;

Axiom denominator-plus { rationalp(r) & integerp(i) -> denominator(r + i) = denominator(r) }
 :rule-classes [:rewrite [:rewrite  
 :corollary rationalp(r) & integerp(i) -> denominator(i + r) = denominator(r)]];

Module { }
{ local 
  
  Theorem div-cancellation-on-right-temp
  { z = x * 1/y & (rationalp(x) & (rationalp(y) & y ~= 0)) -> y * z = x };
  
  Theorem numerator-minus { numerator(-i) = -(numerator(i)) }
   :hints [["Goal" :use [[:instance rational-implies2
                                    [x := i ]]
                         [:instance rational-implies2
                                    [x := -i ]]
                         [:instance denominator-unary-minus
                                    [x := i ]]
                         ]
                   :in-theory disable(rational-implies2, denominator-unary-minus)]];
  
   };

Module { }
{ local 
  
  Theorem numerator-div-x-positive-case { integerp(x) & 0 < x -> numerator(1/x) = 1 }
   :hints [["Goal" :use [[:instance rational-implies2
                                    [x := 1/x ]]
                         [:instance lowest-terms
                                    [n := numerator(1/x), r := 1, q := x, x := 1/x ]]
                         ]
                   :in-theory disable(rational-implies2, mult-r-denominator-r)]];
  
  Theorem numerator-/x { integerp(x) & x ~= 0 -> numerator(1/x) = signum(x) }
   :hints [["Goal" :use [numerator-div-x-positive-case
                         [:instance numerator-div-x-positive-case
                                    [x := -x ]]
                         ]]];
  
   };


