

(in-package "ACL2")
(deflabel typed-lists-section)
(include-book "list-defuns")
(include-book "deflist")


(defmacro naturalp (x)
  `(and (integerp ,x)
	(<= 0 ,x)))
;; :option is not supported in infix acl2

(deflist rational-listp (l) rationalp)
  ;;(:options :omit-defun))

(deflist integer-listp (l) integerp)
  ;;(:options :omit-defun))

(deflist natural-listp (l) (lambda (x) (naturalp x)))

(deflist true-list-listp (l) true-listp)
  ;;(:options :omit-defun))

(deflist symbol-listp (l) symbolp)
  ;;(:options :omit-defun))

(deflist natural-listp (l) (lambda (x) (naturalp x)))
  ;;(:options (:theory nth))

(deflist my-subset (l the-set)
  (declare (xargs :guard (eqlable-listp the-set)))
  (lambda (x) (member x the-set)))

(defun maxlist (l)
  (declare (xargs :guard (and (rational-listp l)
			      (consp l))))
  (cond ((endp (cdr l)) (car l))
	(t (max (car l) (maxlist (cdr l))))))

(defun minlist (l)
  (declare (xargs :guard (and (rational-listp l)
			      (consp l))))
  (cond ((endp (cdr l)) (car l))
	(t (min (car l) (minlist (cdr l))))))

(defun sum (l)
  (declare (xargs :guard (rational-listp l)))
  (cond ((endp l) 0)
	(t (+ (car l) (sum (cdr l))))))

(defun excess-natural (l)
  (declare (xargs :guard (natural-listp l)
		  :verify-guards nil))
  (if (consp l)
      (+ 1 (maxlist l))
      0))

(defun idifference (a b)
  (let ((i-a (ifix a))
	(i-b (ifix b)))
    (if (< i-a i-b) 0 (- i-a i-b))))

(defun make-integer-list (start lub)
  (declare (xargs :guard (and (integerp start)
			      (integerp lub))
		  :measure (idifference lub start)))
  (let ((i-start (ifix start)))
    (cond ((<= (ifix lub) i-start) nil)
	  (t (cons i-start (make-integer-list (+ i-start 1) lub))))))


(defthm rationalp-maxlist
  (implies (and (rational-listp l)
		(consp l))
	   (rationalp (maxlist l)))
  :rule-classes :type-prescription)

(defthm rationalp-minlist
  (implies (and (rational-listp l)
		(consp l))
	   (rationalp (minlist l)))
  :rule-classes :type-prescription)


(defthm natural-listp-forward-to-integer-listp
  (implies (natural-listp l)
	   (integer-listp l))
  :rule-classes :forward-chaining)

(verify-guards excess-natural)

(defthm le-maxlist
  (implies (member-equal x l)
	   (<= x (maxlist l)))
  :rule-classes (:rewrite :linear :forward-chaining))

(defthm minlist-le
  (implies (member-equal x l)
	   (<= (minlist l) x))
  :rule-classes (:rewrite :linear :forward-chaining))


(defthm member-equal-maxlist
  (implies (and (rational-listp l)
		(consp l))
	   (member-equal (maxlist l) l)))

(defthm member-equal-minlist
  (implies (and (rational-listp l)
		(consp l))
	   (member-equal (minlist l) l)))

(defthm member-equal-excess-natural
  (implies (natural-listp l)
	   (not (member-equal (excess-natural l) l)))
  :hints (("Goal" :do-not-induct t
	   :use ((:instance le-maxlist (x (excess-natural l)) (l l)))
	   :in-theory (disable le-maxlist))))


(defthm member-equal-make-integer-listp
  (iff (member-equal x (make-integer-list start lub))
       (and (integerp x)
	    (<= (ifix start) x)
	    (< x (ifix lub)))))


(defthm integer-listp-make-integer-list 
  (integer-listp (make-integer-list start lub)))


(defthm natural-listp-make-integer-list
  (implies (naturalp start)
	   (natural-listp (make-integer-list start lub))))


