

in-package("ACL2");

Label typed-lists-section;

Include("list-defuns");

Include("deflist");

Macro naturalp(x) =  integerp(x) & 0 <= x;

Type rational-listp = (:options :omit-defun) * ;

Type integer-listp = (:options :omit-defun) * ;

Type natural-listp = (lambda (x) (naturalp x)) * ;

Type true-list-listp = (:options :omit-defun) * ;

Type symbol-listp = (:options :omit-defun) * ;

Function maxlist(l)     { IF endp(cdr(l)) then car(l)
                               else max(car(l),maxlist(cdr(l))) }
 :guard rational-listp(l) & consp(l);

Function minlist(l)     { IF endp(cdr(l)) then car(l)
                               else min(car(l),minlist(cdr(l))) }
 :guard rational-listp(l) & consp(l);

Function sum(l)     { IF endp(l) then 0
                           else car(l) + sum(cdr(l)) }
 :guard rational-listp(l);

Function excess-natural(l)     { IF consp(l) then 1 + maxlist(l)
                                      else 0 }
 :guard natural-listp(l);

Function idifference(a,b)
{ i-a := ifix(a);
  { i-b := ifix(b);
    IF i-a < i-b then 0
         else i-a - i-b}};

Function make-integer-list(start,lub)
{ i-start := ifix(start);
  IF ifix(lub) <= i-start then nil
       else i-start :> make-integer-list(i-start + 1,lub)}
 :measure idifference(lub,start)
 :guard integerp(start) & integerp(lub);

Theorem rationalp-maxlist
{ rational-listp(l) & consp(l) -> rationalp(maxlist(l)) }
 :rule-classes :type-prescription;

Theorem rationalp-minlist
{ rational-listp(l) & consp(l) -> rationalp(minlist(l)) }
 :rule-classes :type-prescription;

Theorem natural-listp-forward-to-integer-listp
{ natural-listp(l) -> integer-listp(l) }
 :rule-classes :forward-chaining;

Verify guards excess-natural;

Theorem le-maxlist { x in l -> x <= maxlist(l) }
 :rule-classes [:rewrite :linear :forward-chaining];

Theorem minlist-le { x in l -> minlist(l) <= x }
 :rule-classes [:rewrite :linear :forward-chaining];

Theorem member-equal-maxlist
{ rational-listp(l) & consp(l) -> maxlist(l) in l };

Theorem member-equal-minlist
{ rational-listp(l) & consp(l) -> minlist(l) in l };

Theorem member-equal-excess-natural
{ natural-listp(l) -> excess-natural(l) ~in l }
 :hints [["Goal" :do-not-induct t
                 :use [[:instance le-maxlist
                                  [x := excess-natural(l), l := l ]]
                       ]
                 :in-theory disable(le-maxlist)]];

Theorem member-equal-make-integer-listp
{ x in make-integer-list(start,lub) <-> integerp(x) & (ifix(start) <= x & x < ifix(lub)) };

Theorem integer-listp-make-integer-list
{ integer-listp(make-integer-list(start,lub)) };

Theorem natural-listp-make-integer-list
{ naturalp(start) -> natural-listp(make-integer-list(start,lub)) };










