; ACL2 Version 1.8, 1.9

; Copyright (C) 1989-97 Computational Logic, Inc. (CLI).  All rights reserved.

; Use of this software constitutes agreement with the terms of ACL2
; license agreement, found in the file LICENSE.

;; OUTLINE
;;
;; 1. What we will do for you
;;     Infix-file and its settings.
;; 2. How to run
;;     DOINFIX and its flags
;; 3. Formatting instructions in doc strings and comments
;; 4. Modifying behavior
;;     -syntax files
;;     function-printers

;; MAJOR MODS

;; Apr 25 97 MKS 
;;
;; 3. Added variables to suppress hint and guard printing.
;;    Guards that contribute to type information should be unaffected.
;;
;; 2. Added ASCII and HTML modes.
;;
;;    ASCII mode output should generally be parsable by the infix printer
;;    except that the printer is designed to print arbitrary ACL2
;;    and the parser only accepts a subset.
;;
;; Circa 1995
;;
;; 1. Added ACL2 ~ formatting conventions in doc strings and comments.
;;    DEFAULT SETTING OF VARIABLES BELOW IGNORES THEM:
;;    *acl2-format-comments* => nil,  *acl2-format-doc-strings* => nil.
;;    TO SET:
;;    *acl2-format-comments* => t,  *acl2-format-doc-strings* => t.
;;
;;    We handle doc strings by calling an ACL2 function
;;    that takes a doc string and 2 tables and returns
;;    a doc string, with the ACL2 formatting done.  
;;
;;    In order to handle arbitrary formatting within doc strings `~id[x]` 
;;    will eventually map to `x`.  Thus, in scribe mode, ~id[@section](Testing) 
;;    will come out as @section(Testing).
;;    Whereas @section(Testing) would come out as @@section(Testing)

;; BUGS 

;; 1. In comments !p(implies a b) produces different results from !p(IMPLIES a b)
;;    Case is honored when reading comments.  !p may need to adjust case.
;;    We could just store things redundantly under `IMPLIES' and `implies',
;;    but then we would fail on `Implies'
;;
;; 2. "~ID[{\\it foo}]" does not map to "{\\it foo}".  The braces get quoted.
;;    Matt Kaufmann may fix.  Until then, the user can either not use
;;    any special formatting other that that provided by ! and ~, or can
;;    turn off ACL2 formatting, by setting *acl2-format-doc-strings* and
;;    *acl2-format-comments* to nil.  (See keyword args to INFIX-FILE and
;;    INFIX-SETTINGS.)  With this turned off you can still use ! and whatever
;;    native text formatting conventions you wish.
;;
;; 3. "!cfoo" when processed by nqftm2fmt will result in an attempt to 
;;    eval foo.  This may break, if the user is not taking advantage of
;;    this !c `feature'.

;; Note the difference between the backslash in comments and the backslash
;; in doc strings.  In doc strings, because they are read as strings, they
;; need to be quoted (e.g. doubled).


;; INTERACTIONS

;; We treat comments and doc strings similarly.  The only difference is that
;; the initial comment characters of a comment line determine the format that
;; the comment text is imbedded in (e.g. running text, verbatim, format, etc.
;; See COMMENT-FORMAT documentation).

;; There are 3 (at least) kinds of imbedded formatting commands
;; that can occur in comments and doc strings.  They are handled
;; in distinct phases of processing.
;;
;;   1. ~ ACL2 doc string formatting commands.
;;   2. ! Infix commands.
;;   3.   Text formatter commands (Scribe or LaTeX or HTML)
;;
;; The first two effect the output of this infix processor.

;; To ensure that your text formatter (type 3) commands
;; are preserved during processing steps 1 and 2, the simplest 
;; approach is to enclose the commands that you wish
;; to preserve in the ~id[...] form or between ~bid[] .. ~eid[] pairs.
;; SEE BUG 2 ABOVE.

;; Comments and doc strings are preprocessed to translate their ACL2
;; formatting commands to our target formatter.  
;; Note that the ACL2 doc formatter goes to great lengths to make sure
;; weird characters in the target dialect are quoted.  So if you want
;; such things preserved, you need to use the ~id forms above.  Because
;; we now pass comment text through the ACL2 doc formatter, you
;; will need to treat comments as you treat doc strings.
;; You also need ;; to be careful of occurences of "~".  If a "~" is 
;; not followed by an ACL2 formatting command, we will complain.

#|  Update to formatting command documentation.
    May  8 97 MKS 

There are some complexities to the processing of formatting commands
within comments and doc strings.  The reasons are historical and because
3 distinct kinds of formatting commands are supported!  

We support our own ! commands (see below), the ACL2 documentation
markup commands in doc strings and comments, and of course the user is
free to include target formatter specific commands.  This later case
can be a problem, since the infix printer tries to ensure that the
target formatter won't break, but cannot in general distinguish a
bonifide command from a comment string that needs to be quoted.
E.g. consider

 ;; {\it foo} vs. { foo / bar } 

The first is probably a latex italics command, while the second is the
body of a function or macro representing a division operation.

The advantage of avoiding formatter specific commands (relying
exclusively on ! and ~ commands) is that it is easy to retarget ouput,
say from latex to html.  But for certain fancy effects you may need to
insert formatter specific entries.  You will want to use the -handsoff
option to ensure that formatter specific characters are not quoted.

Phase I:

Important limitation of phase I processing.  In `;' comments you
do not want command arguments to cross line boundaries.

Read line or doc string.

1. If acl2 formatting in effect, do it. If the substitution table 
   is set, this will result in special char quoting.  Which may
   lead to duplicate quotes after complete processing. 
   E.g. \ => \\ => \\\\
   So we leave the subst table empty and handle special chars in 2.

2. If not *hands-off*, quote special characters.

3. Print.

Phase II:

Possible limitation of phase I processing.  The same limitation on `;'
comments may effect you, depending on the choice you have made for
comment formatting.  If the output of phase I contains ! commands whose
arguments are broken by Lisp comment characters, you will get hard errors. 
Thus, multiline ! commands should only appear within \#\| \|\# comments.

1. If *nq-format*, reprocess the file honoring ! commands.


----------------------------------------------------------------------
COMMANDS

!Isym	Italic
!Bsym	Bold

!T(..)	Format inline (term).
	Where fm is an Acl2 term, results in conventional mathematical
        notation for fm, but without any line breaks.  
        Mnemonic: T -- think Term.

!T*(..)	Format (term).
	Where fm is an Acl2 term, results in conventional mathematical
        notation for fm.

!Qform	`form'. Results in form surrounded by single quotes,
        after formatting sensitive characters have been quoted, e.g., !qfoo results 
        in `foo' in TeX.  Useful for distinguishing function symbols from other 
        words in a sentence, since function symbols appear in Roman.  
        Mnemonic: Q -- think Quoted.

!Vsym   Sym is printed as is, but in typewriter font, and with special 
        characters quoted.  Mnemonic: V -- think Verbatim.
!V[..]  Verbatim (inline).
!V*[..] Verbatim.

!section[text]    Format text as a section header.  Read literally.
!subsection[text] Format test as a subsection header. Read literally.

!nsym	[N]ame. Label for Xref
!xsym	[X]ref.

!Lsym 		Literal, ignoring special chars of formatter.
!L[..]	

!Eform - [E]val. 
         The form is evaled in Lisp.  This allows you to do things
         like dynamically change the margin (*rightmost-char-number*)
         turn indexing on (SETQ *DO-NOT-INDEX* NIL) and off
         (SETQ *DO-NOT-INDEX* T), or even redefine a printer.


~-[]    ---
~B[]    bold font
~BF[]   begin format, like verbatim, but if possible don't change font
~EF[]   end format
~BID[]  begin implementation dependent
~EID[]  end implementation dependent
~BQ[]   begin quotation
~EQ[]   end quotation
~BV[]   begin verbatim
~EV[]   end verbatim
~C[]    originally @code, but we don't want `' in info file
~EM[]   emphasis
~GIF[]  gif files, e.g., ~gif[\ foo.gif\  align=top]
~I[]    italics font
~ID[]   implementation dependent  
~NL[]   newline
~PAR[]  paragraph mark, of no significance for latex
~PL[]   used for parenthetical crossrefs
~SC[]   small caps
~ST[]   strong emphasis
~T[]    typewriter font

~IL[aaa rest]   invisible link, for true hypertext environments
~ILC[aaa rest]  invisible link, but use code if such links don't show
~L[aaa rest]    link at beginning of sentence
~TERMINAL[]  terminal only, ignore
~WARN[]      noop
~CLICK-HERE[]   Html specific.  See :DOC ~ss
~PCLICK-HERE[]  Html specific.  see :DOC ~ss )
~FLY[]   Html/ACL2 doc specific.  Flying Tour: see :DOC ~ss )
~WALK[]  Html/ACL2 doc specific.  Walking Tour: see :DOC ~ss )

|#


; NOTE: we depend on the following externally defined package:

(in-package "ACL2")

;; Load packages and macros we depend on.
;; Require must be here.

;; BUILT INTO GCL
;; (require "sloop" "/usr/cli/projects/cacl2/version20/solaris/printer/sloop")
;; (use-package "SLOOP")

(provide "infix")

(eval-when (load eval compile)
 (defmacro sloop (&rest l) `(sloop::sloop ,@l))
)

;; Add space for CFUNs (compiled funs without environment) in GCL.

#+gcl(eval-when (load eval compile)
		(defun pages-allocated (type)
		  (declare (ignore type))
		  (multiple-value-bind 
		   (nfree npages maxpage nppage gcs used) (system:allocated 'cfun)
		   (declare (ignore nfree maxpage nppage gcs used))
		   npages)))

#+gcl(eval-when (load)
		(cond
		 ((boundp 'si::*gcl-major-version*) ;GCL 2.0 or greater
		  (if (< (pages-allocated 'cfun) 200)  (sys:allocate 'cfun 200)))
		 ((< (sys::allocated-pages 'cfun) 200) (sys:allocate 'cfun 200))))

;; ACL2 doc string interaction.  First is for ~ directives, second
;; for special characters.

(defvar acl2-markup-table nil)
(defvar acl2-char-subst-table nil)

;; Read-keyword command uses the acl2 read to suck in keyword
;; commands.

(proclaim '(ftype (function (t) t)
		  acl2-parse-string
		  read-keyword-command))

(defvar a-very-rare-cons nil)
(defvar *function-symbol-out* nil) ;; use to tell when to print append as @ 
                                   ;; cons as :> , etc


(defun read-keyword-form (key n)
  "This function reads the N args for keyword command, KEY, 
where we have used ACL2 mechanisms to compute N.  It returns a form.
If KEY cannot be presented in keyword form, N will be NIL."
  (cond ((null n) key)
	((integerp n)
	 (let (args)
	   (sloop for i from 1 to n 
		  do (setq args
			   (append
			    args
			    (list (readx *standard-input* nil a-very-rare-cons nil)))))
	   (cons (intern (symbol-name key)) args)))
	(t key)))

(eval-when (compile load eval )
;; Set this to be the directory in which we compile this file.
(defparameter *infix-directory* #.(namestring (probe-file "./"))))

(eval-when (load eval compile)

(defvar *loaded* nil)

(defun reset-base () (setq *loaded* nil))

(defun load-base (s &optional force)
  (if (and (not force)
	   (member s *loaded* :test #'string-equal))
      nil
      (progn (pushnew s *loaded* :test #'string-equal)
	     (let ((b1 (concatenate 'string *infix-directory* s))
		   (b2 (concatenate 'string *infix-directory* s ".o"))
		   (b3 (concatenate 'string *infix-directory* s ".lisp")))
	       (cond ((probe-file b1) (load b1))
		     ((probe-file b2) (load b2))
		     ((probe-file b3) (load b3))
		     (t (format t "Failed to find file ~s. Continuing.~%" s)))))))
)

(eval-when (load eval)

(if (find-package "acl2-output-channel")
    (progn (load-base "acl2-formatting.lisp"))

    (progn (defun acl2-parse-string (doc) (cons nil doc))
	   (defun acl2-keywordp (key) (declare (ignore key)) nil)
	   (defun read-keyword-command (key) key)
	   (defun acl2-current-package () *package*)
	   (defun acl2-defpkg (name contents) 
	     (declare (ignore contents))
	     (make-package name :use '(lisp user)))
	   ))

)

(eval-when (load eval compile)

(or (find-package "acl2")
    (make-package "acl2"))

)

(defun symbol-name-equal (a b)
  (if (or (not (atom a)) (not (atom b)))
      nil
    (let ((x (if (stringp a) a (symbol-name a)))
	  (y (if (stringp b) b (symbol-name b))))
      (string-equal x y))))

(defun symbol-name-member (a l)
  (member a l :test #'symbol-name-equal))

#|

     A Conventional Syntax Pretty-Printer for ACL2

       Originially written by Rober Boyer
       Modified for Scribe by Michael K. Smith (2/92,8/93)
       Modified for ACL2   by Michael K. Smith (10/93)
       Modified for interactive ACL2 by Michael K. Smith & Jimmy Wang (10/96)
       Modified for HTML   by Michael K. Smith  Apr 25 97 


			 INTRODUCTION

The functions in this file implement a pretty-printer for Acl2 events.
The syntax produced is conventional mathematical notation.

This file is not automatically compiled or loaded when building Acl2.
To use this printer, after compiling and loading Acl2, compile this file and
load it, i.e., (compile-file "infix.lisp") and (load "infix").  For
more information on installation see the README file in the directory
containing this file.

The following text is, currently, the only documentation for this facility.
Criticism of all sorts solicited.


				  BASIC USE

The intent is to take an ACL2 events file and produce a nicely formatted
document.  Knowing what text formatter you are targeting, you can insert text
formatting commands into comments.  You can also request an infix
transformation of prefix forms in comments (see documentation of function
NQFMT2FMT.

ONE NOTE OF CAUTION:  It is important that you know what text formatter you
are targetting, since the bulk of your comments will be copied literally into
the text input file.  It is up to you to ensure that the interaction of
your comments with the formatting convention of choice for the various
comment types results in legal text formatting commands.  That is, if you
are in Scribe mode and a comment contains an email message with net addresss
you should quote occurences of "@" as "@@".  More importantly, if you decide
that ";\" introduces a line to be formatted as raw LaTex (the default in
"latex" mode), you need to ensure that any occurrence of "_" or other LaTeX
special characters on such a line results in a meaningful formatting
command.  For simple transformations of .event files to LaTeX I suggest you
use the default :COMMENT-FORMAT (= 'BOYER).  This causes most comments to be
formatted within verbatim environments, which are not so picky about special
characters.  Scribe is more forgiving of these problems because it only has
the single special character, "@" that needs to be watched out for. (See
`SETTINGS THAT MAY BE MODIFIED IN MODE-SYNTAX.LISP'.)

There are 4 basic modes, "ascii", "latex", "html" and "scribe".  You
can then build on top of these to customize the printing of functions
in your syntax.  All mode sensitivity should be contained in the file
<mode>-syntax.lisp.  Normally this file also loads a basic file called
<mode>-init.lisp.  The idea is that the `-init' files contain the
minimal required information for INFIX to do its job with respect to a
particular text formatter.  The `-syntax' file contains the details of
how you want the functions in your syntax to be printed.

In order to customize printing for a particular file of events, say
"clock.events", we suggest the following approach.  Each column shows the
procedure for the corresponding text formatter, Latex or Scribe.

First, assume we have a file "clock.events", in proper syntactic form for
acceptance by LD.  That is to say, suppose that the file
"clock.events" contains only legal event commands such as defun's and
defthm's, Lisp style comments, and the few other sorts of miscellaneous
instructions documented as legal instructions to LD.  HTML and ASCII modes 
would be extended in a similar fashion.


1. Create clock-syntax.lisp.  It should have the following form:

-    Tex                                 Scribe
-    ----------------------------------------------------------------
-    (load-base "latex-syntax.lisp")     (load-base "scribe-syntax.lisp")
-     ... 
-     Your extensions and/or redefinitions.  
-     See in particular the documentation on make-infix et.al.
-     under `SIX GENERAL OPERATOR SCHEMAS', and the examples at the
-     end of this file and in scribe-syntax.lisp and latex-syntax.lisp.
-     ...
-     INFIX-SETTINGS provides some simple control over an assortment of 
-     formatting options.  See `SETTINGS THAT MAY BE MODIFIED IN MODE-SYNTAX.LISP'. 
-     ...
-    (infix-settings :mode "clock"       (infix-settings :mode "clock" 
-               :extension "tex" ...)               :extension "mss" ...)

2. Save clock-syntax.lisp, preferably in the same directory with clock.events.

3. Call infix-file.  The simplest call on infix-file would be just

-    (infix-file "clock").

   which will result in the file "clock.tex" or "clock.mss" 

4. Run the appropriate formatter.

5. Special considerations for latex vs. scribe.
   
-  To get an index in LaTeX.             To avoid an index in Scribe
-  ----------------------------------------------------------------
-  %latex clock                          Insert @disable(Index) in clock.mss.
-  %makeindex clock
-  %latex clock


A full blown call to infix-file includes a file root and keyword arguments.
Below is such a call with keywords supplied with their defaults.

-  (infix-file fl :print-case :downcase :mode nil :chars-wide 77 :comment t)
-
-   :print-case - must be one of :downcase, :upcase, or :capitalize.
-                 DEFAULT: :downcase.
-
-   :mode       - if not provided (thus defaulting to nil) we look for
-                 "fl-syntax.lisp" and load it if present.  If not, we use 
-                 the mode of the last successfull call to infix-file or 
-                 query which of the 4 known modes you want to use.
-                 Simplest is to create a trivial
-                 -syntax file in the same directory as the event files that 
-                 just loads appropriate predefined -syntax file.
-                 DEFAULT: fl root name.  If not found, queries the user.
-
-   :chars-wide - approximate width in characters of the formatted output.
-                 Controls where infix inserts line breaks within expressions.
-                 DEFAULT: 77.
-
-   :comment    - If t, then certain specially marked Acl2 expressions in 
-                 comments are replaced with their conventional notation.  
-                 See the documentation of the function `nqfmt2fmt' below for 
-                 a description of the special syntax for this replacement 
-                 process.  We assume you use this feature.  If not,
-                 set *nq-default* to NIL in your mode file.
-                 This is also modified by the :format-!-in-comments keyword
-                 in (infix-settings ...)
-                 DEFAULT: T.


			       COMMENT HANDLING

- Jul 28 93 MKS - Extended comment handling.  
- Aug  3 93 MKS - Still haven't done anything with internal comments.

Modified Treatment of comments: `infix-file' preserves comments
between events, but skips over comments within events.  We completely
skip comments within events because we have not figured out how to
place them appropriately mechanically.  Documentations strings in
events are handled, normally by pulling them out of the event and
inserting them before the event itself.

Comments are formatted in several ways, depending on the character immediately
following the semi-colon or OPEN-BAR-COMMENT.  A comment may be turned into:

- 1. Running TEXT.  The comment chars (see definition following) are
- eliminated and the text is copied to the output file.
- 
- 2. A FORMATted environment.  The comment chars (see definition
- following) are eliminated, line breaks and spaces are preserved, and
- the font is the default document font.
- 
- 3. A VERBATIM environment. The comment chars may or may not be preserved, 
- line breaks and spaces are PRESERVED and the font is a fixed width font.
- 
- 4. An EMPHASIS environment. Like format, but the font is italic.
- 
- 5. A COMMENT environment. Typically wraps a comment form around text.
-    The Latex version will almost certainly fail.  The C version, for
-    example, puts /* ... */ around the text.
-
- 6. A SECTION heading. Should only be a one-line comment.

This set, which is used by the named formats in *comment-format-alist* can
be extended by modifying the value of *comment-environment-mapping* in your
syntax file.

To replace the comment format conventions, use (define-comment-format name
format).

The format argument has two parts, one for semi-colon comments
and the other for \#\| ... \|\# comments.  The assignment below corresponds to
Boyer's initial setting.

- (define-comment-format 'boyer
-   '((#\; (("\\"   nil   text)) nil verbatim #\;)
-     (#\# (("\\"   nil   text))   t verbatim )))

The structure of each of these lists is
-  type      ::= (label (substring*) spaces-p format [char])
-  substring ::= (string spaces-p format [char])

LABEL indicates which of the two types of comments we are looking at, either
those that start with a SEMI-COLON or those that start with LB VERTICAL-BAR.
We determine what to do in the various cases (which amounts to choosing
SPACES-P, FORMAT, and CHAR) based on whether the comment type indicated by
LABEL is followed by any of the strings that begin SUBSTRINGS or not.  If it
matches, we use the components of the matching substring, otherwise we use
the default for the comment type, i.e.  the elements of the type list.

-  If SPACES-P, consume one space if there is one.
-  Begin formatting according to FORMAT.
-  Insert CHAR.

So, for the example above, the first sublist, whose car is the semi-colon
character, describes how to format comments that begin with a semi-colon followed
by specific strings.  There are two possibilites.  If the semi-colon is not
followed by a back-slash (\), we don't look for a space, we ensure we are in a
verbatim environment, and print a semi-colon.  If semi-colon is followed by a
back-slash, we don't look for a space and ensure that we are in a text environment.

Thus, if we encounter a comment beginning ";\", the comment should be
inserted as top level text with no special formatting.  The ";\" will not show
up in the output.


- COMMENT TRANSFORMATIONS:

There are 4 versions.  One, text, reflects MKSmith's preferences, one, verbatim, Boyer's,
one Common Lisp defaults, and one C.  "text" is the default.  To get "verbatim"
do (setup-comment-format 'verbatim).  To get Common Lisp's, do
(setup-comment-format 'cl).  You can insert this form in your syntax file.
To create your own conventions, see DEFINE-COMMENT-FORMAT.  

- Description:

-  BT begins running text, with no environment modifiers.  
-  BF ... EF corresponds  to <begin-format> ... <end-format>
-  BV ... EV corresponds  to <begin-verbatim> ... <end-verbatim> 
-  BE ... EE corresponds  to <begin-emphasis> ... <end-emphasis> 
-  BS ... ES corresponds  to <begin-section-name> ... <end-section-name> 
-  BC ... EC corresponds  to <begin-comment> ... <end-comment> 
- 
-              MKS           Boyer             CL              C
-  
-  \#\|  ... \|\#  BT...         BV ... EV         BT...            BC ... EC
-  \#\|\ ... \|\#  BT...         BT ...            BT...   
-  \#\|- ... \|\#  BF... EF      BV- ... EV        BF... EF
-  \#\|; ... \|\#  BV... EV      BV; ... EV        BV... BV       
-  
-  ; ...       BT...         BV; ... EV        BE... EE         BC ... EC
-  ;; ...      BT...         BV;; ... EV       BF... EF
-  ;;; ...     BV... EV      BV;;; ... EV      BT...   
-  ;;;; ...    BV;... EV     BV;;;; ... EV     BS... ES

-  ;# ...      BC... EC        
-  ;\ ...      BT...         BT ...            BT...   
-  ;- ...      BF... EF      BV;- ... EV       BF... EF
-  ;+ ...      BV... EV      BV;+ ... EV       BV... EV
-  ;! ...      BE... EE      BV;! ... EV       BE... EE

-  ;;- ...     BF; ... EF    BV;;- ... EV      BF; ... EF
-  ;;+ ...     BV; ... EV    BV;;+ ... EV      BV; ... EV
-  ;;! ...     BE; ... EE    BV;;! ... EV      BE; ... EE



				   COVERAGE

The `infix-file' function should handle the entirety of the Acl2 term syntax
checked by ld.  We DO print out the `hint' parts of events.



				  MOTIVATION

We hope this notation will facilitate the communication of work with Acl2 to
those who do not happily read Lisp notation.  But we have no expectation that
this notation will make it easier for the Acl2 user to formulate or to prove
theorems.


			      NO ERROR CHECKING

Warning about the absence of error checking: In general, user-callable
subroutines of Acl2 do extensive syntactic checking on their input and
explicitly report syntactic errors.  But this pretty printer does not do such
syntactic checking.  Rather, we assume the given input is known to be
syntactically correct, namely as though checked by `LD'.  Failure to
provide input in correct syntactic form can result in nasty, brutish, and
short Lisp errors.


			  OTHER USER-LEVEL FUNCTIONS

Besides `infix-file', here are the other user-level functions supported by
this file.

(a) (print-examples) creates a stand-alone, file named
"infix-examples.tex" or "infix-examples.mss", which is a summary of
the syntax we print in terms of the official Acl2 syntax.  This file
will also correctly report any user modifications made to the syntax
by invocation of the make... functions described later.  We hope that
users will want to include this file in reports about Acl2 use to
make clear what syntactic conventions they are using.

(b) (NQFMT2FMT "foo") copies the file "foo.lisp" to the file
"foo.tex" or "foo.mss", but along the way, Acl2 terms preceded by an
exclamation mark and certain alphabetic characters are replaced with
the formatting commands for the conventional notation this printer
generates.  For example, when in latex mode, if nqfmt2fmt encounters
!t(gcd x (difference y x)) in a file, it will replace it with
{\rm{gcd}}\,({\it{x\/}}, {\it{y\/}} $-$ {\it{x\/}}).  We find the
former much easier to read in a file.  nqfmt2fmt thus permits one to
keep Acl2 forms in files to be read and edited by humans, e.g., in
comments in Acl2 event files.  Ordinary uses of !, e.g., uses of it
followed by white space or punctuation characters, are, of course,
unaltered.  

Let ev  be an Acl2 event form, e.g., (defun foo (x) 3)
    fm  be an Acl2 term, e.g., (plus x y)
    foo be a symbol

Summary:

!Eev  - Format event.
!Pfm  - Pretty print.
!Tfm  - Pretty print but without any line breaks.  
!Vfoo - Verbatim.
!Qfn  - `fn'. 
!Bsymbol - bold
!Ifoo - Identity, handling special chars of formatter.
!section(text)    - Format text as a section header.
!subsection(text) - Format test as a subsection header.
!Cform - [C]ommand evaluate.  This should really be E for EVAL.

Detail:

!Eev  - Event. Results in conventional notation for ev.

!Ifoo - Identity. Results in foo, but with with formatting sensitive
        characters quoted.

!Pfm  - Pretty print.  Results in conventional mathematical notation.

!Qfn  - where fn is a symbol, results in fn surrounded by single gritches,
        after formatting sensitive characters have been quoted, e.g., !qfoo results 
        in `foo' in TeX.  Useful for distinguishing function symbols from other 
        words in a sentence, since function symbols appear in Roman.  
        Mnemonic: Q -- think Quoted.

!Tfm  - where fm is an Acl2 term, results in conventional mathematical
        notation for fm, but without any line breaks.  
        Mnemonic: T -- think Term.

!Vfoo - foo is printed as is, but in typewriter font, and with special characters quoted.  
        Mnemonic: V -- think Verbatim.

!section(text)    - Format text as a section header.
!subsection(text) - Format test as a subsection header.

Begin HACK ALERT

!Cform - [C]ommand evaluate.  This should really be E for EVAL.
         The form is evaled in Lisp.  This allows you to do things
         like dynamically change the margin (*rightmost-char-number*)
         turn indexing on (SETQ *DO-NOT-INDEX* NIL) and off
         (SETQ *DO-NOT-INDEX* T), or even redefine a printer.
 
End HACK ALERT


! followed by anything else is left alone, along with the exclamation mark.

See the comment at the beginning of the definition of nqfmt2fmt, below, for
details on the syntax and replacements.  There is also an option to nqfmt2fmt
for simply stripping out the !commands.

(c) (infix-form fm) prints (to *standard-output*) the formatting input for the
conventional notation for the Acl2 term fm.  `infix-form' and `infix-event' 
can be used to generate Latex or Scribe to be inserted manually into
papers, but we recommend the use of nqfmt2fmt, described above, for this
purpose.

(d) (infix-event ev) prints (to *standard-output*) the Latex or Scribe for the
conventional notation for the Acl2 event ev.


			   USER EXTENSION OF SYNTAX

`infix-file' is table-driven, and it is very easy to extend in certain ways,
e.g., to introduce a new infix operator.  See the very end of this file, at
`USER MODIFIABLE TABLE SETUP', for examples of how to establish new syntax.

Also see `SETTINGS THAT MAY BE MODIFIED IN MODE-SYNTAX.LISP' to see how to
control additional features of the printing process, e.g. indexing, details of 
comment handling, parentheses around expressions, etc.


                          PARENTHESES and PRECEDENCE

This pretty-printer provides a facility for the suppression of
parentheses via the declaration of precedence for operators.  An
objective in printing a formula is clarity.  We know of very, very few
cases (e.g., + and *) where precedence is something on which people
agree.  We drop the outermost parentheses of a formula when the
formula is an argument of a function that is being printed in the
usual f(x,y) notation, with subterms separated by parentheses and
commas.


                               OTHER FORMATTERS

There are some functions in this file that take advantage of similarities
between LaTeX and Scribe.  They are marked with `WARNING: Latex/Scribe
dependency'.  If you want to generate input to some other formatter you may
want to take a look at these.  Not guaranteed to be complete.  In order to
built a -init.lisp file for some other formatter make sure you look at
`SPECIAL VARIABLES THAT MUST BE DEFINED IN MODE-INIT.LISP'.


                            END OF COMMENTS ON USE

|#

#| ---------------------------------------------------------------------------------

                          COMPILATION DEPENDENCIES               
|#

;; Check that we are in a compatible Acl2.
;; This is handled in save-acl2.lisp

;; Not used.

; (defun untranslate-event (form) (acl2::untranslate form nil))

;; Included from Nqthm.

(defun our-flatc (x)
  (cond ((stringp x) (+ 2 (length x)))
        ((symbolp x) (length (symbol-name x)))
        ((integerp x) (our-flatc-number x))
        (t (let ((*print-base* 10)
                 (*print-pretty* nil)
                 (*print-radix* nil)
                 (*print-level* nil)
                 (*print-length* nil)
                 (*print-case* :upcase))
             (length (format nil "~a" x))))))

(defun our-flatc-number (n)
  (cond ((< n 0) (1+ (our-flatc-number (- n))))
        ((< n 10) 1)
        (t (1+ (our-flatc-number (floor (/ n 10)))))))

(defvar a-very-rare-cons (cons nil nil))


#| ---------------------------------------------------------------------------------

             SETTINGS THAT MAY BE MODIFIED IN MODE-SYNTAX.LISP. 

Use INFIX-SETTINGS to set this first set of properties.
See files ascii-mode.lisp, latex-mode.lisp and scribe-mode.lisp for examples.

The things you can control with INFIX-SETTINGS are listed below.  The first
choice is the default, so if you are happy with the default you don't need to
fool with the setting.  See examples after the enumeration.  These are
all keyword arguments to INFIX-SETTINGS.

1. MODE: a string naming the syntax we are in.  The default is constructed
   the the name of the events file.  If no corresponding -syntax file is found,
   query the user.

2. EXTENSION: type of output file ("mss", "tex", ...)

3. OP-LOCATION: ['FRONT, 'BACK]
   Controls where operator is printed when printing on multiple lines
   'FRONT - put operator at beginning of line (Smith's way)
   'BACK  - put operator at end of line (Boyer's way)

4. COMMENT-FORMAT: ['text, 'verbatim, 'CL 'C]
   Chooses from one of the predefined conventions for determining from the character
   following the comment character how to format the comment.  This interacts
   with your use of native-mode formatting commands in comments in the .events file.

   For your own control over this, use (DEFINE-COMMENT-FORMAT name format).
   See the description of comment handling for more information.

5a. FORMAT-!-IN-COMMENTS: [T, nil]
   If true, run nqfmt2fmt over comments.

5b. ACL2-FORMAT-COMMENTS: [T, nil]
   If true, run the acl2 doc string formatting process over comments.

5c. ACL2-FORMAT-DOC-STRINGS: [T, nil]
   If true, run the acl2 doc string formatting process over doc strings.

6. NO-INDEX: boolean [NIL, t]. If T, then no index is generated.

7. NO-INDEX-CALLS: boolean [NIL, T] or list
   If you want all function calls indexed, NIL. you do not want any function use indexed, T.
   If you want to exclude certain function calls, provide a list of function 
   names to be ignored.

8. HINTS: boolean [NIL, t]. If T, print hints.

9. NO-GUARDS: boolean [NIL, t]. If T, suppress guard printing.

If you do not provide a keyword value pair, the settings remains unchanged.
Thus, you really don't have to call this function.  Though typically you want
to change the :mode if you have created a special syntax extension on top of
Scribe or Latex.

The minimal call on INFIX-SETTINGS requires a mode and extension.

(INFIX-SETTINGS :MODE                   "scribe"
		:EXTENSION              "mss"   )
	
The maximal call, setting everything explicitly.
The following shows infix-settings with all of the default settings as
arguments.  The comments indicate required types of values.  `...' indicates
settings that the user may extend with more work.

(INFIX-SETTINGS :MODE                   "scribe" ; string ["scribe","latex","html"...]
		:EXTENSION              "mss"    ; string ["mss","tex","html"]
		:OP-LOCATION            'FRONT   ; ['FRONT,'BACK]
		:COMMENT-FORMAT         'TEXT    ; ['TEXT,'VERBATIM,'CL,'C...]
		:FORMAT-!-IN-COMMENTS   NIL      ; [t,nil]
		:ACL2-FORMAT-COMMENTS   T        ; [t,nil]
		:ACL2-FORMAT-DOC-STRINGS T       ; [t,nil]
		:NO-INDEX               NIL      ; [t,nil]
		:NO-INDEX-CALLS         NIL      ; [t,nil,l]
		:NO-GUARDS              NIL      ; [t,NIL]
		:HINTS                  NIL      ; [t,NIL]
		)

|#


; Variable controlling handling of special formatting within comments.  See `NQFMT2FMT'.

(defvar *nq-default* t)

;; I think this variable is OBSOLETE.  Jimmy has modified
;; this away.

;; If *INFIX-OP-LOCATION* (:OP-LOCATION arg to INFIX-SETTINGS) is 'BACK 
;; then you get Boyer's style of printing a list of infix operators and 
;; arguments.  If 'FRONT, you get Smiths.  Smiths is the default.  
;; You can tell who last editted this file.

;- BACK form is      e.g
;-  arg1 op               foo(a,b,c) &
;-  arg2 op               bar(a,1) &
;-  argn                  some-long-function-name(a,b,1)

;- FRONT form is       e.g
;-     arg1                 foo(a,b,c)
;-  op arg2               & bar(a,1)
;-  op argn               & some-long-function-name(a,b,1)


; Either FRONT or BACK.
(defvar *infix-op-location* 'front)

; Extension of input file.
(defvar infix-input-file-type "lisp")


#|
             SETTINGS THAT MAY BE RESET IN MODE-SYNTAX.LISP. (2)

The following variables do not NEED to be reset in your mode file, but they may be.

INDEXING

If you do not any index (SETQ *DO-NOT-INDEX* T).
If you do not want occurences of functions indexed (SETQ *DO-NOT-INDEX-CALLS* T).
If you want to exclude certain functions, add them to the list *DO-NOT-INDEX-CALLS-OF*.
If you want no index, see comments at beginning of file. 


DEBUGGING

Setting *INFIX-TRACE* to T will provide some debugging help when testing new mode files.


                     END OF SETTINGS FOR MODE-INIT.LISP.

---------------------------------------------------------------------------------
|#


; ---------------------------------------------------------------------------------
;
;             SPECIAL VARIABLES THAT MUST BE DEFINED IN MODE-INIT.LISP.
;
; Use INFIX-SETTINGS to set this variable.  See Introduction.

; One of NIL, "latex", "scribe", or another string.
(defvar *infix-mode* nil)          

;; STRINGS BASED ON THE TARGET FORMATTER (LaTeX, Scribe, ...)

;; Default extension of created files.
(defvar *mode-extension* nil)

(defvar *standard-prelude* nil)
(defvar *standard-postlude* nil)

(defvar *example-prelude* nil)
(defvar *begin-example-table*  nil)
(defvar *end-example-table*  nil)
(defvar *example-table-size*  nil)
(defvar *example-postlude*  nil)

;; BASIC BRACKETS AND THEIR QUOTED VERSION.

(defvar *begin*  "")
(defvar *end*    "")
(defvar *lbrace*  "")
(defvar *rbrace*  "")
(defvar *lbracket* "[")
(defvar *rbracket* "]")

;; ENVIRONMENT BEGIN-END PAIRS

(defvar *begin-index*  "")
(defvar *end-index*  "")

(defvar *begin-label*  "")
(defvar *end-label*  "")

(defvar *begin-reference*  "")
(defvar *end-reference*  "")

(defvar *begin-text-env* "")
(defvar *end-text-env* "")

(defvar *begin-verbatim-env* "")
(defvar *end-verbatim-env* "")

(defvar *begin-verb-env* "")        ;; This is the latex \verb;..;
(defvar *end-verb-env* "")		

(defvar *begin-format-env* "")
(defvar *end-format-env* "")

(defvar *begin-emphasis-env*  "")
(defvar *end-emphasis-env*  "")

(defvar *begin-comment-env*  "")
(defvar *end-comment-env*  "")

(defvar *begin-section-env*  "")
(defvar *end-section-env*  "")

(defvar *begin-subsection-env*  "")
(defvar *end-subsection-env*  "")

(defvar *begin-tt-env*  "")
(defvar *end-tt-env*    "")

(defvar *begin-string-env*  "")
(defvar *end-string-env*    "")

(defvar *begin-bold-env*  "")
(defvar *end-bold-env*    "")

(defvar *begin-italic-env*  "")
(defvar *end-italic-env*    "")

(defvar *begin-sc-env*  "")
(defvar *end-sc-env*    "")

(defvar *begin-enumerate-env*  "")
(defvar *end-enumerate-env*  "")
(defvar *begin-item*  "")
(defvar *end-item*  "")

(defvar *forall* "")
(defvar *exists* "")

;; TABBING AND INDENTING ENVIRONMENT AND TAB OPERATIONS

(defvar *begin-tabbing-env*  nil)
(defvar *end-tabbing-env*  nil)
(defvar *new-tab-row*  nil)

;; Needs to be redefined in <mode>-init.lisp
;; No longer used.  A LIE!
(defmacro new-tab-row (&optional followed-by-infix-print-term)
  (declare (ignore followed-by-infix-print-term))
  '(pprinc *new-tab-row*))

(defvar *tab*  nil)

(defvar *column-separator* nil)

; *tabs-list* is a text-formatter specific variable.  Typically of the form of a 
; list of pairs, either (tab . n) or (lm . n), where n is the value of
; *infix-loc* when we set tabs and margins.

(defvar *tab-list* nil)
(defvar *tab-stack* nil)

(defvar *set-margin* nil)
(defvar *pop-margin*  nil)
(defvar *set-tab* nil)

(defvar *default-op-tab-space* "")

(defvar *adjust-tab-before-margin* nil)

;; FONTS

(defvar *function-font*  nil)
(defvar *neg-str* nil)

;; MATH ENV AND OPERATORS

(defvar *math-format*  nil)
(defvar *math-begin*  nil)
(defvar *math-end*  nil)

(defvar *math-thick-space*  nil)
(defvar *math-thin-space*  nil)

(defvar *subscript*  nil)

(defvar *begin-subscript* nil)
(defvar *end-subscript* nil)

;; MISC.

(defvar *last-operator* nil)
(defvar *newpage*  nil)

(defvar *comma-atsign*  nil)
(defvar *caret*  nil)
(defvar *backslash* nil)

(defvar *list-seperator* ", ")
(defvar *dotted-pair-seperator* " . ")

(defvar *no-tab-event-trailer*  nil)
(defvar *print-default-event-header*  nil)
(defvar *print-default-lisp-header*  nil)

(defvar *print-default-command-header*  nil)
(defvar *no-tab-command-trailer*  nil)

;; ACL2 RELATED.

;; We conditionally apply the ACL2 formatting conventions.

(defvar *acl2-format-comments* nil)
(defvar *acl2-format-doc-strings* nil)

;; EVENT-PRINTERS depend on these settings.

(defvar *print-hints* t)
(defvar *print-no-guards* nil)
(defvar *print-verify-guards* t)
(defvar *print-top-locals* t)
(defvar *print-keys* t)


#|\---------------------------------------------------------------------------------

              FUNCTIONS THAT MUST BE DEFINED IN MODE-INIT.LISP.

Signatures as passed to (proclaim '(ftype ...) ..)

() -> t : function of no args, returning arbitrary type
 
            begin-tabbing                end-tabbing
            begin-normal-text            end-normal-text
            set-margin                   pop-margin
            get-margin                   pop-tab
            do-tab
            begin-flushright             end-flushright

            ;; to-current-margin            ;; newline-to-current-margin    
            ;; force-newline

(t) -> t : function of one arbitray arg, returning arbitrary type

            flushright
            roman-font

(t) -> t : function of one optional arg, returning arbitrary type

            set-tab

(character) -> t : function of one character arg, returning arbitrary type

            handle-special-chars
            handle-special-chars-in-string
            char                         { Why is this in this list? }


---------------------------------------------------------------------------------

                           IMPLEMENTATION COMMENTS

The 4 `tables' that govern the printing are the variables: 

1. *atom-alist*, which governs printing of variables, numbers, T, F, and NIL.

2. *fn-alist*, which governs the printing of any term that starts with a
function symbol, including LET, COND, CASE, LIST, LIST*, and FOR.

3. *event-printer-alist*, which governs the printing of events.

4. *special-quoted-forms-alist*, which governs the special printing of selected 
quoted symbols.

Each table is an alist.  Each member of any of these alists is a list (symbol
fn), where symbol is the car of a form (or in the case of a non-consp form,
the form itself) which is about to be printed.  fn is a Common Lisp function
of one argument which is called on the form in question to do the printing.
For each alist, there is a default function that is returned if a symbol is
not explicitly represented.  One such default is the function
default-fn-printer, which is a good function to study before contemplating
serious modifications to this file.

Although adding new members to these alists, and defining corresponding
special purpose functions, is certainly sensible and easy, there are several
points to be made.

1.  It is unlikely that there will be any call to edit `*atom-alist*' until
and unless TRANSLATE is changed.

2.  *fn-alist* can be most easily modified by the use of the macros
make-infix-op, make-unary-prefix-op, make-unary-suffix-op,
make-infix-multiple-op, and make-prefix-multiple-op.  See the very end of the
file for many examples of how syntax operators can be easily established.

We really do assume throughout this code that the input file has passed
through LD, e.g., we assume that the last test in a `cond' is on T,
the last test in a case is an `otherwise', and that the third argument to
`defthm' is something that translate can accept.


STANDARD OUTPUT USED

Printing.  We do *all* of our printing of formulas to *standard-output*, so we
call princ and write-char on just one argument, the thing to be printed.

|#

;                                   PRINTING

; The setting of the *left-margin* causes only the stuff within tabbing
; environments to be moved over.  Acl2 event forms that do not use that
; tabbing environment should be adjusted by other means by the user if desired.
; *left-margin* may be set before invoking infix-form or infix-event.

(defvar *left-margin* 0)

; *rightmost-char-number* is bound sometimes to alter subsidiary printing
; operations that more text needs to be printed on the same line after they
; finish.

(defvar *rightmost-char-number* 77)

; *infix-loc* is a good estimate of how far in from the left margin we
; currently are, counting as 1 any character, or any symbol not wired in.

(defvar *infix-loc* 0)

; If *testing* is t, then we are not really printing but only counting the
; characters we would print, trying to see if a line break is necessary.

(defvar *testing* nil)

(defvar *latex-indent-number-limit* 13)

; In *tabs-in* we keep track of how deep into tabs we are so that we can punt
; if necessary.

(defvar *tabs-in* 0)

(defvar *do-not-use-tabs* nil)

; We cannot place defparameters for the following three special symbols at this
; place in the file because their initial values contain `functions' of
; functions to be defined later.

(proclaim '(special *atom-alist* *fn-alist*))

(defvar *event-printer-alist* nil)

(defvar *user-package* (find-package "user"))

(defun eventp (x)
  (and (consp x)
       (not (consp (car x)))       ;;lambda case ((lambda ...) ..)
       (assoc (car x) *event-printer-alist*)))


(defvar *index-entry-max* 100)

(defvar *negative-constant-table* nil)

(defvar *negative-infix-table* nil)

(defvar *constant-ops* nil)

(defvar *infix-ops* nil)

(defvar *infix-multiple-ops* nil)

(defvar *prefix-multiple-ops* nil)

(defvar *suffix-multiple-ops* nil)

(defvar *unary-prefix-ops* nil)

(defvar *negative-unary-prefix-table* nil)

(defvar *unary-suffix-ops* nil)

(defvar *negative-unary-suffix-table* nil)

(defvar *unary-abs-ops* nil)

(defvar *tracing-advise-break* nil)

(defvar *white-space* '(#\Space #\Newline #\Tab #\Page))

(defvar *started-a-verbatim* nil)

(defvar *started-a-format* nil)

(defvar *reported-tabs* nil)

;  This `secret' function symbol is used to print out integers generated by
; readins #b, #o, or #x syntax.

(defvar *infix-radix* (make-symbol "*infix-radix*"))

; One should add to this list if any other special forms are hard wired into
; this printer.

(defvar *wired-in-infix-examples*
  '((if x y z)
    (cond (test1 value1) (test2 value2) (t value3))
    (case x (key1 answer1) (key2 answer2) (otherwise default))
;    (for x in l when (test x) collect (fn x))
    (let ((var1 val1) (var2 val2)) form)
    (let* ((var1 val1) (var2 val2)) form)
    (forall (x y) (p x))
    (exists (x y) (p x))
    (not x)))

; Severe warning on printing.  It is illegal for any function in this
; file to do any printing except via these four printing macros:
; pprinc, pprin1, pformat, and pwrite-char.  This may be true
; transitively, e.g., pprinci calls pprinc.

; This rule is imposed so that the `hack' of printing invisibly (see
; *testing*) will work.  The point is that any printing operation may
; be called many times!  But we do not want to print unless the
; special *testing* is bound to nil!  Key fact: if *testing* is t, we
; DO NOT print.

; A very sloppy fudge factor to account for the fact that in \tt mode,
; characters are fatter.

(defvar *tt-size* 1.3)

(defvar *do-not-index* nil)

(defvar *do-not-index-calls* nil)

(defvar *infix-comma* (make-symbol "comma"))

(defvar *infix-comma-atsign* (make-symbol "comma-atsign"))

(defvar *infix-backquote* (make-symbol "backquote"))

(defvar *do-not-index-calls-of*
  (list *infix-radix* *infix-comma* *infix-comma-atsign* *infix-backquote*))

(eval-when (load compile eval)

(defmacro pprinc (x) `(or *testing* (princ ,x)))

(defmacro pprin1 (x) `(or *testing* (prin1 ,x)))

(defmacro pformat (&rest x)  `(or *testing* (format ,@x)))

(defmacro ppformat (&rest x) `(or *testing* (format *standard-output* ,@x)))

(defmacro pwrite-char (x) `(or *testing* (write-char ,x)))

)

;; PRINTING ATOMIC FORMS AND STRINGS.

; It is absolutely desireable that any printing done by any function inside this
; file, within the scope of a tabbing environment, be done with with pprinci,
; pprin1, or print-atom IF the printing is to contribute `real', i.e.,
; non-formatting, characters to the final output.  The optional second argument
; specifies how many characters are being contributed, with the default being 
; the length of the first argument.  
; If *testing* is t, not only do we not want to print, but we want to throw 
; out to advise-break if we have exceeded the *rightmost-char-number*.

(defvar *newline-in-text* "")
(defvar *newline-in-env*  "")

(defvar *force-newline-in-text* "")
(defvar *force-newline-in-env*  "")

(defvar *prev-line-loc* nil)

(defun line-return ()
  (setq *prev-line-loc* *infix-loc*)
  (pwrite-char #\Newline)
  (setq *infix-loc* (get-margin)))

(defun fake-line-return ()
  (setq *prev-line-loc* *infix-loc*)
  (setq *infix-loc* (get-margin)))

(defun newline-in-env (&optional force) 
  (if force (pprinc *force-newline-in-env*) (pprinc *newline-in-env*))
  (line-return))

(defun newline-in-text (&optional force) 
  (if force (pprinc *force-newline-in-text*) (pprinc *newline-in-text*))
  (line-return))

(eval-when (load compile eval)

(defmacro begin-text () `(begin-normal-text))

(defmacro end-text () '(end-normal-text))

(defun last-line-feed-position (x)
  (if (stringp x) (position #\linefeed x :from-end t)))

(defun pwrite-chari (c)
  (setq *infix-loc* (+ *infix-loc* 1))
  (cond ((and *testing*
	      (> *infix-loc* *rightmost-char-number*))
	 (throw 'advise-break *infix-loc*)))
  (pwrite-char c))


(defmacro pprinci (x &optional i)
  `(let* ((x ,x)
	 (i ,(if (and i (numberp i)) i))
	 n)
     (if (null i)
	 (setq i (cond ((stringp x) (length x))
		       ((symbolp x) (length (symbol-name x)))
		       (t 1))))
     (pprinc x)
     (setq *prev-line-loc* *infix-loc*)
     (cond ((setq n (last-line-feed-position x))
	    (if (> (- (length x) 1) n)
		(setq *infix-loc* (- (length x) 1 n))
		(setq *infix-loc* (get-margin))))
	   (t (incf *infix-loc* i)))
     (cond ((and *testing*
		 (> *infix-loc* *rightmost-char-number*))
	    (throw 'advise-break *infix-loc*)))))
          
(defmacro pprin1i (x)
  `(progn (let ((x ,x))
            (pprin1 x)
            (incf *infix-loc* (our-flatc x)))
          (cond ((and *testing*
                      (> *infix-loc* *rightmost-char-number*))
                 (throw 'advise-break *infix-loc*)))))

)

(defun newlinex (&optional force)
  (cond (*do-not-use-tabs* (pprinci " "))
	((null *tab-stack*) (newline-in-text force))
	(t (newline-in-env force))))

(defun to-current-margin ()
  (cond ((eql *infix-loc* (get-margin)) nil)
	(t (newlinex))))

(defun blankline ()
  ;; CHANGED T (forced) in all NEWLINEX below to NIL.
  (cond ((and (eql *infix-loc* (get-margin))
	      (eql *prev-line-loc* *infix-loc*)))
	((eql *infix-loc* (get-margin)) (newlinex t))
	(t (newlinex t) (newlinex t))))

(defvar *default-indent-spaces* 4)


;                    SIX GENERAL OPERATOR SCHEMAS

; The following make-... macros are used to generate functions and entries for
; the *fn-alist*.  See the end of this file for many examples of usage which can
; be easily extended.

(defun clean-up (fn)

; This function is supposed to remove completely all trace of any prior establishment
; of any syntax for the function symbol fn.

  (or (symbolp fn)
      (error (format nil "Illegal function symbol name: ~a." fn)))
  ;; DELTA !!!!
  (setf (get fn 'literalform) nil)
  (sloop for lst in '(*constant-ops* *infix-ops* *unary-prefix-ops* *unary-suffix-ops* *unary-abs-ops*)
           do  (set lst (remove fn (eval lst))))
  (sloop for alist in '(*fn-alist* *negative-constant-table* *negative-infix-table* *negative-unary-prefix-table*
                                     *negative-unary-suffix-table* *prefix-multiple-ops*
				     *suffix-multiple-ops*
                                     *infix-multiple-ops*)
           do  (set alist (sloop for pair in (eval alist)
                                   unless (eq fn (car pair))
                                   collect pair))))

;; Used to reinitialize during clean-up-everything
(defvar *save-fn-alist* nil)

(defun clean-up-everything ()
  (sloop for alist in '(*fn-alist* *negative-constant-table*
				     *negative-infix-table* 
				     *negative-unary-prefix-table*
                                     *negative-unary-suffix-table*
				     *prefix-multiple-ops*
				     *suffix-multiple-ops*
                                     *infix-multiple-ops*) 
           do (progn
                (sloop for pair in (eval alist)
                         do (clean-up (car pair)))
                (set alist nil)))
  ;; Reinitialize 
  (setq *fn-alist* *save-fn-alist*))
                 
(defmacro make-constant-op (name str &optional neg-str opsize)
  (let ((fn-name (intern (format nil "~s-constant-op-printer" name))))
    `(progn
       (clean-up ',name)
       (setf (get ',name 'literalform) ,(format nil *math-format* str))
       (defun ,fn-name
         (term)
         (declare (ignore term))
	 (pprinci ,(format nil *math-format* str) ,opsize))
       (push (list ',name (function ,fn-name))
             *fn-alist*)
       (push ',name *constant-ops*)
       ,(cond (neg-str `(push (list ',name ',(format nil *math-format* neg-str))
                              *negative-constant-table*)))
       ',name)))

(defmacro make-infix-op (name str &optional neg-str precedence opsize)
  (let ((fn-name (intern (format nil "~s-infix-op-printer" name))))
    `(progn
       (clean-up ',name)
       (setf (get ',name 'literalform) ,(format nil *math-format* str))
       ,(cond (precedence `(setf (get ',name 'precedence) ,precedence))
	      ((get name 'precedence) nil)
	      (t `(setf (get ',name 'precedence) 0)))
       (defun ,fn-name
         (term)
         (default-infix-printer
           term
           ,(format nil *math-format* str) ,opsize))
       (push (list ',name (function ,fn-name))
             *fn-alist*)
       (push ',name *infix-ops*)
       ,(cond (neg-str `(push (list ',name ',(format nil *math-format* neg-str))
                              *negative-infix-table*)))
       ',name)))

(defmacro make-infix-multiple-op (name &rest strs)
  (let ((fn-name (intern (format nil "~s-infix-multiple-op-printer" name))))
    `(progn
       (clean-up ',name)
       (defun ,fn-name
         (term)
         (default-infix-multiple-printer
           term
           ',(sloop for str in strs collect
                      (format nil *math-format* str))))
       (push (list ',name (function ,fn-name))
             *fn-alist*)
       (push (cons ',name ,(length strs)) *infix-multiple-ops*)
       ',name)))

(defmacro make-prefix-multiple-op (name &rest strs)
  (let ((fn-name (intern (format nil "~s-prefix-multiple-op-printer" name))))
    `(progn
       (clean-up ',name)
       (defun ,fn-name
         (term)
         (default-prefix-multiple-printer
           term
           ',(sloop for str in strs collect
                      (format nil *math-format* str))))
       (push (list ',name (function ,fn-name))
             *fn-alist*)
       (push (cons ',name ,(length strs)) *prefix-multiple-ops*)
       ',name)))

(defmacro make-suffix-multiple-op (name &rest strs)
  (let ((fn-name (intern (format nil "~s-prefix-multiple-op-printer" name))))
    `(progn
       (clean-up ',name)
       (defun ,fn-name
         (term)
         (default-suffix-multiple-printer
           term
           ',(sloop for str in strs collect
                      (format nil *math-format* str))))
       (push (list ',name (function ,fn-name))
             *fn-alist*)
       (push (cons ',name ,(length strs)) *suffix-multiple-ops*)
       ',name)))
                 
;; Needs opsize
(defmacro make-unary-prefix-op (name str &optional neg-str)
  (let ((fn-name (intern (format nil "~s-unary-prefix-op-printer" name))))
    `(progn
       (clean-up ',name)
       (defun ,fn-name
         (term)
         (default-unary-prefix-printer
           term
           ,(format nil *math-format* str)))
       (push (list ',name (function ,fn-name))
             *fn-alist*)
       (push ',name *unary-prefix-ops*)
       ,(cond (neg-str `(push (list ',name ',(format nil *math-format* neg-str))
                              *negative-unary-prefix-table*)))
       ',name)))

;; Needs opsize
(defmacro make-unary-suffix-op (name str &optional neg-str)
  (let ((fn-name (intern (format nil "~s-unary-suffix-op-printer" name))))
    `(progn
       (clean-up ',name)
       (defun ,fn-name
         (term)
         (default-unary-suffix-printer
           term
           ,(format nil *math-format* str)))
       (push (list ',name (function ,fn-name))
             *fn-alist*)
       (push ',name *unary-suffix-ops*)
       ,(cond (neg-str `(push (list ',name ',(format nil *math-format* neg-str))
                              *negative-unary-suffix-table*)))
       ',name)))

;; Needs opsize
(defmacro make-unary-abs-op (name lhs-str rhs-str)
  (let ((fn-name (intern (format nil "~s-unary-abs-op-printer" name))))
    `(progn
       (clean-up ',name)
       (defun ,fn-name
         (term)
         (default-unary-abs-printer
           term
           ,(concatenate 'string *math-begin* lhs-str)
           ,(concatenate 'string rhs-str *math-end*)))
       (push (list ',name (function ,fn-name))
             *fn-alist*)
       (push ',name *unary-abs-ops*)
       ',name)))

(defmacro declare-atom-printer (x v)
  `(let ((temp (assoc ',x *atom-alist*)))
     (if (null temp)
         (setq *atom-alist* (cons (list ',x ,v) *atom-alist*))
         (rplacd temp (list ,v)))))

(defmacro declare-fn-printer (x v)
  `(let ((temp (assoc ',x *fn-alist*)))
     (if (null temp)
         (setq *fn-alist* (cons (list ',x ,v) *fn-alist*))
         (rplacd temp (list ,v)))))

(defmacro declare-event-printer (x v)
  `(let ((temp (assoc ',x *event-printer-alist*)))
     (if (null temp)
         (setq *event-printer-alist* (cons (list ',x ,v) *event-printer-alist*))
         (rplacd temp (list ,v)))))

(defmacro declare-command-printer (x v)
  `(let ((temp (assoc ',x *event-printer-alist*)))
     (if (null temp)
         (setq *event-printer-alist* (cons (list ',x ,v) *event-printer-alist*))
         (rplacd temp (list ,v)))))

(defmacro def-atom-printer (x args body)
  (let ((printer (intern (concatenate 'string (symbol-name x) "-atom-printer"))))
    `(let ((temp (assoc ',x *atom-alist*)))
       (defun ,printer ,args ,body)
       (if (null temp)
	   (setq *atom-alist* (cons (list ',x (function ,printer)) *atom-alist*))
         (rplacd temp (list (function ,printer)))))))

(defmacro def-fn-printer (x args body)
  (let ((printer (intern (concatenate 'string (symbol-name x) "-fn-printer"))))
    `(let ((temp (assoc ',x *fn-alist*)))
       (defun ,printer ,args ,body)
       (if (null temp)
	   (setq *fn-alist* (cons (list ',x (function ,printer)) *fn-alist*))
         (rplacd temp (list (function ,printer)))))))

(defmacro def-event-printer (x args body)
  (let ((printer (intern (concatenate 'string (symbol-name x) "-event-printer"))))
    `(let ((temp (assoc ',x *event-printer-alist*)))
       (defun ,printer ,args ,body)
       (if (null temp)
	   (setq *event-printer-alist*
		 (cons (list ',x (function ,printer)) *event-printer-alist*))
         (rplacd temp (list (function ,printer)))))))


;                                    TABBING

; Infix-File generates text that uses the Latex `tabbing' or Scribe `format' environment, 
; setting a tab for each new level of indentation.  We find this a convenient
; sublanguage to target.  

; It appears based upon various experiment that perhaps Latex cannot handle tabs
; more than about 14 deep, or so.  

; The parameter, *latex-indent-number-limit*, could perhaps be increased if one 
; had a Latex wherein this limit has been raised.  However, it is a relatively rare
; function that needs terms that are more than 13 function calls deep.  When
; infix-file hits this limit in Latex mode, it falls back upon the standard Acl2
; pretty-printer, and puts the result in a verbatim environment.

;; All of the following should be defined in the text-formatting init file,
;; e.g., latex-init.lisp or scribe-init.lisp.

(proclaim '(ftype (function nil t)
                  begin-tabbing		;enter a tabbing environment
		  begin-group-tabbing	;enter a tabbing environment, no page breaks
                  end-tabbing		;exit tabbing env
		  begin-normal-text	;
		  end-normal-text
                  set-margin		;set margin to current line location
                  pop-margin		;pop to previous margin
                  get-margin		;value
                  set-tab		;set at tab at current line location
                  pop-tab		;remove a tab
                  do-tab		;tab to next tab
		  begin-flushright
		  end-flushright
                  ;; to-current-margin
		  ;; newlinex-to-current-margin ;
		  ;; force-newlinex
		  ))

(proclaim '(ftype (function (t) t)
		  flushright roman-font print-function-name))

(proclaim '(ftype (function (&optional t) t)
                  set-tab))

(proclaim '(ftype (function (character) t)
                  handle-special-chars
                  handle-special-chars-in-string char))


;                                  PRINT-ATOM

; We want to slashify special characters in the following three lists in
; case they appear in an Acl2 symbol.  Used only by print-atom and index.

(defvar doc-special-chars  nil)
(defvar doc-other-chars    nil)
(defvar doc-index-specials nil)

; We also to handle the characters in doc-other-chars specially, by going into
; math mode, since slashification with backslash does not work.

(defun print-string (str &optional i)

; Our own printer, which slashifies (or otherwise quotes) the doc-special-chars and
; doc-other-chars in strings.  We print all Acl2 symbols with this
; function and print-atom because we want to avoid generating stuff that will make 
; the text formatter barf, e.g., in Latex, a symbol with an unslashified $, <, or { in it.

  (cond ((stringp str)
         (sloop for j below (length (the string str))
                  for char = (char (the string str) (the fixnum j))
                  do  (handle-special-chars-in-string char)
                  finally (incf *infix-loc* (or i (length str)))))
        ((symbolp str)
         (print-atom str i))
	((characterp str) (print-character str i))
        (t (pprin1i str)))
  (cond ((and *testing*
              (> *infix-loc* *rightmost-char-number*))
         (throw 'advise-break *infix-loc*))))

(defvar *hands-off* nil)

;; Print documentation string.
;; This will normally be called from within a tabbing environment.
;; Needs to be sensitive to CR, esp. for Latex.
(defun print-text-from-string (str &optional i)

; Our own printer, which slashifies (or otherwise quotes) the doc-special-chars and
; doc-other-chars in strings.  We print all Acl2 doc strings with this
; function so that Latex won't barf, e.g.,
; in Latex, a symbol with an unslashified $, <, or { or & in it.

  (cond ((stringp str)
	 (if *hands-off*
	     (progn (princ str) (setq *infix-loc* (+ *infix-loc* (length str))))
	     (sloop for j below (length (the string str))
		    for char = (char (the string str) (the fixnum j))
		    do  (if (char= char #\Newline)
			    ;; Newlinex is sensistive to whether we are 
			    ;; in a tabbing env or not.
			    (newlinex)
			    (handle-special-chars char))
		    finally (incf *infix-loc* (or i (length str))))))
        ((symbolp str)
	 (if *hands-off*
	     (prin1 str)
	     (print-atom str i)))
	((characterp str) (print-character str i))
        (t (pprin1i str)))
  (cond ((and *testing*
              (> *infix-loc* *rightmost-char-number*))
         (throw 'advise-break *infix-loc*))))

(defun print-symbol-body (atm i)
  (let ((str (if (stringp atm) atm (symbol-name atm))))
    (sloop for j below (length (the string str))
	   for char = (char (the string str) (the fixnum j))
	   do  (handle-special-chars char))
    (incf *infix-loc* (or i (length str)))))

(defvar *lisp-package* (find-package "LISP"))

;; (defun print-character (c &optional i)
;; 
;; ; Our own character printer, quotes the doc-special-chars and
;; ; doc-other-chars used to print characters.  We print all Acl2 characters with this
;; ; function because we want to avoid generating stuff that will make the text formatter barf,
;; ; e.g., in Latex, a symbol with an unslashified $, <, or { in it.
;; ; If i is present is is meant to be the actual printed width of the formatted output, e.g.
;; ; stripped of formatting commands.
;; 
;;   (handle-special-chars #\#)
;;   (handle-special-chars #\\)
;;   (handle-special-chars c)
;;   (incf *infix-loc* (or i 3))
;;   (cond ((and *testing*
;;               (> *infix-loc* *rightmost-char-number*))
;;          (throw 'advise-break *infix-loc*))))

(defun print-character (c &optional i)

; Our own character printer, quotes the doc-special-chars and
; doc-other-chars used to print characters.  We print all Acl2 characters with this
; function because we want to avoid generating stuff that will make the text formatter barf,
; e.g., in Latex, a symbol with an unslashified $, <, or { in it.
; If i is present is is meant to be the actual printed width of the formatted output, e.g.
; stripped of formatting commands.

  (handle-special-chars #\')
  (translate-char-to-infix-and-print c)
  (handle-special-chars #\')
  (incf *infix-loc* (or i 3))
  (cond ((and *testing*
              (> *infix-loc* *rightmost-char-number*))
         (throw 'advise-break *infix-loc*))))

(defun translate-char-to-infix-and-print (c)
  ;; These are the only special case characters in ACL2.
  (case c
	(#\Newline (pprinci *backslash*)(pprinci "n" 1))
	(#\Tab     (pprinci *backslash*)(pprinci "t" 1))
	(#\Space   (pprinci " " 1))
	(#\Page    (pprinci *backslash* 1)(pprinci "f" 1))
	(#\Rubout  (pprinci *backslash* 1)(pprinci "b" 1))
	(otherwise (handle-special-chars c))))

(defvar *package-use-list* (package-use-list *package*))

;; Why does Lisp complain if I have the following declaration 
;;   (defvar *the-live-state*)
;; and load into Lisp in the ACL2 image. And how do I proclaim a 
;; constant? And will that do the right thing anyway?
;; Feb 10 97 MKS 

(defvar *infix-package-alist* nil "From acl2-formatting.lisp")

(defvar *infix-file* nil
  "This lets us know whether we are in the acl2 loop or
in the midst of a call to infix-file.  See HANDLE-IN-PACKAGE.")

;; Is the symbol in the current package?

(defun symbol-in-current-package (a)
  ;; IMPORTANT: it has already been determined that a is not 
  ;; a reserved word nor a reserved string.
  (let ((sp (symbol-package a)))
    ;; This may not be right!!!  But let us wait and see. 
    ;; The problem is that when we see a previously unknown
    ;; package we do a (make-package unknown :use ("ACL2" "LISP"))
    ;; So when running in such a mode we need to use this non-acl2
    ;; mode.  What should happen is that in handle-in-package 
    ;; we should invoke the acl2 defpkg.  Probably?  

    ;; If in the middle of a call to infix-file, use
    ;; the lisp package tests.  
    (if *infix-file*
	(or (equal sp *package*)
	    (member a (cdr (assoc-equal *package* *infix-package-alist*))))
	;; Now ACL2.
	(let ((the-current-package (f-get-global 'current-package *the-live-state*)))      
	  (or (equal (symbol-package-name a) the-current-package)
	      (member-eq a
			 (cdr (assoc-equal
			       the-current-package
			       (known-package-alist *the-live-state*)))))))))

(defun find-acl2-package-symbol (all-packages-symbols)
  (if (consp all-packages-symbols)
      (if (member-equal "ACL2-USER" all-packages-symbols)
	  all-packages-symbols
	(if (consp (car all-packages-symbols))
	    (if (member-equal "ACL2-USER" (car all-packages-symbols))
		(car all-packages-symbols)
	      (find-acl2-package-symbol (cdr all-packages-symbols)))))
    nil))

;; (find-acl2-package-symbol (know-pcakage-alist *the-live-state*))
;; gives a list ACL2 defined symbols including importing symbols
;; Jimmy 4-11-97

(defvar *acl2-reserved-words*
  (cdr (find-acl2-package-symbol (known-package-alist *the-live-state*))))

(defvar *acl2-reserved-strings*
  (mapcar #'symbol-name *acl2-reserved-words*))

(defun acl2-reserved-wordp(atm) (member atm *acl2-reserved-words*))

(defun acl2-reserved-stringp(str) (member-equal str *acl2-reserved-strings*))

;;  printing xxx::yyy
;; 
;;  ppp = currrent-package                        ppp =      acl2     foo       bar
;; 
;;  if reserved-word(xxx::yyy) print yyy          acl2::if   if       if        if
;;  if reserved(yyy)           print xxx::yyy     foo::if    foo::if  foo::if   foo::if
;;  if ppp = xxx               print yyy          acl2::foo  foo      -         -
;;                                                foo::bim   -        bim       -
;;  else                       print xxx::yyy     acl2::bim  -        acl2::bim acl2::bim
;;                                                foo::bim   foo::bim -         foo::bim 

(defun print-atom (atm &optional i)

  ;; Our own atom printer, which slashifies (or otherwise quotes) 
  ;; the doc-special-chars and doc-other-chars in symbols and strings.  
  ;; We print all Acl2 symbols with this function because we want
  ;; to avoid generating stuff that will make the text formatter barf,
  ;; e.g., in Latex, a symbol with an unslashified $, <, or { in it.
  ;; If i is present it is meant to be the actual printed width of 
  ;; the formatted output, e.g. stripped of formatting commands.

  (cond ((symbolp atm)
	 (let* ((atm1 atm)
		(symb (get atm 'literalform))
		(atm (if (and symb *function-symbol-out*) symb atm1)))
	   (cond ((keywordp atm1)    (pprinci ":") (print-symbol-body atm i))
		 ((acl2-reserved-wordp atm1)       (print-symbol-body atm i))
		 ((acl2-reserved-stringp (symbol-name atm1))
		  (print-symbol-body-with-pkg atm i))
		 ((symbol-in-current-package atm1) (print-symbol-body atm i))
		 ;; Else print package prefix.  
		 (t (print-symbol-body-with-pkg atm i)))))
	((stringp atm)
	 (incf *infix-loc* (or i (+ 2 (length atm))))
	 (pprinc *begin-string-env*)	;was *begin-tt-env*
	 (pprinc "\"")
	 (sloop for i below (length atm)
		for char = (char (the string atm) (the fixnum i))
		do  (handle-special-chars-in-string char))
	 (pprinc "\"")
	 (pprinc *end-string-env*))
	(t (pprin1i atm)))
  (cond ((and *testing*
	      (> *infix-loc* *rightmost-char-number*))
	 (throw 'advise-break *infix-loc*))))

(defun print-symbol-body-with-pkg (atm i)
  (let ((pkg (package-name (symbol-package atm))))
    (print-symbol-body pkg (length pkg))
    (pprinci "::")
    (print-symbol-body atm i)))

(defun print-atoms (l)
  (cond ((cdr l)
	 (sloop for x on l
		do (cond ((cddr x)
			  (print-atom (car x))
			  (pprinc ", "))
			 ((cdr x)
			  (print-atom (car x))
			  (pprinc " and "))			 
			 (t (print-atom (car x))))))
	(t (print-atom (car l)))))

(defun print-bare-function-name (term)
  ;(pprinc "`")
  (print-atom term)
  ;(pprinc "'")
  )

(defun print-bare-function-names (l)
  (cond ((cdr l)
	 (sloop for x on l
		do (cond ((cddr x)
			  (print-bare-function-name (car x))
			  (pprinc ", "))
			 ((cdr x)
			  (print-bare-function-name (car x))
			  (pprinc " and "))			 
			 (t (print-bare-function-name (car x))))))
	(t (print-bare-function-name (car l)))))

(load-base "font-printers")

;                                   *ATOM-ALIST*

; *atom-alist* is one of the three tables off of which this printer is driven.
; It determines how a variable symbol is printed.  It is unlikely that anyone
; would want to change this unless translate changes because t, f, and nil are
; the only symbols that translate treats specially as constants instead of as
; variable symbols.

; We would like to put this at the beginning of the file but cannot, so we put
; a special declaration up there.  This is defined ni font-printers.

(defvar *atom-alist* nil)

(defun default-atom-printer (var)

; We put variables in italics, strings in typewriter.

  (cond ((keywordp var) (print-atom var))
	((symbolp var)  (italic-sym-printer var)) 
        ((stringp var)  (print-atom var))
        ((numberp var)  (tt-sym-printer var))
	((characterp var) (print-character var))
        (t (pprin1 var))))

(defun get-atom-printer (sym)
  (let ((a (assoc sym *atom-alist*)))
    (cond (a (cadr a))
          (t (function default-atom-printer)))))


;; NEED to have read, read in the user package, where all of our
;; variables are compiled!.  So use readx.
;; ACTUALLY, this should no longer matter.  Apr 18 97 MKS 

(defun readx (str a b c)
  (let ((val (read str a b c)))
    (if (acl2-keywordp val)
	(read-keyword-command val)
	val)))

(defun read-from-stringx (str) (read-from-string str))

(defparameter *testread* nil)

(defun testread ()
  (or (boundp '*testread*)
      (let ((*readtable* (copy-readtable nil)))
	(smash-infix-readtable)
	(setq *testread* *readtable*)))
  (let ((*readtable* *testread*))
    (readx t t nil t)))

; We arrange to read #b, #o, and #x syntax preserving what was read and to
; print it in the conventional notation whereby #o10 comes out as

;      10
;        8

; We do this by having READ fake numbers into the parse tree looking like
; function calls of the function whose name is the uninterned symbol that is
; the value of *infix-radix*, and by supplying a printer for this symbol.

; The 6 read macros for this syntax:

;; !!! LEAVE THIS AROUND, BUT TRY NOT USING.

(defun smash-infix-readtable ()
  (sloop for c in '(#\B #\b #\O #\o #\X #\x)
           as n in '(2 2 8 8 16 16)
           for l = (case n
                         (2 '(#\0 #\1))
                         (8 '(#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7))
                         (16 '(#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9
                               #\A #\B #\C #\D #\E #\F
                               #\a #\b #\c #\d #\e #\f)))
           do
           (set-dispatch-macro-character
            #\#
            c
            (let ((l l)
                  (nn n)
                  (arithmetic-signs '(#\+ #\- #\/)))
              (function (lambda (stream char n)
                          (declare (ignore char n))
                          (list* *infix-radix* nn
                                 (sloop for ch = (read-char stream)
                                          with ll = nil
                                          do
                                          (cond ((or (member ch l)
                                                     (member ch arithmetic-signs))
                                                 (push ch ll))
                                                (t (unread-char ch stream)
                                                   (return (nreverse ll)))))))))))

; Also setup the backquote reader.

  (set-macro-character #\`
    #'(lambda (stream char)
       (declare (ignore char))
       (list *infix-backquote* (readx stream t nil t))))
  (set-macro-character #\,
   #'(lambda (stream char)
       (declare (ignore char))
       (case (peek-char nil stream t nil t)
             ((#\@ #\.)
              (read-char stream)
              (list *infix-comma-atsign* (readx stream t nil t)))
             (t (list *infix-comma* (readx stream t nil t)))))))

;; FROM NQTHM

(eval-when (load compile eval)

(defmacro match (term pattern) (match-macro term pattern))

(defun match-macro (term pat)
  (cond ((atom term)
         (match1-macro term pat))
        (t (let ((sym (gensym "match")))
             `(let ((,sym ,term))
                ,(match1-macro sym pat))))))

(defvar setq-lst nil)
(defvar test-lst nil)

(defun match1-macro (term pat)
  (let (test-lst setq-lst)
    (match2-macro term pat)
    (list (quote cond)
          (cons
           (cond ((null test-lst) t)
                 ((null (cdr test-lst)) (car test-lst))
                 (t (cons (quote and) test-lst)))
           (nconc setq-lst (list t))))))

(defun match2-macro (term pat)
  (cond ((atom pat)
         (cond ((equal pat (quote &)) nil)
               ((or (equal pat t) (equal pat nil))
                (error "attempt to smash t or nil."))
               ((symbolp pat)
		(setq setq-lst (nconc setq-lst (list (list (quote setq) pat term)))))
               (t
		  (setq test-lst
			(nconc test-lst (list (list (quote equal) pat term)))))))
        ((equal (quote cons) (car pat))
         (setq test-lst (nconc test-lst (list (list (quote consp) term))))
         (match2-macro (list (quote car) term) (cadr pat))
         (match2-macro (list (quote cdr) term) (caddr pat)))
        ((equal (quote quote) (car pat))
	 (setq test-lst (nconc test-lst
			       (list (list (quote equal)
					   (list (quote quote) (cadr pat))
					   term)))))
        (t (cond ((not (equal (quote list) (car pat)))
                  (setq pat (cons (quote list)
                                  (cons (list (quote quote) (car pat))
                                        (cdr pat))))))
           (sloop for subpat in (cdr pat) do
                    (progn (setq test-lst
                                 (nconc test-lst (list (list (quote consp) term))))
                           (match2-macro (list (quote car) term) subpat)
                           (setq term (list (quote cdr) term))))
           (setq test-lst (nconc test-lst (list (list (quote equal) term nil)))))))

)

;; END NQTHM importation

(load-base "function-printers")

;                                   *FN-ALIST*

; *fn-alist* is considerably extended via calls to make-... at the end of this
; file.  This initial value given here picks up the very few entries for which
; we have totally ad hoc handling.  Although LIST and LIST* are essentially
; macros, we handle them by the default-fn-printer, since they evaluate all
; their arguments.  We could have handled IF this default way, too.  It was
; essential that we handle QUOTE, COND, CASE, LET, and FOR specially because
; they do not evaluate all their arguments but `parse' them to some extent.

; We would like to put this at the top but have to put it after the functions
; are defined.

(defvar *fn-alist* nil)


(Defun get-real-name (name)
    (if (symbol-name-member name '("ALTERNATIVE-LAMBDA-NAME"))
       'lambda
      (if (symbol-name-member name '("*do-not-print-fn-name*"))
	  nil
	name)))

(defun default-fn-printer (term)

; This function is a good function to study if one finds it necessary to define
; by hand a special handler for a function symbol.  We annotate rather
; verbosely as a pedagogical device.

; In general, we know that term is a lisp object on which TRANSLATE does not
; cause an error.

  (let ((advice (advise-break term)))

    (cond ((atom (car term))

; First put out the function symbol, in roman.
; Functions of no arguments are printed as constants, without parentheses,
; rather than with, e.g., foo().

           (cond ((null (cdr term))
                  (print-function-name (get-real-name (car term)))
		  (math-thin-space)
		  (pprinci "()")
                  (return-from default-fn-printer nil)))

; We want a very tiny space in front of the open parenthesis because otherwise,
; it looks like some function symbols touch the open parenthesis.  In some
; cases, this results in a touch more space than we like, and perhaps there
; exists a more delicate kerning command.

;; get-real-name(x) => if x = *do-not-print-fn-name* then nil
;;                     or when x = lambda-alternate-name return 'lambda
;;                     else x

	   (let ((name (get-real-name (car term))))
	     (if name (print-function-name name)))
           (math-thin-space)
           (pprinci "("))
          (t (pprinci "(")
             (setq term (cons 'foo term))))
    (cond ((null (cdr term))
           (pprinci ")"))
	  ((not (consp (cdr term)))
	   (pprinci *dotted-pair-seperator*)
	   (infix-print-term1 (cdr term))
           (pprinci ")"))
          ((null (cddr term))
           (infix-print-term1 (cadr term))
           (pprinci ")"))

	  ;; The coder of the printer for each function should consider whether to print
	  ;; flat or not, by calling advise-break.  This is a somewhat aesthetic and
	  ;; heuristic decision.

          (t
	   ;; We have collapsed the advice=t and nil cases.  Rather than
	   ;; seperate branches of the cond that are nearly identical, we
	   ;; check advice to set, goto and pop the margin.

	   ;; If it is decided not to print flat, one needs somewhere early to 
	   ;; set a tab stop to which to return.

           (if advice (set-margin))
	   ;; If term is of the form (name a b :c 4)
	   ;; We don't know whether :c is a keyword or 
	   ;; a keyword parameter, so instead of
	   ;; infix-print-keyword-value-pair, we just
	   ;; simply use infix-print-term1  Apr 22 97 Jimmy/MKS 
	   (sloop for tail on (cdr term)
		  do (progn (infix-print-term1 (car tail))
			    (cond ((cdr tail)
				   (if (consp (cdr tail)) 
				       (if (member (car tail)
						   '(&OPTIONAL
						     &REST
						     &KEY
						     &WHOLE
						     &BODY
						     &ALLOW-OTHER-KEYS ))
					   (pprinci " ")
					   (pprinci ", "))
				       (progn (pprinci *dotted-pair-seperator*)
					      (infix-print-term1 (cdr tail))
					      (pprinci ")")
					      (loop-return nil)))
				   ;; Each call of newlinex-to-current-magin will
				   ;; return to the indent we set.
				   (if advice (to-current-margin)))
				  (t (pprinci ")") ))))
	   ;; Now we get rid of the indent.
	   (if advice (pop-margin))))))

;;;((lambda foo(x) x) 3)
;;;(lambda-signature (x y ) (3 1)) => x := 3, y := 1

(defun lambda-signature-printer (arg value)
  (if (and (consp arg) (consp value))
      (progn (infix-print-keyword-value-pair (car arg) (car value))
	     (if (consp (cdr arg))
		 (pprinci ", "))
	     (lambda-signature-printer (cdr arg) (cdr value)))
    nil))


(defun infix-print-keyword-value-pair (key value)
  (progn (infix-print-term1 key)
	 (pprinci " := ")
	 (infix-print-term1 value)))

(defun get-fn-printer (sym)
  (or (symbolp sym)
      (error (format nil "Illegal object where function symbol expected : ~a." sym)))
  (let ((a (assoc sym *fn-alist*)))
    (if a
	(cadr a)
      (function default-fn-printer))))

(defun defun-call-part-printer (term)
  (let ((advice (advise-break term)))
    (cond ((atom (car term))

; First put out the function symbol, in roman.
; Functions of no arguments are printed as constants, without parentheses,
; rather than with, e.g., foo().

           (cond ((null (cdr term))
                  (roman-font (car term))
                  (return-from defun-call-part-printer nil)))

; We want a very tiny space in front of the open parenthesis because otherwise,
; it looks like some function symbols touch the open parenthesis.  In some
; cases, this results in a touch more space than we like, and perhaps there
; exists a more delicate kerning command.

           (roman-font (car term))
           (math-thin-space)
           (pprinci "("))
          (t (pprinci "(")
             (setq term (cons 'foo term))))
    (cond ((null (cdr term))
           (pprinci ")"))
          ((null (cddr term))
	   ;; infix-print-arg takes a list of args and prints the first.
	   ;; This allows it to handle &optional.
           (infix-print-args (cdr term))
           (pprinci ")"))

; The coder of the printer for each function should consider whether to print
; flat or not, by calling advise-break.  This is a somewhat aesthetic and
; heuristic decision.

          (advice

; If it is decided not to print flat, one needs somewhere early to set a tab
; stop to which to return.

           (set-margin)
	   (infix-print-args (cdr term) t)
	   (pprinci ")")
	   (pop-margin))
          (t (infix-print-args (cdr term))
	     (pprinci ")")))))


;                                    BREAKS

(defun advise-break (term)

; advise-break is one of two places where *testing* is bound, and here it is bound
; to t, meaning that we want no printing, just want to know if we can print
; term flat.  We also bind, just to restore, the current *infix-loc* and
; *tab-list*.

; This first cond is only for debugging purposes.  Same for the second value
; of the prog1.

  (cond (*tracing-advise-break*
         (format *terminal-io* "~%Entering *infix-loc* = ~a, *testing* = ~a, margin = ~a~%"
		 *infix-loc* *testing* (get-margin))))
  (prog1
      (let ((*infix-loc* *infix-loc*)
            (*left-margin* *left-margin*)
            (*tab-stack* *tab-stack*)
            (*tab-list* *tab-list*)
            (*left-margin* *left-margin*))
        (cond (*testing* nil)
	      (t (catch 'advise-break
                   (let ((*testing* t))
                     (infix-print-term1 term)
                     nil)))))
    (cond (*tracing-advise-break*
           (format *terminal-io* "~%Exiting *infix-loc* = ~a, margin = ~a~%"
		   *infix-loc* (get-margin))))))

(defvar *tracing-advise-flatsize* nil)

(defun advise-flatsize (term)

; advise-flatsize is the other place that *testing* is bound, and here it is bound
; to 'count, meaning that we want no printing, just want to know what column
; we would end up in if we printed term flat.  We also bind, just to restore, 
; the current *infix-loc* and *tab-list*.

; This first cond is only for debugging purposes.  Same for the second value
; of the prog1.

  (cond (*tracing-advise-flatsize*
         (format *terminal-io* "~%Entering *infix-loc* = ~a, *testing* = ~a~%" *infix-loc* *testing*)))
  (prog1
      (let ((*infix-loc* *infix-loc*)
            (*left-margin* *left-margin*)
            (*tab-stack* *tab-stack*)
            (*tab-list* *tab-list*)
            (*left-margin* *left-margin*))
        (if *testing*
	    nil
	    (let ((count (catch 'advise-break
			   (let ((*testing* 'count))
			     (infix-print-term1 term)
			     *infix-loc*))))
	      count)))
    (cond (*tracing-advise-flatsize*
           (format *terminal-io* "~%Exiting *infix-loc* = ~a~%" *infix-loc*)))))


;; Next 3 from Nqthm.
(defvar *extra-propertyless-symbols-alist* nil)

(defun propertyless-symbolp (x)
  (or (car-cdrp x)
      (member x (quote (nil quote list let case cond t f
                               list*)))
      (assoc x *extra-propertyless-symbols-alist*)))

(defun car-cdrp (x)
  (let ((str (symbol-name x)))
    (cond ((and (> (length str) 2)
                (eql (aref str 0) #\c)
                (eql (aref str (1- (length str))) #\r)
                (sloop for i from 1 to  (- (length str) 2)
                         always (or (eql (aref str i) #\a)
                                    (eql (aref str i) #\d))))
           (sloop for i downfrom (- (length str) 2) to 1 collect
                    (aref str i)))
          (t nil))))

(defun advise-break-if-testing ()

; A printer function that is sure that it will break should short circuit the
; expense of calculating whether printing flat is ok.

  (cond (*testing*
         (throw 'advise-break *infix-loc*))))

(defun do-not-index-call-of (fn)
  (or *do-not-index*
      *do-not-index-calls*
      (propertyless-symbolp fn)
      ;; (eq 'ground-zero (get fn 'main-event))
      (get fn '*predefined*)		;seems appropriate for Acl2.
      (member fn *do-not-index-calls-of*)))
      
(defun index-call (fn &optional kind)
  (cond (*testing* nil)
	(*do-not-index* nil)
        ((do-not-index-call-of fn) nil)
        (t (index fn kind))))

(defun do-not-index-event-of (fn)
  (or *do-not-index*
      ;;seems appropriate for Acl2.
      (get fn '*predefined*) ))

(defun index-event (fn &optional kind)
  (cond (*testing* nil)
	(*do-not-index* nil)
        ((do-not-index-event-of fn) nil)
        (t (index fn kind))))

;; Do we really want to add this?  Or should we
;; just use     index-call  and index-event     to be
;; respectively label-event     reference-event
;; ?????  Apr 24 97 MKS 

(defun label-event (fn)
  (cond (*testing* nil)
	(*do-not-index* nil)
        ((do-not-index-event-of fn) nil)
        (t (define-label fn))))

(defun reference-event (fn)
  (cond (*testing* nil)
	(*do-not-index* nil)
        ((do-not-index-event-of fn) nil)
        (t (reference-label fn))))

(defun print-dot-list(term)
  ;;;(a b c . d)
  ;;;(a . d)
  (cond ((atom (cdr term))
	 (infix-print-term1 (car term))
	 (pprinc " . ")
	 (infix-print-term1 (cdr term)))
	(t (infix-print-term1 (car term))
	   (pprinci ", ")
	   (print-dot-list (cdr term)))))

(Defmacro my-funcall (fn term)
  `(,fn ,term))

(defun infix-print-term1 (term) 
  (cond 
   ((atom term)
    (funcall (get-atom-printer term) term))
	
   ((not (true-listp term))
    ;; If (evisceratep term) just print the cdr 
    (cond ((or (evisceratep term) (evisceratep (Car term))) (evisc-printer term))
	  ((or (atom (cdr term)) (atom (cdr (last (cdr term)))));;check dot list
	   (pprinci *lbracket*)
	   (print-dot-list term)
	   (pprinci *rbracket*))
	  (t (infix-print-term1 (car term))
	     (infix-print-term1 (cdr term)))))
   ((or (evisceratep term) (evisceratep (Car term))) (evisc-printer term))
   ((and (consp term) (keywordp (car term)) (= (length term) 2))
    (pprinci *lbracket*) (print-rune-string term) (pprinci *rbracket*))
   ((and (consp (car term)) (equal (caar term) 'LAMBDA))
    (lambda-printer term))
   ((consp (car term)) 
    (pprinci "(")
    (set-margin)
    (sloop for x on term do (progn (infix-print-term1 (car x)) (if (cdr x) (pprinci " "))))
    (pop-margin)
    (pprinci ")"))
   ((not (symbolp (car term)))
    (pprinci "(")
    (set-margin)
    (sloop for x on term do (progn (infix-print-term1 (car x)) (if (cdr x) (pprinci " "))))
    (pop-margin)
    (pprinci ")"))
   (t
    ;; Feb 19 97 MKS deleted.  Should be redundant.
    ;; (let ((funct (get-fn-printer (car term)))) (if (equal (Car term) 'quote) (Quote-printer term) ... ))
    (funcall (get-fn-printer (car term)) term)
    ;(index-call (car term))
    )))

(defun infix-print-term (term)
  (newlinex)
  (infix-print-term1 term)
  (newlinex)
  nil)

(defun print-list-item (term &optional sep newline)
  (if (eq newline 'check)
      (progn (setq newline nil)
	     (if (symbolp term)
		 (if (> (+ (length (symbol-name term)) *infix-loc*)
			*rightmost-char-number*)
		     (to-current-margin))
		 (if (> *infix-loc* *rightmost-char-number*)
		     (to-current-margin))))
      (if newline (to-current-margin)))
  (infix-print-term1 term)
  (if sep (pprinci sep))
  (if newline (to-current-margin))
  nil)

(defun print-list (l &optional (sep *list-seperator*)
		     (dot *dotted-pair-seperator*)
		     break
		     (printer (function print-list-item)))
  ;; Printer should take args of form (term &optional (seperator " ") newline)
  (let ((newline (or break (advise-break l)))
	(lastitem (last l))
	dotted)
    ;; Does list end in . x
    (if (and lastitem (symbolp (car lastitem)) (not (null (cdr lastitem))))
	(setq dotted t)
	(setq lastitem (car lastitem)))
    (if (and newline (every #'symbolp (cdr l)))
	(setq newline 'check))

    (set-margin)
    (if (not (consp l))
	(progn (pprinci dot) (print-list-item l nil newline))
	(sloop for tail on l
	       do (cond ((consp tail) 
			 (funcall printer (car tail)
				  (if (cdr tail)
				      (if (consp (cdr tail)) sep dot))
				  newline)
			 ;; The dot was handled above.  If the cdr is a symbol
			 ;; print it (with no seperator or eol) and exit loop.
			 (cond ((and (cdr tail) (not (consp (cdr tail))))
				(funcall printer (cdr tail) nil newline)
				(return nil))))
			;; Should never get to this.
			;; Sloop doesn't deal with non-consp cdrs.
			(t (pprinci dot)
			   (print-list-item tail nil 
					    (if (not (equal (car tail) (car (last l))))
						newline nil))))))
    ;; Now we get rid of the indent.
    (pop-margin)))


;;   &optional var || &optional (var [literal [varp]]) 
;;   &rest v
;;   &key foo bar
;;   &whole l
;;   &body l
;;   &allow-other-keys

(defvar infix-lambda-list-keywords
  '(&optional &rest &key &whole &body &allow-other-keys))

(defun infix-print-args (l &optional terpri-flag)
  ;; Whether the args will fit was decided by the caller.
  ;; If not, terpri-flag is T and we emit a newline between args.
  ;; Margin is set at current position.
  (sloop for tail on l
	 do (let ((arg (car tail)))
	      (cond ((not (member arg infix-lambda-list-keywords))
		     (cond ((symbolp arg) (italic-sym-printer arg))
			   ((and (consp arg) (equal (length arg) 2))
			    (italic-sym-printer (car arg))
			    (math-space 1)
			    (pprinci ":=")
			    (math-space 1)
			    (infix-print-term1 (cadr arg)))
			   ((and (consp arg) (equal (length arg) 3))
			    (italic-sym-printer (car arg))
			    (math-space 1)
			    (pprinci ":=")
			    (math-space 1)
			    (infix-print-term1 (cadr arg))
			    (math-space 1)
			    (pprinci " flag ")
			    (math-space 1)
			    (infix-print-term1 (caddr arg)))
			 (t (format t "Unrecognized argument type ~s" arg)))
		     (if (cdr tail) (pprinci ", ")))
		  (t (small-caps-sym-printer arg) (pprinci " ")))
	      (if terpri-flag (line-return)))) ;newlinex
  )

;; Let set-margin and set-tab know if they are in a context where
;; we are using a left margin followed by a tab.
(defvar *left-margin-tab-context* nil)

;; This seems bogus.  But in the interest of backward compatiblity...
;; Ascii mode sets it to nil.
(defvar *op-length* 3)

;; PRECEDENCE

(defun precedence (term)
  (cond ((not (consp term)) 80)      ;;80 number larger than precedence 
	((not (symbolp (car term))) 80)
	((get-operator-precedence (car term)))
	(t 80)))

(defun get-operator-precedence (op)
  (and (symbolp op)
       (get op 'precedence)))

(defun parens-print-term1 (term)
    (pprinci "(")
    (infix-print-term1 term)
    (pprinci ")"))

(defun make-1/x (term)
   (if (not (consp term))
	term
     (if (equal (car term) :evisceration-mark)
	 term
	 ;; Added true-lisp check to guard length.
       (if (and (equal (car term) '/) (true-listp term) (= (length term) 2))
	   (cons (car term) (cons 1 (rest term)))
	 term))))

(defun move-left-one-space? (name)
  (member name '(let let* mv-let)))

(defconstant *right-associative-operators*
  '(CONS IFF IF COND IMPLIES CASE CASE-MATCH ELIF CASE APPEND AND OR + *))

(defun print-term-according-to-precedence (prec term &optional (which-term 'left))
  (let* ((op-prec (precedence term))
	 (term (make-1/x term))      ;;if term is (/ x)
	 (advice (advise-break term)))
    ;; (CAR TERM) in move-left-one-space? was being called on atoms, e.g. (CAR #\l)
    (if (and advice (consp term) (move-left-one-space? (car term)))
	(setq *infix-loc* (+ *infix-loc* 1)))
    ;; The first cond is not really necessary, but its clearer.
    ;; If term in not consp, then its precedence should be maximum.
    (cond ((not (consp term))  (infix-print-term1 term))
	  ((evisceratep term)  (infix-print-term1 term))
	  ((= (length term) 2) (infix-print-term1 term))
	  ((= prec op-prec)
	   (cond ((equal which-term 'right)
		  ;; (if advice (setq *infix-loc* (+ *infix-loc* 1)))
		  (parens-print-term1 term))
		 ((member (car term) *right-associative-operators*);;right assoc
		  (cond ((equal which-term 'left)              
			 ;;(if advice (setq *infix-loc* (+ *infix-loc* 1)))
			 (parens-print-term1 term))
			(t (infix-print-term1 term))))
		 (t (infix-print-term1 term))))
	  ((< op-prec prec)
	   ;;(if advice (setq *infix-loc* (+ *infix-loc* 1)))
	   (parens-print-term1 term))
	  (t (infix-print-term1 term)))))


;; WE CAN'T USE THIS SPECIAL CASE STUFF FOR /!!!
;; Prints (/ (/ (/ 1 2) 3) 4) as 1/2/3/4 which parses as
;; (/ 1/2 3/4)
;; Apr  4 97 MKS 

(defun print-op (op)
  (cond ((equal op "/")	 (pprinci op))
	(t (pprinci " ") (pprinci op) (pprinci " "))))

(defun pre-op (op)
  (cond ((equal op "/")	 )
	(t (pprinci " "))))

(defun default-infix-printer (term op &optional opsize)
  ;; Note that this INFIX printer expects at least two arguments.
  ;; One problem is the width of OP, which is used in (SET-TAB OP).
  ;; The op in latex mode is typically something grotesque, like
  ;; "\\rightarrow" or "{\\rm\\bf{less}}", whose string length is
  ;; immaterial.  Optional OPSIZE input provides a mech. to deal
  ;; with this.
  (let* ((advice (advise-break term))
	 (op-prec (precedence term))
	 (term1 term))
    ;; We hate to see (member x (foo ...)) broken right after the x, 
    ;; where x is an atom.
    (cond ((= (length term) 2)
	   ;; We have captured a unary op.  Since 
	   ;; we assume these will behave like +/-, e.g. if 
	   ;; (op x y) prints as "x O y" then (op x) prints as "O x".
	   (default-unary-prefix-printer term  op))                 
	   ;; NORMAL CASE.
          (t
           ;; Normalization based on ACL2 default interpretation.
	   ;; (append a (append b (append c d)) => (append a b c d)
	   (if (and (= (length term) 3) (member (car term) '(+ * append and or)))
	       (setf term1 (cons (car term) (unary-op-transformation term (car term)))))
	   ;;
	   (if advice (progn (set-margin) (set-tab (or opsize op))
			     ;; (pre-op op) - Special case for / doesn't work
			     ;; The pprinci prints a spurious space.
			     ;; I think.  Apr 25 97 MKS 
			     ;; (pprinci " ")
			     ))

	   (print-term-according-to-precedence op-prec (cadr term1) 'left)
	   (if advice
	       (to-current-margin)
	       (pprinci " "))

	   (sloop for tail on (cddr term1)
		  do (progn 
		       ;; (print-op op) - Special case for / doesn't work
		       (pprinci op opsize) (pprinci " ")
		       (if advice (do-tab))
		       (print-term-according-to-precedence op-prec
							   (car tail)
							   (if (null (cdr tail))
							       'right 'left))
		       ;; The comparison below was equal.
		       ;; BUT EQ is faster.
		       (if (and advice (not (eq tail (last term1))))
			   (to-current-margin))
		       (if (not advice) (pprinci " "))))
	  ;; EVERY set-margin needs a corresponding pop-margin !!!
	  ;; Feb 19 97 MKS 
	  (if advice (pop-margin))
	  ))))


;;;(+ x (+ y (+ z m))) => (+ x y z m)

(defun unary-op-transformation (term op)
  (let* ((lt (cadr term)) (rt (caddr term)))
    (cons lt (unary-op-transformation2 rt op))))

(defun unary-op-transformation2 (term op)
  (cond ((consp term)
	 (cond ((and (eq (car term) op) (> (len term) 2))
		(unary-op-transformation term op))
	       (t (list term))))
	(t (list term))))
	       
(defun find-which-term (x term)
  (if (equal x (car (last term)))
      'right
   'left))



; See `SETTINGS THAT MAY BE MODIFIED IN MODE-SYNTAX.LISP' for a description
; of the difference between these two modes of printing.

(defun boyer-infix-print-tail (op args &optional (op-prec 0))
  (set-tab)
  (sloop for tail on args
	 do (cond ((cdr tail)
		   (print-term-according-to-precedence op-prec (car tail))
		   (newlinex) (do-tab)	;force-newline
		   (math-space 5)
		   (pprinci op *op-length*)
		   (newlinex) (do-tab))	;force-newline
		  (t (print-term-according-to-precedence op-prec (car tail))
		     (pop-tab)))))

;; ( + a b c (* d e) (+ f g) (+ m n))
;; a + b + d * e + ( f + b ) + ( m + n )

(defun smith-infix-print-tail (op args &optional (op-prec 0))
  ;; Does this assume we are in a tabbing env?
  (set-margin)
  (set-tab op)
  (print-term-according-to-precedence op-prec (car args))
  (sloop for tail on (cdr args)
	 do
	 (progn (to-current-margin)
		(pprinci op)
		(do-tab)
		(print-term-according-to-precedence op-prec (car tail))
		(if (cdr tail)
		    (to-current-margin))))
  
  (pop-margin))

;; Odd unary cases
;; (and x) x
;; (or  x) x
;; (+ x)   (+ x)
;; (- x)   (- x)
;; (/ x)   1 / x

(defun default-unary-prefix-printer (term op)
  (let ((op-prec (precedence term)))          
    (cond ((equal (car term) '/) ;;(/x) => (/ 1 x)
	   (default-infix-printer (cons (car term) (cons 1 (rest term))) op))
	  ((member (car term) '(and or))
	   (infix-print-term1 (cadr term)))
	  ;; WARNING: Implementation dependency re text formatters.
	  ;; We assume a bare '+' will print ok.
	  ;; MKS/JW - 4/24/97
	  ((eq (car term) '+)
	   (pprinci "(+ ")
	   (infix-print-term1 (cadr term))
	   (pprinci ")"))
	  (t (pprinci op)
	     (cond ((and (= (precedence (cadr term)) op-prec) ;;(op x)
			 (consp (cadr term))
			 (= (length (cadr term)) 2))          ;; ??? >>> x in unary 
		    (infix-print-term1 (cadr term)))
		   ((consp (cadr term))                       ;;(op x), x is a list
		    (parens-print-term1 (cadr term)))
		   (t (infix-print-term1 (cadr term))))))))
    

(defun default-infix-multiple-printer (term strs)
  (let* ((advice (advise-break term)))
    (if (eq *infix-op-location* 'FRONT)
        (smith-infix-multiple-printer term strs advice)
        (boyer-infix-multiple-printer term strs advice))))

(defun boyer-infix-multiple-printer (term strs advice)
    (set-tab)
    (infix-print-term1 (cadr term))
    (sloop for arg in (cddr term)
             as str in strs do
             (cond (advice (newlinex) (do-tab)) ;force-newline
                   (t (pprinci " ")))
             (pprinci str)
             (cond (advice (newlinex) (do-tab)) ;force-newline
                   (t (pprinci " ")))
             (infix-print-term1 arg))
    (pop-tab))

(defun smith-infix-multiple-printer (term strs advice)
  (set-margin)
  (infix-print-term1 (cadr term))
  (sloop for arg in (cddr term)
           as str in strs 
           do (progn (cond (advice (to-current-margin))	
                           (t (pprinci " ")))
                     (pprinci str)
                     (pprinci " ")
                     (infix-print-term1 arg)))
  (pop-margin))

(defun default-prefix-multiple-printer (term strs)
  (let* ((advice (advise-break term)))
    (set-margin)
    (sloop for tail on (cdr term)
             as str in strs
             do (progn (pprinci str)
                       (pprinci " ")
                       (infix-print-term1 (car tail))
                       (cond ((null tail) nil) 
                             (advice (to-current-margin)) ;newlinex
                             (t (pprinci " ")))))
    (pop-margin)))

(defun default-suffix-multiple-printer (term strs)
  (let* ((advice (advise-break term)))
    (set-margin)
    (sloop for tail on (cdr term)
             as str in strs
             do (progn (infix-print-term1 (car tail))
                       (pprinci str)
                       (cond ((null (cdr tail)))
                             (advice (to-current-margin)) ;newlinex
                             (t nil))))
    (pop-margin)))

(defun default-unary-suffix-printer (term op)
  (let ((op-prec (precedence (car term))))
    (set-margin)
    (if (< (precedence (cadr term)) op-prec)
	(parens-print-term1 (cadr term))      
	(infix-print-term1 (cadr term)))
    (pprinci " ")
    (pprinci op)
    (pop-margin)))
        
(defun default-unary-abs-printer (term lhs-str rhs-str)
    (set-margin)
    (pprinci lhs-str)
    (infix-print-term1 (cadr term))
    (pprinci rhs-str)
    (pop-margin))

;                      PRINTING INDEX and XREF ENTRIES

; Who could ever have guessed that it would take this much code to print out a
; simple \index{foo} command for an arbitrary Acl2 function symbol foo.  There
; are so many special cases one can hardly believe one's eyes.

; Should be re-defined in mode-init.lisp, e.g. latex-init.lisp.

(defvar index-subitem-length 30)

(defun index (x &optional subkind)
  ;; Subkind must be string or atom.
  (if *do-not-index*
      nil
      (progn (pprinc *begin-index*)
	     (cond ((stringp x) (print-string x 0))
		   (t (print-atom x 0)))
	     ;; WARNING: Latex/Scribe dependency.
	     (if subkind
		 (cond ((stringp subkind) (pprinc ", ") (print-string subkind 0))
		       ((symbolp subkind) (pprinc ", ") (print-atom subkind 0))
		       (t nil)))
	     (pprinc *end-index*))))

;; Not yet certain we need these.
;;  Apr 24 97 MKS 
;; YES.  !x and !n

(defun define-label (x)
  (if *do-not-index*
      nil
      (progn (pprinci *begin-label* 0)
	     (cond ((stringp x) (print-string x 0))
		   (t (print-atom x 0)))
	     (pprinci *end-label* 0))))

(defun reference-label (x)
  (if *do-not-index*
      nil
      (progn (pprinci *begin-reference* 0)
	     (cond ((stringp x) (print-string x 0))
		   (t (print-atom x 0)))
	     (pprinci *end-reference* 0))))


;                                EVENT PRINTERS

(defvar *local-context* nil

  "Indicates when we are in a `local' context, e.g. in encapsulate.")

(defvar *local-handlers* nil 
  "List of events that are sensitive to *local-context*")
;; Every -printer that calls event-label printer is so sensitive.

(defmacro local-print (x)
  `(or (member ',x  *local-handlers*)
       (setq *local-handlers* (cons ',x *local-handlers*))))

(defvar *minimize-event* nil)

(defun event-doc-string (doc)
  (let (pair)
    (cond ((or *minimize-event*
	       (null doc)
	       (not (stringp doc))
	       (string= doc "")))
	  ((and *acl2-format-doc-strings* 
		(progn (setq pair (acl2-parse-string doc))
		       ;; Car pair is nil if successfull
		       (not (car pair))))	
	   (begin-text)
	   ;; DON'T replicate special char quoting at this point.
	   (pprinc (cdr pair))
	   (end-text)
	   (blankline))
	  (t (begin-text)
	     ;; PPRINC is NOT the right thing below because of special chars?
	     ;; Defined print-text-from-string - Apr 18 97 MKS 
	     ;; HACK.  Seems to add an extra linefeed on end of doc string.
	     ;; SO DELETE IT.
	     (let ((len (length doc)))
	       (if (and (> len 1) (char= (aref doc (- len 1)) #\newline))
		   (print-text-from-string (subseq doc 0 (- len 1)))
		   (print-text-from-string doc)))
	     (end-text)
	     (blankline)))))

(defun print-default-event-header ()
  (ppformat *print-default-event-header*)
  (reset-loc-and-margin 0))

(defun no-tab-event-trailer ()
  (ppformat *no-tab-event-trailer*))

(defun print-default-command-header ()
  (ppformat *print-default-command-header*)
  (reset-loc-and-margin 0))

(defun reset-loc-and-margin (&optional (n 0))
  (cond (*local-context* nil)
	((equal n 0)
	 (setq *tab-list* nil)
	 (setq *tab-stack* nil)
	 (setq *tabs-in* 0)
	 (setq *infix-loc* 0)
	 (setq *left-margin* 0))
	(t (setq *infix-loc* n)
	   (set-margin))))

(defun no-tab-command-trailer ()
  "May occur in a tabbing env. due to an encapsulate.
In which case we don't want to print *no-tab-command-trailer*."
  (if *tab-stack* (blankline) (ppformat *no-tab-command-trailer*)))

(defun event-label-printer (label &optional lowercase)
  (if *local-context* (small-caps-sym-printer "local "))
  (small-caps-sym-printer label)
  (if lowercase (pprinc lowercase)))

(eval-when (load compile eval)

(defvar *suppress-trailing-semi-colon* nil)
(defvar *acl2-in-output* nil)

(defun begin-event (&optional doc) 
  ;; The first test really means we are in a context which
  ;; will handle the printing of whitespace around the form,
  ;; so that this is a noop.
  (cond (*suppress-trailing-semi-colon* nil)
	(*minimize-event* (to-current-margin))
	(*acl2-in-output* (set-margin))
	(t (blankline)
	   (begin-tabbing)
	   (if doc (event-doc-string doc)))))

;(defmacro end-event ()
;  '(progn (or *suppress-trailing-semi-colon* (pprinci ";"))
;	  (if *minimize-event* nil (progn (end-tabbing) (blankline)))))

(defun end-event ()
  (cond (*suppress-trailing-semi-colon* nil)
	(t (pprinci ";")
	   (cond (*minimize-event* nil)
		 (*acl2-in-output* (pop-margin))
		 (t (end-tabbing) (blankline))))))

(defmacro begin-paragraph () 
  `(progn (blankline) (begin-text)))

(defmacro end-paragraph () 
  `(progn (end-text) (blankline)))

)

;; acl2    == segment*
;; segment == text blankline || event blankline
;; event   == [ doc blankline ] begin-event body [ keys ] end-event
;; body    == type header = form || type header acl2

(defun default-event-printer (event)
  (begin-event)
  (if *local-context* (event-label-printer "Local "))
  ; (quote-printer1 event) 6/4/96 swapped out. MKS
  (infix-print-term1 event)
  (end-event))

(defvar *defining-events* nil "Defined in event-printers.lisp")

(defvar *prefix-events-with-arguments*
  '(defun def defmacro defstub signature defun-sk defloop))

(defvar *begin-event-name* "")
(defvar *end-event-name* "")

(defun begin-event-name () (princ *begin-event-name*))
(defun end-event-name () (princ *end-event-name*))

(defun event-name-printer (x)
  (princ *begin-event-name*) (print-atom x) (princ *end-event-name*))

(defun prefix-event-printer (event)
  (cond ((not (consp event)) (print-atom event))
	((and (not *print-top-locals*) (eq (car event) 'local)))
	((and (not *print-verify-guards*)
	      (member (car event) '(verify-guards verify-termination))))
	(t (let ((event-name (car event))
		 (name (cadr event))
		 (rest (cddr event))
		 (*print-pretty* t))
	     (terpri)
	     (princ *begin-verbatim-env*)
	     (cond ((member event-name *defining-events*)
		    (princ "(")
		    (event-name-printer event-name)
		    (princ " ")
		    (print-function-name name)
		    (princ " ")
		    (cond ((member event-name *prefix-events-with-arguments*)
			   (prin1 (car rest))
			   (setq rest (cdr rest))))
		    (sloop while rest
			   do
			   (let ((x (car rest)))
			     (if (keywordp x)
				 (if *print-keys*
				     (if (or *print-hints*
					     (not (eq x :hints)))
					 (progn (terpri)
						(princ " ")
						(prin1 x)
						(setq rest (cdr rest))
						(if rest
						    (progn (princ " ")
							   (prin1 (car rest))
							   (setq rest (cdr rest)))))
					 (setq rest (cddr rest)))
				     (progn (setq rest (cdr rest))
					    (if rest (setq rest (cdr rest)))))
				 (progn (terpri)
					(princ " ")
					(prin1 x)
					(setq rest (cdr rest))))))
		    (princ ")"))
		   (t (print event))))
	   (princ *end-verbatim-env*))))

(defun extract-keywords (l keys)
  (cond ((not (consp keys)) nil)
	((member (car keys) l)
	 (cons (cadr (member (car keys) l)) 
	       (extract-keywords l (cdr keys))))
	(t (cons nil (extract-keywords l (cdr keys))))))

;; 7/1/96: Jimmy
;; (extract2 '(:stop a :scheme b) '(:stop :scheme))
;; => ((:Stop a) (:schemem b))

(defun extract-keywords2 (l keys)
  ;; Includes the keyword in the result.
  (let* (;(result (member (car keys) l))
	 )
    (cond ((not (consp keys)) nil)
	  ((member (car keys) l)
	   ;; The cons below was (list (car keys) (cadr (member (car keys) l)))
	   ;; But I think we want cons.  mks.
	   ;; 9/19 if b = (:stop ((...) (...))) using cons is (:stop (...) (...))
	   ;; so if b is alist of lists, we need to use list ????
	   (cons (cons (car keys) (Cadr (member (car keys) l)))
		   (extract-keywords2 l (cdr keys))))
	   (t (cons nil (extract-keywords2 l (cdr keys)))))))

;; '(0 (a b) c) => ((a b c))
;; '(0 0 c) => nil
(defun list-of-lists? (l)
  (member-if #'consp l))


(load-base "event-printers")

(defvar *save-event-printer-alist* *event-printer-alist*)

(defun get-event-printer (sym)
  (or (symbolp sym)
      (error (format nil "Illegal object where event symbol expected : ~a." sym)))
  (let ((a (assoc sym *event-printer-alist*)))
    (if a
	(cadr a)
      (function default-event-printer))))

;;                          KEYWORD PRINTERS

(defvar *keyword-printer-alist* nil)

;; Argument to keyword printers is a pair (key . value)

(defun get-keyword-printer (sym)
  (let ((a (assoc sym *keyword-printer-alist*)))
    (cond (a (cadr a))
          (t (function default-keyword-printer)))))

(defun default-keyword-printer (pair)
  (ppformat " ~a is ~a" (car pair) (cdr pair)))

(defun standard-co-keyword-printer (pair)
  ;; "foo.out"
  (ppformat " standard output to ~a" (cdr pair)))

(defun proofs-co-keyword-printer (pair)
  ;; "proofs.out"
  (ppformat " proof output to ~a" (cdr pair)))

(defun current-package-keyword-printer (pair)
  ;; 'acl2
  (let ((name (if (equal (car (cdr pair)) 'quote)
		  (cadr (cdr pair))
		  (cdr pair))))
    (ppformat " in the package ~a" name)))

(defun ld-skip-proofsp-keyword-printer (pair)
  ;; 'include-book
  (cond ((null (cdr pair))
	 (ppformat " doing proofs"))
	((equal (cdr pair) t)
	 (ppformat " skipping termination proofs"))
	((equal (cdr pair) '(quote include-book))
	 (ppformat " assuming proofs in included books"))
	((equal (cdr pair) '(quote include-book-with-locals))
	 (ppformat " assuming proofs in included books and executing local events"))
	((equal (cdr pair) '(quote initialize-acl2))
	 (ppformat " skipping local proofs"))
	(t nil)))

(defun ld-redefinition-action-keyword-printer (pair)
  ;; nil
  (cond ((null (cdr pair))
	 (ppformat " redefinition prohibited"))
	(t (let ((a (car (cdr pair)))
		 (b (cdr (cdr pair))))
	     (cond ((eq a ':query)
		    (ppformat " query on redefinition "))
		   ((eq a ':warn)
		    (ppformat " warn on redefinition"))
		   ((eq a ':doit)
		    (ppformat " redefinition ok"))
		   (t			; :warn! :query!
		    (ppformat " redefinition ~a" a)))
	     (cond ((eq b ':erase)
		    (ppformat " erase properties on redefinition "))
		   ((eq a ':overwrite)
		    (ppformat " overwrite properties on redefinition"))
		   (t
		    (ppformat " redefinition properties ~a" a)))))))

(defun ld-prompt-keyword-printer (pair)
  ;; t
  (cond ((null (cdr pair))
	 (ppformat " no prompt"))
	((eq (cdr pair) t)
	 (ppformat " default prompt"))
	(t
	 (ppformat " prompt function = ~a" (cdr pair)))))

(defun ld-keyword-aliases-keyword-printer (pair)
  ;; pair = (:ld-keyword-aliases '((:q 0 q-fn) (:e 0 exit-acl2-macro)))
  (let ((x (cdr pair)))
    (if (equal (car x) 'quote)
	(setq x (cadr x)))
    (ppformat " define keyword aliases (")
    (sloop for tail on x do
	   (cond ((cddr tail)
		  (ppformat "~s = ~a of arity ~d, "
			    (car (car tail)) (caddr (car tail)) (cadr (car tail))))
		 ((cdr tail)
		  (ppformat "~s = ~a of arity ~d and "
			    (car (car tail)) (caddr (car tail)) (cadr (car tail))))
		 (t (ppformat "~s = ~a of arity ~d"
			      (car (car tail)) (caddr (car tail)) (cadr (car tail))))))
    (ppformat ")")))

(defun ld-pre-eval-filter-keyword-printer (pair)
  ;; :all
  (cond ((null (cdr pair)))
	((equal (cdr pair) ':all)
	 (ppformat " every form read is evaled"))
	((equal (cdr pair) ':query)
	 (ppformat " user queried whether form read is evaled"))
	(t nil)))

(defun ld-pre-eval-print-keyword-printer (pair)
  ;; nil
  (if (cdr pair)
      (ppformat " forms read are printed")
      (ppformat " forms read are not printed")))

(defun ld-post-eval-print-keyword-printer (pair)
  ;; :all
  (cond ((null (cdr pair))
	 (ppformat " results are not printed"))
	((equal (cdr pair) t)
	 (ppformat " results are printed"))
	((equal (cdr pair) ':command-conventions)
	 (ppformat " conditionally print error triples"))
	(t nil)))

(defun ld-evisc-tuple-keyword-printer (pair)
  ;; '(alist nil nil level length)
  (let (value)
    (cond ((null (cdr pair)) (setq value nil))
	  ((not (consp (cdr pair))) (setq value (cdr pair)))
	  ((equal (car (cdr pair)) 'quote) (setq value (cadr (cdr pair))))
	  (t (setq value (cdr pair))))
    (cond ((null value)
	   (ppformat " print results fully"))
	  (t (let ((alist (cadr value))
		   (level (car (cdddr value)))
		   (length (car (cdr (cdddr value)))))
	       (ppformat " eviscerate results (level = ~d, length = ~d)" level length)
	       (if alist
		   (progn (ppformat "whose cars are in (" )
			  (sloop for tail on alist do
				 (cond ((cddr tail)
					(ppformat"~a, " (car (car tail))))
				       ((cdr tail)
					(ppformat " ~a, and" (car (car tail))))
				       (t (ppformat " ~a)" (car (car tail)))))))))))))

(defun ld-error-triples-keyword-printer (pair)
  ;; t
  (cond ((null (cdr pair))
	 (ppformat " errors printed as triples"))
	((equal (cdr pair) t)
	 (ppformat " errors roll back to state before call"))
	(t nil)))

(defun ld-error-action-keyword-printer (pair)
  ;; :return
  (cond ((null (cdr pair)))
	((equal (cdr pair) ':continue)
	 (ppformat " continue after errors"))
	((equal (cdr pair) ':return)
	 (ppformat " return after errors"))
	((equal (cdr pair) ':error)
	 (ppformat " signal errors"))
	(t nil)))

(defun ld-query-control-alist-keyword-printer (pair)
  ;; nil
  (cond ((null (cdr pair))
	 (ppformat " handle queries interactively"))
	((equal (cdr pair) 't)
	 (ppformat " queries default to first accepted response"))
	(t
	 (ppformat "queries default according to ~s"  (cdr pair)))))

(defun ld-verbose-keyword-printer (pair)
  ;; nil
  (cond ((null (cdr pair))
	 (ppformat"verbose mode off "))
	(t
	 (ppformat " verbose mode on"))))

(defparameter *keyword-printer-alist*
  (list (list ':standard-co (function standard-co-keyword-printer))
	(list ':proofs-co (function proofs-co-keyword-printer))
	(list ':current-package (function current-package-keyword-printer))
	(list ':ld-skip-proofsp (function ld-skip-proofsp-keyword-printer))
	(list ':ld-redefinition-action (function ld-redefinition-action-keyword-printer))
	(list ':ld-prompt (function ld-prompt-keyword-printer))
	(list ':ld-keyword-aliases (function ld-keyword-aliases-keyword-printer))
	(list ':ld-pre-eval-filter (function ld-pre-eval-filter-keyword-printer))
	(list ':ld-pre-eval-print (function ld-pre-eval-print-keyword-printer))
	(list ':ld-post-eval-print (function ld-post-eval-print-keyword-printer))
	(list ':ld-evisc-tuple (function ld-evisc-tuple-keyword-printer))
	(list ':ld-error-triples (function ld-error-triples-keyword-printer))
	(list ':ld-error-action (function ld-error-action-keyword-printer))
	(list ':ld-query-control-alist (function ld-query-control-alist-keyword-printer))
	(list ':ld-verbose (function ld-verbose-keyword-printer))))


;                           COPY COMMENTS

;; MKS Tue Jul 13 1993
;; Make ;- comments and BAR-COMMENTs followed by a dash appear in
;; a formatted (in Latex, verbatim) environment.

(defvar *current-environment* nil)

(defvar *comment-semi-net* nil)
(defvar *comment-lb-net* nil)

;; We use this two stage template/eval kludge so that if the template can 
;; be used to reset the variable after the user has defined these variables.

;; Note a problem with this flexibility.  In some contexts we need to format 
;; special characters and in others we don't.  Thus in LaTeX, in a Verbatim
;; we don't need to quote `_', but in running text we do need to quote it.

;; Comment actions in (text format verbatim emphasis comment section)

(defvar *comment-environment-mapping-template*
  (quote `((text     "" "")
	   (format   ,*begin-format-env*   ,*end-format-env*)
           (verbatim ,*begin-verbatim-env* ,*end-verbatim-env*)
           (emphasis ,*begin-emphasis-env* ,*end-emphasis-env*)
           (comment  ,*begin-comment-env*  ,*end-comment-env*)
	   (section  ,*begin-section-env*  ,*end-section-env*))))

(defvar *comment-environment-mapping*
  (eval *comment-environment-mapping-template*))

(defvar *saved-whitespace* nil)

(defun add-saved-whitespace (form)
  ;; Form is string or char
  (setq *saved-whitespace*
	(append *saved-whitespace* (list form))))

(defun print-saved-whitespace ()
  (sloop for l on *saved-whitespace*
	 do (let (skip (c (car l)))
	      (cond (skip (setq skip nil))
		    ((stringp c)
		     (setq skip nil)
		     (pprinc c))
		    ((and (characterp c) (char-equal c #\Newline))
		     ;; The setting of infix-loc may be bogus.
		     ;; Was it previously set by the "virtual" linefeed
		     ;; saved earlier?
		     (setq skip nil)
		     (if (and (cdr l)
			      (char-equal (cadr l) #\Newline))
			 (progn (blankline) (setq skip t))
			 (to-current-margin)))
		    (t
		     (setq skip nil)
		     (pwrite-char c)))))
  (setq *saved-whitespace* nil))				     

(defun wrap-up-copy-comments ()
  ;; MOD - X
  (end-environment)
  (print-saved-whitespace)
  (throw 'copy-comments nil))

(defun begin-environment (env)
  (setq *current-environment* env)
  (ppformat
           (or (cadr (assoc env *comment-environment-mapping*)) "")))

(defun end-environment ()
  (if *current-environment*
      (ppformat (or (caddr (assoc *current-environment*
				  *comment-environment-mapping*))
		    "")))
  (setq *current-environment* nil))

(defun end-environment-and-write (c)
  (end-environment)
  (pwrite-char c))

(defun pop-environment-write-and-push (string)
  (let ((saved-env *current-environment*))
    (if (not saved-env)
        (pprinc string)
        (progn (end-environment)
               (pprinc string)       
               (begin-environment saved-env)))))

(defun end-environment-if-not (env)
  (if (not (equal *current-environment* env))
      (end-environment)
      (print-saved-whitespace)))

(defun white-space-string-p (s)
  (let ((x t))
    (sloop for i from 0 to (- (length s) 1) 
             do (setq x (and x (member (char s i) *white-space*))))
    x))

(defun insert-newpage ()
  (pop-environment-write-and-push (pformat nil "~%~a~%" *newpage*)))

(defun verbatim*-print-term (term)
  (if *hands-off*
      (progn (princ *begin-verbatim-env*) (print term) (princ *end-verbatim-env*))
      (quote-printer1 term)))
  
(defun verbatim-print-term (term)
  (let ((*do-not-use-tabs* t))
    (if *hands-off*
	(progn (princ *begin-verb-env*)
	       (prin1 term)
	       (princ *end-verb-env*))
	(quote-printer1 term))))

;; (defvar *format-tag-char* #\~)

(defun check-environment-and-write (env ch)

  ;; Note: Latex causes an error on an empty VERBATIM environment, so we watch out
  ;; for that as a special case.  Also, Latex forbids control-l in a verbatim
  ;; environment, so we watch out for that, too.

  ;; Thus, we forbid any empty environments, and we always pull a page break 
  ;; out of the current environment.

  ;; First, end an existing environment if it is not ENV.

  (end-environment-if-not env)

  (cond ((not *current-environment*)    ; We are in no environment.  Enter one.
         (cond ((and ch (not (member ch *white-space*)))
		;; Assuming *saved-whitespace* can't intersect
		;; characters handled by (handle-special-chars-in-string ch)
		(print-saved-whitespace)
                (begin-environment env))

;; Does this work? Printing of saved whitespace complicated by the
;; verbatim restrictions of latex.  I.e., can't have an empty verbatim.

               ((and ch (eql ch #\Page))
		(ppformat t "~%~a~%" *newpage*))
               (ch (setq *saved-whitespace*
                        (append *saved-whitespace* (list ch)))))))
  ;; Tabs
  (cond ((and (equal env 'verbatim)
              (eql ch #\Tab) 
              (not *reported-tabs*))
         (setq *reported-tabs* t)
         (pformat *terminal-io* 
                  "WARNING: about tabs!~%We note the presence of a tab ~
                  in a comment that we are copying~%into a verbatim ~
                  environment.  Tabs will be treated just like ~%single spaces ~
                  by Latex.  Consider removing all tabs, e.g., ~%with the ~
                  Emacs command M-x untabify.~%")))

  ;; Print it or not, checking #\Page.
  (cond (*saved-whitespace*)
	((null ch))
        ((and (eql ch #\Page) (equal env 'verbatim))
         (end-environment)
         (ppformat "~%~a~%"  *newpage*)
         (begin-environment env))               
	(*hands-off* (pwrite-char ch))
        (ch
	 ;; Switched (pwrite-char ch)
	 (handle-special-chars-in-string ch))))

(defun unread-chars (l stream)
  (sloop for c in l do (unread-char c stream)))

(defun read-char-in-comment ()
       (let ((c (read-char *standard-input* nil a-very-rare-cons)))
         (cond ((eq c a-very-rare-cons)
                ;; EOF.  We are no longer in a comment.  
                ;; Exit whatever environment we are in, which will be either
                ;; verbatim, format, or none.
                (wrap-up-copy-comments))
               (t c))))

(defun copy-comments-read-net (net)
  ;; Returns (env char), where env is the environment to 
  ;; enter and char is nil or a char to unread.
  ;; Net is cdr of a net whose car = #\;
  ;; Have already read a #\; 
  ;;
  ;; Test (progn (read-char)(COPY-COMMENTS-READ-NET *comment-semi-net*))
  ;;
  (let* ((subnet (car net))
         (action (cadr net))
	 ;; on EOF does a throw out once it cleans up.
         (c (read-char-in-comment)) 
         (branch (assoc c subnet)))
    (cond ((null branch) 
           (unread-char c *standard-input*)
           (list (car action) (cadr action)))
          (t (copy-comments-read-net (cdr branch))))))

(defvar number-deep nil "Measure of depth of imbedding in \#\| \|\# comments")

(defvar *handling-newlines-in-lb-comment* nil)

(defun normalize-lb-comment-line (str action)
  (let ((beg (search "#|" str))
	(end (search "|#" str))
	(*handling-newlines-in-lb-comment* t))
    (cond ((and beg end)
	   (cond ((< beg end)
		  (normalize-lb-comment-line
		   (concatenate 'string
				(subseq str 0 beg)
				(subseq str (+ beg 2) end)
				(subseq str (+ end 2)))
		   action))
		 (t (format-comment-string (subseq str 0 end) action)
		    (decf number-deep 1)
		    (cond ((= number-deep 0)
			   (end-environment)
			   (unread-chars (coerce (subseq str (+ end 2)) 'list)
					 str))
			  (t (normalize-lb-comment-line (read-line *standard-input*) action))))))
	  (beg (incf number-deep 1)
	       (format-comment-string (subseq str 0 beg) action)
	       (normalize-lb-comment-line (subseq str (+ beg 2)) action))
	  (end (format-comment-string (subseq str 0 end) action)
	       (decf number-deep 1)
	       (cond ((= number-deep 0)
		      (end-environment)
		      (subseq str (+ end 2)))
		     (t (normalize-lb-comment-line (subseq str (+ end 2)) action))))
	  (t (format-comment-string str action)
	     (pwrite-char #\Newline)
	     (normalize-lb-comment-line (read-line *standard-input*) action)))))

(defun format-comment-string (line mode)
  ;; Mode in (text format verbatim emphasis comment section)
  (let (pair)
  (cond ((and *acl2-format-comments*
	      (search "~" line)
	      (progn (setq pair (acl2-parse-string line))
		     (not (car pair))))	;failed
	 (setq line (cdr pair))
	 (let ((max (- (length line) 1)))
	   (pprinc line)
	   (setq *infix-loc* (+  *infix-loc* max))
	   (if *handling-newlines-in-lb-comment*
	       nil
	       (if (> max 0)
		   (if (not (char-equal (aref line max) #\Newline))
		       (add-saved-whitespace  #\Newline)
		       nil)
		   (add-saved-whitespace  #\Newline)))
	   nil))	 
	(t (let ((max (- (length line) 1)))
	     (sloop for i from 0 to max
		    do (let ((ch (aref line i)))
			 (cond ((eql ch #\Page) (insert-newpage))
			       ;; Hands off special chars in comments.
			       (*hands-off* (check-environment-and-write mode ch))
			       (*current-environment* (handle-special-chars-in-string ch))
			       (t (check-environment-and-write mode ch)))))
	     (setq *infix-loc* (+  *infix-loc* max))
	     (if *handling-newlines-in-lb-comment*
		 nil
		 (if (or (<= max 0)
			 (not (char-equal (aref line max) #\Newline)))
		     (add-saved-whitespace  #\Newline)
		     nil))
	     nil)))))

(defun copy-comments ()

; This function tries to sneak up to the next top-level open parenthesis,
; parsing all of the Lisp comments up till there.

; NOTE:  Jul 13 93 MKS 
; Random atoms, numbers, strings and characters are treated as if they were in
; comments.  And are printed in a FORMAT environment.

; NOTE:  Nov 30 94 MKS 
; This sneaking up doesn't work quite the way we would like. If we have
; a comment line, followed by a blank line, followed by (foo ..) then we
; put out the formatted comment line, followed by a blank line, followed
; by a close-environment, followed by the formatted (foo ..).

  ;; (format t "~%Entering copy-comments~%")

  (let (*current-environment*)
    (catch 'copy-comments
      (sloop
       for ch = (read-char-in-comment)

; The top level loop for simulating the skimming of whitespace and comments
; that READ would perform to get to the next open parenthesis.

       (case ch

; Semicolon starts a comment.  
; We use *COMMENT-SEMI-NET* to determine how to format the comment based on the 
; immediately following characters.  This is user-setable to produce
; running text, a format environment (honors spaces and newlinexs, but probably
; does not produce a fixed width font), a verbatim environment (like format, but
; fixed width font), an emphasis environment (typically italics), or a title
; environment (like a section name).

         (#\;
	  ;; Do one line.
          (let ((action (copy-comments-read-net *comment-semi-net*)))
            (check-environment-and-write (car action) (cadr action))
	    ;; COULD ADD ! PROCESSING HERE
	    (format-comment-string (read-line *standard-input*) (car action))
	    ;(format t "~%semi-colon comment, action = ~s~%" action)
	    ))
	 

; \#\| starts a comment.  
; As above, we use the *comment-lb-net* to determine what formatting action to take.
; \|# ends one. 

         (#\#
          (setq ch (read-char-in-comment))
          (cond ((eql ch #\+)
		 (let ((label (readx *standard-input* nil a-very-rare-cons nil)))
		   (pprinci "#ifdef ")
		   (infix-print-term1 label)
		   (print (readx *standard-input* nil a-very-rare-cons nil))
		   (pwrite-char #\newline)
		   (pprinci "#endif ")
		   (to-current-margin)))
		((eql ch #\-)
		 (let ((label (readx *standard-input* nil a-very-rare-cons nil)))
		   (pprinci "#ifdef ")
		   (infix-print-term1 label)
		   (print (readx *standard-input* nil a-very-rare-cons nil))
		   (pwrite-char #\newline)
		   (pprinci "#endif ")
		   (to-current-margin)))
		((not (eql ch #\|))
                 (error"Do not know how to handle #~s while copying at the top level." ch))
		(t
		 (let ((action (copy-comments-read-net *comment-lb-net*)))
            
		   ;; The following may not put us into an env if (cadr
		   ;; action) is whitespace.

		   (check-environment-and-write (car action) (cadr action))

		   ;; We ignore formatting changes within imbedded \#\| \|\# comments.
		   ;; They are stuck with whatever the outermost comment
		   ;; decreed.

		   (setq number-deep 1)
		   (let* ((line (read-line *standard-input*))
			  (rest (normalize-lb-comment-line line (car action))))
		     (cond ((and rest (not (string-equal rest "")))
			    ;; Sep 22 95 MKS reversed order of the 2 forms below.
			    ;; Wrap-up throws out of copy-comments.
			    (unread-chars (nreverse (coerce rest 'list))
					  *standard-input*)
			    (wrap-up-copy-comments))))
		   ;;(format t "~%lb-vbar comment, level = ~d, action = ~s~%" number-deep action)
		   ))))

         ;; A raw ^L at the top level, not in a comment.
         (#\Page (end-environment)
          (ppformat "~%~a~%"  *newpage*))

	 (#\Newline
	  ;;(format t "~%newline, ending = ~s~%" *current-environment*)
	  (add-saved-whitespace #\Newline)
	  ;(end-environment)
	  ;(print-saved-whitespace)
	  ;(newlinex)
	  )
         (#\(
          (unread-char #\( *standard-input*)
          (wrap-up-copy-comments))
	 ;; Handle keywords like parenthesized forms, because they may
	 ;; need to read their arguments.
         (#\:
          (unread-char #\: *standard-input*)
          (wrap-up-copy-comments))
         (otherwise ;; switched (pwrite-char ch)
	  ;; MOD -  Sep 21 95 MKS - Added the following:
	  ;; (print-saved-whitespace)
	  ;; (handle-special-chars-in-string ch)
	  ;; MOD - Replaced the two lines above with those
	  ;; below.  The problem is acl::T, acl2::NIL, 'foo::bar
	  ;; being read at the top level.  This branch would
	  ;; consume the package name, the preceding would stop
	  ;; and unread a : and we would try to read ::bar.
	  ;; Which gives an error message about an undefined,
	  ;; unstated package.
          (unread-char ch *standard-input*)
          (wrap-up-copy-comments))
	  ))))
    ;;(format t "~%Exiting copy-comments~%")
    )


;;--------------------------------------------------------------------------------
;
;                       COMMENT FORMATS
;

(defvar *comment-format-alist* nil)
(defvar *comment-format* 'text)	

(defun update-alist (al key value)
  (cond ((not (consp al)) (list (cons key key value)))
        ((eq key (caar al)) (cons (cons key value) (cdr al)))
        (t (cons (car al) (update-alist (cdr al) key value)))))

(defun define-comment-format (n format)
  ;; Last call to this sets *comment-format*.
  ;; Can be overruled by 
  ;;  1. assigning directly to *comment-format*,
  ;;  2. calling (setup-comment-format format-name), or 
  ;;  3. calling infix-setup with the appropriate arguments.
  (if (not (check-comment-character-net format))
      (format *terminal-io* "~%Ill formed definition for comment format ~a" n)
      (progn 
	(setq *comment-format* n)
	(cond ((assoc n *comment-format-alist*)
	       (setq  *comment-format-alist* (update-alist *comment-format-alist* n format)))
	      (t (setq *comment-format-alist*
		       (cons (cons n format) *comment-format-alist*)))))))

(defun setup-comment-format (&optional n)
  (cond ((and n (assoc n *comment-format-alist*))
	 ;; A named format.
	 (setq *comment-format* n))
	((and n *comment-format*)
	 ;; Not defined.  Use existing setting.
	 (format *terminal-io*
		 "~%No comment format named ~a.  Left as ~a." n *comment-format*))
	(n (setq *comment-format* 'text)
	   (format *terminal-io*
		   "~%No comment format named ~a.  Defaluting to ~a." n *comment-format*))
        ((null *comment-format*)
         (cond ((eq *infix-op-location* 'FRONT)
                (setq *comment-format* 'text))    
               ((eq *infix-op-location* 'BACK)
                (setq *comment-format* 'verbatim))    
               (t (setq *comment-format* 'text))))
        ((assoc *comment-format* *comment-format-alist*))
	(*comment-format-alist* 
	 (setq *comment-format* (caar *comment-format-alist*))
	 (format *terminal-io* "~%Defaluting to first format in alist, ~a." *comment-format*))
        (t ;; Should never get here.
	 (format *terminal-io* "~%*** No comment formats defined!!! ***")))
  (compute-comment-character-net *comment-format*)
  ;; We have side-effected *comment-semi-net* and *comment-lb-net*
  ;; Update mapping info AFTER users syntax file loaded.
  (setq *comment-environment-mapping* (eval *comment-environment-mapping-template*)))

(defun check-comment-character-net (l)
  (if (null l)
      (format *terminal-io* "*COMMENT-FORMAT* is not present in *COMMENT-FORMAT-ALIST*."))
  (if (not (assoc #\; l))
      (format *terminal-io* "Selected comment format should include a list labelled \"\;\"."))
  (if (not (assoc #\# l))
      (format *terminal-io* "Selected comment format  should include a list labelled \"\#\"."))
  ;; Each branch is of the form (string flag environment [echo-character])
  (check-comment-character-net2 l))

(defun check-comment-character-net2 (l)
  (cond ((null l) t)
        ((and (listp l)
              (listp (car l))
              (characterp (caar l))
              (every
	       (function
		(lambda (branch)
		  (if (check-comment-character-branch branch)
		      t
		      (format *terminal-io* "Ill-formed branch in *COMMENT-FORMAT-ALIST*.~
                                                       ~%~a" branch))))
	       (cadr (car l)))
              (let ((top (cddr (car l))))
                (and (or (equal (car top) 't) (null (car top)))
                     ;; Must be known environment.
                     (assoc (cadr top) *comment-environment-mapping*)
                     (or (null (cddr top)) (characterp (caddr top))))))
         ;; Do the other one, if there.
         (check-comment-character-net2 (cdr l)))
        (t nil)))

(defun check-comment-character-branch (b)
  (and (listp b)
       (> (length b) 2)
       (stringp (car b))
       (or (equal (cadr b) 't) (null (cadr b)))
       ;; Must be known environment.
       (assoc (caddr b) *comment-environment-mapping*)
       (or (null (cdddr b))
           (characterp (cadddr b)))))

(defun compute-comment-character-net (name)
  (let ((l (cdr (assoc name  *comment-format-alist*))))
    (let ((net (assoc #\; l)))
      (setq *comment-semi-net*
            (if net
                (cdr (compute-net -1 (car net) (cadr net) (caddr net) (cdddr net)))
                nil)))
    (let ((net (assoc #\# l)))
      (setq *comment-lb-net*
            (if net
                (cdr (compute-net -1 (car net) (cadr net) (caddr net) (cdddr net)))
                nil)))))

(defun compute-net (n char net skip-one-blank-p default)
  (list char
        (append (if skip-one-blank-p
                    `((#\  nil ,default))
                    nil)
                (compute-branches (+ n 1) net))
        default))

(defun compute-branches (n net)
  (cond ((null net) nil)
        (t (merge-net (compute-branch n (car net))
                      (compute-branches n (cdr net))))))

(defun compute-branch (n branch)
  ;; branch is of the form (string flag . default)
  (let ((string (car branch))
        (flag (cadr branch))
        (default (cddr branch)))
  (cond ((> n (length string)) nil)
        ((= n (length string))
         (if flag
             `(#\  nil ,default)
             nil))
        (t (append (list (char string n)
                         (list (compute-branch (+ n 1) branch)))
                   (if (= (+ n 1) (length string)) (list default) nil))))))

(defun merge-net (branch net)
  ;; All branches of net begin with a unique character.
  ;; As does the result.
  (cond ((null net) (list branch))
        ((char= (car branch) (caar net))
         (let ((def1 (caddr branch))
               (def2 (caddr (car net))))
           (cons
	    (list (car branch)
		  (merge-net (caadr branch) (cadr (car net)))
		  (cond ((equal def1 def2) def1)
			((and def1 def2)
			 (format *terminal-io*
				 "Your comment network is in conflict ~a ~a." branch (car net))
			 def1)
			(def1 def1)
			(t def2)))
	    (cdr net))))
        (t (cons (car net) (merge-net branch (cdr net))))))

(define-comment-format 'verbatim
  '((#\; (("\\"   t   text))
     nil verbatim #\;)
    (#\# (("\\"   t   text))
     nil verbatim)))

(define-comment-format 'text
  '((#\; (("\;"   t   text   )
          ("\#"   t   comment )
          ("\;\;" t   verbatim )
          ("\\"   t   text     )
          ("-"    t   format   )
          ("+"    t   verbatim )
          ("!"    t   emphasis )
          ("\;\\" nil   text     #\;)
          ("\;-"  nil   format   #\;)
          ("\;+"  nil   verbatim #\;)
          ("\;!"  nil   emphasis #\;))
     t text)
    (#\# (("\\"   t   text     )
          ("\#"   t   comment  )
          ("-"    t   format   )
          ("\;"   t   verbatim ))
     t text )))

(define-comment-format 'c
  '((#\; (("\;\;" t comment)
	  ("\;" t comment))
     t comment)
    (#\# nil t comment )))

(define-comment-format 'CL
  '((#\; (("\;"   t    format )
          ("\;\;" t    text   )
          ("\;\;\;" t  section)

          ("\\"   t   text     )
          ("-"    t   format   )
          ("+"    t   verbatim )
          ("!"    t   emphasis )
          ("\;\\" nil   text     #\;)
          ("\;-"  nil   format   #\;)
          ("\;+"  nil   verbatim #\;)
          ("\;!"  nil   emphasis #\;))
     t emphasis)
    (#\# (("\\"   t   text     )
          ("-"    t   format   )
          ("\;"   t   verbatim ))
     t text )))

(setup-comment-format 'text)

;; End of comment stuff.

(defun infix-form (form &key ((:print-case *print-case*) :downcase))
  ;; We cannot recover well from this case since we don't
  ;; know where we are.  All of the I/O that provides that info
  ;; is in the top level loop.  So these inner forms are just out of luck.
  (let ((*print-pretty* nil)
	;; (*saved-tab-stack* *tab-stack*)
	)
    ;; (cond ((catch 'taboverflow
    ;;              (let ((*tabs-in* 1))
    ;;                (or *do-not-use-tabs* (begin-tabbing))
    ;;                (infix-print-term1 form)
    ;;                (or *do-not-use-tabs* (end-tabbing))
    ;;                nil))
    ;;            (pformat *terminal-io*
    ;; 		    "~%Sorry. Exceeded tabbing limit (1). ~a needs hand massaging.~%"
    ;; 		    (car form))
    ;; 	   (setq *tab-stack* *saved-tab-stack*)
    ;; 	   (newlinex)))
    (let ((*tabs-in* 1))
      (or *do-not-use-tabs* (begin-tabbing))
      (infix-print-term1 form)
      (or *do-not-use-tabs* (end-tabbing))
      nil)))

(defvar *defining-events* nil)

(defun infix-event (form &optional *suppress-trailing-semi-colon*
			 &key ((:print-case *print-case*) :downcase))
  ;; We cannot recover well from this case since we don't
  ;; know where we are.  All of the I/O that provides that info
  ;; is in the top level loop.  So these inner forms are just out of luck.
  (let ((*print-pretty* nil)
	(*tabs-in* 1))
    (funcall (get-event-printer (car form)) form)
    ;; Do it after, so the event doesn't self-reference.
    (check-to-save-url form)))

(defun check-to-save-url (form)
  (cond (*do-not-index*)
	((not (consp form)))
	((eq (car form) 'local)
	 (let ((form (cadr form)))
	   (if (and (consp form)
		    (member (car form) *defining-events*))
	       (note-definition-location (car form) (cadr form)))))
	((member (car form) *defining-events*)
	 (note-definition-location (car form) (cadr form)))))

(defvar *saved-urls* nil "Alist of name . file")
(defvar *current-url-file* nil "Save in var above")

(defun clear-urls () (setq *saved-urls* nil))

(defun note-definition-location (type name)
  (declare (ignore type))
  (let ((pair (assoc name *saved-urls*)))
    (if pair
	(setf (cdr pair) *current-url-file*)
	(setq *saved-urls*
	      (cons (cons name *current-url-file*)
		    *saved-urls*)))))

(defvar *first-term* 0)

(defvar *infix-trace* nil)

(defun current-directory ()
  ;; This is somewhat redundant.  
  ;; That is (probe-file file) should equal 
  ;; (probe-file (concatenate 'string (current-directory) file))
  ;; But we let *current-directory* also be set by the input file.
  (probe-file "./"))

;; This may be set by the function above or based on the directory of the input file.
(defvar *current-directory* nil)

(defun probe-syntax (fl)
  (let ((name (concatenate 'string (pathname-name fl) "-syntax")))
    (or (probe-file (make-pathname :name name
				   :type "lisp" :defaults fl))
	(probe-file (make-pathname :name name
				   :type "lisp"
				   :directory (pathname-directory *current-directory*)
				   :defaults fl))
	(probe-file (make-pathname :name name
				   :type "lisp"
				   :directory (pathname-directory *infix-directory*)
				   :defaults fl)))))

;; Check that *infix-directory* we can find at least a latex and scribe syntax.
(eval-when (load eval)
 (if (not (and (probe-syntax "scribe") (probe-syntax "latex")))
     (format *terminal-io* "~%Seem to be missing syntax of scribe or latex or both.~%")))

(defun type-file-name (file type &optional force)
  ;;if extension = nil, return file.
  ;;if file already has type and force = nil, return file.
  (cond ((null type) file)
	((and (not force) (pathname-type file)) file)
	(t (make-pathname :type type :defaults file))))

(defvar *default-chars-wide* 77)

(defun directory-or-current (fl)
  (let ((dir (directory-namestring fl)))
    (cond ((null dir) (current-directory))
	  ((not (equal dir "")) dir)
	  (t (current-directory)))))

(defun probe-cert-for-packages (fl)
  (let ((cert (probe-file (make-pathname :type "cert" :defaults fl)))
	doit)
    (if cert
	(progn 
	  (format t "Checking ~s for defpkg forms.~%" cert)
	  (with-open-file
	   (*standard-input* cert :direction :input)
	   (sloop for form = (read *standard-input* nil a-very-rare-cons nil)
		  until (or (eq form a-very-rare-cons)
			    (eq form :end-portcullis-cmds))
		  do
		  (cond ((eq form :begin-portcullis-cmds)
			 (setq doit t))
			((and doit (consp form) (eq (car form) 'defpkg))
			 (if (not (find-package (cadr form)))
			     ;; EITHER
			     ;; Could use the lisp make-package, e.g.
			     ;; (make-package (cadr form) :use '("LISP" "ACL2"))
			     ;; OR 
			     ;; Use ACL2 version.  Though note that the 
			     ;; context part of the defpkg cannot be evaluated
			     ;; accurately, so in acl2-defpkg it will get
			     ;; some default.
			     (acl2-defpkg (cadr form) nil)))
			(t nil))))))))

(defvar *infix-acl2-package* nil
  "Try to maintain proper synchronization with ACL2")

(defun handle-in-package (s &optional f)
  (let ((p (find-package s)))
    (cond ((not p) (format *terminal-io* "~%~%WARNING: The package, ~s, is not defined
You should probably make sure you have a ~s-syntax file for the input file
and insert the appropriate DEFPKG forms in it.  See example-syntax.lisp~%~%" s f)
	   (setq p (acl2-defpkg s nil))))
    (in-package s)
    (setq *package* p)
    ;; Now, synchronize with ACL2 loop.  DOESN'T WORK.
    ;; (in-package-fn s *the-live-state*)    
    ))

(defvar *print-as-prefix* nil 
  "Overrides the normal prefix printing and prints
in God's syntax (e.g. prefix).  Still format comments,
suppresses hints, etc.")

(defun infix-file (fl &key ((:print-case *print-case*) :downcase)
                      (mode nil)
                      (chars-wide *default-chars-wide*)
                      (comment *nq-default*)
		      ;; For include files.
		      (omit-prelude nil)) 
  (let ((*infix-file* t)
	(*current-directory* (directory-or-current fl))
	(fl-root (pathname-name fl))
	(*rightmost-char-number* chars-wide))
    (cond ((and mode (string= mode *infix-mode*))
           (format t "~%Processing in ~a mode.~%" mode))
          ((stringp mode)
           (setq *infix-mode* mode)
           (format t "~%Entering ~a mode.~%" *infix-mode*)
           (load-infix-init-file))
          (mode
           (setq *infix-mode* (string mode))
           (format t "~%Entering ~a mode.~%" *infix-mode*)
           (load-infix-init-file))
          ((probe-syntax fl-root)
           (setq *infix-mode* (pathname-name fl-root))
           (format t "~%Entering ~a mode.~%" *infix-mode*)
           (load-infix-init-file))
          ((null *infix-mode*)
           (cond ((y-or-n-p "Enter Ascii mode? ")
                  (setq *infix-mode* "ascii"))
		 ((y-or-n-p "Enter Latex mode? ")
                  (setq *infix-mode* "latex"))
		 ((y-or-n-p "Enter HTML mode? ")
                  (setq *infix-mode* "html"))
                 ((y-or-n-p "Enter Scribe mode? ")
                  (setq *infix-mode* "scribe"))
                 (t (setq *infix-mode* nil)))
	   (if *infix-mode*
	       (progn (format t "~%Entering ~a mode.~%" *infix-mode*)
		      (load-infix-init-file))))
          (t (format t "~%In ~a mode.~%" *infix-mode*)))

; infix-file takes a root file name, e.g., foo, reads the file foo.lisp,
; which we suppose has been previously checked by LD, and creates the
; file foo.tex, which the user can then run through Latex, etc.  By default, we
; lowercase all variable and function symbols, but the user can override this
; with the keyword parameter.

; If the keyword comment is given as true, then we first generate fl.nqxxx and then
; invoke nqfmt2fmt, generating fl.xxx (.tex or .mss).

    ;; Update comment information AFTER users syntax file loaded.
    (setup-comment-format)

    (if *infix-mode*
        (let ((infl  (type-file-name fl infix-input-file-type))  ; ".lisp"
              ;; .mss, .tex, .nqmss, .nqtex
              (outfl (type-file-name fl (fmtfile-extension *infix-mode* comment) t))  
              (a-very-rare-cons (cons nil nil))
              (*package* *package*)
              (*print-pretty* nil)
              (*readtable* (copy-readtable nil))
	      (*tab-list* nil)
	      (*tab-stack* nil)
              (*reported-tabs* nil)
              (*infix-loc* 0)
              (*left-margin* 0)
              (count 1)
              inpos)

	  ;; (format t "~%% chars-wide: ~3d" chars-wide)
	  (setq *current-url-file* (namestring outfl))
	  (probe-cert-for-packages fl)

          ;(smash-infix-readtable)
	  
          (with-open-file
           (*standard-input* infl :direction :input)
           (with-open-file

; We do *all* of our printing of terms to *standard-output*, giving princ only
; one argument.

            (*standard-output* outfl :direction :output :if-exists :rename-and-delete)

; The formatting system opening.

            (or omit-prelude (ppformat *standard-prelude*))
            (sloop for form = (progn (copy-comments)
				     ;; Set this here so we don't rewrite preceding comment, 
				     ;; if tabs overflow.
				     (setq inpos (file-position *standard-input*))
				     (readx *standard-input* nil a-very-rare-cons nil))

					; We remember where we are in the output file as part of our mechanism for
					; recovering from the very small finiteness of the Latex tabbing mechanism.  We
					; will rewrite to the current position and start printing in verbatim mode with
					; PPR if we exceed the Latex tabbing mechanism.

		   for outpos = (file-position *standard-output*)
		   until (eq form a-very-rare-cons)
		   do
					; (ppformat "\\filbreak %\\item~%")
		   (progn
		     (if (and (consp form)
			      (eq (car form) 'in-package)
			      (consp (cdr form))
			      (stringp (cadr form)))
			 (handle-in-package (cadr form)))
		     (COND
		      ((not (consp form))
		       ;; event-seperator?
		       (print-atom form) (pprinci " "))
		      ((or (eq (car form) 'comment)
			   (COND
			    ((catch 'taboverflow
			       (let ((*tabs-in* 1))
				 (if *infix-trace*
				     (format *terminal-io* "~% ~a " (car form)))
				 ;; Let the user know we are making some
				 ;; kind of progress, every 10 events.
				 (if (= count 10)
				     (progn (format *terminal-io* ".")
					    (setq count 1))
				     (incf count))
				 (progn
				   (if *print-as-prefix*
				       (prefix-event-printer form)
				       (funcall (get-event-printer (car form)) form))
				   (check-to-save-url form)))
			       nil)
			     (pformat *terminal-io*
				      "~%Warning: Tab limit reached in event ~a~
                                                 ~%Hand massaging required.~%"
				      (or (and (consp form) (cdr form) (cadr form))
					  (car form)))
			     t)))
		       ;; If taboverflow, go back to where we started printing and
		       ;; print over the previous stuff.  In the 'commment case our
		       ;; position is unchanged.
		       (file-position *standard-output* outpos)
		       ;; Reinitialize to 0 points.
		       (reset-loc-and-margin)
		       ;; Print out literally.
		       (begin-environment 'verbatim)
		       (let ((stop (file-position *standard-input*)))
			 (file-position *standard-input* inpos)
			 (sloop while (< (file-position *standard-input*) stop)
				do (check-environment-and-write 'verbatim
								(read-char *standard-input*))))
		       (end-environment))
		      (t nil))))
            (or omit-prelude (ppformat *standard-postlude*))))
          (if comment
              (nqfmt2fmt fl)
              outfl))
        (format t "~%No mode specified (e.g. Latex or Scribe), aborting.~%"))))

(defun load-obj-or-lisp (file)
  (let ((object (make-pathname :type "o" :defaults file))
	(lisp   (make-pathname :type "lisp" :defaults file)))
    (cond ((and (probe-file object)
		(> (file-write-date object) (file-write-date file)))
	   (load object))	  
	  ((probe-file file) (load file))
	  ((probe-file lisp) (load lisp))
	  (t (error (format nil "~%No syntax or init file mathcing ~f~%" file))))))

;; (defun load-obj-or-lisp (file)
;;   (let* ((name2 (if (member 'sparc *features*)
;; 		    (concatenate 'string (pathname-name file) "-sparc")
;; 		    (concatenate 'string (pathname-name file) "-sun")))
;; 	 (file2   (probe-file (make-pathname :name name2 :defaults file)))
;;          (object  (probe-file (make-pathname :type "o" :defaults file)))
;;          (object2 (if file2 (probe-file (make-pathname :type "o" :defaults file2)))))
;;     (cond ((and object2 (> (file-write-date object2) (file-write-date file2)))
;; 	   (load object2))
;; 	  (file2 (load file2))
;; 	  ((and object (> (file-write-date object) (file-write-date file)))
;; 	   (load object))	  
;; 	  ((probe-file file) (load file))
;; 	  (t (error (format nil "~%No syntax or init file mathcing ~f~%" file))))))

(defun load-syntax-or-init (dir)
  (let* ((initfile   (make-pathname :name (concatenate 'string *infix-mode* "-init")
				    :type "lisp"
				    :directory (pathname-directory dir)))
         (syntaxfile (make-pathname :name (concatenate 'string *infix-mode* "-syntax")
				    :type "lisp"
				    :directory (pathname-directory dir))))
    ;; We assume that, if present, the -syntax file loads the -init file.
    (cond ((probe-file syntaxfile) (load-obj-or-lisp syntaxfile) t)
          ((probe-file initfile)   (load-obj-or-lisp initfile) t)
          (t nil))))

(defun load-infix-init-file ()
  (clean-up-everything)
  (cond ((null *infix-mode*)
         (format t "~%Failed to initialize.  *infix-mode* is NIL.~%"))
        ((not (stringp *infix-mode*))
         (format t "~%Failed to initialize.~
                    ~%*infix-mode* (~a) is not a string.~%" *infix-mode*))
        ((and *current-directory* (load-syntax-or-init *current-directory*)))
        ((load-syntax-or-init *infix-directory*))
        (t (format t "~%Failed to initialize mode ~a." *infix-mode*)
	   (if *current-directory* (format t "~%No init or syntax file in ~a.~%"
                   *infix-mode* *current-directory*))
	   (format t "~%No init or syntax file in ~a.~%"
                   *infix-mode* *infix-directory*)
           (setq *infix-mode* nil))))

(defun fmtfile-extension (mode comment)
  (cond ((and *mode-extension* (stringp *mode-extension*))
         (if comment (concatenate 'string "nq" *mode-extension*) *mode-extension*))
        ((string= mode "latex")
         (if comment "nqtex" "tex"))
        ((string= mode "scribe")
         (if comment "nqmss" "mss"))
        (t (if comment "nqnq" "nq"))))

(defun fmt2fmt-extension (remove! mode)
  (cond (remove! "stripped")
        ((and *mode-extension* (stringp *mode-extension*))
         (concatenate 'string "nq" *mode-extension*))
        ((string= mode "latex")  "nqtex")
        ((string= mode "scribe") "nqmss")
        (t "nqnq")))

(defvar nqread-default-white-space '(#\Space #\Newline #\Tab #\Page #\Return))
(defvar nqread-default-normal-clause-enders '(#\. #\! #\? #\, #\; #\:))
(defvar nqread-default-break-chars '(#\( #\) #\` #\' #\" #\; #\,))

(defvar nqread-white-space nqread-default-white-space)
(defvar nqread-normal-clause-enders nqread-default-normal-clause-enders)
(defvar nqread-break-chars nqread-default-break-chars)

(defun acl2-read-preserving-whitespace-error ()
  (error "A character or an integer or an Acl2 symbol was expected at location ~a of input."
         (file-position *standard-input*)))

(defun acl2-read-preserving-whitespace ()

; This function does the READing right after a ! command.  This
; function is almost the same as read-preserving-whitespace.  It is
; different only because of the peculiar problem of trailing
; punctuation marks.  We sometimes stop the reading before Common Lisp
; would.

; In processing ! commands, we call READ on the following text.  If
; the text starts with an open parenthesis or string-quote, it is
; clear where the READing should stop.  However, if a mere symbol or
; number follows a ! command, then when READing the symbol we treat
; an exclamation mark, colon, question mark, or period that is
; followed by whitespace as a punctuation mark, not as part of the
; symbol that is to be READ.  The other ordinary English clause ending
; punctuation characters (comma and semicolon) already naturally
; terminate a READ, as do parentheses, quotation marks, and
; whitespace.

; Example.  When we do a READ while processing a ! command under
; nqtex2tex, we have to decide what to do about text such as

;    !tnil.

; The question is, does the period belong to the token to be read, or
; is it just some user punctuation?  We take the attitude that the
; answer is punctuation.  Now, this attitude is a bit arbitrary.  nil.
; is a legal Common Lisp symbol.  Unfortunately it is ALSO a legal Acl2 symbol.
; But we are just going to assume that it is atypical.  And likewise for 
; `foo.bar'.  We also run into problems reading things like
; xxx.@end{text}@begin{format}@tabclear{}.  This is a fine atom as far as
; ACL2 is concerned. Likewise 
; xxx.\begin{verbatim}\hfill
; So the text-formatting -init file will need to extend nqread-normal-clause-enders
; to account for this.

; One might ask, who cares?  The reason we care is that nil, and other 
; symbols on *atom-alist*, get printed specially.  For example, nil is 
; printed in bold, not italics.  If we read nil. as one symbol, it would 
; come out in intalics because nil. is not on *atom-alist*.

; The idea of fiddling with READ so that it is `smart' about not
; reading a trailing punctuation character is weird.  But then,
; calling READ in the middle of Tex file is weird, too.  We have found from
; experience that it is very hard to write sentences that have
; whitespace after every symbol.  We want to be able to write things
; like !tnil, and !tnil.  So here is the general procedure for how far
; we READ after a ! command.  When we say READ below, we really mean
; read-preserving-whitespace.

; We peek at the first nonwhitespace character after the ! command,
; and we consider the cases on this character.

; If it is ( or " we simply call READ, knowing that upon encountering
; the closing " or ) READ will not look at any trailing punctuation.

; If it is ' or ` we recursively read further with this function and
; then quote or backquote the result.  This is so that `foo. will come
; out right.

; Otherwise, we call READ on the string consisting of all of the
; characters up to but not including the first (a) whitespace, (b)
; terminating readmacro Common Lisp character, i.e., ()`'"";,, or (c)
; normal English clause ending character, i.e., .!?:, that is followed
; by a whitespace.

; Known ambiguity.  Although periods are permitted at the ends of
; numbers in Acl2 syntax, we treat them as ends of sentences, if they
; are followed by white space.  Thus in reading !t5. , the read
; would not pick up the period.  So the period would appear in the
; final text.  It is hard to see whether this is a bug, a feature, or
; a problem that simply never arises.

; Known ambiguity.  Because the quotation mark is a legal character
; in Acl2 symbols, a minor question arises about the handling of
; a terminal question mark in an Acl2 symbol; we treat it as punctuation.
; Thus !qfoo? will come out as `foo'? rather than `foo?'.

; All this peeking stuff doesn't work really correctly if there are
; comments in the way, so we adopt this rule: don't put comments in
; expressions after ! commands.  Typically, this function is called
; inside of a comment.  If the text to be read extends over a line and
; the next line begins with a ; character, you may not get at all what
; you want because the text on the line after the ; will be skipped.

  (case (peek-char t *standard-input*)
        ((#\( #\") (read-preserving-whitespace *standard-input*))
        (#\'
         (read-char *standard-input*)
         (list 'quote (acl2-read-preserving-whitespace)))
        (#\`
         (read-char *standard-input*)
         (list *infix-backquote* (acl2-read-preserving-whitespace)))
        (otherwise 
         (read-from-string
          (coerce
           (nconc
            (prog (c ans c2)
                  loop
                  (setq c (peek-char nil *standard-input* nil a-very-rare-cons))
                  (cond ((or (eq c a-very-rare-cons)
                             (member c nqread-white-space)
                             (member c nqread-break-chars))
                         (cond ((null ans)
                                (acl2-read-preserving-whitespace-error)))
                         (return (nreverse ans)))
                        ((member c nqread-normal-clause-enders)
                         (read-char *standard-input*)
                         (setq c2 (peek-char nil *standard-input* nil a-very-rare-cons))
                         (cond ((or (member c2 nqread-white-space)
                                    (eq c2 a-very-rare-cons))
                                (unread-char c *standard-input*)
                                (cond ((null ans)
                                       (acl2-read-preserving-whitespace-error)))
                                (return (nreverse ans)))
			       ;; Handle !qfoo.\begin{bar}
			       ((char= c2 #\\)
                                (unread-char c *standard-input*)
                                (cond ((null ans)
                                       (acl2-read-preserving-whitespace-error)))
                                (return (nreverse ans)))
                               (t (push c ans))))
                        ((member c '(#\| #\; #\\))
                         (acl2-read-preserving-whitespace-error))
                        (t (read-char *standard-input*)
                           (push c ans)))
                  (go loop))

; Sticking on this extra space is not strictly necessary.  We do it to
; work around a bug in AKCL 1-605.

            (list #\Space))
           'string)))))

(defun read-literally ()

; This function does the READing right after a !I command, returning a string.
; If no left delimiter after the I, it does an acl2-read-preserving-whitespace.
; If left delimiter after the I, it reads char by char until it
; hits a corresponding right delimeter.
; Note that a similar effect can be achieved by setting *hands-off* to T.

; The distinction is that 
;  \bf{foo}       prints as   bf\{foo\}
;  !i[\bf{foo}]   prints as   \bf{foo}

; We warn if the number of chars between delimiters exceeds a limit.

; We peek at the first nonwhitespace character after the ! command,
; and we consider the cases on this character.

; If it is (, { or [ we "read literally" up to the first ), }, or ],
; respectively.

; Othewise, acl2-read-preserving-whitespace.

  (case (peek-char t *standard-input*)
	(#\( (read-char *standard-input*) (read-to-first #\( #\) *standard-input*))
        (#\{ (read-char *standard-input*) (read-to-first #\[ #\} *standard-input*))
	(#\[ (read-char *standard-input*) (read-to-first #\{ #\] *standard-input*))
        (otherwise (acl2-read-preserving-whitespace)))
  )

(defvar *literal-char-warning-limit* 200)

(defun read-to-first (left right &optional (*standard-input* *standard-input*))
  (let (done ans (limit 0) (front ""))
    (sloop while (not done)
	   do (let ((c (peek-char nil *standard-input* nil a-very-rare-cons)))
		(cond ((eq c a-very-rare-cons)
		       (setq done t)
		       (cond ((null ans) (setq ans ""))
			     (t (setq ans (coerce (nreverse ans) 'string)))))
		      ((char= c right)
		       (read-char *standard-input*)
		       (setq done t)
		       (cond ((null ans) (setq ans ""))
			     (t (setq ans (coerce (nreverse ans) 'string)))))
		      (t (push (read-char *standard-input*) ans)))
		(setq limit (1+ limit))
		(cond ((> limit *literal-char-warning-limit*)
		       (setq limit 11)
		       (format *terminal-io*
			       "Likely missing delimiter on !I~c~a" left front))
		      ((= limit 10)
		       (setq front (coerce (reverse ans) 'string))))))
    ans))


(defvar nqtex2tex-chars
  (coerce "eipqtv" 'list))

;             NQFMT2FMT

(defun nqfmt2fmt (fl &key
		     ((:print-case *print-case*) :downcase)
		     ((:left-margin *left-margin*) 5)
		     (just-remove-! nil))

  ;; Copies the file fl.nqxxx file to the file fl.xxx file and honors ! 
  ;; and ~ formatting commands, replacing Acl2 forms
  ;; preceded by a formatting instructions with the results described below.  
  ;; If an exclamation mark or ~ is not followed by one
  ;; of these special characters, then the following form is preserved unchanged,
  ;; and the mark and the character following it are preserved, too.

  ;; Although we may extend this set of replacement commands, we *promise* to give
  ;; special meanings only to alphabetic characters after !.  Thus we promise
  ;; never to give !!  a replacement effect.  ~ forms currently include ~-, but
  ;; that is the only non-alphabetic formatting instruction.

  ;; In every case, for the replacement instruction, upper or lower case has
  ;; the same effect.

  ;; !Bform, prints form in bold.

  ;; !Eev, where ev is an Acl2 event form, e.g., (defun foo (x) 3), results in
  ;; conventional notation for ev.  We may introduce line breaks via tabbing commands.
  ;; Mnemonic: E -- think Event.  BETTER NOT RESULT IN AN ERROR.

  ;; !Ifoo, where foo is a symbol, results in foo, but with with formatting sensitive
  ;; characters quoted.  For example, in TeX, !Ia$b would result in a\$b.  
  ;; Mnemonic: I -- think Identity.

  ;; !Pfm, where fm is an Acl2 term, e.g., (plus x y), results in conventional
  ;; mathematical notation for fm.  May introduce line breaks via tabbing.
  ;; Mnemonic: P -- think Pretty print.

  ;; !Qfn, where fn is a symbol, results in fn surrounded by single gritches,
  ;; after formatting sensitive characters have been quoted, e.g., !qfoo results in
  ;; `foo' in TeX.  Useful for distinguishing function symbols from other words in a
  ;; sentence, since function symbols appear in Roman.  
  ;; Mnemonic: Q -- think Quoted.

  ;; !Tfm, where fm is an Acl2 term, results in conventional mathematical
  ;; notation for fm, but without any line breaks.  
  ;; Mnemonic: T -- think Term.

  ;; !Vfoo means that foo is printed as is, but in typewriter font, and with
  ;; special characters quoted.  
  ;; Mnemonic: V -- think Verbatim.

  ;; ! followed by anything else is left alone, along with the exclamation mark.

  ;; One can certainly use nqfmt2fmt on the result of running infix-file, but one
  ;; must do so deliberately by first running infix-file, then renaming the
  ;; resulting file, say foo.tex, to be foo.nqtex, and then running nqfmt2fmt.
  ;; More convenient is to run infix-file with the :comment keyword parameter set to t,
  ;; which causes infix-file first to generate a .nqtex file and second to run
  ;; nqfmt2fmt on that file.

  ;; If the :just-remove-! keyword is t, then a file named root.stripped is
  ;; created, with all of our special ! commands options removed.

  ;; Implementation note.  In all the cases we treat explicitly, the characters
  ;; after !? are read with the Lisp function READ-PRESERVING-WHITESPACE, which
  ;; is just the same as READ except that it doesn't gratuitously consume whitespace
  ;; at the end of a READ.

  ;; Warning: Because we use a relative of READ to obtain the forms immediately
  ;; after the two character exclamation commands, the user must beware that if a
  ;; form to be read extends for more than a line, and if a semicolon (comment
  ;; character) is encountered on the next line, say at the beginning, READ will
  ;; skip that line from the semicolon on, which may not be what the user intends.
  ;; Thus it can be safer to use the \#\| ... \|\# construct for comments containing
  ;; !, especially if one is in the habit of using the Emacs command M-x fill
  ;; paragraph to rearrange paragraphs that begin with the comment delimiter
  ;; semicolon.

  (let ((infl (type-file-name fl (fmt2fmt-extension just-remove-! *infix-mode*) t))
        (*print-pretty* nil)
        (outfl (type-file-name fl (fmtfile-extension *infix-mode* nil) t))
        (a-very-rare-cons (cons nil nil))
        (count 1)
        (*readtable* (copy-readtable nil)))
    ;;(smash-infix-readtable)
    (with-open-file
     (*standard-input* infl :direction :input)
     (with-open-file
      (*standard-output* outfl :direction :output :if-exists :rename-and-delete)
      (catch 'eof
	(sloop for c = (read-char *standard-input* nil a-very-rare-cons)
	       (let ()
		 (cond ((eq c a-very-rare-cons) (return-from nqfmt2fmt outfl))
		       ;; The Latex indexing routines may insert new exclamation points!!!
		       ((eql c #\\)
			(cond ((skip-index-entries) nil)
			      ;; I am inserting a gigantic hack here because
			      ;; I can't figure out a more principled, simple 
			      ;; way to get the effect I want in LaTeX.
			      ;; The problem is that at the end of a tabbing
			      ;; environment we often have two lines of the form:
			      ;;   ....  \\
			      ;;   \end{tabbing}
			      ;; which causes the event formatted in the tabbing
			      ;; env to appear to be followed by two blank lines
			      ;; rather than one.
			      ((adjust-tabbing-env) nil)
			      (t (pwrite-char c))))

		       ((eql c #\!) (format-!-command just-remove-!))

		       ;; Let the user know we are making some kind 
		       ;; of progress, every 60 lines
		       ((eql c #\Newline)
			(if (= count 60)
			    (progn (format *terminal-io* "-") (setq count 1))
			    (incf count))
			(pwrite-char c))
		       (t (pwrite-char c)))
		 )))
      ))))
                              
(defun read-!-argument ()
  (let ((c (peek-char nil *standard-input* nil a-very-rare-cons)))
    (cond ((eq c a-very-rare-cons) (throw 'eof nil))
	  ((member c '(#\{ #\[))
	   (read-literally))
	  (t (acl2-read-preserving-whitespace)))))

(defun format-!-command (just-remove-!)
  (let (term)
    (cond
     (just-remove-! nil)
     ((catch 'taboverflow
	(let ((c (read-char *standard-input* nil a-very-rare-cons)))
	  (cond ((eq c a-very-rare-cons)
		 (pwrite-char #\!)
		 (throw 'eof nil))
		(t (case c
		     ((#\B #\b)
		      (setq term (acl2-read-preserving-whitespace))
		      (bold-sym-printer term))
		     ((#\I #\i)
		      (setq term (acl2-read-preserving-whitespace))
		      (italic-sym-printer term))
		     ((#\E #\e)
		      (setq term (acl2-read-preserving-whitespace))
		      (eval term))
		     ((#\T #\t)
		      (let ((c2 (peek-char nil *standard-input* nil a-very-rare-cons)))
			(if (eq c2 a-very-rare-cons)
			    (progn (pwrite-char #\!)
				   (pwrite-char c)
				   (throw 'eof nil))
			    (let ()
			      (if (char= c2 #\*)
				  (read-char *standard-input* nil a-very-rare-cons))
			      (setq term (acl2-read-preserving-whitespace))
			      (if (char= c2 #\*)
				  ;; * says "not inline"
				  (if (eventp term)
				      (infix-event term)
				      (infix-print-term term))
				  (let ((*do-not-use-tabs* t))
				  (if (eventp term)
				      (infix-event term)
				      (infix-print-term1 term))))))))
		     ((#\L #\l)
		      (setq term (read-literally))
		      (princ term))
		     ((#\N #\n)
		      (setq term (acl2-read-preserving-whitespace))
		      (define-label term))
		     ((#\X #\x)
		      (setq term (acl2-read-preserving-whitespace))
		      (reference-label term))
		     ((#\Q #\q)
		      (setq term (acl2-read-preserving-whitespace))
		      (pprinc "`")
		      (print-bare-function-name term)
		      (pprinc "'"))
		     ((#\S #\s)
		      (setq term (acl2-read-preserving-whitespace))
		      (sectionize term))
		     ((#\H #\h)
		      (setq term (acl2-read-preserving-whitespace))
		      (handsoff-or-on term))
		     ((#\V #\v)
		      (let* ((c2 (peek-char nil *standard-input* nil a-very-rare-cons)))
			(if (eq c2 a-very-rare-cons)
			    (let ()
			      (pwrite-char #\!)
			      (pwrite-char c)
			      (throw 'eof nil))
			    (let ()
			      (if (char= c2 #\*)
				  (read-char *standard-input* nil a-very-rare-cons))
			      (let ((term (read-!-argument)))
				(if (char= c2 #\*)
				    ;; * says "not inline"
				    (verbatim*-print-term term)
				    (verbatim-print-term term)))))))

		     ((#\Space #\Tab #\Newline)
		      (pwrite-char #\!)
		      (pwrite-char c))
		     (#\!
		      (pwrite-char #\!))
		     (otherwise
		      (pformat *terminal-io*
			       "Surprising character after ! ~a.~%" c)
		      (pwrite-char #\!)
		      (pwrite-char c))))))
	nil)
      (pformat *terminal-io*
	       "~%Sorry. Exceeded tabbing limit (2). 
We can't handle this large a form in running text.  ~a needs hand massaging.~%"
	       term)
      ;; Reinitialize to 0 points.
      (reset-loc-and-margin)
      (newlinex)))))

(defvar balanced-parens '((#\( . #\))
			  (#\[ . #\])
			  (#\{ . #\})))

(defun sectionize (rest)
  ;; sleazy hack to make sections independent of mode (scribe or tex).
  ;; if just-remove-! is T, then this is not called.
  (let (left right)
    (setq rest (intern (string-upcase rest)))
    (cond ((equal rest 'ection)
	   (setq left (read-char *standard-input* nil a-very-rare-cons))	 
	   (setq right (cdr (assoc left balanced-parens)))
	   (cond ((null right) (pprinc "section") (pwrite-char left))
		 (t (pprinc *begin-section-env*)
		    (sloop for c = (read-char *standard-input* nil a-very-rare-cons)
			   until (equal c right) 
			   do (pwrite-char c))
		    (pprinc *end-section-env*))))
	  ((equal rest 'ubsection) 
	   (setq left (read-char *standard-input* nil a-very-rare-cons))	 
	   (setq right (cdr (assoc left balanced-parens)))
	   (cond ((null right) (pprinc "subsection") (pwrite-char left))
		 (t (pprinc *begin-subsection-env*)
		    (sloop for c = (read-char *standard-input* nil a-very-rare-cons)
			   until (equal c right) 
			   do (pwrite-char c))
		    (pprinc *end-subsection-env*))))
	  (t (pformat *terminal-io* "Surprising string after !s ~s.~%" rest)
	     (pwrite-char #\!)
	     (pwrite-char #\s)
	     (pprinc rest)))))

(defun handsoff-or-on (rest)
  (let (sym)
    (setq sym (intern (string-upcase rest)))
    (cond ((equal sym 'andsoff)
	   (setq *hands-off* t))
	  ((equal sym 'andson) 
	   (setq *hands-off* nil))
	  (t (pformat *terminal-io* "Surprising string after !h ~s.~%" rest)
	     (pwrite-char #\!)
	     (pwrite-char #\h)
	     (if *hands-off*
		 (princ rest)
		 (print-atom rest))))))

(defun skip-index-entries ()
  ;; We are looking at a backslash. In Tex mode we need to skip to the end 
  ;; of the entry, because we may add !'s.  In Scribe mode this is just NIL.
  nil)

(defun adjust-tabbing-env ()
  ;; We are looking at a backslash. In Tex mode we may need to replace 
  ;;   ....  \\
  ;;   \end{tabbing}
  ;; with
  ;;   ....  
  ;;   \end{tabbing}
  ;; NOTE: this will only work if NQFMT2FMT is run.
  ;; In Scribe mode this is just NIL.
  nil)


;                        INFIX SETTINGS

; This function should be called by the <mode>-init file, or may be called by
; the <mode>-syntax file to override the <mode>-init settings.

(defun infix-settings
  (&key (mode           nil p-mode)                  ; string ["SCRIBE","latex",...]
        (extension              nil p-extension)              ; string ["MSS","tex"]
        (op-location            nil p-op-location)            ; ['FRONT,'BACK]
        (comment-format         nil p-comment-format)         ; ['text,'verbatim, cl,c]
        (format-!-in-comments   nil p-format-!-in-comments)   ; [T,nil]
        (acl2-format-comments   nil p-acl2-format-comments)   ; [T,nil]
        (acl2-format-doc-strings nil p-acl2-format-doc-strings) ; [T,nil]
        (no-index               nil p-no-index)               ; [t,NIL]
        (no-index-calls         nil p-no-index-calls)         ; [t,NIL,l]
        (suppress               nil p-suppress)  	      ; 
        (no-keys                nil p-no-keys)                ; [t,NIL]
        (no-guards              nil p-no-guards)              ; [t,NIL]
        (no-top-locals          nil p-no-top-locals)          ; [t,NIL]
        (no-verify-guards       nil p-no-verify-guards)	      ; [t,NIL]
        (hints                  nil p-hints))                 ; [t,NIL]

  (if p-mode           (setq *infix-mode* mode))
  (if p-extension      (setq *mode-extension* extension))
  (if p-op-location    (setq *infix-op-location* op-location))
  (if p-comment-format (setup-comment-format comment-format))
  (if p-format-!-in-comments   (setq *nq-default* format-!-in-comments))
  (if p-acl2-format-comments   (setq *acl2-format-comments* acl2-format-comments))
  (if p-acl2-format-doc-strings (setq *acl2-format-doc-strings* acl2-format-doc-strings))
  (if p-no-index
      (setq *do-not-index* no-index))
  (if p-no-index-calls
      (cond ((consp no-index-calls)
             (setq *do-not-index-calls-of* (append no-index-calls *do-not-index-calls-of*))
             (setq *do-not-index-calls* nil))
            (t  (setq *do-not-index-calls* no-index-calls))))
  (if p-no-keys (setq *print-keys* (not no-keys)))
  (if p-no-guards (setq *print-no-guards* no-guards))
  (if p-no-top-locals (setq *print-top-locals* (not no-top-locals)))
  (if p-no-verify-guards (setq *print-verify-guards* (not no-verify-guards)))
  (if p-hints (setq *print-hints* hints))
  (if p-suppress
      (progn (setq *print-hints* (not suppress))
	     (setq *print-keys* (not suppress))
	     (setq *print-no-guards* suppress)
	     (setq *print-top-locals* (not suppress))
	     (setq *print-verify-guards* (not suppress)))))

(defun will-will-not (x)
  (if x "will" "will not"))

(defun show-infix-settings ()
  (format *terminal-io* "~%Expecting a .~a file to be mapped to .~a file to be formatted by ~a."
          infix-input-file-type *mode-extension* *infix-mode*)
  (format *terminal-io* "~%Multiline infix ops will be printed at the ~a of the line." *infix-op-location*)
  (format *terminal-io* "~%Comment format is ~a." *comment-format*)
  (format *terminal-io* "~%Hints ~a be printed." (will-will-not *print-hints*))
  (format *terminal-io* "~%Guards ~a be printed." (will-will-not (not *print-no-guards*)))
  (format *terminal-io* "~%Other keyword arguments ~a be printed." (will-will-not *print-keys*))
  (format *terminal-io* "~%Top LOCALS ~a be printed." (will-will-not *print-top-locals*))
  (format *terminal-io* "~%!formatting ~a be in effect." (will-will-not *nq-default*))
  (format *terminal-io* "~%Index ~a be created." (will-will-not (not *do-not-index*)))
  (format *terminal-io* "~%Calls ~a be indexed." (will-will-not (not *do-not-index-calls*))))

(defun help-infix-settings ()
  (format *terminal-io* "~%To see current settings use (SHOW-INFIX-SETTINGS).
To change settings use INFIX-SETTINGS and supply the keyword arguments
for settings you wish to modify. Defaults in caps.~%")
  (format *terminal-io* "~%:mode           : string - formatting style [\"SCRIBE\",\"latex\",...]")
  (format *terminal-io* "~%:extension      : string - output file extension [\"MSS\",\"tex\"]")
  (format *terminal-io* "~%:op-location    : ['FRONT, 'back]
                - Multiline infix operators will be printed at the front 
                - or back of the line according to this setting.")
  (format *terminal-io* "~%:comment-format : ['TEXT, 'verbatim, 'cl, 'c] - Comment format")
  (format *terminal-io* "~%:no-index: [t,NIL] - Index will/will not be created")
  (format *terminal-io* "~%:no-index-calls : [t,NIL,1ist]
                - Calls will/will not be indexed.  If a list,
                - these functions will not be indexed.")
  (format *terminal-io* "~%:hints                   : [t, NIL] - print hints.")
  (format *terminal-io* "~%:no-guards               : [t, NIL] - suppress guards.")
  (format *terminal-io* "~%:no-keys                 : [t, NIL] - suppress keyword args.")
  (format *terminal-io* "~%:no-top-locals           : [t, NIL] - suppress toplevel LOCALs.")
  (format *terminal-io* "~%:no-verify-guards        : [t, NIL] - suppress VERIFY-GUARDS.")
  (format *terminal-io* "~%:suppress                : [t, NIL] - suppress above 5.")

  (format *terminal-io* "~%:format-!-in-comments    : [T, nil] - !formatting in effect.")
  (format *terminal-io* "~%:acl2-format-comments    : [T, nil] - acl2 comment formatting.")
  (format *terminal-io* "~%:acl2-format-doc-strings : [T, nil] - acl2 formatting doc strings."))

                                   
;                             DEFINITION BY EXAMPLES

; Anyone extending the syntax by hand rather than by use of one of the make...
; functions ought also to add something to this list of examples to illustrate
; the new syntax.

(defun functify (l)
  ;; Removes redundant elements from an alist.
  (sloop for tail on l with ans
           do (cond ((null (assoc (car (car tail)) ans))
                     (push (car tail) ans)))
           finally (return (nreverse ans))))

(defvar *infix-test-directory*
  (concatenate 'string *infix-directory* "test/"))

(defun scrunch (l)
  (sloop for tail on l unless (member (car tail) (cdr tail))
	 collect (car tail)))

(defun print-examples (&optional mode)

; Illustrates the current syntax via a brief sample document.

  (cond (mode
	 (cond ((and (stringp mode)
		     (or (equal mode "latex")
			 (equal mode "ascii")
			 (equal mode "html")
			 (equal mode "scribe")))
		(setq *infix-mode* mode)
		(format t "~%Entering ~a mode.~%" *infix-mode*)
		(load-infix-init-file))
	       (t (error (format nil "Unknown mode ~s" mode)))))
	((stringp *infix-mode*)
         (format t "~%Remaining in ~a mode.~%" *infix-mode*))
        ((null *infix-mode*)
         (format t "~%Provide a mode, e.g. (print-examples \"latex\").~%"))
        (t (format t "In ~a mode." *infix-mode*)))

  (let ((*print-pretty* nil)
        (*print-case* :downcase))
    (with-open-file
     (*standard-output* (type-file-name "infix-examples"
					(fmt2fmt-extension nil *infix-mode*) t)
                        :direction :output
                        :if-exists :rename-and-delete)
     (ppformat *example-prelude* *infix-mode*)
     (to-current-margin)
     (sloop for form in (functify *wired-in-infix-examples*)
              do (let ((*do-not-use-tabs* t))
                   (pprinc *begin-item*)
                   (special-char-printer form)
                   (ppformat " is printed as ~%~%")
                   (infix-form form)
                   (ppformat ".~%")
                   (pprinc *end-item*))
              (ppformat "~%"))
     (pprinc *begin-item*)
     (ppformat "The remaining symbols that are printed specially are
described in the following table.")
     (pprinc *end-item*)
     (pprinc *end-enumerate-env*)
     (ppformat "~%~%")
     (let ((*tab-stack* t)
	   (table-number 1)
	   (example-number 1))
       ;; need to set *tab-list* non-nil to ensure (newlinex) works properly.
       (ppformat *begin-example-table* table-number)
       (sloop for form in
              (append (sloop for pair in (functify *atom-alist*)
			     collect (car pair))
                      (sloop for name in (scrunch *constant-ops*)
			     collect (list name))
                      (sloop for name in (scrunch *infix-ops*)
			     collect (list name 'x 'y))
                      (sloop for pair in (functify *negative-constant-table*)
			     collect (list 'not (list (car pair))))
                      (sloop for pair in (functify *negative-infix-table*)
			     collect (list 'not (list (car pair) 'x 'y)))
                      (sloop for name in (scrunch *unary-prefix-ops*)
			     collect (list name 'x))
                      (sloop for pair in (functify *negative-unary-prefix-table*)
			     collect (list 'not (list (car pair) 'x)))
                      (sloop for name in (scrunch *unary-suffix-ops*)
			     collect (list name 'x))
                      (sloop for pair in (functify *negative-unary-suffix-table*)
			     collect (list 'not (list (car pair) 'x)))
                      (sloop for name in (scrunch *unary-abs-ops*)
			     collect (list name 'x))
                      (sloop for pair in (functify *prefix-multiple-ops*)
			     collect (cons (car pair)
					   (sloop for i from 1 to (cdr pair) collect
						  (intern (format nil "X~a" i)))))
                      (sloop for pair in (functify *suffix-multiple-ops*)
			     collect (cons (car pair)
					   (sloop for i from 1 to (cdr pair) collect
						  (intern (format nil "X~a" i)))))
                      (sloop for pair in (functify *infix-multiple-ops*)
			     collect (cons (car pair)
					   (sloop for i from 1 to (1+ (cdr pair)) collect
						  (intern (format nil "X~a" i))))))
              do (let ((*do-not-use-tabs* t))
		   (cond ((> example-number *example-table-size*)
			  (ppformat *end-example-table*)
			  (ppformat *begin-example-table* table-number)
			  (setq example-number 1)))
                   (special-char-printer form)
                   (pprinc *column-separator*)
                   (infix-form form)
		   (new-tab-row nil)
		   (line-return)
		   (setq table-number (+ 1 table-number))
		   (setq example-number (+ 1 example-number))
                   (setq *infix-loc* *left-margin*)))
       (ppformat *end-example-table*))
     (pprinc *example-postlude*))
    (nqfmt2fmt "infix-examples")))

; The following should be modified to interact with the vaiables that set 
; parens printing.  In particular, this seems to be the piece of precedence that
; is most easily screwed up.

(defun special-char-printer (x &optional listp)
  (cond (listp
	 (cond ((null x) nil)
	       ((not (consp x))
		(pprinci " . ")
		(print-atom x))
	       (t (special-char-printer (car x) nil)
		  (and (consp (cdr x)) (pprinci " "))
		  (special-char-printer (cdr x) t))))
	((not (consp x)) (print-atom x))
	(t (pprinci "(")
	   (special-char-printer x t)
	   (pprinci ")"))))

;                                      NOT

; The following code is for the special handling of NOT, which involves diving
; into the term negated to turn a predicate into one that has a slash through
; it.  We advise that the casual user not touch this.

(defun not-printer (term)
  (let (x)
    (cond ((atom (cadr term))
           (default-unary-prefix-printer term *neg-str*))
          ((setq x (assoc (car (cadr term)) *negative-constant-table*))
           (pprinc (cadr x)))
          ((setq x (assoc (car (cadr term)) *negative-infix-table*))
           (default-infix-printer (cadr term) (cadr x)))
          ((setq x (assoc (car (cadr term)) *negative-unary-prefix-table*))
           (default-unary-prefix-printer (cadr term) (cadr x)))
          ((setq x (assoc (car (cadr term)) *negative-unary-suffix-table*))
           (default-unary-suffix-printer (cadr term) (cadr x)))
          (t (default-unary-prefix-printer term *neg-str*)))))

(declare-fn-printer not (function not-printer))

;; REAL INITIALIZATION.  Used to reinitialize during clean-up-everything
(setq *save-fn-alist* *fn-alist*)


;                          USER MODIFIABLE TABLE SETUP

; It is easy to augment, or even to modify, the syntax by calling one of the
; make-... functions illustrated below.  The non-initial arguments to these
; make-...  functions are strings to be printed by Latex to generate operators
; and other noise words when printing a term whose function symbol is the
; intial argument of the call to make-...

; make-infix-op, make-unary-prefix-op, and make-unary-suffix-op take an
; optional second argument, *neg-str*, which indicates how to print an the
; negation of an application of the function symbol in question.

; In TeX or Latex, one can do astonishingly clever things.  But the strings
; that we have in mind should do nothing clever involving motion, they should
; only result in characters being placed at the current location.  While being
; printed, the strings will be passed no arguments or information about the
; context in which printing is to take place. Typically, these strings should
; be nothing more than instructions to print a single symbol. The strings are
; processed in `math mode', and in fact, they are auomatically embedded in
; $...$.

; None of the operators below are built into this printer anywhere else except
; by the code below.  The meaning of `not', defined above, is wired in because
; it gives the meaning to the optional *neg-str* arguments.


;                                     CONSTANT-OPS

; Sometimes you want to print a function as a constant, particularly if it is one.
; (make-constant-op op str) causes (op ..) to print as str. 

;                                     INFIX-OPS

; infix-ops (infix operators) should be function symbols of two or more
; arguments for which it is desired that one symbol come out between every
; adjacent pair of arguments.  E.g., invoking (make-infix-op plus "+") causes
; the term (plus a b c d) to be printed as (a $+$ b $+$ c $+$ d).  Invoking
; (make-infix-op equal "=" "\\not=") causes the term (equal x y) to be printed
; as (x $=$ y) and it also causes the term (not (equal x y)) to be printed as
; (x $\not= y).

; Thus, for example, if one introduces a new function, say join, and wants to
; print terms of the form (join x y) as (x \bigtriangledown y), cf. p. 44 of
; the Latex manual, then one should invoke:

;   (make-infix-op join "\\bigtriangledown")

; from Lisp.  That is all that need be done to cause infix-file to subsequently
; print `join' terms this way.

; Note that throughout the following examples, we have used two backslashes to
; get one because, in Common Lisp, backslash is a character for quoting other
; characters.

; Examples of make-infix-op are contained in latex-syntax.lisp.  Look for INFIX OPERATORS.


;             UNARY-PREFIX-OPS, UNARY-SUFFIX-OPS, and UNARY-ABS-OPS

; Use make-unary-prefix-op and make-unary-suffix-op only for function symbols
; of one argument.  The string str (or *neg-str*) will be printed before or
; after the argument.

; For more examples, see latex-syntax.lisp.

; unary-suffix-ops should be unary function symbols.

; (make-unary-suffix-op foo x str) makes (foo x) print as (x $str$).

; Examples of make-unary-suffix-op.

; unary-prefix-ops should be unary function symbols. 

; (make-unary-prefix-op foo str) makes (foo x) print as ($str$ x).

; unary-abs-ops should be unary function symbols.

; To create syntax like that for absolute value, use (make-unary-absolute-op
; lhs-str rhs-str), where lhs-str and rhs-str are the strings to print on the
; left and right of the argument.  (make-unary-abs-op foo str1 str2) makes (foo
; x) print as (str1 x str2).  See the example for abs below.


;                           SOME POSSIBLE EXTENSIONS


;; (simple-extension)           ; see latex-syntax.lisp
;; (dmg-syntax)                 ; see latex-syntax.lisp

; Undoing.  To cause applications of a function symbol fn to be printed in the
; default way, i.e., fn(x, y), invoke (clean-up 'fn).


;; TESTING

;; Lines in the test file are the following form:

;;    filename 
;; or (filename mode)
;; or (filename mode comment)

;; Comment defaults to T.

(defvar *mode-list* '("latex" "scribe"))

(defvar *test-file* "testfile")

;; Better to use test-directory
;; (defun test-infix ()
;;   (let ((files (read-file-list *test-file*)))
;;     (sloop for test in files
;;              do (cond ((or (stringp test)
;;                            (and (consp test) (null (cdr test))))
;;                        ;; "file" or (file)
;;                        (if (consp test) (setq test (car test)))
;;                        (sloop for mode in *mode-list*
;;                                 do (progn
;; 				     (format *terminal-io*
;; 					     "~%Translating ~a in ~a mode.~%" test mode)
;; 				     (infix-file test :mode mode :comment t))))
;;                       ((and (consp test) (eql (length test) 2))
;;                        ;; (file mode)
;;                        (format *terminal-io*
;; 			       "~%Translating ~a in ~a mode.~%" (car test) (cadr test))
;;                        (infix-file (car test) :mode (cadr test) :comment t))
;;                       ((and (consp test) (eql (length test) 3))
;;                        ;; (file mode comment)
;;                        (format *terminal-io* 
;;                                "~%Translating ~a in ~a mode, with comment = ~a.~%"
;;                                (car test) (cadr test) (caddr test))
;;                        (infix-file (car test) :mode (cadr test) :comment (caddr test)))
;;                       (t (format *terminal-io* "~%BAD TEST FILE SPEC: ~ad.~%" test))))))

;; (defun read-file-list (file)
;;   (cond ((null (probe-file file))
;;          (format t "~%Failed to find file: ~a~%" file))
;;         (t (with-open-file
;;             (*standard-input* file :direction :input)
;;             (sloop for form = (readx *standard-input* nil a-very-rare-cons nil)
;;                      until (eq form a-very-rare-cons)
;;                      collect form)))))


;; Testing functions

;; MODIFY and USE:
;; mks: script test.log
;; mks: 
;; >(load "infix")
;; >(test-directory "scribe")
;; >(bye)
;; mks: foreach f (*.mss)
;; ? scribe $f
;; ? end
;; mks: ^D
;; mks: sed -e s/^V^M// test.log > test-scribe.log

(defun test-directory (mode &optional directory (comment nil comment-p) omit-prelude)
  ;; ONLY EXPECTING mode = "latex", "scribe", or "ascii"
  (let (type
	(dir (or directory *infix-test-directory*))
	;; Default testing excludes nqfmt2fmt translations
	;; for latex, since it is so much more sensitive
	;; to character weirdness.
	(com (if (string= mode "latex") nil t)))
    (cond ((string= mode "latex") (setq type "tex"))
	  ((string= mode "tex") (setq type "tex"))
	  ((string= mode "scribe") (setq type "mss"))
	  ((string= mode "ascii") (setq type "doc")))
    (if comment-p (setq com comment))
    (setq dir (concatenate 'string dir "*.lisp"))
    (mapc (function (lambda (f) 
                      (format *terminal-io* "~%Infixing ~a.lisp." (pathname-name f))
		      (reset-loc-and-margin)
		      (if (not (probe-file f))
			  (format *terminal-io* "~%~a does not exist. Skipping.~%" f)
			  (if (probe-file (make-pathname :type type :defaults f))
			      (format *terminal-io* "~%~a.~a already exists. Skipping.~%"
				      (pathname-name f) type)
			      (infix-file f :mode mode :omit-prelude omit-prelude)))))
          (directory dir))))

(format *terminal-io* "~%--- Do (help-infix-settings) for help. ---~%~%")



;; infix-ascii , infix-latex , infix-scribe, infix

(defmacro infix-ascii (&rest files)
  `(do-infix-ascii ',files))

(defun do-infix-ascii (files) 
  (load-base "ascii-syntax.lisp" t)
  (mapc (function (lambda (f) (infix-file f))) files))

(defmacro infix-latex (&rest files)
  `(do-infix-latex ',files))

(defun do-infix-latex (files)
  (load-base "latex-syntax.lisp" t)
  (mapc (function (lambda (f) (infix-file f))) files)
  (load-base "ascii-syntax.lisp" t))

(defmacro infix-html (&rest files)
  `(do-infix-html ',files))

(defun do-infix-html (files)
  (load-base "html-syntax.lisp" t)
  (mapc (function (lambda (f) (infix-file f))) files)
  (load-base "ascii-syntax.lisp" t))

(defmacro infix-scribe (&rest files)
  `(do-infix-scribe ',files))

(defun do-infix-scribe (files) 
  (load-base "scribe-syntax.lisp" t)
  (mapc (function (lambda (f) (infix-file f))) files)
  (load-base "ascii-syntax.lisp" t))

(defmacro infix (&rest files)
  `(do-infix-root ',files))

(defun do-infix-root (files) 
  (mapc (function
	 (lambda (f)
	   (let ((*current-directory* (directory-or-current f))
		 (f-root (pathname-name f)))
	     (cond ((probe-syntax f-root)
		    (setq *infix-mode* (pathname-name f-root))
		    (format t "~%Entering ~a mode.~%" *infix-mode*)
		    (load-infix-init-file)
		    (infix-file f))
		   (t (format t
			      "~%~%!!!Formatting aborted. No syntax file for s.~%"
			      f-root))))))
	files)
  (load-base "ascii-syntax.lisp" t))





#| Note on math printing from LSmith.  Adopted.

Bill and I had troubles printing with infix because nesting math modes
is not allowed in latex. Our latex output from infix is filled with 

     {\ifmmode <form>\else$<form>$\fi}

with the two <form>s identical and often containing more ifmmodes.

I happened across a solution to this problem in the Latex manual.

     \mbox{$<form>$}

works whether of not we are currently in mathmode. In other words, it
is possible to nest math modes in latex after all. This feature is
documented in section 3.4.1, Defining Commands, in the Latex manual. 

|#
